
# If it turns out that some optimization could help, one of the first steps
# to do would be to reduce the length of symbol names used routinely
# or even behind the scenes - such as XSUB names and Perl-space variable names.
# C-space variable name lengths don't make a difference at runtime (with static linking).


# The following XSUBs comprise a basic interface from Perl space 
# to Kludge objects

int
Path2Ptr(path_str, this_object)
   char* path_str
   int this_object
   CODE:
   Kludge_Thread* this_thread = Kludge_Thread::Lookup_By_ID(pthread_self());
   string* tmp_str = new string(path_str);
   Kludge_Path* tmp_path = Kludge_CLI_String2path(tmp_str, this_thread->CLI->output_stream);
   RETVAL = (int) Kludge_CLI_Path2Ptr(tmp_path, this_thread->CLI,(Kludge_Object*)this_object);
   delete tmp_str;
   delete tmp_path;
   OUTPUT:
   RETVAL


char*
get_object_name(this_object)
   int this_object
   CODE:
   char* output;
   string* tmp_ptr = ((Kludge_Object*)this_object)->My_Name();
   if (tmp_ptr == NULL)
   {
      output = "";
   }
   else
   {
      output = (char*) tmp_ptr->c_str(); // the typecast just zaps a warning about "const"
      // must not delete tmp_ptr - it's a pointer to the object's very own "string Name" 
      // or "string RDN";
   }
   RETVAL = output;
   OUTPUT:
   RETVAL


SV*
get_object_type(this_object)
   int this_object
   CODE:
   if (this_object != 0)
   {
      switch (((Kludge_Object*)this_object)->Type)
      {
	 case Kludge_Object::KLUDGE_ROOT : RETVAL = newSVpv("ROOT",0);
	 break;
	 case Kludge_Object::KLUDGE_CLASS : RETVAL = newSVpv("CLASS",0);
	 break;
	 case Kludge_Object::KLUDGE_CLASS_PROPERTY :
	 {
	    switch (((Kludge_Class_Property*)this_object)->SubType)
	    {
	       case Kludge_Class_Property::SUBTYPE_PROPERTY : RETVAL = newSVpv("CLASS_PROPERTY",0);
	       break;
	       case Kludge_Class_Property::SUBTYPE_BLOB : RETVAL = newSVpv("CLS_BLOB",0);
	       break;
	       case Kludge_Class_Property::SUBTYPE_TXT_BLOB : RETVAL = newSVpv("CLS_TXT_BLOB",0);
	       break;
	       case Kludge_Class_Property::SUBTYPE_METHOD : RETVAL = newSVpv("METHOD",0);
	       break;
	       default: RETVAL = newSVpv("CLASS_PROPERTY_UNKNOWN",0);
	       break;
	    }
	 }
	 break;
	 case Kludge_Object::KLUDGE_CLASS_REL_STUB : RETVAL = newSVpv("CLASS_RELATIONSHIP",0);
	 break;
	 case Kludge_Object::KLUDGE_INSTANCE : RETVAL = newSVpv("INSTANCE",0);
	 break;
	 case Kludge_Object::KLUDGE_INSTANCE_PROPERTY :
	 {
	    switch (((Kludge_Instance_Property*)this_object)->SubType)
	    {
	       case Kludge_Instance_Property::SUBTYPE_PROPERTY : RETVAL = newSVpv("INSTANCE_PROPERTY",0);
	       break;
	       case Kludge_Instance_Property::SUBTYPE_BLOB : RETVAL = newSVpv("INST_BLOB",0);
	       break;
	       case Kludge_Instance_Property::SUBTYPE_TXT_BLOB : RETVAL = newSVpv("INST_TXT_BLOB",0);
	       break;
	       default: RETVAL = newSVpv("INSTANCE_PROPERTY_UNKNOWN",0);
	       break;
	    }
	 }
	 break;
	 case Kludge_Object::KLUDGE_INST_REL_STUB : RETVAL = newSVpv("INSTANCE_RELATIONSHIP",0);
	 break;
	 default: RETVAL = newSVpv("UNKNOWN",0);
	 break;
      }
   }
   else RETVAL = newSVpv("",0);
   OUTPUT:
   RETVAL


int
resolve_content_name(this_object, content_name, own_only)
   int this_object
   char* content_name
   int own_only
   CODE:
   Kludge_Thread* this_thread = Kludge_Thread::Lookup_By_ID(pthread_self());
   if (this_object != 0)
   {
      string tmp_name(content_name);
      RETVAL = (int) ((Kludge_Object*)this_object)->Resolve_Content_Name(
                       &tmp_name,
       	               Kludge_Thread::Lookup_By_ID(pthread_self())->CLI,
		       (own_only == 0) ? false : true
                                                                        );
   }
   else
   {
      RETVAL = -1;
   }
   OUTPUT:
   RETVAL

# The purpose of the following XSUB is to explicitly flush a given Perl IO handle,
# which is especially useful (or even necessary) when serving an AF_INET socket,
# originally open by the host program in C space and then dup()ped and open()ed
# in an embedded Perl space (or hooked up straight into the guts the way we do it).
# More on the internals: the file/socket access stack in Perl on UNIX looks like this:
# 
# 4) Perl script space - file handles, with no special leading characters.
#    Can only be passed into an XSUB as part of a typeglob - if you try to pass
#    the handle name alone, without the leading asterisk, the XSUB obtains
#    a string (a scalar value).
#    
# 3) Perl guts - struct IO - another `struct SV' look-alike, wrapping two
#    `struct FILE's: one for input and one for output.
#    
# 2) <stdio.h> - struct FILE (AKA file stream), used by functions such as 
#    fopen(), fclose(), fread(), fwrite() - and fflush(), as this is precisely
#    where this sort of buffering happens.
#    THIS IS WHERE WE CAN CALL FFLUSH.
#    Note: in Perl, when using <stdio.h>, `struct PerlIO' is typedeffed to 
#    `stuct FILE'. Alternatively, perl can use <sfio.h> - in that case, PerlIO
#    is typedeffed to Sfio_t - dunno how that one works.
#    Note2: this is also what GNU C++ is using as a backend to its <streambuf>
#    library - hence, this is where the `<< flush' manipulator does its dirty
#    job :)
#    
# 1) <stdio.h> - (int)file_descriptor used by C-space functions such as 
#    open(), close(), read(), write().
#    Alternatively, perl can use sfio.h - dunno how that one works. 
int
flush_file(file_to_flush)
   SV* file_to_flush
   CODE:
   #   PerlIO_flush() equals fflush(), unless you're using the alternative io library.
   #   ST(0) = first parameter on stack - i.e., file_to_flush. It's a glob value.
   #           The (int) type declared for file_to_flush is completely bogus.
   #   GvIOp() = unwrap an IO* out of a GV*. `struct IO' is another `struct SV' look-alike.
   #   IoOFP() = unwrap the output FILE* out of an IO*. Even if there's really a single
   #             low-level (int) file descriptor or socket open in read+write mode,
   #	     there are two FILEs in a `struct IO', one for reading and one for
   #	     writing. Check (FILE*)->_fileno  to see for yourself. 
   if (SvTYPE(file_to_flush) == SVt_RV)
   {
      PerlIO_flush( IoOFP(GvIOp(SvRV(file_to_flush))) );
   }
   else if (SvTYPE(file_to_flush) == SVt_PVGV)
   {
      PerlIO_flush( IoOFP(GvIOp(file_to_flush)) );
   }
   else
   {
      cout << "Error - in Kludge_pl::flush_file(): object to flush is not a file handle.\n";
   }

   RETVAL = 0;
   OUTPUT:
   RETVAL


# The following two XSUBs are only here for debugging purposes.
# They don't have any actual use at the moment.
# They helped me a great deal while I was figuring out how to hook up
# the sockets to Perl IO.

int
get_current_fd()
   CODE:
   /* RETVAL = dup(Kludge_Thread::Lookup_By_ID(pthread_self())->socket); */
   /* work behind the scenes on Sockets */
   RETVAL = Kludge_Thread::Lookup_By_ID(pthread_self())->socket;
   OUTPUT:
   RETVAL


int
close_fd(sock_to_close)
   int sock_to_close
   CODE:
   RETVAL = close(sock_to_close);
   OUTPUT:
   RETVAL


# The following function runs the Kludge CLI to serve the particular client whose
# arrival has invoked the current child thread.
# Originally, this function was called from C-space, directly from the top thread
# function (from the very thread body). Now this call is wrapped inside an XSUB
# inside a Perl-space wrapper function, that sets up some thread-local Perl-space
# variables, before the CLI is unleashed, and does some cleanup after the CLI is
# finished with its job.
# Messy as this may sound (but hey, this prog is called a kludge, right?), it is
# probably the only way to get things set up smooth and quick. Calling some
# low-level Perl guts from C space proved to be an infeasibly complex task.
# Namely, each call to perl_eval_sv() invokes its own scope, as a result of which
# doing a separate lead-in and lead-out eval is not a useful way to initialize
# the root scope of a particular thread.
int   
run_Kludge_CLI()
   CODE:
   /* NOW FINALLY RUN THE CLI :) */
   RETVAL = Kludge_Thread::Lookup_By_ID(pthread_self())->CLI->Run();
   OUTPUT:
   RETVAL


# The following three functions help to simulate a thread-local socket handle :)
# - create_thr_socket_handle() creates an anonymous glob and hooks it up to this thread's socket
# - get_thr_socket_handle() looks up the current thread's socket filehandle (GV) reference
# - destroy_thr_socket_handle() destroys the anonymous glob, after it's been close()'d in Perl space
# The `create' and `destroy' functions are used by the Perl-space thread init wrapper sub.
# The `get' function is used silently at the beginning of each Kludge_pl_eval_sub()
# to setup `my $CLIENT_SOCKET'.
SV*
create_thr_socket_handle()
   CODE:
   Kludge_Thread* current_thread = Kludge_Thread::Lookup_By_ID(pthread_self());
   /* glob name generation principle cannibalized from newGVgen() */
   current_thread->socket_glob = gv_fetchpv(
                   form("Kludge_pl::THR_%d", current_thread->pl_thr->tid),
                   TRUE,  /* add a new one if doesn't exist yet? */
                   SVt_PVGV
                                           );
   
   /* Some work behind the scenes on Sockets. */
   /* This is exactly what perl-space open(int fildes) does, */
   /* plus we also tag the handle as a socket - something that */
   /* Perl only does with its Socket module/objects (normally). */
   IO* io = GvIOn(current_thread->socket_glob);
   IoIFP(io) = PerlIO_fdopen(current_thread->socket, "r");
   IoOFP(io) = PerlIO_fdopen(current_thread->socket, "w");
   IoTYPE(io) = 's';
   /* Each fdopen() on a socket generates a failed seek op, */
   /* which can be seen in the errno variable and externally in an strace output. */
 
   RETVAL = newRV_inc((SV*) current_thread->socket_glob);
   OUTPUT:
   RETVAL

SV*
get_thr_socket_handle()
   CODE:
   GV* glob_containing_handle = Kludge_Thread::Lookup_By_ID(pthread_self())->socket_glob;
   if (glob_containing_handle != NULL)
   {
      RETVAL = newRV_inc( (SV*) glob_containing_handle );
   }
   else
   {
      // A dummy value needs to be inserted to prevent misbehavior 
      // when this XSUB is automagically inserted within Kludge_pl_thr_Run()
      RETVAL = newSViv(0);
   }
   OUTPUT:
   RETVAL

int
destroy_thr_socket_handle()
   CODE:
   SvREFCNT_dec( (SV*) Kludge_Thread::Lookup_By_ID(pthread_self())->socket_glob );
   RETVAL = 0;
   OUTPUT:
   RETVAL

