/*

For Kludge, this base64 library has been adapted from Perl's MIME::Base64.

The tables and some of the code that used to be here was borrowed from
metamail, which comes with this message:

  Copyright (c) 1991 Bell Communications Research, Inc. (Bellcore)

  Permission to use, copy, modify, and distribute this material 
  for any purpose and without fee is hereby granted, provided 
  that the above copyright notice and this permission notice 
  appear in all copies, and that the name of Bellcore not be 
  used in advertising or publicity pertaining to this 
  material without the specific, prior written permission 
  of an authorized representative of Bellcore.	BELLCORE 
  MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY 
  OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", 
  WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.

*/


#include <base64.h>

int byte_buf::set(void* init_buffer, unsigned int init_len, bool copy_buf = true, bool _delete_buf = true)
{
 int result = 0;

   delete_buf = _delete_buf;
   len = init_len;

   if (init_buffer != NULL)
   {
      if (copy_buf)
      {
         buf = new unsigned char[init_len];
         memcpy(buf, init_buffer, init_len);
      }
      else
      {
         buf = (unsigned char*) init_buffer;
      }
   }
   else
   {
      buf = NULL;
      cout << "byte_buf::set() obtained a null pointer for init!\n";
      result = 1;
   }

   return(result);
}






#define MAX_LINE  76 /* size of encoded lines */

static char basis_64[] =
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

#define XX      255	/* illegal base64 char */
#define EQ      254	/* padding */
#define INVALID XX

static unsigned char index_64[256] = {
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,62, XX,XX,XX,63,
    52,53,54,55, 56,57,58,59, 60,61,XX,XX, XX,EQ,XX,XX,
    XX, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25,XX, XX,XX,XX,XX,
    XX,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51,XX, XX,XX,XX,XX,

    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
};



string* encode_base64(byte_buf* data_to_encode)
{
 unsigned char *str;  /* string to encode */
 int len;             /* length of the string */
 char *eol;           /* the end-of-line sequence to use */
 unsigned int eollen; /* length of the EOL sequence */
 char *r;             /* result string */
 unsigned int rlen;   /* length of result string */
 unsigned char c1, c2, c3;
 int chunk;
 string* result = NULL;

   str = data_to_encode->buf;
   len = data_to_encode->len;
   eol = "\n";
   eollen = 1;

   /* calculate the length of the result */
   rlen = (len+2) / 3 * 4;	 /* encoded bytes */
   if (rlen) {
       /* add space for EOL */
       rlen += ((rlen-1) / MAX_LINE + 1) * eollen;
   }

   /* allocate a result buffer */
   result = new string();
   result->resize(rlen ? rlen : 1);
   r = (char*) result->c_str(); // the typecast suppresses a warning

   /* encode */
   for (chunk=0; len > 0; len -= 3, chunk++) {
       if (chunk == (MAX_LINE/4)) {
           char *c = eol;
           char *e = eol + eollen;
           while (c < e)
               *r++ = *c++;
           chunk = 0;
       }
       c1 = *str++;
       c2 = *str++;
       *r++ = basis_64[c1>>2];
       *r++ = basis_64[((c1 & 0x3)<< 4) | ((c2 & 0xF0) >> 4)];
       if (len > 2) {
           c3 = *str++;
           *r++ = basis_64[((c2 & 0xF) << 2) | ((c3 & 0xC0) >>6)];
           *r++ = basis_64[c3 & 0x3F];
       } else if (len == 2) {
           *r++ = basis_64[(c2 & 0xF) << 2];
           *r++ = '=';
       } else { /* len == 1 */
           *r++ = '=';
           *r++ = '=';
       }
   }
   if (rlen) {
       /* append eol to the result string */
       char *c = eol;
       char *e = eol + eollen;
       while (c < e)
           *r++ = *c++;
   }
   *r = '\0';  /* every SV in perl should be NUL-terminated */

   return(result);
}



byte_buf* decode_base64(string* string_to_decode)
{
 unsigned int len = string_to_decode->size();
 register unsigned char *str = (unsigned char*) string_to_decode->c_str(); // the typecast suppresses a warning
 unsigned char const* end = str + len;
 unsigned char *r;
 unsigned char c[4];
 byte_buf* result = NULL;

     {
         /* always enough, but might be too much */
         unsigned int rlen = len * 3 / 4;
         result = new byte_buf(rlen ? rlen : 1); 
     }
     r = result->buf;

     while (str < end) {
         int i = 0;
         do {
             unsigned char uc = index_64[*str++];
             if (uc != INVALID)
                 c[i++] = uc;

             if (str == end) {
                 if (i < 4) {
                     if (i)
                         cout << "Premature end of base64 data.\n";
                     if (i < 2) goto thats_it;
                     if (i == 2) c[2] = EQ;
                     c[3] = EQ;
                 }
                 break;
             }
         } while (i < 4);
         
         if (c[0] == EQ || c[1] == EQ) {
             cout << "Premature padding of base64 data.\n";
             break;
         }
         /* printf("c0=%d,c1=%d,c2=%d,c3=%d\n", c[0],c[1],c[2],c[3]); */

         *r++ = (c[0] << 2) | ((c[1] & 0x30) >> 4);

         if (c[2] == EQ)
             break;
         *r++ = ((c[1] & 0x0F) << 4) | ((c[2] & 0x3C) >> 2);

         if (c[3] == EQ)
             break;
         *r++ = ((c[2] & 0x03) << 6) | c[3];
     }

   thats_it:
     result->len = r - result->buf;
     *r = '\0';

   return(result);
}
