/*

This library defines perhaps the simplest possible base64 encoding and decoding
routines.
At the decoded side, we're talking about binary data, rather than null-terminated
textual strings - for this purpose, this library also offers a raw binary buffer
class: a buffer accompanied by a length counter (like a Borland Pascal `string')
with customizable auto alloc/free and customizable cctor/`ptr assignment' behavior.

For Kludge, this base64 library has been adapted from Perl's MIME::Base64.

Lazy as I am, I've evaluated several base64 routines available - namely the
Perl's Base64.xs and Randy Charles Morin's Base64 C++ class, along with several
extensive MIME libraries. I had to conclude that the Perl's Base64 XSUBs are
probably the simplest and fastest - implemented in plain C, with no additional
code I wouldn't need. I merely had to convert the XS language hooks into C grammar
and replace Perl guts types with C++ types. The little amount of C++ in the
resulting library only serves to ensure proper construction and destruction
- it should not result in significant processing overhead.

*/


#ifndef _BASE64_H
#define _BASE64_H

#include <string>

class byte_buf
{
public:
   unsigned char* buf;
   unsigned int len;

   byte_buf(unsigned int init_len) : delete_buf(true)
      { buf = new unsigned char[init_len]; len = init_len; }
   byte_buf(void* init_buffer, unsigned int init_len, bool copy_buf = true, bool _delete_buf = true)
      { set(init_buffer, init_len, copy_buf, _delete_buf); }
   ~byte_buf() { if (delete_buf && (buf != NULL)) delete buf; }
   int set(void* init_buffer, unsigned int init_len, bool copy_buf = true, bool _delete_buf = true);
   // copy_buf: 
   //     true  = "copy-construct buffer contents (deep copy)", 
   //     false = "only copy the pointer (shallow copy)"
   // delete_buf: 
   //     true  = "deallocate the buffer when the byte_buf object is destroyed", 
   //     false = "don't deallocate the buffer when the byte_buf object is destroyed"
private:
   bool delete_buf;
};

string* encode_base64(byte_buf* data_to_encode);
byte_buf* decode_base64(string* string_to_decode);

#endif /* _BASE64_H */
