#include "kludge_CLI.h"


string blanks = " \t",                      // "stuffing" between punctuation and data tokens
       punctuation = "={}();",              // constitute punctuation tokens
       delimiters = blanks + punctuation;   // separate data tokens
// ^^^ Caution! Each punctuation character must have its own section
// ^^^ in Kludge_Tokenizer::Get_Delimiters() !
//     If you add a new punctuator in here only, you get "unknown punctuation
//     character" errors.

extern pthread_t master_thread;




/* ### class Kludge_Path ### */

Kludge_Path::Kludge_Path()
 : leading_slash(false), trailing_slash(false)
{
   elements.clear();
}



Kludge_Path::~Kludge_Path()
{
   for (int i=elements.size()-1;i>=0;i--)
   {
      delete elements[i];
   }
   elements.clear();
}






/* ### class Kludge_CLI_CMD_PARM ### */

Kludge_CLI_CMD_PARM::Kludge_CLI_CMD_PARM(Kludge_CLI_CMD_PARM_Type parm_type)
 : type(parm_type)
{
   content = "";
}



Kludge_CLI_CMD_PARM::~Kludge_CLI_CMD_PARM()
{
   for (size_t i=0; i<contents.size(); i++)
   {
      delete contents[i];
   }
}



int Kludge_CLI_CMD_PARM::add_content(Kludge_CLI_CMD_PARM* cont_to_add)
{
   contents.push_back(cont_to_add);
   return(0);
}






/* ### class Kludge_CLI_Command ### */

Kludge_CLI_Command::~Kludge_CLI_Command()
{
   for (size_t i=0; i<params.size(); i++)
   {
      delete params[i];
   }
}



int Kludge_CLI_Command::add_param(Kludge_CLI_CMD_PARM* param)
{
   params.push_back(param);
   return(0);
}






/* ###   class Kludge_Tokenizer   ### */

Kludge_Tokenizer::Kludge_Tokenizer(iostream* strmin, iostream* strmout = NULL)
    : sin(strmin), sout(strmout), line_pos(0), done(false) 
{
   if (strmout == NULL)
   { 
      sout = sin;
   }
   CLI_input_line = new char[MAX_CLI_INPUT_LINE_LEN+1];
   line_chain.clear();
}



Kludge_Tokenizer::~Kludge_Tokenizer()
{
   delete CLI_input_line;
   line_chain.clear();
}



int Kludge_Tokenizer::Get_Line()
{
 char char_to_examine;

   // As a result of a number of fdopen(int socket) calls at startup
   // (two from the fstream constructors and two from perl IO init), 
   // at this point errno reports an invalid seek(). Fdopen() just does
   // a seek on every fd it gets, even after it's obtained the information
   // that it's really a socket from a fstat() call (found this out using strace).
   // The resulting nonzero errno, if I didn't take care of it, would
   // be detected after the following call to [fstream]->get().
   if (errno != 0) 
   {
      errno = 0;
   }
   
   char_to_examine = sin->get();
   if ((unsigned char)char_to_examine == 255)
   {
      if ((errno == 0) && (! sin->fail() )) // is that a telnet option or an error? (e.g. a signal or TCP connection closed)
      {
         cout
       << " Oh bugger. There's a telnet client talking to us. We gonna have fun.\n"
       << "  Escape character (should be 255): " << dec << char_to_examine << "\n"
       << "  Option: " << dec << sin->get() << "\n"
       << "  Value:  " << dec << sin->get() << "\n" ;
      }
      else
      {
         // either errno or sin->fail() signal something.
         cout << 
           "Error in thread #" << pthread_self() << 
           " while waiting for [fstream]->get().\n";

         if (errno != 0)
         {
            perror(" Errno reports: ");
            cout << " Errno: " << errno << endl;
            if (errno == EINTR) cout << 
               " This thread has been requested to shut down (SIG_QUIT from the master thread).\n";
         }
         if (sin->fail()) 
         {
            // or should we better call sin->clear() ?
            cout <<
               " The input stream signals failure. [fstream]->fail() == "
                 << sin->fail() << endl;
            if (done)
            {
               cout <<
               " This thread has been requested to shut down (SIG_QUIT from the master thread).\n";
            }
            else
            {
               cout <<
               " Socket failed - perhaps the client closed connection without saying \"quit\".\n ";
            }
         }
         cout << flush;
         throw Ex_Shutdown();
      }
   }
   else sin->unget();

   sin->getline(CLI_input_line,MAX_CLI_INPUT_LINE_LEN);

   return(0);
}



Kludge_Token* Kludge_Tokenizer::Get_Token(string* line_parsed, size_t begin_here)
{
 size_t we_got_here = 0;
 bool finished = false, error = false;
 Kludge_Token *result = NULL;

   if (line_parsed->find_first_of(delimiters,begin_here) != begin_here)
   // just a safety check - we expect to start at a non-delimiter
   {
      result = new Kludge_Token;
      result->punctuation = false;
      we_got_here = begin_here;

      if (line_parsed->find_first_of("\"\'",begin_here) != begin_here)
      // not a quoted string
      {
         while (! finished)
         {
            we_got_here = line_parsed->find_first_of(delimiters+"\"\'\\",we_got_here);
            if (we_got_here != string::npos)
            {
               switch ((*line_parsed)[we_got_here])
               {
                  case '\\' :
                  {
                     // note that as long as we chop all trailing blank space,
                     // we can't escape a trailing space onto the end of the last
                     // unquoted token on a line 

                     if (we_got_here != line_parsed->length() - 1)
                     // we've got a correct escape sequence here...
                     // Perhaps we should add another check to verify
                     //  that the escape sequence is valid (one of a list).
                     // The tokenizer does not convert escape sequences - let's leave
                     //  this up to higher-level functions working on top of the tokens.
                     {
                        if (we_got_here + 1 != line_parsed->length() - 1)
                        // the escape sequence is not at the very end of the line
                        {
                           we_got_here += 2;
                        }
                        else
                        // the escape sequence is at the end of line
                        {
                           result->end_pos = string::npos;
                           result->token_data = line_parsed->substr(
                                         begin_here,
                                         (we_got_here + 2 - begin_here) 
                                                                  );
                           finished = true;
                        }
                     }
                     else
                     {
                        *sout << "Currently we don't support lines continued with a backslash\n";
                        finished = true;
                        error = true;
                     }
                  }
                  break;
                  case '\"' :
                  {
                     *sout << "Unexpected quotation mark (\") in the middle of a token.\n";
                     finished = true;
                     error = true;
                  }
                  break;
                  case '\'' :
                  {
                     *sout << "Unexpected quotation mark (\') in the middle of a token.\n";
                     finished = true;
                     error = true;
                  } 
                  break;
                  default :           // any other delimiter
                  {
                     result->end_pos = we_got_here;
                     result->token_data = line_parsed->substr(
                                   begin_here,
                                   (we_got_here - begin_here) 
                                                            );
                     finished = true;
                  } 
               }
            }
            else
            // this is the last token on the line
            {
               result->end_pos = string::npos;
               result->token_data = line_parsed->substr(
                                      begin_here,
                                      (line_parsed->length() - begin_here) 
                                                      );
               finished = true;
            }
         }         
      }
      else
      // yep, we have a quoted string here
      {
         switch ((*line_parsed)[begin_here])
         {
            case '\"' :
            // okay, it's quoted - we should take care of escape sequences
            {
               we_got_here++;        // skip the quotes for now
               while (! finished)
               {
                  we_got_here = line_parsed->find_first_of("\"\\",we_got_here);
                  if (we_got_here != string::npos)
                  {
                     switch ((*line_parsed)[we_got_here])
                     {
                        case '\\' :
                        {
                           if (we_got_here != line_parsed->length() - 1)
                           // we've got a correct escape sequence here...
                           // Perhaps we should add another check to verify
                           //  that the escape sequence is valid (one of a list).
                           // The tokenizer does not convert escape sequences - let's leave
                           //  this up to higher-level functions working on top of the tokens.
                           {
                              if (we_got_here + 1 != line_parsed->length() - 1)
                              // the escape sequence is not at the very end of the line
                              {
                                 we_got_here += 2;
                              }
                              else
                              // the escape sequence is at the end of line
                              {
                                 *sout << "Unmatched quotation mark (\").\n";
                                 finished = true;
                                 error = true;
                              }
                           }
                           else
                           {
                              *sout << "Currently we don't support lines continued with a backslash\n"
                                            << "within a quoted string.\n";
                              finished = true;
                              error = true;
                           }
                        }
                        break;
                        case '\"'  :
                        {
                           if (we_got_here != line_parsed->length() - 1)
                           // the closing quotation mark is not followed by EOL
                           {
                              if (line_parsed->find_first_of(delimiters,we_got_here) == we_got_here + 1)
                              //okay, the quoted string is followed by a delimiter
                              {
                                 result->end_pos = we_got_here + 1;
                                 result->token_data = line_parsed->substr(
                                               begin_here + 1,
                                               (we_got_here - begin_here - 1) 
                                                                        );
                                 we_got_here++;
                                 finished = true;
                              }
                              else
                              //the quoted string is followed by a non-delimiter!
                              {
                                 *sout << "A closing quotation mark (\") is supposed to be followed by a delimiter.\n";
                                 finished = true;
                                 error = true;
                              }
                           }
                           else
                           // the closing quotation mark is followed by EOL
                           {
                              result->end_pos = string::npos;
                              result->token_data = line_parsed->substr(
                                      begin_here+1,
                                      (line_parsed->length() - begin_here - 2) 
                                                                     );
                              finished = true;
                           }
                        }
                        break;
                        default   :
                        {
                           *sout
                           << "Internal tokenizer error - backslash or quotes expected.";
                           error = true;
                        }
                     }
                  }
                  else
                  {
                     *sout << "Unmatched quotation mark (\").\n";
                     finished = true;
                     error = true;
                  }
               }
            }
            break;
            case '\'' :
            // It's enclosed in apostrophes
            //  - we needn't take care of escape sequences.
            {
               we_got_here++;           // skip the quotes for now
               we_got_here = line_parsed->find('\'',we_got_here);
               if (we_got_here != string::npos)
               {
                  if (we_got_here != line_parsed->length() - 1)
                  // the closing quotation mark is not followed by EOL
                  {
                     if (line_parsed->find_first_of(delimiters,we_got_here) == we_got_here + 1)
                     //okay, the quoted string is followed by a delimiter
                     {
                        result->end_pos = we_got_here + 1;
                        result->token_data = line_parsed->substr(
                                      begin_here + 1,
                                      (we_got_here - begin_here - 1) 
                                                               );
                        we_got_here++;
                        finished = true;
                     }
                     else
                     //the quoted string is followed by a non-delimiter!
                     {
                        *sout << "A closing quotation mark (\') is supposed to be followed by a delimiter.\n";
                        finished = true;
                        error = true;
                     }
                  }
                  else
                  // the closing quotation mark is followed by EOL
                  {
                     result->end_pos = string::npos;
                     result->token_data = line_parsed->substr(
                             begin_here+1,
                             (line_parsed->length() - begin_here - 2) 
                                                            );
                     finished = true;
                  }
               }
               else
               {
                  *sout << "Unmatched quotation mark (\').\n";
                  finished = true;
                  error = true;
               }
            }
            break;
            default   :
            {
               *sout
               << "Internal tokenizer error - quotation mark expected ( \" or \' )\n";
               error = true;
            }
         }
      }
      if (! error) return(result);
      else
      {
         delete result;
         result = NULL;
      }
   }
   else
   {
      *sout << "Internal error - In Kludge_CLI_Get_Token: non-delimiter expected.\n";
      result = NULL;
   }
   return(result);
}



size_t Kludge_Tokenizer::Get_Delimiters(string* line_parsed, size_t begin_here)
{
 size_t position,end_pos,result;
 Kludge_Token *new_punct = NULL;

   result = line_parsed->find_first_not_of(delimiters, begin_here);

   if (result != string::npos)
   // there's some more data after this set of delimiters
   {
      end_pos = result;  // actually the ending position + 1
   }
   else
   // no more data tokens until the end of this line
   {
      end_pos = line_parsed->length();
   }

   position = begin_here;
   while ((position < end_pos) && (position != string::npos))
   {
      position = line_parsed->find_first_of(punctuation,position);
      if ((position != string::npos) && (position < end_pos))
      // yes there is some punctuation before the next token
      {
         new_punct = new Kludge_Token;
         new_punct->punctuation = true;
         switch ((*line_parsed)[position])
         // in the future, we may need to implement two-character punctuation
         // signs, such as ">=", "<=", "!=", "<>" etc...
         // That's why the explicit literal assignments in the indiv. sections
         {
            case '=' :
            {
               new_punct->punct_type = '=';
               line_chain.push_back(*new_punct);
            }
            break;
            case '{' :
            {
               new_punct->punct_type = '{';
               line_chain.push_back(*new_punct);
            }
            break;
            case '}' :
            {
               new_punct->punct_type = '}';
               line_chain.push_back(*new_punct);
            }
            break;
            case '(' :
            {
               new_punct->punct_type = '(';
               line_chain.push_back(*new_punct);
            }
            break;
            case ')' :
            {
               new_punct->punct_type = ')';
               line_chain.push_back(*new_punct);
            }
            break;
            case ';' :
            {
               new_punct->punct_type = ';';
               line_chain.push_back(*new_punct);
            }
            break;
            default :
            {
               *sout << "Internal parser error: unknown punctuation character.\n";
            }
            delete new_punct;
            new_punct = NULL;
         }

         position++;
      }
      // else there are no punctuation characters until the next token
   }

   return(result);
}



bool Kludge_Tokenizer::Tokenize_Line()
{
 size_t position = 0; //current_position on the line
 Kludge_Token *token_popped = NULL;

   // line_chain.clear();  // not necessary - see Peek_Token()
   string string_to_parse(CLI_input_line);

   // this is to prepare ground for the tokenizer cycle
   position = Get_Delimiters(&string_to_parse, position);
   if (position != string::npos)
   {
      // the tokenizer cycle
      while (position != string::npos)
      {

         token_popped = Get_Token(&string_to_parse, position);
         if (token_popped != NULL)
         {
            line_chain.push_back(*token_popped);
            position = token_popped->end_pos;
            delete token_popped;
            token_popped = NULL;
         }
         else
         {
            *sout << "Data token null. The tokenizer probably detected bad input.\n";
            line_chain.clear();
            token_popped = NULL;
            position = string::npos;
         }

         if (position != string::npos)
         {
            position = Get_Delimiters(&string_to_parse, position);
         }
      }
      // ^^^ the tokenizer cycle

   }
   // else ignore empty lines


   return(0);
} 



Kludge_Token* Kludge_Tokenizer::Peek_Token()
{
   if (line_pos >= line_chain.size())
   // hmm, we need to fetch another line of data
   {
      line_chain.clear();
      line_pos = 0;
      while (line_chain.empty())
      {
         *sout << "> " << flush;   // send a prompt
         Get_Line();
         cout << "Got: " << CLI_input_line << "<eol>\n";
         Tokenize_Line();
      }
   }
   return(&line_chain[line_pos]);
}



Kludge_Token* Kludge_Tokenizer::Pop_Token()
{
   Kludge_Token* result = Peek_Token();
   line_pos++;
   return(result);
}



// for external callers, mainly from class Kludge_CLI
int Kludge_Tokenizer::Reset()
{
   line_chain.clear();
   line_pos = 0;
   return(0);
}






/* ### class Kludge_CLI ### */

Kludge_CLI::Kludge_CLI(iostream* input_str, iostream* output_str, Kludge_Thread* thread)
 : input_stream(input_str), output_stream(output_str), ind_spaces(""), ind_depth(1), indent(true), pwd(Kludge_Root), thr(thread), CLI_done(false), tokenizer(NULL)
{
   if ((input_stream != NULL) && (output_stream != NULL))
   {
      ind_spaces.reserve(50);
      tokenizer = new Kludge_Tokenizer(input_stream, output_stream);
   }
   else CLI_done = true;
}



Kludge_CLI::~Kludge_CLI()
{
   if (tokenizer != NULL)
   {
      delete tokenizer;
   }
}



int Kludge_CLI::Welcome_Message()
{
   *output_stream 
        << "This is Kludge - an object database/directory server (cough).\n"
        << "If you're duly kind to me, I won't crash on you, I promise.\n"
        << "\n"
        << flush;

   return(0);
}



int Kludge_CLI::Indent()
{
   if (indent)
   {
      ind_spaces.append(ind_depth,' ');
   }
   return(0);
}



int Kludge_CLI::Unindent()
{
   if (indent)
   {
      ind_spaces.erase(ind_spaces.size() - ind_depth - 1, string::npos);
   }
   return(0);
}



// Parses a variable assignment out of tokens.
// Used exclusively by Kludge_CLI::Parse_Obj_Def() 
Kludge_CLI_CMD_PARM* Kludge_CLI::Parse_Var_Def()
{
 Kludge_CLI_CMD_PARM *result = NULL;
 Kludge_Token* tmp_token = NULL;
   
   tmp_token = tokenizer->Pop_Token();
   if ( ! tmp_token->punctuation )
   {
      result = new Kludge_CLI_CMD_PARM(PARM_VARIABLE);
      result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));
      
      tmp_token = tokenizer->Pop_Token();
      if ( tmp_token->punctuation )
      {
         if (tmp_token->punct_type == '=')
         {
            tmp_token = tokenizer->Pop_Token();
            if ( ! tmp_token->punctuation )
            {
               result->add_content(new Kludge_CLI_CMD_PARM(PARM_VALUE,tmp_token->token_data));

               tmp_token = tokenizer->Pop_Token();
               if (tmp_token->punctuation)
               {
                  if (tmp_token->punct_type == ';')
                  {
                     // do nothing - just return result
                  }
                  else
                  {
                     *output_stream <<
                      "Parse_Var_Def(): Unexpected punctuation token.\n"
                      "After specifying the variable name and value, you have to\n"
                      "finish the definition using a semicolon.\n";
                     delete result; result = NULL;
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Var_Def(): Unexpected non-punctuation token.\n"
                   "After specifying the variable name and value, you have to\n"
                   "finish the definition using a semicolon.\n";
                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream << 
                  "A variable definition must contain a string value - at least an empty string\n"
                  " (=\"\"). Punctuation character encountered after equality mark (=): \'"
                  << tmp_token->punct_type << "\'.\n";
               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream
             << "Variable name should be separated from the associated value by an \'=\'\n"
                 " (\"equality mark\") .\n";
            delete result; result = NULL;
         }
      }
      else
      {
         *output_stream
          << "Variable name should be separated from the associated value by an \'=\'\n"
              " (\"equality mark\") .\n";
         delete result; result = NULL;
      }
   }
   else
   {
      *output_stream <<
         "A variable definition can't start with a punctuation character.\n"
         " Parser encountered \'" << tmp_token->punct_type << "\'. Variable name expected.\n";
   }

   return(result);
}



// Parses a Blob body. Circumvents the tokenizer and reads line by line.
// Terminates when an empty line is encountered.
// Used exclusively by Kludge_CLI::Parse_Obj_Def() 
Kludge_CLI_CMD_PARM* Kludge_CLI::Parse_Blob_Def()
{
 bool blob_finished = false;
 Kludge_Token *tmp_token = NULL;
 Kludge_CLI_CMD_PARM* result = NULL;

   tmp_token = tokenizer->Pop_Token(); // the opening brace

   if (tmp_token->punctuation)
   {
      if (tmp_token->punct_type == '{')
      {
         // Some heavy hacks into the tokenizer follow.
         // Discard any remaining tokens on this line (perhaps we'd better check them too?)
         tokenizer->Reset();
         
         // Create a new empty parameter of type VALUE
         result = new Kludge_CLI_CMD_PARM(PARM_BLOB_VALUE);

         cout << "Reading BLOB contents.\n";
         // Read lines until an empty line
         while (! blob_finished)
         {
            tokenizer->Get_Line();
            if ((*(tokenizer->CLI_input_line) == '\n') || (*(tokenizer->CLI_input_line) == 0))
            {
               blob_finished = true;
            }
            else
            {
               cout << ".";
               result->content += tokenizer->CLI_input_line;
               result->content += '\n';
            }
         }
         cout << "\nBlob contents read.\n";

         tmp_token = tokenizer->Pop_Token(); // the closing brace
         if (tmp_token->punctuation)
         {
            if (tmp_token->punct_type == '}')
            {
               tmp_token = tokenizer->Pop_Token(); // the closing semicolon
               if (tmp_token->punctuation)
               {
                  if (tmp_token->punct_type == ';')
                  {
                     // do nothing - just return result
                  }
                  else
                  {
                     *output_stream <<
                      "Parse_Blob_Def(): Unexpected punctuation token.\n"
                      " After the blob value is entered, after the closing brace,\n"
                      " a semicolon is expected.\n"
                      " Another punctuation token encountered instead: \""
                      << tmp_token->punct_type << "\"\n";
                     delete result; result = NULL;
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Blob_Def(): Unexpected non-punctuation token.\n"
                   " After the blob value is entered, after the closing brace,\n"
                   " a semicolon is expected.\n"
                   " A generic text token encountered instead.\n";
                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream <<
                "Parse_Blob_Def(): Unexpected punctuation token.\n"
                "After blob content, a closing brace ('}') should follow.\n"
                " Another punctuation token encountered instead: \""
                << tmp_token->punct_type << "\"\n";
               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream <<
             "Parse_Blob_Def(): Unexpected non-punctuation token.\n"
             "After blob content, a closing brace ('}') should follow.\n"
             "A generic text token encountered instead.\n";
            delete result; result = NULL;
         }
      }
      else if (tmp_token->punct_type == ';')
      {
         result = new Kludge_CLI_CMD_PARM(PARM_BLOB_VALUE);
         // an empty blob - return an empty result
      }
      else
      {
         *output_stream <<
          "BLOB content expected - starting with an opening brace ('{').\n"
          " Alternatively, an empty BLOB can be entered using just a semicolon (';').\n"
          " Another punctuation token encountered instead: \""
          << tmp_token->punct_type << "\"\n";
      }
   }
   else
   {
      *output_stream << 
       "BLOB content expected - starting with an opening brace ('{').\n"
       " Alternatively, an empty BLOB can be entered using just a semicolon (';').\n"
       " A generic text token encountered instead.\n";
   }

   return(result);
}



// Parses a structured kludge object definition out of tokens.
// Used exclusively by Kludge_CLI::Prepare_Command()
Kludge_CLI_CMD_PARM* Kludge_CLI::Parse_Obj_Def()
{
#define IS_OBJ(OBJ_STRING) tmp_token->token_data == OBJ_STRING
// defines case-sensitive string comparison
 Kludge_CLI_CMD_PARM *tmp_parm = NULL, *result = NULL;
 Kludge_Token *tmp_token = NULL;
 
   tmp_token = tokenizer->Pop_Token();
   if ( ! tmp_token->punctuation )
   {
      if ( IS_OBJ("INSTANCE_PROPERTY") )
      {
         tmp_token = tokenizer->Pop_Token(); // name
         if ( ! tmp_token->punctuation )
         {
            result = new Kludge_CLI_CMD_PARM(PARM_INSTPROP);
            result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token(); // class
            if ( ! tmp_token->punctuation )
            {
               result->add_content(new Kludge_CLI_CMD_PARM(PARM_OBJ_CLASS,tmp_token->token_data));

               tmp_token = tokenizer->Pop_Token(); // class relationship
               if ( ! tmp_token->punctuation )
               {
                  result->add_content(new Kludge_CLI_CMD_PARM(PARM_OBJ_CLSREL,tmp_token->token_data));

                  tmp_token = tokenizer->Pop_Token(); // either a value or a semicolon
                  if (! tmp_token->punctuation)
                  {  // it's a value
                     result->add_content(new Kludge_CLI_CMD_PARM(PARM_VALUE,tmp_token->token_data));

                     tmp_token = tokenizer->Pop_Token(); // the closing semicolon
                     if (tmp_token->punctuation)
                     {
                        if (tmp_token->punct_type == ';')
                        {
                           // do nothing - just return result
                        }
                        else
                        {
                           *output_stream <<
                            "Parse_Obj_Def(): Unexpected punctuation token.\n"
                            "After \'INSTANCE_PROPERTY <name> <class> <cls_rel> <value>\',\n"
                            "you must finish the object definition using a semicolon.\n"
                            "This is the correct syntax:\n"
                            " INSTANCE_PROPERTY <name> <class> <cls_rel> <value>;\n";
                           delete result; result = NULL;
                        }
                     }
                     else
                     {
                        *output_stream <<
                         "Parse_Obj_Def(): Unexpected non-punctuation token.\n"
                         "After \'INSTANCE_PROPERTY <name> <class> <cls_rel> <value>\',\n"
                         "you must finish the object definition using a semicolon.\n"
                         "This is the correct syntax:\n"
                         " INSTANCE_PROPERTY <name> <class> <cls_rel> <value>;\n";
                        delete result; result = NULL;
                     }
                  }
                  else
                  {
                     if (tmp_token->punct_type == ';')
                     {  // it's a semicolon
                        // insert empty value
                        result->add_content(new Kludge_CLI_CMD_PARM(PARM_VALUE));
                     }
                     else
                     {
                        *output_stream <<
                         "Parse_Obj_Def(): Unexpected punctuation character.\n"
                         "After \'INSTANCE_PROPERTY <name> <class> <cls_rel>\',\n"
                         "you can either finish the object definition using a semicolon,\n"
                         "or assign a value. This is the correct syntax:\n"
                         " INSTANCE_PROPERTY <name> <class> <cls_rel> [value] ;\n";

                        delete result; result = NULL;
                     }
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Obj_Def(): Unexpected punctuation token.\n"
                   "After specifying \'INSTANCE_PROPERTY <name> <class>\' you need to continue\n"
                   "further with the object declaration. The next parameter expected is <cls_rel>\n"
                   "- as per the basic syntax template:\n"
                   " INSTANCE_PROPERTY <name> <class> <cls_rel> [value];\n";

                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream <<
                "Parse_Obj_Def(): Unexpected punctuation token.\n"
                "After specifying \'INSTANCE_PROPERTY <name>\' you need to continue further\n"
                "with the object declaration. The next parameter expected is <class>\n"
                "- as per the basic syntax template:\n"
                " INSTANCE_PROPERTY <name> <class> <cls_rel> [value];\n";

               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream <<
             "Cannot insert an empty kludge instance property.\n"
             "After specifying \'INSTANCE_PROPERTY\' you need to continue further\n"
             "with the object declaration. The next parameter expected is <name>\n"
             "- as per the basic syntax template:\n"
             " INSTANCE_PROPERTY <name> <class> <cls_rel> [value];\n";
         }
      }
      if ( IS_OBJ("INSTANCE_RELATIONSHIP") )
      {
         tmp_token = tokenizer->Pop_Token();
         
         if (tmp_token->punctuation)
         {
            if (tmp_token->punct_type == '{') // instance relationship attributes follow
            {
               result = new Kludge_CLI_CMD_PARM(PARM_INSTREL);
               
               tmp_token = tokenizer->Peek_Token();
               while (
                        ! (
                         tmp_token->punctuation
                         && (tmp_token->punct_type == '}')
                        )
                        && (result != NULL)
                     )
               {
                  tmp_parm = Parse_Var_Def(); // will also filter improper punctuation
                  if (tmp_parm != NULL)
                  {
                     result->add_content(tmp_parm);
                     tmp_token = tokenizer->Peek_Token();
                     if ((tmp_token->punctuation) && (tmp_token->punct_type != '}'))
                     {
                        *output_stream
                         << "Unexpected punctuation character. Closing brace (\"}\") expected.\n";
                        delete result; result = NULL;
                     }
                  }
                  else
                  // basic syntax error
                  {
                     delete result; result = NULL;
                  }
               }
               // OK - variables read.
               
               // Of course there's quite some call for various semantic checks - but let's leave
               // these up to Kludge_Instance::Insert() - it has to read the variable values in anyway
               // and is perhaps the correct place to check for semantic sanity.
               // Here we are in a low-level syntax parser.

               if (result != NULL)   // finish reading the INSTANCE_RELATIONSHIP
               {
                  tokenizer->Pop_Token(); // chop the closing brace - now that we know it's there

                  tmp_token = tokenizer->Pop_Token();
                  if (tmp_token->punctuation)
                  {
                     if (tmp_token->punct_type == ';')
                     {
                        // OK. Do nothing - just return result.
                     }
                     else
                     {
                        *output_stream <<
                        "The closing brace is supposed to be followed by a semicolon.\n";
                        delete result; result = NULL;
                     }
                  }
                  else
                  {
                     *output_stream <<
                     "The closing brace is supposed to be followed by a semicolon.\n";
                     delete result; result = NULL;
                  }
               }
               // done with the contents.
            }
            else if (tmp_token->punct_type == ';') // a closing semicolon - no relationship contents
            {
               *output_stream <<
                "Parse_Obj_Def(): unexpected closing semicolon after \"INSERT INSTANCE_RELATIONSHIP\".\n"
                " An empty INSTANCE_RELATIONSHIP doesn't make sense - there are mandatory attributes.\n"
                " A definition of attributes should follow, starting with an opening brace (\'{\').\n";
            }
            else
            {
               *output_stream <<
                "Parse_Obj_Def(): unexpected punctuation after \"INSTANCE_RELATIONSHIP\": \'"
                << tmp_token->punct_type << "\'.\n" 
                " A definition of attributes should follow, starting with an opening\n"
                " brace (\'{\').\n";
            }
         }
         else
         {
            *output_stream <<
             "Parse_Obj_Def(): Unexpected non-punctuation token.\n"
             "After declaring the insertion of an \'INSTANCE_RELATIONSHIP\',\n"
             "you have to define its mandatory attributes, enclosed in curly braces.\n";
         }
      }
      else if ( IS_OBJ("INSTANCE") )
      {
         tmp_token = tokenizer->Pop_Token(); // name
         if ( ! tmp_token->punctuation )
         {
            result = new Kludge_CLI_CMD_PARM(PARM_INSTANCE);
            result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token(); // class
            if ( ! tmp_token->punctuation )
            {
               result->add_content(new Kludge_CLI_CMD_PARM(PARM_OBJ_CLASS,tmp_token->token_data));

               tmp_token = tokenizer->Pop_Token(); // class relationship
               if ( ! tmp_token->punctuation )
               {
                  result->add_content(new Kludge_CLI_CMD_PARM(PARM_OBJ_CLSREL,tmp_token->token_data));

                  tmp_token = tokenizer->Peek_Token(); // Either an opening brace, 
                                                       // or an opening parenthesis,
                                                       // or a semicolon.
                                                       // Parse_Arg_List expects to see the parenthesis.
                  if (tmp_token->punctuation)
                  {
                     // continue here - insert the parsing of an argument list
                    
                     if (tmp_token->punct_type == '(')
                     {
                        tmp_parm = Parse_Arg_List(/*check semicolon?*/false);
                        if (tmp_parm == NULL)
                        {
                           delete result; result = NULL;
                        }
                        else
                        {
                           result->add_content(tmp_parm);
                           tmp_token = tokenizer->Peek_Token();
                        }
                     }
                    
                     if (result != NULL)
                     {
                        if (tmp_token->punct_type == ';') // the closing semicolon - no nested contents
                        {
                           tokenizer->Pop_Token(); // Pop the semicolon - now that we know it's there
                           // OK - return result.
                        }
                        else if (tmp_token->punct_type == '{') // nested instance contents follow
                        {
                           tokenizer->Pop_Token(); // Pop the opening brace - it's no use anymore

                           tmp_token = tokenizer->Peek_Token();
                           while (
                                    ! (
                                     tmp_token->punctuation
                                     && (tmp_token->punct_type == '}')
                                    )
                                    && (result != NULL)
                                 )
                           {
                              tmp_parm = Parse_Obj_Def(); // will also filter improper punctuation
                              if (tmp_parm != NULL)
                              {
                                 result->add_content(tmp_parm);
                                 tmp_token = tokenizer->Peek_Token();
                                 if ((tmp_token->punctuation) && (tmp_token->punct_type != '}'))
                                 {
                                    *output_stream
                                     << "Unexpected punctuation character. Closing brace (\"}\") expected.\n";
                                    delete result; result = NULL;
                                 }
                              }
                              else
                              {
                                 // nested Parse_Obj_Def() detected an error and reported it.
                                 delete result; result = NULL;
                              }
                           }

                           if (result != NULL)
                           {
                              tokenizer->Pop_Token(); // chop the closing brace - now that we know it's there

                              tmp_token = tokenizer->Pop_Token();
                              if (tmp_token->punctuation)
                              {
                                 if (tmp_token->punct_type == ';')
                                 {
                                    // OK. Do nothing - just return result.
                                 }
                                 else
                                 {
                                    *output_stream <<
                                    "The closing brace is supposed to be followed by a semicolon.\n";
                                    delete result; result = NULL;
                                 }
                              }
                              else
                              {
                                 *output_stream <<
                                 "The closing brace is supposed to be followed by a semicolon.\n";
                                 delete result; result = NULL;
                              }
                           }
                           // else no processing to be done - error already reported, just quit
                        }
                        else
                        {
                           *output_stream <<
                            "Parse_Obj_Def(): unexpected punctuation after \""
                            << result->contents[0]->content <<
                            "\".\n Permitted punctuation characters are \';\', \'{\', \'(\'\n"; 

                           delete result; result = NULL;
                        }
                     }
                  }
                  else
                  {
                     *output_stream <<
                      "Parse_Obj_Def(): Unexpected non-punctuation token.\n"
                      "After \'INSTANCE <name> <class> <cls_rel>\',\n"
                      "you can either finish the object definition using a semicolon, or\n"
                      "you can optionally add CTOR arguments in parentheses, or"
                      "you can proceed to define nested objects using curly braces:\n"
                      " INSTANCE <name> <class> <cls_rel>;\n"
                      " INSTANCE <name> <class> <cls_rel> ( arg_name1=value1; arg_name2=value2; ... );\n"
                      " INSTANCE <name> <class> <cls_rel> { <content>; [<content>;]* };\n"
                      "You can supply a list of CTOR arguments followed by nested contents.\n";

                     delete result; result = NULL;
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Obj_Def(): Unexpected punctuation token.\n"
                   "After specifying \'INSTANCE <name> <class>\' you need to continue further\n"
                   "with the object declaration. The next parameter expected is <cls_rel>\n"
                   "- as per the basic syntax templates:\n"
                   " INSTANCE <name> <class> <cls_rel>;\n"
                   " INSTANCE <name> <class> <cls_rel> ( arg_name1=value1; arg_name2=value2; ... );\n"
                   " INSTANCE <name> <class> <cls_rel> { <content>; [<content>;]* };\n"
                   "You can supply a list of CTOR arguments followed by nested contents.\n";

                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream <<
                "Parse_Obj_Def(): Unexpected punctuation token.\n"
                "After specifying \'INSTANCE <name>\' you need to continue further\n"
                "with the object declaration. The next parameter expected is <class>\n"
                "- as per the basic syntax templates:\n"
                " INSTANCE <name> <class> <cls_rel>;\n"
                " INSTANCE <name> <class> <cls_rel> ( arg_name1=value1; arg_name2=value2; ... );\n"
                " INSTANCE <name> <class> <cls_rel> { <content>; [<content>;]* };\n"
                "You can supply a list of CTOR arguments followed by nested contents.\n";

               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream <<
             "Cannot insert an empty kludge instance.\n"
             "After specifying \'INSTANCE\' you need to continue further\n"
             "with the object declaration. The next parameter expected is <name>\n"
             "- as per the basic syntax templates:\n"
             " INSTANCE <name> <class> <cls_rel>;\n"
             " INSTANCE <name> <class> <cls_rel> ( arg_name1=value1; arg_name2=value2; ... );\n"
             " INSTANCE <name> <class> <cls_rel> { <content>; [<content>;]* };\n"
             "You can supply a list of CTOR arguments followed by nested contents.\n";
         }
      }
      else if ( (IS_OBJ("INST_BLOB")) || (IS_OBJ("INST_TXT_BLOB")) )
      {
         result = new Kludge_CLI_CMD_PARM(PARM_INST_BLOB);
         result->content = tmp_token->token_data;  // store BLOB subtype - yet another kludge :)

         tmp_token = tokenizer->Pop_Token(); // name
         if ( ! tmp_token->punctuation )
         {
            result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token(); // class
            if ( ! tmp_token->punctuation )
            {
               result->add_content(new Kludge_CLI_CMD_PARM(PARM_OBJ_CLASS,tmp_token->token_data));

               tmp_token = tokenizer->Pop_Token(); // class relationship
               if ( ! tmp_token->punctuation )
               {
                  result->add_content(new Kludge_CLI_CMD_PARM(PARM_OBJ_CLSREL,tmp_token->token_data));

                  tmp_parm = Parse_Blob_Def();
                  if (tmp_parm != NULL) result->add_content(tmp_parm);
                  else
                  {
                     *output_stream <<
                      "Error inserting an instance BLOB thing. Failed to parse BLOB data definition.\n";

                     delete result; result = NULL;
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Obj_Def(): Unexpected punctuation token.\n"
                   "Cannot insert an instance BLOB with no class relationship specified.\n"
                   " After `INSERT <some_type_of_inst_blob> <blob_name> <class>',\n"
                   " the relevant class relationship should follow - the one that governs\n"
                   " this particular ownership between the instance and the BLOB.\n"
                   " A punctuation character encountered instead: \""
                   << tmp_token->punct_type << "\"\n";
                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream <<
                "Parse_Obj_Def(): Unexpected punctuation token.\n"
                "Cannot insert an instance BLOB with no class.\n"
                " After `INSERT <some_type_of_inst_blob> <blob_name>', the class of the item\n"
                " has to follow. A punctuation character encountered instead: \""
                << tmp_token->punct_type << "\"\n";
               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream <<
             "Cannot insert a BLOB with no name.\n"
             " After `INSERT <some_type_of_inst_blob>', the name of the item has to follow.\n"
             " A punctuation character encountered instead: \""
             << tmp_token->punct_type << "\"\n";
            delete result; result = NULL;
         }
      }
      else if ( (IS_OBJ("CLS_BLOB")) || (IS_OBJ("CLS_TXT_BLOB")) || (IS_OBJ("METHOD")) )
      {
         result = new Kludge_CLI_CMD_PARM(PARM_CLS_BLOB);
         result->content = tmp_token->token_data;  // store BLOB subtype - yet another kludge :)
         
         tmp_token = tokenizer->Pop_Token(); // name
         if ( ! tmp_token->punctuation )
         {
            result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

            tmp_parm = Parse_Blob_Def();
            if (tmp_parm != NULL) result->add_content(tmp_parm);
            else
            {
               *output_stream <<
                "Error inserting a class BLOB thing. Failed to parse BLOB data definition.\n";

               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream <<
                "Cannot insert a BLOB with no name.\n"
                " After `INSERT <some_type_of_cls_blob>', the name of the item has to follow.\n"
                " A punctuation character encountered instead: \""
                << tmp_token->punct_type << "\"\n";
            delete result; result = NULL;
         }
      }
      else if ( IS_OBJ("CLASS_RELATIONSHIP") )
      {
         tmp_token = tokenizer->Pop_Token();
         
         if (tmp_token->punctuation)
         {
            if (tmp_token->punct_type == '{') // class relationship attributes follow
            {
               result = new Kludge_CLI_CMD_PARM(PARM_CLSREL);
               
               tmp_token = tokenizer->Peek_Token();
               while (
                        ! (
                         tmp_token->punctuation
                         && (tmp_token->punct_type == '}')
                        )
                        && (result != NULL)
                     )
               {
                  tmp_parm = Parse_Var_Def(); // will also filter improper punctuation
                  if (tmp_parm != NULL)
                  {
                     result->add_content(tmp_parm);
                     tmp_token = tokenizer->Peek_Token();
                     if ((tmp_token->punctuation) && (tmp_token->punct_type != '}'))
                     {
                        *output_stream
                         << "Unexpected punctuation character. Closing brace (\"}\") expected.\n";
                        delete result; result = NULL;
                     }
                  }
                  else
                  // basic syntax error
                  {
                     delete result; result = NULL;
                  }
               }
               // OK - variables read.
               
               // Of course there's quite some call for various semantic checks - but let's leave
               // these up to Kludge_Class::Insert() - it has to read the variable values in anyway
               // and is perhaps the correct place to check for semantic sanity.
               // Here we are in a low-level syntax parser.

               if (result != NULL)   // finish reading the CLASS_RELATIONSHIP
               {

                  tokenizer->Pop_Token(); // chop the closing brace - now that we know it's there

                  tmp_token = tokenizer->Pop_Token();
                  if (tmp_token->punctuation)
                  {
                     if (tmp_token->punct_type == ';')
                     {
                        // OK. Do nothing - just return result.
                     }
                     else
                     {
                        *output_stream <<
                        "The closing brace is supposed to be followed by a semicolon.\n";
                        delete result; result = NULL;
                     }
                  }
                  else
                  {
                     *output_stream <<
                     "The closing brace is supposed to be followed by a semicolon.\n";
                     delete result; result = NULL;
                  }
               }
               // done with the contents.
            }
            else if (tmp_token->punct_type == ';') // a closing semicolon - no relationship contents
            {
               *output_stream <<
                "Parse_Obj_Def(): unexpected closing semicolon after \"INSERT CLASS_RELATIONSHIP\".\n"
                " Empty CLASS_RELATIONSHIP doesn't make sense - there are mandatory attributes.\n"
                " Definition of mandatory attributes should follow, starting with an opening\n"
                " brace (\'{\').\n";
            }
            else
            {
               *output_stream <<
                "Parse_Obj_Def(): unexpected punctuation after \"CLASS_RELATIONSHIP\": \'"
                << tmp_token->punct_type << "\'.\n" 
                " Definition of mandatory attributes should follow, starting with an opening\n"
                " brace (\'{\').\n";
            }
         }
         else
         {
            *output_stream <<
             "Parse_Obj_Def(): Unexpected non-punctuation token.\n"
             "After declaring the insertion of \'CLASS_RELATIONSHIP\',\n"
             "you have to define its mandatory attributes, enclosed in curly braces.\n";
         }
      }
      else if ( IS_OBJ("CLASS_PROPERTY") )
      {
         tmp_token = tokenizer->Pop_Token(); // name
         if ( ! tmp_token->punctuation )
         {
            result = new Kludge_CLI_CMD_PARM(PARM_CLSPROP);
            result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token(); // value

            if (! tmp_token->punctuation)
            {
               result->add_content(new Kludge_CLI_CMD_PARM(PARM_VALUE,tmp_token->token_data));
               
               tmp_token = tokenizer->Pop_Token(); // the closing semicolon
               if (tmp_token->punctuation)
               {
                  if (tmp_token->punct_type == ';')
                  {
                     // do nothing - just return result
                  }
                  else
                  {
                     *output_stream <<
                      "Parse_Obj_Def(): Unexpected punctuation token.\n"
                      "After specifying the Class Property name and value, you have to\n"
                      "finish the object definition using a semicolon.\n";
                     delete result; result = NULL;
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Obj_Def(): Unexpected non-punctuation token.\n"
                   "After specifying the Class Property name and value, you have to\n"
                   "finish the object definition using a semicolon.\n";
                  delete result; result = NULL;
               }
            }
            else
            {
               if (tmp_token->punct_type == ';')
               {
                  // result->add_content(new Kludge_CLI_CMD_PARM(PARM_VALUE,""));
                  // just return result
               }
               else
               {
                  *output_stream
                   << "Unexpected punctuation character encountered after class property name.\n";
                  delete result; result = NULL;
               }
            }
         }
         else
         {
            *output_stream << 
                "Cannot insert a kludge class property with no name and value.\n"
                " After `INSERT CLASS_PROPERTY', a name of the property has to follow.\n"
                " A punctuation character encountered instead: \""
                << tmp_token->punct_type << "\"\n";
         }
      }
      else if ( IS_OBJ("CLASS") )
      {
         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            result = new Kludge_CLI_CMD_PARM(PARM_CLASS);
            result->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token();
            if (tmp_token->punctuation)
            {
               if (tmp_token->punct_type == ';') // the closing semicolon - no class contents
               {
                  // OK - nothing to do - just return result.
               }
               else if (tmp_token->punct_type == '{') // class contents follow
               {
                  tmp_token = tokenizer->Peek_Token();
                  while (
                           ! (
                            tmp_token->punctuation
                            && (tmp_token->punct_type == '}')
                           )
                           && (result != NULL)
                        )
                  {
                     tmp_parm = Parse_Obj_Def(); // will also filter improper punctuation
                     if (tmp_parm != NULL)
                     {
                        result->add_content(tmp_parm);
                        tmp_token = tokenizer->Peek_Token();
                        if ((tmp_token->punctuation) && (tmp_token->punct_type != '}'))
                        {
                           *output_stream
                            << "Unexpected punctuation character. Closing brace (\"}\") expected.\n";
                           delete result; result = NULL;
                        }
                     }
                     else
                     {
                        delete result; result = NULL;
                     }
                  }

                  if (result != NULL)
                  {
                     tokenizer->Pop_Token(); // chop the closing brace - now that we know it's there

                     tmp_token = tokenizer->Pop_Token();
                     if (tmp_token->punctuation)
                     {
                        if (tmp_token->punct_type == ';')
                        {
                           // OK. Do nothing - just return result.
                        }
                        else
                        {
                           *output_stream <<
                           "The closing brace is supposed to be followed by a semicolon.\n";
                           delete result; result = NULL;
                        }
                     }
                     else
                     {
                        *output_stream <<
                        "The closing brace is supposed to be followed by a semicolon.\n";
                        delete result; result = NULL;
                     }
                  }
               }
               else
               {
                  *output_stream <<
                   "Parse_Obj_Def(): unexpected punctuation after \""
                   << result->contents[0]->content <<
                   "\".\n Permitted punctuation characters are \';\' and \'{\'\n"; 

                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream <<
                "Parse_Obj_Def(): Unexpected non-punctuation token.\n"
                "After \'CLASS <name>\' you can either finish the object definition\n"
                "using a semicolon, or you can proceed to define nested objects using\n"
                "curly braces.\n";

               delete result; result = NULL;
            }
         }
         else
         {
            *output_stream
             << "Cannot insert an empty kludge class.\n";
         }
      }
      else
      {
         *output_stream
          << "Unknown Kludge Object type: \"" << tmp_token->token_data << "\".\n"
             "Only the following types are permitted:\n"
             " CLASS, CLASS_PROPERTY, CLASS_RELATIONSHIP,\n"
             " INSTANCE, INSTANCE_PROPERTY, INSTANCE_RELATIONSHIP\n";
      }
   }
   else
   {
      *output_stream
       << "An object definition can't start with a punctuation character.\n";
   }

   return(result);
#undef IS_OBJ(OBJ_STRING)
}



// Parses a method's argument list out of tokens.
// Used by Kludge_CLI::Prepare_Command() and Kludge_CLI::Parse_Obj_Def()
Kludge_CLI_CMD_PARM* Kludge_CLI::Parse_Arg_List(bool check_semicolon = true)
{
 Kludge_CLI_CMD_PARM *tmp_parm = NULL, *result = NULL;
 Kludge_Token *tmp_token = NULL;

   tmp_token = tokenizer->Pop_Token();
   
   if (tmp_token->punctuation)
   {
      if (tmp_token->punct_type == '(') // variable definitions follow
      {
         result = new Kludge_CLI_CMD_PARM(PARM_ARGLIST);

         tmp_token = tokenizer->Peek_Token();
         while (
                  ! (
                   tmp_token->punctuation
                   && (tmp_token->punct_type == ')')
                  )
                  && (result != NULL)
               )
         {
            tmp_parm = Parse_Var_Def(); // will also filter improper punctuation
            if (tmp_parm != NULL)
            {
               result->add_content(tmp_parm);
               tmp_token = tokenizer->Peek_Token();
               if ((tmp_token->punctuation) && (tmp_token->punct_type != ')'))
               {
                  *output_stream
                   << "Unexpected punctuation character. Closing parenthesis \")\" expected.\n";
                  delete result; result = NULL;
               }
            }
            else
            // basic syntax error
            {
               delete result; result = NULL;
            }
         }
         // OK - variables read.
         
         if (result != NULL)   // finish reading the arguments list
         {
            tokenizer->Pop_Token(); // chop the closing parenthesis - now that we know it's there

            if (check_semicolon)
            {
               tmp_token = tokenizer->Pop_Token();
               if (tmp_token->punctuation)
               {
                  if (tmp_token->punct_type == ';')
                  {
                     // OK. Do nothing - just return result.
                  }
                  else
                  {
                     *output_stream <<
                     "The closing parenthesis of an argument list\n"
                     " is supposed to be followed by a semicolon.\n";
                     delete result; result = NULL;
                  }
               }
               else
               {
                  *output_stream <<
                  "The closing parenthesis of an argument list\n"
                  " is supposed to be followed by a semicolon.\n";
                  delete result; result = NULL;
               }
            }
            // else don't pop() the next token - the calling function will take care of the
            // closing semicolon.
            // When Parse_Arg_List() is used to pass arguments to an instance's CTOR, there
            // can be further nested contents on the command line.
         }
         // done with the arguments list.
      }
      else if (tmp_token->punct_type == ';') // a closing semicolon - no variables defined
      {
         result = new Kludge_CLI_CMD_PARM(PARM_ARGLIST); // return an empty set
         // this is to distinguish an empty result from an error (NULL result).
      }
      else
      {
         *output_stream <<
          "Parse_Arg_List(): unexpected punctuation after `RUN <some_method>\': \'"
          << tmp_token->punct_type << "\'.\n" 
          " A definition of an argument list (a set of variables) should follow,\n"
          " starting with an opening parenthesis \'(\'.\n"
          " Or just a semicolon for no arguments.";
      }
   }
   else
   {
      *output_stream <<
       "Parse_Arg_List(): Unexpected non-punctuation token.\n"
       "After the RUN command and the path to the method to run,\n"
       "you can optinally specify a list of method arguments, like this:\n"
       " RUN <some_method> ( some_var_name=some_value;\n"
       "                     another_var_name=another_value );\n"
       "Or just finish with a semicolon:\n"
       " RUN <some_method>;\n";
   }

   return(result);
}



// Does some basic syntax parsing and stores the data in a tree-style structure
// for further processing by Execute_Command().
// Walks through the stream of tokens and barks at straight syntax nonsense.
Kludge_CLI_Command* Kludge_CLI::Prepare_Command()
{
#define IS_CMD(CMD_STRING) Bruce_Eckel_stringCmpi(tmp_token->token_data, CMD_STRING) == 0
// defines case-INsensitive string comparison
 Kludge_CLI_Command* result = NULL;
 Kludge_CLI_CMD_PARM* tmp_parm = NULL;
 Kludge_Token *tmp_token = NULL;
 string tmp_string;

   tmp_token = tokenizer->Pop_Token();
   if ( ! tmp_token->punctuation )
   {
      // hook up the basic commands in here

      if ( IS_CMD("GET") )
      // 
      // synopsis:
      // GET <kludge path> ;
      //
      {
         result = new Kludge_CLI_Command(CMD_GET);

         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            tmp_parm = new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data);
            result->add_param(tmp_parm);

            tmp_token = tokenizer->Pop_Token();
            if (! ((tmp_token->punctuation) && (tmp_token->punct_type == ';')))
            {
               delete result;
               result = NULL;
 
               *output_stream
                << "After \"GET\" and the necessary path specifier,\n"
                << "one would expect a semicolon (;) .\n";
            }
            // else everything OK, return the GET command with one parameter
            // of type "path"
         }
         else
         {
            delete result; result = NULL;
 
            *output_stream
             << "After \"GET\", one would expect a path to the object\n"
             << "to be retrieved - not a punctuation character.\n";
         }
      }
      else if ( IS_CMD("INSERT") )
      // 
      // synopsis:
      // INSERT <kludge path : where> <kludge object : what> ;
      //
      {
         result = new Kludge_CLI_Command(CMD_INSERT);
         
         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            tmp_parm = new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data);
            result->add_param(tmp_parm);    // where

            tmp_token = tokenizer->Peek_Token();
            if ( ! tmp_token->punctuation )
            {
               tmp_parm = Parse_Obj_Def();  // what

               if (tmp_parm == NULL)
               {
                  delete result;
                  result = NULL;
    
                  *output_stream
                   << "Error parsing an object definition within an INSERT command.\n";
               }
               else
               {
                  result->add_param(tmp_parm);
                  // everything OK, return the INSERT command with two
                  // parameters as per synopsis.
                  // Don't check for the trailing semicolon - Parse_Obj_Def
                  // does that and pops it.
               }
            }
            else
            {
               delete result;
               result = NULL;
    
               *output_stream
                << "After \"INSERT\" and a path specifying where to insert,\n"
                   " one would expect a kludge object definition.\n"
                   " A punctuation character certainly wouldn't do.\n";
            }
         }
         else
         {
            delete result;
            result = NULL;
 
            *output_stream
             << "After \"INSERT\", one would expect a path specifying\n"
                " where to insert the following object. A punctuation character\n"
                " certainly wouldn't do.\n";
         }
      }
      else if ( IS_CMD("SET") )
      // 
      // synopsis:
      // SET <kludge path : property> <string : value> ;
      // SET <kludge path : blob thingy> <blob contents definition : value> ;
      // SET <kludge path : relationship> <string : attribute name>=<string : value> ;
      //
      {
         result = new Kludge_CLI_Command(CMD_SET);
         
         tmp_token = tokenizer->Pop_Token(); // what to set
         if ( ! tmp_token->punctuation )
         {
            result->add_param(new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data));

            // As Parse_Blob_Def() expects to pop the opening brace of a BLOB
            // definition, at this point we may only take a peep at the next token
            tmp_token = tokenizer->Peek_Token(); // value - a plain string or a BLOB's opening brace

            if ( ! tmp_token->punctuation ) // a plain string
            {
               tokenizer->Pop_Token();     // now we can safely pop it off the que
               // We have yet to find out what version of SET we deal with
               // (property or relationship).
               // We should not use Kludge_CLI_Path2ptr()->Type as a hint.
               // Therefore we need temporary storage for the token just read.
               tmp_string = tmp_token->token_data;

               tmp_token = tokenizer->Pop_Token(); // ";" or "="
               if ( tmp_token->punctuation )
               {
                  if ( tmp_token->punct_type == ';')  // property
                  {
                     // SET <property> <value>;
                     result->add_param(new Kludge_CLI_CMD_PARM(PARM_VALUE,tmp_string));
                     // everything all right, command valid - return result.
                  }
                  else if ( tmp_token->punct_type == '=') // relationship
                  {
                     // SET <relationship> <attribute name>=<value>;
                     result->add_param(new Kludge_CLI_CMD_PARM(PARM_VARIABLE));
                     result->params[1]->add_content(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_string));

                     tmp_token = tokenizer->Pop_Token(); // value
                     if ( ! tmp_token->punctuation )
                     {
                        result->params[1]->add_content(
                           new Kludge_CLI_CMD_PARM(PARM_VALUE,tmp_token->token_data)
                        );

                        tmp_token = tokenizer->Pop_Token(); // closing semicolon
                        if ( tmp_token->punctuation )
                        {
                           if ( tmp_token->punct_type == ';')
                           {
                              // everything all right, command valid - return result.
                           }
                           else
                           {
                              delete result;
                              result = NULL;
                   
                              *output_stream <<
                                 "After \"SET\", a path to the relationship to set and\n"
                                 " the variable definition (<NAME>=<VALUE>), a closing semicolon\n"
                                 " should follow. The parser encountered a different punctuation character:\n"
                                 " \"" << tmp_token->punct_type << "\" .\n";
                           }
                        }
                        else
                        {
                           delete result;
                           result = NULL;
                
                           *output_stream << 
                              "After \"SET\", a path to the relationship to set and\n"
                              " the variable definition (<NAME>=<VALUE>), a closing semicolon\n"
                              " should follow. Instead, the parser encountered further string data.\n";
                        }
                     }
                     else
                     {
                        delete result;
                        result = NULL;
             
                        *output_stream << 
                           "After \"SET\", a path to the relationship to set,\n"
                           " the attribute name and the equal mark (\"=\"), an attribute value\n"
                           " should follow. Punctuation character encountered instead.\n";
                     }
                  }
                  else
                  {
                     delete result;
                     result = NULL;
          
                     *output_stream << 
                        "After \"SET\", a path to the property to set\n"
                        " and the new value, a closing semicolon (\";\") should follow\n"
                        " (or an equal mark (\"=\") when setting a relationship attribute).\n"
                        " Unexpected punctuation character encountered instead: \""
                        << tmp_token->punct_type << "\" .\n";
                  }
               }
               else
               {
                  delete result;
                  result = NULL;
       
                  *output_stream << 
                     "After \"SET\", a path to the property to set\n"
                     " and the new value, a closing semicolon (\";\") should follow.\n"
                     " Alternatively, when setting a CLASS_RELATIONSHIP attribute,\n"
                     " the second argument is a variable definition (<NAME>=<VALUE>).\n"
                     " Parser encountered further non-punctuation data (unexpected).\n";
               }
            }
            else
            {
               // punctuation encountered - possibly the opening brace of a BLOB definition
               if (tmp_token->punct_type == '{')
               {
                  // yes indeedy - seems to be a BLOB definition
                  tmp_parm = Parse_Blob_Def();

                  if (tmp_parm != NULL) // BLOB content parsed fine
                  {
                     result->add_param(tmp_parm);
                  }
                  else
                  {
                     *output_stream <<
                      "The command parser understood that the SET command was intended\n"
                      " to assign a new value to a BLOB (encountered an opening brace)\n"
                      " - but has failed to read the BLOB definition that should've followed.\n"
                      " This regards the mere reading of lines, NOT the Base64 decoding.\n";
                     delete result; result = NULL;
                  }
               }
               else
               {
                  // nay, not a BLOB definition
                  delete result;
                  result = NULL;
       
                  *output_stream <<
                     "After \"SET\" and a path specifying a property to set,\n"
                     " one would expect one of the following:\n"
                     " - a string representing the new value of a plain property\n"
                     " - a variable definition, when setting a CLASS_RELATIONSHIP attribute:\n"
                     "   <NAME>=<VALUE>\n"
                     " - a BLOB definition, when setting the value of a BLOB:\n"
                     "   {\n"
                     "    <BASE64 BLOB CONTENT - multiple lines>\n"
                     "\n"
                     "   };\n"
                     " An unrelated punctuation character encountered instead: \""
                     << tmp_token->punct_type << "\"\n";
               }
            }
         }
         else
         {
            delete result;
            result = NULL;
 
            *output_stream <<
               "After \"SET\", one would expect a path specifying which object\n"
               " to update - a property to set a new value to. A punctuation character\n"
               " certainly wouldn't do.\n";
         }
      }
      else if ( IS_CMD("DELETE") )
      // 
      // synopsis:
      // DELETE <kludge path> [RECURSIVE] ;
      //
      {
         result = new Kludge_CLI_Command(CMD_DELETE);

         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            result->add_param(new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token(); // the closing semicolon or "RECURSIVE"
            if (!tmp_token->punctuation)
            {
               if ( IS_CMD("RECURSIVE") )
               {
                  result->add_param(new Kludge_CLI_CMD_PARM(PARM_OPTION,"R"));
                  tmp_token = tokenizer->Pop_Token(); // the closing semicolon
               }
               else
               {
                  *output_stream <<
                   "After \"DELETE\" and the necessary path specifier,\n"
                   " a semicolon (;) should follow, optionally being preceded\n"
                   " by the 'RECURSIVE' tag.\n"
                   " Unexpected non-punctuation token encountered instead.\n";

                  delete result; result = NULL;
               }
            }

            if (tmp_token->punctuation)
            {
               if (tmp_token->punct_type == ';')
               {
                  // everything OK, return the DELETE command with one parameter
                  // of type PATH and an optional parameter of type OPTION
               }
               else
               {
                  *output_stream <<
                   "After \"DELETE\" and the necessary path specifier,\n"
                   " a semicolon (;) should follow, optionally being preceded\n"
                   " by the 'RECURSIVE' tag.\n"
                   " Unexpected punctuation character encountered instead: "
                   << tmp_token->punct_type << endl;

                  delete result; result = NULL;
               }
            }
            else
            {
               *output_stream <<
                "After \"DELETE\" and the necessary path specifier,\n"
                " a semicolon (;) should follow, optionally being preceded\n"
                " by the 'RECURSIVE' tag.\n"
                " Unexpected non-punctuation token encountered instead.\n";

               delete result; result = NULL;
            }
         }
         else
         {
            delete result; result = NULL;
 
            *output_stream
             << "After \"DELETE\", one would expect a path to the object\n"
             << "to be erased - not a punctuation character.\n";
         }
      }
      else if ( IS_CMD("MOVE") )
      // 
      // synopsis:
      // MOVE <kludge path : what> <kludge path : where> ;
      //
      {
         result = new Kludge_CLI_Command(CMD_MOVE);
         
         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            result->add_param(new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token();
            if ( ! tmp_token->punctuation )
            {
               result->add_param(new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data));

               tmp_token = tokenizer->Pop_Token();
               if ( tmp_token->punctuation )
               {
                  if ( tmp_token->punct_type != ';')
                  {
                     delete result;
                     result = NULL;
          
                     *output_stream
                      << "After \"MOVE\", a path to the object to move and a path\n"
                         " to destination, a closing semicolon (\";\") should follow.\n"
                         " Unexpected punctuation character encountered instead: \""
                         << tmp_token->punct_type << "\" .\n";
                  }
                  // else everything all right, command valid - return result.
               }
               else
               {
                  delete result;
                  result = NULL;
       
                  *output_stream
                   << "After \"MOVE\", a path to the object to move and a path\n"
                      " to destination, a closing semicolon (\";\") should follow.\n"
                      " Parser encountered further data (unexpected).\n";
               }
            }
            else
            {
               delete result;
               result = NULL;
    
               *output_stream
                << "After \"MOVE\" and a path specifying the object to move,\n"
                   " enter a destination path.\n"
                   " A punctuation character certainly wouldn't do.\n";
            }
         }
         else
         {
            delete result;
            result = NULL;
 
            *output_stream
             << "After \"MOVE\", one would expect a path specifying which object\n"
                " to move. A punctuation character certainly wouldn't do.\n";
         }
      }
      else if ( IS_CMD("RENAME") )
      // 
      // synopsis:
      // RENAME <kludge path : what> <string : new name> ;
      //
      {
         result = new Kludge_CLI_Command(CMD_RENAME);
         
         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            result->add_param(new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data));

            tmp_token = tokenizer->Pop_Token();
            if ( ! tmp_token->punctuation )
            {
               result->add_param(new Kludge_CLI_CMD_PARM(PARM_NAME,tmp_token->token_data));

               tmp_token = tokenizer->Pop_Token();
               if ( tmp_token->punctuation )
               {
                  if ( tmp_token->punct_type != ';')
                  {
                     delete result;
                     result = NULL;
          
                     *output_stream
                      << "After \"RENAME\", a path to the object to be renamed\n"
                         " and the new name, a closing semicolon (\";\") should follow.\n"
                         " Unexpected punctuation character encountered instead: \""
                         << tmp_token->punct_type << "\" .\n";
                  }
                  // else everything all right, command valid - return result.
               }
               else
               {
                  delete result;
                  result = NULL;
       
                  *output_stream
                   << "After \"RENAME\", a path to the object to be renamed\n"
                      " and the new name, a closing semicolon (\";\") should follow.\n"
                      " Parser encountered further data (unexpected).\n";
               }
            }
            else
            {
               delete result;
               result = NULL;
    
               *output_stream
                << "After \"RENAME\" and a path to the object to be renamed,\n"
                   " one would expect a string representing the new name.\n"
                   " A punctuation character certainly wouldn't do.\n";
            }
         }
         else
         {
            delete result;
            result = NULL;
 
            *output_stream
             << "After \"RENAME\", one would expect a path specifying the object\n"
                " to rename. A punctuation character certainly wouldn't do.\n";
         }
      }
      else if ( IS_CMD("CD") )
      // 
      // synopsis:
      // CD <kludge path> ;
      //
      {
         result = new Kludge_CLI_Command(CMD_CD);

         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            tmp_parm = new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data);
            result->add_param(tmp_parm);

            tmp_token = tokenizer->Pop_Token();
            if (! ((tmp_token->punctuation) && (tmp_token->punct_type == ';')))
            {
               delete result;
               result = NULL;
 
               *output_stream
                << "After \"CD\" and a path to switch to,\n"
                << "one would expect a semicolon (;) .\n";
            }
            // else everything OK, return the CD command with one parameter
            // of type "path"
         }
         else
         {
            delete result;
            result = NULL;
 
            *output_stream
             << "After \"CD\", one would expect a path to switch to.\n";
         }
      }
      else if (( IS_CMD("EXIT") )||( IS_CMD("QUIT") ))
      // 
      // synopsis:
      // EXIT ;
      // QUIT ;
      //
      {
         result = new Kludge_CLI_Command(CMD_EXIT);

         tmp_token = tokenizer->Pop_Token();
         if ( ! ((tmp_token->punctuation) && (tmp_token->punct_type == ';')))
         {
            delete result;
            result = NULL;
            *output_stream
             << "After EXIT or QUIT there still needs to be a semicolon.\n";
         }
         // else OK, return the EXIT command

      }
      else if ( IS_CMD("SHUTDOWN") )
      // 
      // synopsis:
      // SHUTDOWN ;
      //
      {
         result = new Kludge_CLI_Command(CMD_SHUTDOWN);

         tmp_token = tokenizer->Pop_Token();
         if ( ! ((tmp_token->punctuation) && (tmp_token->punct_type == ';')))
         {
            delete result;
            result = NULL;
            *output_stream
             << "After SHUTDOWN there still needs to be a semicolon.\n";
         }
         // else OK, return the SHUTDOWN command

      }
      else if ( IS_CMD("PWD") ) 
      // 
      // synopsis:
      // PWD ;
      //
      {
         result = new Kludge_CLI_Command(CMD_PWD);
         
         tmp_token = tokenizer->Pop_Token();
         if ( ! ((tmp_token->punctuation) && (tmp_token->punct_type == ';')))
         {
            delete result;
            result = NULL;
            *output_stream
             << "After PWD there still needs to be a semicolon.\n";
         }
         // else OK, return the PWD command
      }
      else if ( IS_CMD("RUN") )
      // 
      // synopsis:
      // RUN <kludge path> ;
      // RUN <kludge path> ( var_name=value; var_name=value; ... );
      //
      {
         result = new Kludge_CLI_Command(CMD_RUN);

         tmp_token = tokenizer->Pop_Token();
         if ( ! tmp_token->punctuation )
         {
            result->add_param(new Kludge_CLI_CMD_PARM(PARM_PATH,tmp_token->token_data));

            tmp_parm = Parse_Arg_List();
            if (tmp_parm != NULL)
            {
               result->add_param(tmp_parm);
            }
            else // there was a parse error - already reported by Parse_Arg_List().
            {
               delete result; result = NULL;
            }
         }
         else
         {
            delete result; result = NULL;
 
            *output_stream
             << "After \"RUN\", one would expect a path to the method\n"
             << "to be called/executed - not a punctuation character.\n";
         }
      }
      else
      // 
      // Unknown command.
      //
      {
         *output_stream
          << "Unknown command: "
          << tmp_token->token_data << endl;
      }
   }
   else
   {
      *output_stream
        << "Any command should start with a string identifier.\n"
        << "A punctuation character certainly wouldn't do :)\n";
   }

   // Reset the tokenizer. Drop any garbage trailing the semicolon on this line.
   tokenizer->Reset();

   return(result);
#undef IS_CMD(CMD_STRING)
}



//static
Kludge_pl_params* Kludge_CLI::Prepare_Method_Params(Kludge_CLI_CMD_PARM* params)
{
   Kludge_pl_params* result = new Kludge_pl_params();

   if (params != NULL)
   {
      if (params->type == PARM_ARGLIST)
      {
         for (unsigned int i=0; i< params->contents.size(); i++)
         {
            result->push_back(
                              new Kludge_pl_param(
              (char*) params->contents[i]->contents[0]->content.c_str(),
              (char*) params->contents[i]->contents[1]->content.c_str()
                                                 )
                             );
         }
      }
      else
      {
         cout <<
            "Internal error - in Kludge_CLI::Prepare_Method_Params:\n"
            " obtained params container of type other than PARM_ARGLIST.\n";
      }
   }
   else
   {
      cout <<
         "Internal error - in Kludge_CLI::Prepare_Method_Params:\n"
         " obtained NULL params pointer.\n";
   }
   
   return(result);
}



int Kludge_CLI::Execute_Command(Kludge_CLI_Command* cmd)
{
 string* tmp_path_str = NULL;
 Kludge_Path *tmp_path = NULL;
 Kludge_Object *tmp_ptr = NULL, *tmp_ptr2 = NULL;
 Kludge_pl_params* tmp_params = NULL;

   switch (cmd->type)
   {
      case CMD_GET :
      {
         tmp_path = Kludge_CLI_String2path(
                               &cmd->params[0]->content,
                               output_stream
                                          );
         if (tmp_path != NULL)
         {
            tmp_ptr = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);

            if (tmp_ptr != NULL)
            {
               tmp_ptr->List_Contents(this);
               tmp_ptr = NULL;
            }
            else
            {
               *output_stream
                 << "Execute_Command: Should've retrieved \""
                 << cmd->params[0]->content
                 << "\" .\n"
                 << " No such object.\n";
            }

            delete tmp_path; tmp_path = NULL;
         }
         else
         {
            *output_stream
             << "Execute_Command: Something wrong with string2path...\n"
             << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
         }
      }
      break;
      case CMD_INSERT :
      {
         if (cmd->params.size() == 2) // where, what to insert
         {
            tmp_path = Kludge_CLI_String2path(
                                  &cmd->params[0]->content,
                                  output_stream
                                             );
            if (tmp_path != NULL)
            {            
               tmp_ptr = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);
               delete tmp_path; tmp_path = NULL;            

               if (tmp_ptr != NULL)
               {
                  tmp_ptr->Insert(cmd->params[1], this);
                  tmp_ptr = NULL;
               }
               else
               {
                  *output_stream
                   << "Execute_Command: Should've inserted a new object under\n \""
                   << cmd->params[0]->content
                   << "\" .\n"
                   << " No such existing object.\n";
               }
            }
            else
            {
               *output_stream
                << "Execute_Command: Something wrong with string2path...\n"
                << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
            }         
         }
         else
         {
            cout <<
            "Internal Kludge_CLI error - in Execute_Command():\n"
            " obtained an INSERT command with param. count other than 2.\n"
            " Perhaps Prepare_Command() goofed up.\n";
         }
      }
      break;
      case CMD_SET :
      {
         if (cmd->params.size() == 2) // property, value
         {
            tmp_path = Kludge_CLI_String2path(
                                  &cmd->params[0]->content,
                                  output_stream
                                             );
            if (tmp_path != NULL)
            {            
               tmp_ptr = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);
               delete tmp_path; tmp_path = NULL;            

               if (tmp_ptr != NULL)
               {
                  tmp_ptr->Set(cmd->params[1], this);
                  tmp_ptr = NULL;
               }
               else
               {
                  *output_stream
                   << "Execute_Command: Should've set a new value to\n \""
                   << cmd->params[0]->content
                   << "\".\n"
                   << " No such existing object.\n";
               }
            }
            else
            {
               *output_stream
                << "Execute_Command: Something wrong with string2path...\n"
                << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
            }         
         }
         else
         {
            cout <<
            "Internal Kludge_CLI error - in Execute_Command():\n"
            " obtained a SET command with param. count other than 2.\n"
            " Perhaps Prepare_Command() goofed up.\n";
         }
      }
      break;
      case CMD_DELETE :
      {
         tmp_path = Kludge_CLI_String2path(
                               &cmd->params[0]->content,
                               output_stream
                                          );
         if (tmp_path != NULL)
         {
            // We'd better only delete own class properties and relationships
            // (not the inherited ones). 
            // To ensure that, we need to make sure that the last path element 
            // resolves to an object reference only if the object is not inherited.
            tmp_ptr2 = Kludge_CLI_Path2Ptr(tmp_path,this,pwd,/*skip last*/true);
            if (tmp_ptr2 != NULL)
            {
               tmp_ptr = tmp_ptr2->Resolve_Content_Name
                         (
                          tmp_path->elements[tmp_path->elements.size()-1],
                          this,
                          /*own only?*/true
                         );
            }

            if (tmp_ptr != NULL)
            {
               if (! (*tmp_ptr >= *pwd))
               // we can't cut the branch we're sitting on
               {
                  // check for the recursive option
                  bool tmp_recursive_tag = false;
                  if (cmd->params.size() > 1)
                  {
                     if (cmd->params[1]->type == PARM_OPTION)
                     {
                        if (cmd->params[1]->content == "R")
                        {
                           tmp_recursive_tag = true;
                        }
                        else
                        {
                           cout <<
                              "Internal error - in Execute_Command(DELETE):\n"
                              " unknown option received - \'R\' expected.\n";
                        }
                     }
                     else
                     {
                        cout <<
                           "Internal error - in Execute_Command(DELETE):\n"
                           " unknown parameter received at position [1].\n"
                           " - the RECURSIVE option was expected\n";
                     }
                  }

                  // the following line means
                  // "Erase the object pointed to by tmp_ptr and report errors 
                  // to *this CLI."
                  tmp_ptr->Delete(this, tmp_recursive_tag);
                  tmp_ptr = NULL;
               }
               else
               {
                  *output_stream
                     << "In Kludge_CLI::Execute_Command():\n"
                        " Can't delete an object that's equal to or superior to the current path.\n";
               }
            }
            else   // Okay - the name could still point to an inherited object.
            {      // Let's check this out, just to have a meaningful error message to report.
               if (
                     (tmp_ptr2 != NULL) &&  // if the last-but-one path element is valid
                     (
                        tmp_ptr2->Resolve_Content_Name // and there is some inherited stuff,
                        (
                            tmp_path->elements[tmp_path->elements.size()-1],
                            this
                        )   // by default, own_only == false => we're looking for inherited objects
                        != NULL
                     )
                     // i.e., if the name is an inherited property
                  )
               {
                  *output_stream
                   << "Execute_Command: Should've deleted \""
                   << cmd->params[0]->content
                   << "\" .\n" <<
                   "Given the context of this path, the object that you're trying\n"
                   " to delete is an inherited member. Inherited members can only\n"
                   " be deleted in their original owner class. In any of the heir\n"
                   " classes, inherited members can only be suppressed or overloaded.\n";
               }
               else
               {
                  *output_stream
                   << "Execute_Command: Should've deleted \""
                   << cmd->params[0]->content
                   << "\" .\n"
                   << " No such object.\n";
               }
            }
            
            tmp_ptr2 = NULL;
            delete tmp_path; tmp_path = NULL;
         }
         else
         {
            *output_stream 
             << "Internal error - in Execute_Command: Something wrong with string2path...\n"
             << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
         }
      }
      break;
      case CMD_MOVE :
      {
         if (cmd->params.size() == 2) // property, value
         {
            tmp_path = Kludge_CLI_String2path(
                                  &cmd->params[0]->content,
                                  output_stream
                                             );
            if (tmp_path != NULL)
            {            
               tmp_ptr = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);
               delete tmp_path; tmp_path = NULL;            

               if (tmp_ptr != NULL)  // object to move OK 
               {

                  tmp_path = Kludge_CLI_String2path(
                                        &cmd->params[1]->content,
                                        output_stream
                                                   );
                  if (tmp_path != NULL)
                  {            
                     tmp_ptr2 = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);
                     delete tmp_path; tmp_path = NULL;            

                     if (tmp_ptr2 != NULL)
                     {
                        // can't make an object subordinate to itself
                        if (!(*tmp_ptr >= *tmp_ptr2))
                        {
                           // don't exercise moving an object to its current location.
                           if (tmp_ptr->Superior() != tmp_ptr2)
                           {
                              tmp_ptr->Move(tmp_ptr2, this);
                              tmp_ptr2 = NULL;
                           }
                           else
                           {
                              *output_stream <<
                                 "You've just asked Kludge to move an object to a destination\n"
                                 " that's equal to the object's current superior.\n"
                                 " No action taken. Not really an error.";
                           }
                        }
                        else
                        {
                           *output_stream <<
                              "When moving, destination must not be subordinate\n"
                              " to the object being moved. Exciting as cycles are,\n"
                              " Kludge considers them clearly undesirable.\n";
                        }
                     }
                     else
                     {
                        *output_stream
                         << "Execute_Command: Should've moved an object to:\n \""
                         << cmd->params[0]->content
                         << "\".\n"
                         << " No such existing object / destination.\n";
                     }
                  }
                  else
                  {
                     *output_stream
                      << "Execute_Command: Something wrong with string2path...\n"
                      << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
                  }         
                  
                  tmp_ptr = NULL;
               }
               else
               {
                  *output_stream
                   << "Execute_Command: Should've moved the following object:\n \""
                   << cmd->params[0]->content
                   << "\".\n"
                   << " No such existing object.\n";
               }
            }
            else
            {
               *output_stream
                << "Execute_Command: Something wrong with string2path...\n"
                << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
            }         
         }
         else
         {
            cout <<
            "Internal Kludge_CLI error - in Execute_Command():\n"
            " obtained a MOVE command with param. count other than 2.\n"
            " Perhaps Prepare_Command() goofed up.\n";
         }
      }
      break;
      case CMD_RENAME :
      {
         if (cmd->params.size() == 2) // object, name
         {
            tmp_path = Kludge_CLI_String2path(
                                  &cmd->params[0]->content,
                                  output_stream
                                             );
            if (tmp_path != NULL)
            {            
               tmp_ptr = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);
               delete tmp_path; tmp_path = NULL;            

               if (tmp_ptr != NULL)
               {
                  tmp_ptr->Rename(cmd->params[1], this);
                  tmp_ptr = NULL;
               }
               else
               {
                  *output_stream
                   << "Execute_Command: Should've set a new name to\n \""
                   << cmd->params[0]->content
                   << "\".\n"
                   << " No such existing object.\n";
               }
            }
            else
            {
               *output_stream
                << "Execute_Command: Something wrong with string2path...\n"
                << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
            }         
         }
         else
         {
            cout <<
            "Internal Kludge_CLI error - in Execute_Command():\n"
            " obtained a RENAME command with param. count other than 2.\n"
            " Perhaps Prepare_Command() goofed up.\n";
         }
      }
      break;
      case CMD_CD :
      {
         tmp_path = Kludge_CLI_String2path(
                               &cmd->params[0]->content,
                               output_stream
                                          );
         if (tmp_path != NULL)
         {
            tmp_ptr = Kludge_CLI_Path2Ptr(tmp_path, this, pwd);

            if (tmp_ptr != NULL)
            {
               if 
               ( 
                  (tmp_ptr->Type == Kludge_Object::KLUDGE_CLASS)
                  || (tmp_ptr->Type == Kludge_Object::KLUDGE_INSTANCE)
                  || (tmp_ptr->Type == Kludge_Object::KLUDGE_ROOT)
               )
               {
                  pwd = tmp_ptr;
               }
               else
               {
                  *output_stream << "Can only CD to a class or instance.\n";
                  tmp_ptr = NULL;
               }
            }
            else
            {
               *output_stream
                << "Execute_Command: Should've CD'ed to \""
                << cmd->params[0]->content
                << "\" .\n"
                << " No such class or instance.\n";
            }

            delete tmp_path; tmp_path = NULL;
         }
         else
         {
            *output_stream
             << "Execute_Command: Something wrong with string2path...\n"
             << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
         }
      }
      break;
      case CMD_EXIT :
      {
         *output_stream <<
             "Bye :)\n" << flush;
         CLI_done = true;
      }
      break;
      case CMD_SHUTDOWN :
      {
         pthread_kill(master_thread,SIGINT);
      }
      break;
      case CMD_PWD :
      {
         tmp_path_str = Kludge_CLI_PWD(pwd, output_stream);
         if (tmp_path_str != NULL)
         {
            *output_stream
             << "Current Kludge Path is \"" << *tmp_path_str << "\".\n";
            delete tmp_path_str;
         }
         else
         {
            *output_stream <<
             "Execute_Command: Kludge_CLI_PWD() failed - returned null string*."
             " Alloc error?\n";
         }
      }
      break;
      case CMD_RUN :
      {
         tmp_path = Kludge_CLI_String2path(
                               &cmd->params[0]->content,
                               output_stream
                                          );
         if (tmp_path != NULL)
         {
            // Let's process the last path element in our own way.
            tmp_ptr2 = Kludge_CLI_Path2Ptr(tmp_path,this,pwd,/*skip last*/true);
            if (tmp_ptr2 != NULL)
            {
               if (tmp_ptr2->Type == Kludge_Object::KLUDGE_INSTANCE)
               {
                  // instance.member() call style
                  tmp_ptr = ((Kludge_Instance*)tmp_ptr2)->Class->Resolve_Content_Name
                               (
                                tmp_path->elements[tmp_path->elements.size()-1],
                                this,
                                /*own only?*/false
                               );
               }
               else if (tmp_ptr2->Type == Kludge_Object::KLUDGE_CLASS)
               {
                  // class::method() call style (like a static method)
                  tmp_ptr = ((Kludge_Class*)tmp_ptr2)->Resolve_Content_Name
                               (
                                tmp_path->elements[tmp_path->elements.size()-1],
                                this,
                                /*own only?*/false
                               );
               }
               else
               {
                  *output_stream << 
                   "Error - asked to run \""
                   << cmd->params[0]->content <<
                   "\".\n"
                   " Failed to resolve that path - the last-but-one path element\n"
                   " does exist, but it's neither a Kludge Class nor an Instance.\n";
               }
              
               // by now we should have a particular Kludge_Object resolved.
               if (tmp_ptr != NULL)
               {            
                  if (tmp_ptr->Type == Kludge_Object::KLUDGE_CLASS_PROPERTY)
                  {
                     if ( ((Kludge_Class_Property*)tmp_ptr)->SubType == 
                               Kludge_Class_Property::SUBTYPE_METHOD )
                     {
                        //the following function creates the vector of tmp_params
                        tmp_params = Prepare_Method_Params(cmd->params[1]);

                        if (tmp_ptr2->Type == Kludge_Object::KLUDGE_INSTANCE)
                        {  // instance.member() call style - we need to pass the "this" pointer
                           tmp_params->push_back(new Kludge_pl_param("this", tmp_ptr2));
                        }
                        
                        ((Kludge_Class_Property*)tmp_ptr)->Run_Method(this,tmp_params);
                        // the tmp_params vector gets deleted by Run_Method
                     }
                     else
                     {
                        *output_stream << 
                         "Error - asked to run something called \""
                         << *tmp_path->elements[tmp_path->elements.size()-1] <<
                         "\",\n"
                         " which is a not a METHOD. It *IS* a CLASS_PROPERTY style thingy.\n";
                     }
                  }
                  else
                  {
                     *output_stream << 
                      "Error - asked to run something called \""
                      << *tmp_path->elements[tmp_path->elements.size()-1] <<
                      "\",\n"
                      " which is not a METHOD. Not even a CLASS_PROPERTY style thingy.\n";
                  }
               }
               else
               {
                  if ( (tmp_ptr2->Type == Kludge_Object::KLUDGE_INSTANCE)
                       || (tmp_ptr2->Type == Kludge_Object::KLUDGE_CLASS) )
                  {
                     *output_stream << 
                      "Error - asked to run something called \""
                      << *tmp_path->elements[tmp_path->elements.size()-1] <<
                      "\".\n"
                      " No such name in the respective class.\n";
                  }
                  // else error already reported
               }
            }
            else
            {
               *output_stream << 
                "Error - asked to run \""
                << cmd->params[0]->content <<
                "\".\n"
                " Failed to resolve that path - even the last-but-one path element\n"
                " doesn't seem to exist.\n";
            }
         }
         else
         {
            *output_stream 
             << "Internal error - in Execute_Command: Something wrong with string2path...\n"
             << " cannot convert \"" << cmd->params[0]->content << "\" to a path.\n";
         }
      }
      break;
   }

   return(0);
}



int Kludge_CLI::Process_Command()
{
 Kludge_CLI_Command* tmp_command = NULL;

   tmp_command = Prepare_Command(); // parse the command line syntax

   if (tmp_command != NULL)
   {
   
      *output_stream << endl;   // send a separating empty line.

      Execute_Command(tmp_command); // react to the pre-parsed command
      delete tmp_command; tmp_command = NULL;

      *output_stream << endl;   // send a separating empty line.
   }
   
   return(0);
}



/*
-- This is the command line interface of Kludge - a function that gets
-- called by a client thread just created. In other words, it is the
-- core functionality of the thread, although the thread itself is
-- pthread_create()d from another function in the kludge_threads module.
*/
int Kludge_CLI::Run()
{
 //string tmp_class_name;
 //Kludge_Class* tmp_class_ptr = NULL;

   Welcome_Message();

   /*
   // this used to be a sorta debugging instrumentation stub

   tmp_class_name = "Object";
   tmp_class_ptr = Kludge_Class::Name2ptr(&tmp_class_name, this);
   if (tmp_class_ptr != NULL)
   {
      tmp_class_ptr->List_Contents(this);
      tmp_class_ptr = NULL; // do NOT `delete` this pointer after use!
   }
   else
   {
      cout << "No such class name.\n";
   }
   */

   try
   {
      while ( ! CLI_done)
      {
         Process_Command();
      }
   }
   catch(Ex_Shutdown)
   {
      cout <<
         "Shutdown exception caught in thread " << pthread_self() << ".\n" << flush;
      *output_stream <<
         "Thread terminated due to server shutdown.\n" << flush;
   }

   return(0);
}








/* ###   global (standalone) functions   ### */

Kludge_Path *Kludge_CLI_String2path(string *str_parsed, iostream* error_output)
{
 Kludge_Path *result = NULL;
 size_t position=0,elem_start = 0;
 bool finished = false;

   if (str_parsed->length() != 0)
   // OK, we have been passed a non-zero token string
   {
      result = new Kludge_Path;
      if (str_parsed->length() != 1)
      // OK, our path is longer than a single character
      {
         if ((*str_parsed)[0] == '/')
         {
            result->leading_slash = true;
            position++;
         }

         elem_start = position;
         while (! finished)
         {
            position = str_parsed->find_first_of("/\\",position);
            if (position != string::npos)
            // found something before the end of the input string
            {
               switch ((*str_parsed)[position])
               {
                  case '/'  :
                     {
                        if (position != elem_start)
                        {
                           result->elements.push_back
                           (
                              new string
                              (
                                 *str_parsed,
                                 elem_start,
                                 position - elem_start
                              )
                           );

                           if (position + 1 != str_parsed->length())
                           // this is not a trailing slash
                           {
                              position++;
                              elem_start = position;
                           }
                           else
                           // the slash we've found is a trailing slash
                           {
                              result->trailing_slash = true;
                              finished = true;
                           }
                        }
                        else
                        {
                           *error_output << "Double slash found - an invalid path construct.\n";
                           delete result;
                           result = NULL;
                           finished = true;
                        }
                     }
                     break;
                  case '\\' :
                     {
                        if (position + 1 != str_parsed->length())
                        // the backslash is not at the end of path
                        {
                           if (position + 2 != str_parsed->length())
                           // the escape sequence is not at the end of path
                           {
                              position += 2;
                           }
                           else
                           // the escape sequence is at the end of path
                           {
                              result->elements.push_back
                              (
                                 new string
                                 (
                                    *str_parsed,
                                    elem_start,
                                    position + 2 - elem_start
                                 )
                              );
                              finished = true;
                           }
                        }
                        else
                        {
                           *error_output << "Paths finished with a backslash are not supported.\n";
                           delete result;
                           result = NULL;
                           finished = true;
                        }
                     }
                     break;
               }
            }
            else
            // this is the last element in the path
            {
               result->elements.push_back
               (
                  new string
                  (
                     *str_parsed,
                     elem_start,
                     str_parsed->length() - elem_start
                  )
               );
               finished = true;
            }
         }
      }
      else
      // this is a single-character path
      {
         if ((*str_parsed)[0] != '/')
         // whatever it is, it's only one character - only one path element.
         {
            result->elements.push_back(new string(*str_parsed));
         }
         else
         // the whole path is only the slash character - i.e., the root node
         {
            result->leading_slash = true;
         }
      }
   }
   else
   {
      *error_output << "Internal parser error:\n"
                    << "zero-length strings cannot be translated to path.\n";
   }
   return(result);
}



// The "base" parameter can be set to $PWD, 
// or can be used for recursive resolution
// (more precisely, in recursive resolution it MUST be set
// to the last element resolved).
// Setting base to NULL or leaving it undefined effectively means `base = Kludge_Root`
Kludge_Object* Kludge_CLI_Path2Ptr(Kludge_Path* path, Kludge_CLI* CLI,
                Kludge_Object* base = NULL, bool skip_last=false, size_t level=0) 
{
 Kludge_Object *tmp_ptr = NULL, *result = NULL;
 size_t skip = 0;

   if (base == NULL) base = Kludge_Root;

   if (path != NULL)
   // yes we do have something to process - that's a good start
   {
      if (skip_last) skip = 1;

      if (path->elements.size() - skip > level)
      // The path elements count in combination with the recursion level
      // must make sense.
      {
         if ((path->leading_slash) && (level == 0))
         {
            base = Kludge_Root;
         }
         
         tmp_ptr = base->Resolve_Content_Name(path->elements[level], CLI);

         if (tmp_ptr == NULL)
         // allright, we can start a path from a class name,
         // but we want that to be the choice of last resort
         {
            if ((level == 0) && (! path->leading_slash))
            {
               tmp_ptr = Kludge_Class::Name2ptr(path->elements[0], CLI);
            }
         }

         if (tmp_ptr != NULL)
         // recurse
         {
            if ((path->elements.size() - skip) > (level +1) )
            // the path has further elements
            {
               result = Kludge_CLI_Path2Ptr(path, CLI,
                        tmp_ptr, skip_last, level + 1);
            }
            else
            // The last element in the path. Done.
            {
               result = tmp_ptr;
            }
         }
         else
         {
            *CLI->output_stream
                    << "Path2ptr: path element \""
                    << *path->elements[level]
                    << "\" not found.\n";
         }
      }
      else
      {
         if ((path->leading_slash) && (path->elements.size() - skip == 0))
         // the whole path is just a slash - the root sign
         {
            result = Kludge_Root;
         }
         else
         {
            if (skip_last && (path->elements.size() == 1) && ! path->leading_slash)
            {
               result = base;
            }
            else
            {
               *CLI->output_stream <<
                     "Internal error - path2ptr:\n"
                     " actual path element count insufficient for the desired recursion level.\n";
            }
         }
      }
   }
   else
   // no path to process :(
   {
      *CLI->output_stream << "Internal error - path2ptr: input path pointer is null.\n";
   }

   // the function calling path2ptr should NOT delete the returned object after use!
   return(result);
}



string* Kludge_CLI_PWD(Kludge_Object* ptr, iostream* error_output)
{
   string* result = new string;
   
   if (result != NULL)
   {
      ptr->pwd(result);
   }
   else
   {
      *error_output << "Kludge_CLI_PWD(): error creating a `new string' for the path.\n";
   }

   return(result);
}



// Case insensitive compare function
// Taken from "Thinking in C++" by Bruce Eckel
int Bruce_Eckel_stringCmpi(const string& s1, const string& s2)
{
   // Select the first element of each string:
   string::const_iterator p1 = s1.begin(), p2 = s2.begin();
   string::const_iterator p1e = s1.end(), p2e = s2.end();
   // Don't run past the end:
   while(p1 != p1e && p2 != p2e)
   {
      // Compare upper-cased chars:
      if(toupper(*p1) != toupper(*p2))
         // Report which was lexically  greater:
         return (toupper(*p1)<toupper(*p2))? -1 : 1;
      p1++;
      p2++;
   }
   // If they match up to the detected eos, say
   // which was longer. Return 0 if the same.
   return(s2.size() - s1.size());
}
