USE mysql;
DROP DATABASE kludge;
CREATE DATABASE kludge;
GRANT Select,Insert,Update,Delete ON kludge.* TO kludge@localhost;
USE kludge;
#SHOW DATABASES;
#SHOW TABLES;


CREATE TABLE Classes (
  Class INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Name VARCHAR(254),
  Parent INT UNSIGNED
);
# key: Class
#
# Low-level unique identifier for DBMS-based operations. Application protocol should use class tree path or unique class name for identification.
# Should be unique, too. The separation of Class_ID from the textual class Name is for the sake of simplified referential integrity at the object level.
# Parent Class_ID

CREATE TABLE Class_Properties (
  Class INT UNSIGNED NOT NULL,
  RDN VARCHAR(254) NOT NULL,
  Value TINYTEXT,
  Suppress TINYINT UNSIGNED,
  Flags TINYINT UNSIGNED,
  PRIMARY KEY(Class, RDN)
);
# key: Class, RDN
#
# Owner class ID
# Relative Distinguished Name - property name, local to owner class
# Property value
# Suppress inherited properties with a matching name
# Blob flags - IS_BLOB, IS_TXT_BLOB and IS_METHOD

CREATE TABLE Class_Relationships (
  RelCLSID INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Master INT UNSIGNED NOT NULL,
  Slave INT UNSIGNED NOT NULL,
  Type TINYINT UNSIGNED,
  Name_M VARCHAR(254) NOT NULL,
  Name_S VARCHAR(254) NOT NULL,
  MinCount_M INT UNSIGNED,
  MaxCount_M INT UNSIGNED,
  DefaultCount_M INT UNSIGNED,
  NamingGen_M VARCHAR(254),
  NamingMatch_M VARCHAR(254),
  MinCount_S INT UNSIGNED,
  MaxCount_S INT UNSIGNED,
  NamingGen_S VARCHAR(254),
  NamingMatch_S VARCHAR(254),
  Suppress TINYINT UNSIGNED
);
# key: Master, Slave
#
# Class Relationship unique ID
# Class_ID of the master (source) class
# Class_ID of the slave (destination) class
# One of a few defined types - 0=equal-level, 1=master/slave, 2=container/member (primary ownership)
# Name at Master 
# Name at Slave
# Minimum number of instances of the slave class within an instance of the master class
# Maximum number of instances of the slave class within an instance of the master class
# Default number of instances of the slave class, created upon the creation of an instance of the master class
# The naming convention for auto-created instances, following the C syntax of printf() - at master
# The name checking regular expression (perl syntax) - at master
# Minimum number of master instances refered to by the slave instance
# Maximum number of master instances refered to by the slave instance
# The naming convention for auto-created instances, following the C syntax of printf() - at slave
# The name checking regular expression (perl syntax) - at slave
# For the suppression of inherited relationships - the same as with the plain class properties

CREATE TABLE Instances (
  OID INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Class INT UNSIGNED,
  ClsRel INT UNSIGNED,
  RDN VARCHAR(254),
  Owner INT UNSIGNED
);
# key: OID
#
# Object ID - Low-level unique identifier for DBMS-based operations. Application protocol should use instance tree path for identification.
# Class ID of the particular instance's Kludge_Class.
# Relationship ID of the respective "OWN" type class relationship (relationship class?)
# Relative Distinguished Name - instance name.
# Instance ID of the owner instance.

CREATE TABLE Instance_Properties (
  OID INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Instance INT UNSIGNED NOT NULL,
  RDN VARCHAR(254) NOT NULL,
  Class INT UNSIGNED,
  ClsRel INT UNSIGNED,
  Value TEXT,
  Flags TINYINT UNSIGNED
);
# key Oid, RDN
#
# Object ID - unique instance identifier - needed exclusively for obscure internal purposes (as a key in a tree counting instances of a class)
# Instance ID of the owner instance
# Property name
# Class ID of the property class
# RelCLSID of the relevant "OWN"-type class relationship
# Property content
# Blob flags - IS_BLOB and IS_TXT_BLOB

CREATE TABLE Instance_Relationships (
  RelOID INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Master INT UNSIGNED NOT NULL,
  Slave INT UNSIGNED NOT NULL,
  ClsRel INT UNSIGNED NOT NULL,
  Name_M VARCHAR(254) NOT NULL,
  Name_S VARCHAR(254) NOT NULL
);
#
# Relationship Object ID - unique instance identifier - needed exclusively for obscure internal purposes (as a key in a tree counting instance relationships of a relationships class)
# Master instance ID
# Slave instance ID
# Instance relationship class
# Name at Master
# Name at Slave

CREATE TABLE Blobs (
  BlobID INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Content MEDIUMBLOB
);
# key: BlobID
#
# unique Blob ID - should be refered to by the value of a special class of properties
# blob content

CREATE TABLE Text_Blobs (
  BlobID INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,
  Content MEDIUMTEXT
);
# to support future full_text searches
# key: BlobID
#
# unique Blob ID - should be refered to by the value of a special class of properties
# blob content

INSERT INTO Classes VALUES (NULL,"Root_Class",0);
INSERT INTO Classes VALUES (NULL,"Root_Instance_Class",1);
INSERT INTO Instances VALUES (NULL,2,0,"Root_Instance",0);

