#include "kludge_perl.h"
#undef assert
#include <fstream>
#include <strstream>

extern "C" void xs_init _((void));  // Provided by xsinit.o, that is statically linked to us.
                                    // See Makefile - xsinit.c refers to XS package names!

static PerlInterpreter *Kludge_pl;  // The global Perl interpreter
static CV* Kludge_pl_thr_wrapper;   // The Perl-space per-thread CLI wrapper (does some init)
                                    // See Kludge_pl_init() for more details.


// ###        class Kludge_pl_param         ###
//           ALL METHODS ARE INLINED :)



// ###    generic perl eval and call wrappers/helpers     ###


int Kludge_pl_eval(char *code_str)
{
 int result = 0;
 SV* sv = newSVpv(code_str,0);

   dSP;                    // declare a few local variables for the following macros
   ENTER;                  // start a new scope
   SAVETMPS;               // mortals declared after this point will be deleted upon FREETMPS
   PUSHMARK(sp);           // mark the start of input arguments on stack
   // add parameters here - if any
   PUTBACK;                // done entering input params on stack; returned results follow (if any)
   result = perl_eval_sv(sv,G_DISCARD); // eval the piece of code, discarding results
   SPAGAIN;                // mark the start of output arguments (results) on stack
   SvREFCNT_dec(sv);                    // delete sv
   // if there are returned results, we'd better POP them off the stack right here!!!
   PUTBACK;                // done parsing output arguments (results) from the stack
   FREETMPS;               // delete mortals created since the last SAVETMPS
   LEAVE;                  // leave scope

   return(result);
}


/*
// no use for this function at the moment - see the header file
int Kludge_pl_call(char *subname, int int_passed)
{
 int result = 0;

   dSP;                    // declare a few local variables for the following macros
   ENTER;                  // start a new scope
   SAVETMPS;               // mortals declared after this point will be deleted upon FREETMPS
   PUSHMARK(sp);           // mark the start of input arguments on stack
   // add parameters here - if any
   XPUSHs(sv_2mortal(newSViv(int_passed)));
   PUTBACK;                // done entering input params on stack; returned results follow (if any)
   result = perl_call_pv(subname,G_DISCARD);  // run the procedure requested, discarding results
   SPAGAIN;                // mark the start of output arguments (results) on stack
   // if there are returned results, we'd better POP them off the stack right here!!!
   PUTBACK;                // done parsing output arguments (results) from the stack
   FREETMPS;               // delete mortals created since the last SAVETMPS
   LEAVE;                  // leave scope

   return(result);
}
*/




// ###     Perl eval and call functions,                             ###
// ###     observing the additional constraints of Kludge Perl       ###
CV* Kludge_pl_eval_sub(char* code_str)
{
 CV* result = NULL;
 int eval_err = 0;
 string code_str_complete;
 SV* code_sv = NULL;
 SV* cv_ref = NULL;
 string* tmp_str;

   if (code_str != NULL)
   {
      code_str_complete = "sub {\n";       // the result from the evaluation will be a CV reference
      code_str_complete += "my $CLIENT_SOCKET = Kludge_pl::get_thr_socket_handle();\n";
      code_str_complete += code_str;
      code_str_complete += "\n}\n";
      code_sv = newSVpv((char*)code_str_complete.c_str(),0); // the typecast gets rid of `const'

      dSP;                    // declare a few local variables for the following macros
      ENTER;                  // start a new scope
      SAVETMPS;               // mortals declared after this point will be deleted upon FREETMPS
      PUSHMARK(sp);           // mark the start of input arguments on stack
      // add parameters here - if any
      PUTBACK;                // done entering input params on stack; returned results follow (if any)
      eval_err = perl_eval_sv(code_sv,G_SCALAR); // eval the piece of code, accept a scalar result (in this case, it's a CV ref)
      SPAGAIN;                // mark the start of output arguments (results) on stack
      SvREFCNT_dec(code_sv);  // we didn't mortalize it, so we have to delete it explicitly

      if (eval_err == 1)      // Should be a one. A zero or a small positive int mean the number
      {                       // of results returned. A negative int means there was an error.
         cv_ref = POPs;       // POP the (CV ref) result off the stack
         //SvREFCNT_inc(cv_ref); // we don't need the reference - let it die (it's a mortal).

         if (SvTYPE(cv_ref) == SVt_RV)
         {
            result = (CV*) SvRV(cv_ref);      // the lack of proper C++ polymorphism is a mess :)
            if (SvTYPE(result) == SVt_PVCV)
            {
               // Okay, so it's indeed a CV.
               // We do need the bare CV - we must prevent its deletion, which would
               // happen when the mortal CV reference would die.
               SvREFCNT_inc(result);
               // and that's it :)
            }
            else
            {
               tmp_str = decipher_SV_type((SV*) result); // the lack of proper C++ polymorphism is a mess :)
               cout << "Error - in Kludge_pl_eval_sub: eval successful, returned a reference,\n"
                       " but the value referred to is not a CV. Value type: " 
                    << *tmp_str << ".\n";
               delete tmp_str;
               result = NULL;
            }
         }
         else
         {
            tmp_str = decipher_SV_type(cv_ref);
            cout << "Error - in Kludge_pl_eval_sub(): eval successful but returned a result\n"
                    " that is not a reference. Result type: " 
                 << *tmp_str << ".\n";
            delete tmp_str;
         }
      }
      else
      {
         if (eval_err >= 0)
         {
            cout << "Error: incorrect number of params returned from an eval: " << eval_err << " !\n";
            for (int i=0; i < eval_err; i++)   // one by one
            {
               cv_ref = POPs; // We must pop the unwanted/incorrect results anyway
                              // - to prevent corruption of the Perl's stack (in this thread)
            }
         }
         else /* eval_err < 0*/ cout << "Error - in Kludge_pl_eval_sub: evaluation failed!\n";
      }
      PUTBACK;                // done parsing output arguments (results) from the stack
      FREETMPS;               // delete mortals created since the last SAVETMPS - e.g., results from perl_eval or perl_call
      LEAVE;                  // leave scope
   }
   else
   {
      cout << "Internal error - in Kludge_pl_eval_sub: NULL char* obtained as input.\n";
   }

   return(result);
}


CV* Kludge_pl_eval_sub(string& code_str)  // a wrapper taking string& instead of char*
{
   return( Kludge_pl_eval_sub((char*)code_str.c_str()) );
}



int Kludge_pl_call_CV(CV* sub_CV, Kludge_pl_params* args)
{
 int result = 0;

   if (sub_CV != NULL)
   {
      dSP;                    // declare a few local variables for the following macros
      ENTER;                  // start a new scope
      SAVETMPS;               // mortals declared after this point will be deleted upon FREETMPS
      PUSHMARK(sp);           // mark the start of input arguments on stack
      // add parameters here - if any
      if (args != NULL)
      {
         for (unsigned int i=0; i < args->size(); i++)
         {
            XPUSHs(sv_2mortal( (*args)[i]->Key ));
            XPUSHs(sv_2mortal( (*args)[i]->Value ));
            delete (*args)[i];   // we can dispose of the "pair wrapper object" right here.
         }
         delete args; // and we can dispose of the whole params array here.
      } // else there are no args to process - run the sub with no args
      PUTBACK;                // done entering input params on stack; returned results follow (if any)
      //result = perl_call_sv(sv_2mortal(newRV_inc((SV*)sub_CV)),G_DISCARD);  // run the procedure requested, discarding re   SPAGAIN;                // mark the start of output arguments (results) on stack
      result = perl_call_sv((SV*)sub_CV,G_DISCARD);  // run the procedure requested, discarding results
      SPAGAIN;                // mark the start of output arguments (results) on stack
      // if there are returned results, we'd better POP them off the stack right here!!!
      PUTBACK;                // done parsing output arguments (results) from the stack
      FREETMPS;               // delete mortals created since the last SAVETMPS
      LEAVE;                  // leave scope
   }
   else
   {
      cout << "Internal error - in Kludge_pl_call_CV: CV* to call is NULL !\n";
      // this is a terrible internal error - let's give up cleaning up the parameters obtained
      result = 1;
   }

   return(result);
}










// ### Threaded perl stuff - extracted from Thread.c,                ###
// ### including the /* comments */                                  ###
// ### This is based on version 5.005_03 - the same extraction       ###
// ### work should be applicable to later versions of Perl5,         ###
// ### compiled to USE_5005THREADS. Or even ITHREADS (threads.c) -   ### 
// ### with a few modifications to the way Perl is used by Kludge.   ###
// ### Unfortunately the extraction process would be difficult       ###
// ### to automate - essentially, I'd have to be parsing C code      ###
// ### and the semantics of Perl guts, which is infeasible with my   ###
// ### limited coding skills. If you need to adapt kludge to another ###
// ### version of Perl or even ITHREADS, you have to do it manually. ###

#include "XSUB.h"
/* Magic signature for Thread's mg_private is "Th" */ 
#define Thread_MAGIC_SIGNATURE 0x5468
#ifdef __cplusplus
#ifdef I_UNISTD
#include <unistd.h>
#endif
#endif
#include <fcntl.h>
                        
#ifndef THREAD_RET_TYPE
#define THREAD_RET_TYPE void *
#define THREAD_RET_CAST(x) ((THREAD_RET_TYPE) x)
#endif


// The following function is taken over from Thread.c -- *UNCHANGED*.
// Yes, indeed.
// The point is that it's easier for me to paste this little piece of code
// here than to patch the Perl source of Thread.c *not* to include the `static'
// keyword in the function declaration. 
// The `static' declaration causes that, even if I re-declared the function
// extern in this file (it's not in the headers), I wouldn't be able to access it
// from here. The symbol is not available in the original Thread.o, even though
// the compiled function is really there.
static void remove_thread(struct perl_thread *t)
{
#ifdef USE_THREADS
    DEBUG_S(WITH_THR(PerlIO_printf(PerlIO_stderr(),
                                   "%p: remove_thread %p\n", thr, t)));
    MUTEX_LOCK(&PL_threads_mutex);
    MUTEX_DESTROY(&t->mutex);
    PL_nthreads--;
    t->prev->next = t->next;
    t->next->prev = t->prev;
    COND_BROADCAST(&PL_nthreads_cond);
    MUTEX_UNLOCK(&PL_threads_mutex);
#endif
}



// In the original Thread.c (5.005 threads), there are two important 
// wrapper functions:
//
// 1) newthread(), a wrapper around pthread_create(), running in the 
//    parent thread
//
// 2) threadstart(), running in the child thread - a wrapper around the
//    Perl-space thread-body sub (the one that's passed to Thread::new).
//
// If you want your child threads (launched by your C++ program) to be able to
// act like Perl threads when dealing with perl stuff, you need the following:
// 
// - you need a single global interpreter (5.005 threads), initialized by
//   the parent thread before the children can start using perl (or even 
//   initialize)
//
// - you need to mimick the contents of newthread() and threadstart().
//
// The simplest way to mimick newthread() and threadstart() is to cannibalize
// their contents. At the same time, you want to do the thread startup and
// perl sub calls from C++ space in your own way. And, you don't want to
// clutter your C++ code with cannibalized snippets of Perl guts - at least,
// no more than absolutely necessary.
// The most logical way to achieve all this is by spliting newthread() and
// threadstart() into the leadin and leadout halves - the split point being
// the call to pthread_create() or perl_call_sv(), respectively.
// You can throw away most of the Perl-guts-wise parameter passing and stack
// manipulation. The thread is not started from Perl space anymore.
// You do need to pass Perl-guts-wise parameters to the Perl-space subs that
// you'll be calling from C++ space within a particular child thread - a simple 
// C++ space wrapper function would come in handy.


// We need to pass the current Perl Thread as an argument to threadstart_LeadIn()
// and threadstart_LeadOut(). The thread is created by newthread_LeadIn - you
// need to pass it into the child thread as part of your custom per-thread
// struct*, that's given as the last parameter to pthread_create().
// Or in some other way - that's up to you, the user of this library.

int threadstart_LeadIn(Thread arg)
{
    Thread thr = arg;

    DEBUG_S(PerlIO_printf(PerlIO_stderr(),
                          "new thread %p waiting to start\n",
			  thr));

    /* Don't call *anything* requiring dTHR until after SET_THR() */
    /*
     * Wait until our creator releases us. If we didn't do this, then
     * it would be potentially possible for out thread to carry on and
     * do stuff before our creator fills in our "self" field. For example,
     * if we went and created another thread which tried to JOIN with us,
     * then we'd be in a mess.
     */
    MUTEX_LOCK(&thr->mutex);
    MUTEX_UNLOCK(&thr->mutex);

    /*
     * It's safe to wait until now to set the thread-specific pointer
     * from our pthread_t structure to our struct perl_thread, since
     * we're the only thread who can get at it anyway.
     */
    SET_THR(thr);

    /* Only now can we use SvPEEK (which calls sv_newmortal which does dTHR) */
    DEBUG_S(PerlIO_printf(PerlIO_stderr(), "new thread %p starting at %s\n",
			  thr, SvPEEK(TOPs)));

    return(0);
}    



int threadstart_LeadOut(Thread arg)
{
    Thread thr = arg;
    SET_THR(thr);

    if (SvCUR(thr->errsv)) {
	MUTEX_LOCK(&thr->mutex);
	thr->flags |= THRf_DID_DIE;
	MUTEX_UNLOCK(&thr->mutex);
	DEBUG_S(PerlIO_printf(PerlIO_stderr(), "%p died: %s\n",
			      thr, SvPV(thr->errsv, n_a)));
    }
    // else { }
#if 0    
    /* removed for debug */
    SvREFCNT_dec(PL_curstack);
#endif
    SvREFCNT_dec(thr->cvcache);
    SvREFCNT_dec(thr->threadsv);
    SvREFCNT_dec(thr->specific);
    SvREFCNT_dec(thr->errsv);
    SvREFCNT_dec(thr->errhv);

    /*Safefree(cxstack);*/
    while (PL_curstackinfo->si_next)
	PL_curstackinfo = PL_curstackinfo->si_next;
    while (PL_curstackinfo) {
	PERL_SI *p = PL_curstackinfo->si_prev;
	SvREFCNT_dec(PL_curstackinfo->si_stack);
	Safefree(PL_curstackinfo->si_cxstack);
	Safefree(PL_curstackinfo);
	PL_curstackinfo = p;
    }    
    Safefree(PL_markstack);
    Safefree(PL_scopestack);
    Safefree(PL_savestack);
    Safefree(PL_retstack);
    Safefree(PL_tmps_stack);
    Safefree(PL_ofs);

    SvREFCNT_dec(PL_rs);
    SvREFCNT_dec(PL_nrs);
    SvREFCNT_dec(PL_statname);
    Safefree(PL_screamfirst);
    Safefree(PL_screamnext);
    Safefree(PL_reg_start_tmp);
    SvREFCNT_dec(PL_lastscream);
    SvREFCNT_dec(PL_defoutgv);

    MUTEX_LOCK(&thr->mutex);
    DEBUG_S(PerlIO_printf(PerlIO_stderr(),
			  "%p: threadstart finishing: state is %u\n",
			  thr, ThrSTATE(thr)));

    // we're detached per definition
    ThrSETSTATE(thr, THRf_DEAD);
    MUTEX_UNLOCK(&thr->mutex);
    DEBUG_S(PerlIO_printf(PerlIO_stderr(),
                          "%p: DETACHED thread finished\n", thr));
    remove_thread(thr);	/* This might trigger main thread to finish */
    
    return(0); 
}



// We need to pass thr, oldmask and pthread_create() result code to the leadout.
// Please declare n_thread, oldmask and thr_perl_err in the calling scope.
// Like this:
//
// Thread n_thr;       // threaded perl stuff - gee, what about using C++ references?
// sigset_t oldmask;   // threaded perl stuff
// int thr_perl_err;   // threaded perl stuff
//
// thr_perl_err = newthread_LeadIn(&n_thr, &oldmask);
// pthread_create( some attributes my C++ code is using );
// newthread_LeadOut(thr_perl_err, &n_thr, &oldmask);

int newthread_LeadIn(Thread* n_thread, sigset_t* oldmask)
{
    dSP; // what? delete this, too?
    void* bogus_ptr = sp; bogus_ptr = NULL; // a crude hack by F.Rysanek to zap the "unused variable" warning
    Thread savethread;
    int err;
    static pthread_attr_t attr;
    static int attr_inited = 0;
    sigset_t fullmask;
    
    savethread = thr;
    thr = new_struct_thread(thr);
    *n_thread = thr;        // added by F.Rysanek
    /* temporarily pretend to be the child thread in case the
     * XPUSHs() below want to grow the child's stack.  This is
     * safe, since the other thread is not yet created, and we
     * are the only ones who know about it */
    SET_THR(thr);
    DEBUG_S(PerlIO_printf(PerlIO_stderr(),
			  "%p: newthread (%p), tid is %u, preparing stack\n",
			  savethread, thr, thr->tid));

    /* On your marks... */
    SET_THR(savethread);
    MUTEX_LOCK(&thr->mutex);

    /* Get set...  */
    sigfillset(&fullmask);
    if (sigprocmask(SIG_SETMASK, &fullmask, oldmask) == -1)
	croak("panic: sigprocmask");
    err = 0;
    if (!attr_inited) {
	attr_inited = 1;
#ifdef OLD_PTHREADS_API
	err = pthread_attr_create(&attr);
#else
	err = pthread_attr_init(&attr);
#endif
#ifdef OLD_PTHREADS_API
#ifdef VMS
/* This is available with the old pthreads API, but only with */
/* DecThreads (VMS and Digital Unix) */
	if (err == 0)
	    err = pthread_attr_setdetach_np(&attr, ATTR_JOINABLE);
#endif
#else
	if (err == 0)
	    err = pthread_attr_setdetachstate(&attr, ATTR_JOINABLE);
#endif
    }

    return(err);
}



int newthread_LeadOut(int err, Thread* n_thread, sigset_t* oldmask)
{
    dSP;                            // Added by F.Rysanek
    void* bogus_ptr = sp; bogus_ptr = NULL; // a crude hack to zap the "unused variable" warning
    thr = *n_thread;                // Added by F.Rysanek

    if (err) {
	MUTEX_UNLOCK(&thr->mutex);
        DEBUG_S(PerlIO_printf(PerlIO_stderr(),
			      "%p: create of %p failed %d\n",
			      savethread, thr, err));
	/* Thread creation failed--clean up */
	SvREFCNT_dec(thr->cvcache);
	remove_thread(thr);
	MUTEX_DESTROY(&thr->mutex);
	return(1);
    }

#ifdef THREAD_POST_CREATE
    THREAD_POST_CREATE(thr);
#else
    if (sigprocmask(SIG_SETMASK, oldmask, 0))
	croak("panic: sigprocmask");
#endif

    /* Go */
    MUTEX_UNLOCK(&thr->mutex);

    return(0);
}

// So much for the cannibalized leadin and leadout code.






// ### The following function inits the global perl interpreter and,  ###
// ### while it's at it, also throws in some necessary bootstrapping. ###

int Kludge_pl_init()
{
 int result = 0;
 char* Kludge_pl_args[] = { "kludge_perl", "-e", "0;" };

   Kludge_pl = perl_alloc();             // Allocate the global interpreter
   perl_construct(Kludge_pl);            // Initialize the global interpreter

   // We can do without an external file being loaded and parsed.
   // To get rid of it, we need the "-e" parameter (read the script code
   // from the next cmdline argument) and we can pass it the shortest possible
   // script, which is "0;" (means: exit with a zero result code).
   if ( perl_parse(Kludge_pl, xs_init, 3, Kludge_pl_args, NULL) != 0 )
   {
      cout << "Parsing of the minimalistic input script has failed.";
      result = 1;
   }
   else // no problem - let's rave on
   {
      // The following piece of code performs a Perl-space init of
      // a statically linked module Kludge_pl and `uses' module Thread - which does
      // the same bootstrap operation plus some additional stuff.
      if (Kludge_pl_eval // -- define code inline -- 
            (
               // Bootstrap Kludge_pl.pm:
               //"use Kludge_pl; "      // Originally I was using this instead of the same
                                        // command that would have to occur in user scripts.
                                        // All this does is: it loads Kludge_pl.pm
               "bootstrap Kludge_pl;"   // Now I don't even need Kludge_pl.pm :) :) :)

               // Prepend the private Perl build into the Perl's include path:
               "sub BEGIN"              // see documentation about BEGIN blocks 
               "{"                      //  - this block is run at compile-time :)
               "   unshift (@INC, \"./perl5.005_03/ext/Thread\");"
               "   unshift (@INC, \"./perl5.005_03/lib\");"
               "}"

               // Now that we have the include path correct, we can load Thread.pm
               // and the modules it loads internally (DynaLoader and maybe more).
               "use Thread;"            // There's more to Thread.pm than just the bootstrap

               // If we ever needed to set some global variables before
               // the "user" scripts would run, we could do it here.

            ) == -1  // can return a zero or a small positive int - the number of results
         )
      {
         cout << "Bootstrap eval unsuccessful.\n";
         result = 1;
      }
   }

   // ### The following sub does some per-thread perl init and shutdown         ###
   // ### - specifically, it arranges a `pseudo-local' IO handle for the client ###
   // ### socket served.                                                        ###
   Kludge_pl_thr_wrapper = Kludge_pl_eval_sub
         (
            // $CLIENT_SOCKET is declared `my' silently upon eval - see Kludge_pl_eval_sub()
            "   $CLIENT_SOCKET = Kludge_pl::create_thr_socket_handle();\n" // get an "anon" GVref

            "   Kludge_pl::run_Kludge_CLI();\n" // THIS IS THE CLI OF THE CURRENT THREAD.
            
            "   close($CLIENT_SOCKET);\n"  // done automagically upon SvREFCNT_dec(GV), anyway
            "   Kludge_pl::destroy_thr_socket_handle();\n" // calls SvREFCNT_dec on the GV
            // The socket handle GV really vanishes when $CLIENT_SOCKET goes out of scope
            // - precisely when the last (lexical) reference is implicitly removed.

            // As the GV containing the handle is destroyed anyway when the scope closes,
            // and closing a Perl-space file handle also silently closes the underlying
            // file descriptor, it is not necessary to close the original (int) socket.
         );
   if (Kludge_pl_thr_wrapper == NULL)
   {
      cout << "The `Perl init, CLI run, Perl shutdown' eval unsuccessful.\n";
      result = 1;
   }

   // In this setup, we don't need to run the script at all.
   // Yeah, right - if we did, we wouldn't do it at the init stage.
   //perl_run(Kludge_pl);                             // Run

   return(result);
}



// A truly mysterious function - no documentation available.
// Makes me wonder what this function is doing.         ;)
void Kludge_pl_shutdown()
{
   perl_destruct(Kludge_pl);
   perl_free(Kludge_pl);
}



// A helper - we also need to know the Thread struct of the default thread.
Thread Kludge_pl_get_deflt_thread()
{
   dSP; // so that `thr' gets declared
   void* bogus_ptr = sp; bogus_ptr = NULL; // a crude hack to zap the "unused variable" warning
   // it would certainly make things faster if I declared thr manually.

   return(thr);
}



// A wrapper around Kludge_CLI::Run(), used to set up some perl-space environment.
// See Kludge_pl_init() for its perl-space contents.
// I've found out that this cannot be e.g. split into two eval's (lead-in and
// lead-out) - each eval inherently creates its own scope, as a result of which
// any `local' or `my' variables vanish at the end of the eval. Thus, any
// attempt to initialize thread-private data in this way is automatically void.
int Kludge_pl_thr_Run()
{
   return Kludge_pl_call_CV(Kludge_pl_thr_wrapper, NULL);
}



// A simple wrapper around SvTYPE - reports the SV type as a character string.
string* decipher_SV_type(SV* data)
{
 string* result = NULL;
 ostrstream tmp_ss;

   if (data != NULL)
   {
      switch (SvTYPE(data))
      {
         case SVt_NULL:  result = new string("a NULL value (null SV* pointer?)");
         break;
         case SVt_IV:    result = new string("an integer");
         break;
         case SVt_NV:    result = new string("a float");
         break;
         case SVt_RV:    result = new string("a reference");
         break;
         case SVt_PV:    result = new string("a character string");
         break;
         case SVt_PVIV:  result = new string("a pointer value to an integer - whatever that is");
         break;
         case SVt_PVNV:  result = new string("a pointer value to a float - whatever that is");
         break;
         case SVt_PVMG:  result = new string("a blessed or magical scalar");
         break;
         case SVt_PVBM:  result = new string("a pointer value to a BM - whatever that is");
         break;
         case SVt_PVLV:  result = new string("a pointer value to an LV - whatever that is");
         break;
         case SVt_PVAV:  result = new string("an array");
         break;
         case SVt_PVHV:  result = new string("a hash");
         break;
         case SVt_PVCV:  result = new string("a code value");
         break;
         case SVt_PVGV:  result = new string("a glob (possibly a file handle)");
         break;
         case SVt_PVFM:  result = new string("a format");
         break;
         case SVt_PVIO:  result = new string("an IO handle");
         break;
         default: 
         {
            tmp_ss << "an unknown value type: " << SvTYPE(data) << ends;
            result = new string(tmp_ss.str());
         } 
         break;
      }
   }
   else result = new string("-- SvTYPE() obtained a null pointer!!! --");

   return(result);
}

