#include "kludge_engine.h"
#include "kludge_threads.h"


// funny thing - the MYSQL_ROW returned by mysql_fetch_row()
// needn't (actually, must not) be manually deallocated,
// as it's really only a pointer to a structure within the
// respective MYSQL_RES - which must be released using
// mysql_free_result().

Kludge_Root_Object* Kludge_Root = NULL;

// mandatory definition of static members:
GTree* Kludge_Class::By_ID = g_tree_new(Kludge_Compare_IDs);
GTree* Kludge_Class::By_Name = g_tree_new(Kludge_Compare_Names);
GTree* Kludge_Class_Relationship::Index = g_tree_new(Kludge_Compare_IDs);
GTree* Kludge_Instance::Index = g_tree_new(Kludge_Compare_IDs);



#define MY_SSCANF(CHAR_PTR, VAR_TO_READ) \
      tmp_ss = new istrstream(CHAR_PTR); \
      *tmp_ss >> VAR_TO_READ; \
      delete tmp_ss; tmp_ss = NULL;




/* ###    class Kludge_Object (abstract)    ### */


// You are required to provide a function body for
// pure virtual destructors (despite them being pure virtual).
Kludge_Object::~Kludge_Object() {}
   


int Kludge_Object::operator>(const Kludge_Object& rhs) const
{
 int result = false;
 Kludge_Object* tmp_obj = (Kludge_Object*)&rhs;

   if (this != tmp_obj)
   {
      if (tmp_obj != Kludge_Root)
      {
         while (tmp_obj != Kludge_Root)
         {
            tmp_obj = tmp_obj->Superior();
            if (tmp_obj == this)
            {
               result = true;
               break;
            }
            // else let's search on - if no match is found until 
            // the root is met, the object &RHS is not subordinate to `this'.
         }
      }
      // else the object &rhs is the Kludge_Root => it is certainly
      // greater than `this'. => the result is false.
   }
   // else the two pointers are equal, pointing to a single object 
   //  => the result is false.

   return(result);
}



int Kludge_Object::operator>=(const Kludge_Object& rhs) const
{
   if (this == &rhs) return(true);
   else
   {
      return( *this > rhs );
   }
}



int Kludge_Object::operator<(const Kludge_Object& rhs) const
{
   return( rhs > *this );
}



int Kludge_Object::operator<=(const Kludge_Object& rhs) const
{
   if (this == &rhs) return(true);
   else return( *this < rhs );
}







/* ###    class Kludge_Class    ### */


// This constructor is used for creating a new class from scratch.
// It cannot be used for loading a class from the DB.
// It actually stores the newly created class in the DB.
//
Kludge_Class::Kludge_Class(
  string* class_name,
  Kludge_Class* parent_class,
  Kludge_CLI* CLI
) :
  Class(0),
  Name(*class_name),
  Parent(parent_class),
  Children(NULL),
  Properties(NULL),
  Inherited_Props(NULL),
  Relationships(NULL),
  Inherited_Rels(NULL),
  Instances(NULL),
  ctor_failed(false),
  relationships_loaded(false),
  inheritance_done(false)
{
 int stor_success = 0;
 ostrstream qry_string;

   Type = KLUDGE_CLASS;

   if ( Parent->Children == NULL ) Parent->Load_Children(CLI);
   
   qry_string << "INSERT INTO Classes VALUES (NULL,\""
              << Name
              << "\",\""
              << Parent->Class
              << "\")"
              <<  ends;
   stor_success = CLI->thr->DB_Cmd(qry_string.str());
   if (stor_success == 0)
   {
      Class = mysql_insert_id(CLI->thr->DB);
      Add_To_Index();
   }
   else
   {
      ctor_failed = true;
   }
}



// This constructor can only be used for loading an existing Kludge_Class
// from a database. It is not intended for creating a new class from scratch.
//
Kludge_Class::Kludge_Class(
  const char *class_name,
  Kludge_Class* parent_class,
  Kludge_CLI* CLI
) : 
  Parent(parent_class),
  Children(NULL),
  Properties(NULL),
  Inherited_Props(NULL),
  Relationships(NULL),
  Inherited_Rels(NULL),
  Instances(NULL),
  ctor_failed(false),
  relationships_loaded(false),
  inheritance_done(false)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;

   Type = KLUDGE_CLASS;
   qry_string << "SELECT Class,Name FROM Classes WHERE Name = \""
              << class_name
              << "\""
              <<  ends;
   mysql_query_result = CLI->thr->DB_Query(qry_string.str());

   if (( result_row = mysql_fetch_row(mysql_query_result) ))
   {
      cout << "Got a class body to read, found by Name.\n";
      if (! Extract_From_MysqlRow(result_row) )
      {
         ctor_failed = true;
      }
      else Add_To_Index();
   }
   else
   {
      ctor_failed = true;
   }
   mysql_free_result(mysql_query_result);
}



// This constructor can only be used for loading an existing Kludge_Class
// from a database. It is not intended for creating a new class from scratch.
//
Kludge_Class::Kludge_Class(
  OID_Type ClassID,
  Kludge_Class* parent_class,
  Kludge_CLI* CLI
) :
  Parent(parent_class),
  Children(NULL),
  Properties(NULL),
  Inherited_Props(NULL),
  Relationships(NULL),
  Inherited_Rels(NULL),
  Instances(NULL),
  ctor_failed(false),
  relationships_loaded(false),
  inheritance_done(false)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;

   Type = KLUDGE_CLASS;
   qry_string << "SELECT Class,Name FROM Classes WHERE Class = \""
              << ClassID
              << "\""
              <<  ends;
   mysql_query_result = CLI->thr->DB_Query(qry_string.str());

   if (( result_row = mysql_fetch_row(mysql_query_result) ))
   {
      cout << "Got a class body to read, found by ClassID.\n";
      if (! Extract_From_MysqlRow(result_row) )
      {
         ctor_failed = true;
      }
      else Add_To_Index();
   }
   else
   {
      ctor_failed = true;
   }
   mysql_free_result(mysql_query_result);
}



// This constructor can only be used for loading an existing Kludge_Class
// from a database. It is not intended for creating a new class from scratch.
//
Kludge_Class::Kludge_Class(
  MYSQL_ROW DB_record,
  Kludge_Class* parent_class,
  Kludge_CLI* CLI
) : 
  Parent(parent_class),
  Children(NULL),
  Properties(NULL),
  Inherited_Props(NULL),
  Relationships(NULL),
  Inherited_Rels(NULL),
  Instances(NULL),
  ctor_failed(false),
  relationships_loaded(false),
  inheritance_done(false)
{
   Type = KLUDGE_CLASS;

   if ( ! Extract_From_MysqlRow(DB_record) )
   {
      ctor_failed = true;
   }
   else Add_To_Index();
}



// Careful - a Kludge_Class should only be deleted from within
// Kludge_Class::Free_Children() - otherwise use Kludge_Class::Zap(),
// which also deletes the respective entry from the parent's children
// list (GTree). Kludge_Class::Free_Children uses automated tree traversal
// - thus, using Kludge_Class::Zap(), it would delete tree entries
// while traversing, which would cause incomplete traversal even
// with Post-Order, as the tree is automatically re-balancing itself
// upon any modification...
//
// This dtor also removes the related entries from the two global class
// indexes (by name and by ID). Therefore, neither this nor
// Kludge_Class::Zap() can be used to release all the classes from
// within a traverse function launched unto one of the main indexes.
// To release all the classes, delete the root class.
//
Kludge_Class::~Kludge_Class()
{
   cout << "Deallocating class: " << Name << "\n" ;
   Free_Properties();
   Free_Children();
   Free_Relationships();
   Free_Instances();
   // we have to be able to delete while traversing Children
   // ==> no Free_From_Index() here. Zap() does that for us.
   Free_From_Global_Index();
}



// Zap() does a nice memory cleanup.
// Only makes sense 
// a) for dynamic partial release from memory 
//     In that case, invalidate the list of children.
// b) when a class is deleted
//     Requires deletion from the SQL-based object repository
//      - i.e., use DB_Delete() and then Zap() (or just Delete())
int Kludge_Class::Zap()
{
 int result = 0;

   cout << "Zapping a class: " << Name << "\n" ;

   if (Instances != NULL)
   {
      if (g_tree_nnodes(Instances) > 0)
      {
         cout << "Cannot zap CLASS " << Name << "\n" <<
            " There are dependent instance-side objects loaded in memory.\n";
         result = 1;
      }
   }
   
   if (result == 0)
   {
      if (Parent != NULL)
      {
         if (Parent->Children != NULL)
         {
            Free_From_Index();
         }
         else
         {
            cout <<
               "Internal error - in Kludge_Class::Zap(): parent's Children index not initialized!\n";
            result = 1;
         }

         delete this;
      }
      else
      {
         if (Name == "Root_Class")
         {
            cout << 
               "In Kludge_Class::Zap(): cannot delete the Root Class.\n";
            result = 1;
         }
         else
         {
            cout << 
               "In Kludge_Class::Zap(): zapping class " << Name << ".\n"
               " This class doesn't seem to have a parent!\n";

            delete this;
            result = 1;
         }
      }
   }
   
   return(result);
}



int Kludge_Class::Load_Children(Kludge_CLI* CLI, bool recursive = false)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;
 Kludge_Class *New_Class;
 int result = 0;

   if ( (Class != 0) && (Children == NULL) )
   {
      Children = g_tree_new(Kludge_Compare_Names);

      qry_string << "SELECT Class,Name FROM Classes WHERE Parent = \""
                 << Class
                 << "\""
                 <<  ends;
      mysql_query_result = CLI->thr->DB_Query(qry_string.str());
      // add a not-NULL check here !

      while (( result_row = mysql_fetch_row(mysql_query_result) ))
      {
         New_Class = new Kludge_Class(result_row, this, CLI);
         if (New_Class != NULL)
         {
            if (! New_Class->ctor_failed)
            {
               if (recursive)
               {
                  New_Class->Load_Children(CLI, /*recursive?*/true);
               }
            }
            else
            {
               cout << "Creation of a child class from a MYSQL_ROW failed,"
                       " although raw malloc() of the class succeeded.\n" ;
               delete New_Class; New_Class = NULL;
               result = 1;
            }
         }
         else
         {
            cout << "Creation of a child class from a MYSQL_ROW failed,"
                    " probably on malloc() of the new class.\n" ;
            result = 1;
         }
      }
      mysql_free_result(mysql_query_result);
   }
   else
   {
      cout << "In Kludge_Class::Load_Children():\n"
      << "the ClassID is equal to zero,\n"
      << "or the Children are already loaded.\n";
      result = 1;
   }

   return(result);
}



bool Kludge_Class::Free_Children()
{
   // I tried zapping the children within the traverse
   // function, but that wouldn't work, not even using post-order
   // - there were skipped entries.
   // (in-order or pre-order gave segmentation faults)
   // If I just deallocate the class structs pointed to, i.e.
   // including the tree keys, everything seems to work, as long
   // as I don't enter/delete/update records during the traversal,
   // which obviously I don't.
   // Still, I'm traversing the tree on keys that I'm deallocating
   // in the process - this might potentially mean a PROBLEM,
   // especially with multi-threading, where the RAM space released
   // may get reused before the tree traversal is finished.
   // For the moment, let me rely on the hypothesis that the tree
   // traversal first decides the order in which to enter the
   // two leaves of a particular btree node and then enter them 
   // without ever looking at the keys again.
   // If this technique starts to cause havoc under load, the only
   // solution is to lock the Kludge Class, stack all the children
   // pointers into a <vector>, destroy the tree and release the
   // children based on the <vector>, rather than the original tree.

   if (Children != NULL)
   {
      g_tree_traverse(          // Delete all the children
            Children,
            Kludge_Class::Free_Children_Traverse,
            G_POST_ORDER,
            NULL);

      g_tree_destroy(Children);
      Children = NULL;
   }
   // else children not loaded - not an error.

   return(true);
}



int Kludge_Class::Load_Properties(Kludge_CLI* CLI, bool initiate_inheritance = true)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;
 Kludge_Class_Property *New_Class_Property;
 int result = 0;

   if (Properties == NULL)
   {
      Properties = g_tree_new(Kludge_Compare_Names);
      qry_string << "SELECT RDN,Value,Suppress,Flags FROM Class_Properties WHERE Class = \""
                 << Class
                 << "\""
                 <<  ends;
      mysql_query_result = CLI->thr->DB_Query(qry_string.str());

      while (( result_row = mysql_fetch_row(mysql_query_result) ))
      {
         New_Class_Property = new Kludge_Class_Property(result_row, this, CLI);
         if (New_Class_Property != NULL)
         { 
            if (New_Class_Property->ctor_failed)
            {
               cout << "Creation of a class property from a MYSQL_ROW failed,"
                       " for reasons other than raw class malloc().\n" ;
               delete New_Class_Property; New_Class_Property = NULL;
               result = 1;
            }
         }
         else
         {
            cout <<
               "Error: in Kludge_Class::Load_Properties():\n"
               " creation of a class property from a MYSQL_ROW failed.\n" ;
            result = 1;
         }
      }
      mysql_free_result(mysql_query_result);

      if (initiate_inheritance) Inherit(CLI);
   }
   else
   {
      cout << "In Kludge_Class::Load_Properties:\n"
           << " properties already loaded!\n";
   }

   return(result);
}



bool Kludge_Class::Free_Properties()
{
   if (Properties != NULL)
   {
      g_tree_traverse(          // Delete all the properties
               Properties,
               Kludge_Class::Free_Properties_Traverse,
               G_POST_ORDER,
               NULL);

      g_tree_destroy(Properties);
      Properties = NULL;
   }
   // else properties not loaded

   if (Inherited_Props != NULL)
   {
      g_tree_destroy(Inherited_Props);
      Inherited_Props = NULL;
   }
   // else inheritance not done yet

   return(true);
}



int Kludge_Class::Init_Relationships()
{
   if (Relationships == NULL)
   {
      Relationships = g_tree_new(Kludge_Compare_Names);
   }
   // else already initialized - nothing to be done.

   return(0);
}



int Kludge_Class::Init_Instances()
{
   if (Instances == NULL)
   {
      Instances = g_tree_new(Kludge_Compare_IDs);
   }
   // else already initialized - nothing to be done.

   return(0);
}



int Kludge_Class::Load_Relationships(Kludge_CLI* CLI, bool initiate_inheritance = true)
{
 int result = 0;
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;
 Kludge_Class_Relationship* new_cls_rel = NULL;
 istrstream *tmp_master = NULL, *tmp_slave = NULL;
 OID_Type tmp_master_ID = 0, tmp_slave_ID = 0;
 string tmp_name_M, tmp_name_S;

   if (! relationships_loaded)
   {
      // If the relationships are partially loaded, the easiest way to prevent
      // cloned relationship entries in memory is to free any previously loaded
      // relationships and reload from scratch.
      // This way however, the relationships index at the other peer of any
      // previously loaded relationship would get invalidated upon loading
      // full index at our end - thus, the approach described above is a no-go.
      // We must not delete previously loaded rels from an incomplete index.

      if (Relationships == NULL) Init_Relationships();

      if (Relationships != NULL)
      {
         qry_string << 
            "SELECT * FROM Class_Relationships WHERE Master = \""
            << Class 
            << "\" OR Slave = \""
            << Class 
            << "\""
            <<  ends;
            
         mysql_query_result = CLI->thr->DB_Query(qry_string.str());

         while (( result_row = mysql_fetch_row(mysql_query_result) ))
         {
            // check that this relationship is not loaded yet
            // first fetch some data to check from the Mysql Row
            tmp_master = new istrstream(result_row[1]);
            tmp_slave = new istrstream(result_row[2]);
            *tmp_master >> tmp_master_ID;
            *tmp_slave >> tmp_slave_ID;
            tmp_name_M = result_row[4];
            tmp_name_S = result_row[5];

            if // check based on relative name and Master/Slave CLSID's
            (!(
               (( tmp_master_ID == Class ) && (Lookup_Relationship_Raw(&tmp_name_M,CLI) != NULL))
               || (( tmp_slave_ID == Class ) && (Lookup_Relationship_Raw(&tmp_name_S,CLI) != NULL))
            ))
            {
               // relationship not loaded yet
               new_cls_rel = new Kludge_Class_Relationship(result_row, CLI);
               if (new_cls_rel != NULL)
               { 
                  if (new_cls_rel->ctor_failed)
                  {
                     cout << "Creation of a class relationship from a MYSQL_ROW failed,"
                             " for reasons other than raw malloc() failure.\n" ;
                     delete new_cls_rel; new_cls_rel = NULL;
                     result = 1;
                  }
                  // else OK, relationship loaded
               }
               else
               {
                  cout <<
                     "Error: in Kludge_Class::Load_Relationships():\n"
                     " creation of a class relationship from a MYSQL_ROW failed.\n" ;
                  result = 1;
               }
            }
            // else relationship already loaded - skip it.

            // clean up the local test variables for another cycle
            tmp_master_ID = 0;
            tmp_slave_ID = 0;
            tmp_name_M = "";
            tmp_name_S = "";
            delete tmp_master; tmp_master = NULL;
            delete tmp_slave; tmp_slave = NULL;
         }

         mysql_free_result(mysql_query_result);
         
         if (result == 0)
         {
            relationships_loaded = true;
            if (initiate_inheritance) Inherit(CLI);
         }
      }
      else
      {
         cout << "In Kludge_Class::Load_Relationships():\n"
              << " couldn't Init_Relationships().\n";
         result = 1;
      }
   }
   else
   {
      cout << "In Kludge_Class::Load_Relationships():\n"
           << " relationships already loaded!\n";
   }

   return(result);
}



bool Kludge_Class::Free_Relationships()
{
 vector<Kludge_Class_Rel_Stub*> rels_to_free;
 GTree* tmp_duplicates_index = NULL;
 bool skip_zap_for_now = false;

   if (Relationships != NULL)
   {
      // First, stuff all the relationship pointers into a throwaway vector
      // (we can't zap them while traversing the btree).
      // No matter if they're completely loaded or not - we're just clearing up space on the heap.
      g_tree_traverse(
               Relationships,
               Kludge_Class::Free_Relationships_Traverse,
               G_POST_ORDER,
               &rels_to_free);

      // We need to keep track of self-referral relationships.
      // Such relationships would have two index entries, both of which
      // will be zapped in a single run of Zap(). An attempt to Zap() 
      // the second entry in our derived throwaway vector would cause a SIGSEGV.
      tmp_duplicates_index = g_tree_new(Kludge_Compare_Names);
      
      for (int i=rels_to_free.size()-1; i>=0; i--) // now zap all the relationships one by one
      {
         skip_zap_for_now = false;
            
         // detect self-referral relationships
         if (rels_to_free[i]->Body->Master_Stub.Owner == rels_to_free[i]->Body->Slave_Stub.Owner)
         {
            // check that it's not already in the duplicates index
            if (g_tree_lookup(tmp_duplicates_index, &rels_to_free[i]->Name) == NULL)
            {
               // OK, we didn't know about this one yet.
               // For now, let's just write it down into the index of duplicates
               // - we'll zap it later, when we encounter the other end.
               skip_zap_for_now = true;

               // find the other end (stub) and store it in the duplicates index
               if (rels_to_free[i] == &rels_to_free[i]->Body->Master_Stub)
               {
                  g_tree_insert(
                     tmp_duplicates_index,
                     &rels_to_free[i]->Body->Slave_Stub.Name,
                     &rels_to_free[i]->Body->Slave_Stub
                  );
               }
               else if (rels_to_free[i] == &rels_to_free[i]->Body->Slave_Stub)
               {
                  g_tree_insert(
                     tmp_duplicates_index,
                     &rels_to_free[i]->Body->Master_Stub.Name,
                     &rels_to_free[i]->Body->Master_Stub
                  );
               }
               else
               {
                  cout <<
                     "Internal error - in Kludge_Class::Free_Relationships():\n"
                     " self-referencing relationship encountered, but couldn't detect the other end.\n";
               }
            }
            else
            {
               // This self-referral relationship has already been noticed.
               // Therefore we can zap it by now. But before we do that, we
               // remove it from the index of duplicates (BTree) to prevent
               // its corruption.
               g_tree_remove(tmp_duplicates_index, &rels_to_free[i]->Name);
            }
         }

         if (skip_zap_for_now = false)
         {
            // Detection of duplicates done - now zap the current relationship.

            // Before actually clearing the struct from memory, invalidate the 
            // list of relationships at the peer Kludge Class
            if (rels_to_free[i] == &rels_to_free[i]->Body->Master_Stub)
            {
               rels_to_free[i]->Body->Slave_Stub.Owner->relationships_loaded = false;
            }
            else if (rels_to_free[i] == &rels_to_free[i]->Body->Slave_Stub)
            {
               rels_to_free[i]->Body->Master_Stub.Owner->relationships_loaded = false;
            }
            else
            {
               cout <<
                  "Internal error - in Kludge_Class::Free_Relationships():\n"
                  " couldn't detect the other end.\n";
            }
            rels_to_free[i]->Zap();
         }   
      }
      // Done with the zapping.
      g_tree_destroy(tmp_duplicates_index);

      // As we need the relationships tree to exist while Zap()ping,
      // we can only destroy it after we've iterated through the throwaway vector.
      g_tree_destroy(Relationships);
      Relationships = NULL;
      relationships_loaded = false;
   }
   // else relationships not loaded

   if (Inherited_Rels != NULL)
   {
      g_tree_destroy(Inherited_Rels);
      Inherited_Rels = NULL;
   }
   // else inheritance not done yet

   return(true);
}



bool Kludge_Class::Free_Instances()
{
   if (Instances != NULL)
   {
      if (g_tree_nnodes(Instances) > 0)
      {
         cout << "In Kludge_Class::Free_Instances():\n"
                 " there are still instances left on the heap!\n";
         // I don't plan to free instances recursively from their classes.
         // On shutdown, instances should be torn down first, then classes.
         // I guess this is more civilized.
         // Let's make this a serious warning.
      }
      
      g_tree_destroy(Instances);
      Instances = NULL;
   }
   // else instances not initialized - probably there was no reason

   return(true);
}



// This makes sense in lookup-focused scenarios, where this could
// become a means of lookup speed optimization... (second only to loading
// all the objects to memory on startup).
int Kludge_Class::ReInherit_Children(Kludge_CLI* CLI)
{
   if (Children != NULL)
   {
      g_tree_traverse(
                      Children, 
                      Kludge_Class::ReInherit_Children_Traverse,
                      G_IN_ORDER,
                      CLI
                     );
   }

   return(0);
}



// Build/rebuild your inheritance indexes - now.
int Kludge_Class::Inherit(Kludge_CLI* CLI)
{
 int result = 0, counter;
 vector<Kludge_Class_Property*> properties_to_inherit;
 vector<Kludge_Class_Rel_Stub*> relationships_to_inherit;

   cout << "Inherit()ing in class " << Name << endl << flush;

   if (Inherited_Props != NULL)
   {
      g_tree_destroy(Inherited_Props);
      Inherited_Props = NULL;
   }
   if (Inherited_Rels != NULL)
   {
      g_tree_destroy(Inherited_Rels);
      Inherited_Rels = NULL;
   }
   inheritance_done = false;
   // ^^^ so that ReInherit doesn't bother (to avoid an endless loop with Parent->Inherit())
   
   if (Properties == NULL) Load_Properties(CLI, /*call Inherit()?*/ false);
   if (! relationships_loaded) Load_Relationships(CLI, /*call Inherit()?*/ false);
   // ^^^ the boolean parameter "call Inherit" prevents mutual recursion

   if (Parent != NULL) 
   // we have a parent (base class) to inherit from - do a full inheritance procedure
   {
      if (! Parent->inheritance_done) Parent->Inherit(CLI);

      Inherited_Props = g_tree_new(Kludge_Compare_Names);
      Inherited_Rels = g_tree_new(Kludge_Compare_Names);

      // build an array of the Parent's properties (own and inherited)
      g_tree_traverse(
                      Parent->Properties, 
                      Kludge_Class::Build_Props_Array_Traverse,
                      G_IN_ORDER,
                      &properties_to_inherit
                     );
      g_tree_traverse(
                      Parent->Inherited_Props, 
                      Kludge_Class::Build_Props_Array_Traverse,
                      G_IN_ORDER,
                      &properties_to_inherit
                     );
      // now browse the array and select properties that are not overloaded here
      for (counter = properties_to_inherit.size()-1; counter >= 0; counter--)
      {
         if ( Resolve_Content_Name_Enh(&properties_to_inherit[counter]->RDN,
              CLI,/*own_only*/false,/*initiate_inheritance*/false) == NULL
            )
         {
            g_tree_insert(
                          Inherited_Props,
                          &properties_to_inherit[counter]->RDN,
                          properties_to_inherit[counter]
                         );
         }
         // else this inheritable property is overloaded in this class
      }

      // build an array of the Parent's relationships (own and inherited)
      g_tree_traverse(
                      Parent->Relationships, 
                      Kludge_Class::Build_Rels_Array_Traverse,
                      G_IN_ORDER,
                      &relationships_to_inherit
                     );
      g_tree_traverse(
                      Parent->Inherited_Rels, 
                      Kludge_Class::Build_Rels_Array_Traverse,
                      G_IN_ORDER,
                      &relationships_to_inherit
                     );
      // now browse the array and select relationships that are not overloaded here
      for (counter = relationships_to_inherit.size()-1; counter >= 0; counter--)
      {
         // This resolution mechanism should change once I implement explicit relationship
         // inheritance (chaining). The one below is a makeshift solution.
         if (Resolve_Content_Name_Enh(&relationships_to_inherit[counter]->Name,
              CLI,/*own_only*/false,/*initiate_inheritance*/false) == NULL)
         {
            g_tree_insert(
                          Inherited_Rels,
                          &relationships_to_inherit[counter]->Name,
                          relationships_to_inherit[counter]
                         );
         }
         // else this inheritable relationship is overloaded in this class
      }

      // done inheriting properties and relationships
      inheritance_done = true;
   }
   else
   // there's no parent - either *this is the root class or something's wrong
   {
      if (Name == "Root_Class")
      // *this is the root class - let's just found empty inheritance indexes
      // for the children to inherit from using the standard procedure.
      {
         Inherited_Props = g_tree_new(Kludge_Compare_Names);
         Inherited_Rels = g_tree_new(Kludge_Compare_Names);
         inheritance_done = true;
      }
      else
      {
         *CLI->output_stream <<
         "Internal error - in Kludge_Class::Inherit(): this Kludge Class has no parent.\n";
      }
   }
   
   // The following only makes sense in scenarios where partial loading is off
   // and all the data is loaded on startup:
   // If children are loaded, tell them all to rebuild their inheritance indexes.
   //ReInherit_Children(CLI);

   // With partial loading on, the following is sufficient:
   // If children are loaded, tell them all to invalidate their inheritance indexes.
   UnInherit_Children();

   return(result);
}



// Something has changed in your ancestry - rebuild your inheritance index,
// if you've already built it. If not, don't bother.
int Kludge_Class::ReInherit(Kludge_CLI* CLI)
{
 int result = 0;

   if (inheritance_done) Inherit(CLI);
   // Will clean up existing inherited contents if necessary.
   // Else the inheritance indexes are not authoritative - don't bother to rebuild.
   
   return(result);
}



int Kludge_Class::UnInherit_Children()
{
   if (Children != NULL)
   {
      g_tree_traverse(
                      Children, 
                      Kludge_Class::UnInherit_Children_Traverse,
                      G_IN_ORDER,
                      NULL
                     );
   }

   return(0);
}



// Something has changed in your ancestry. Invalidate your inheritance index,
// if you've already built it - and tell your children. If not, don't bother.
int Kludge_Class::UnInherit()
{
 int result = 0;

   cout << "UnInherit()ing in class " << Name << endl << flush;
   if (inheritance_done) 
   {
      inheritance_done = false;
      UnInherit_Children();
   }
   // else no need to bother telling the children
   
   return(result);
}



int Kludge_Class::Add_To_Index(Kludge_Class* Where = NULL)
// both to local and global
{
 int result = 0;
 if (Where == NULL) Where = Parent;

   cout << "Adding class " << Name << " to local and global indexes.\n";
   g_tree_insert(By_Name,&Name,this);
   g_tree_insert(By_ID,&Class,this);
   if (Where != NULL)
   {
      if (Where->Children != NULL)
      {
         g_tree_insert(Where->Children,&Name,this);
         Where->UnInherit();
      }
      else
      {
         cout << "Kludge_Class::Add_To_Index(): `brothers' not loaded.\n";
         result = 1;
      }
   }
   // else this class is probably the Root_Class (or something's broken)
   
   return(result);
}



int Kludge_Class::Free_From_Index(Kludge_Class* Where = NULL)
// from local index only
{
 int result = 0;
 if (Where == NULL) Where = Parent;

   cout << "Releasing class " << Name << " from local index.\n";
   if (Where != NULL)
   {
      if (Where->Children != NULL)
      {
         g_tree_remove(Where->Children,&Name);
         Where->UnInherit();
      }
      else
      {
         cout << "In Kludge_Class::Free_From_Index():\n"
                 " the parent's Children index is not initialized.\n";
         result = 1;
      }
   }
   // else the class is probably the Root_Class (or something's broken)
   
   return(result);
}



bool Kludge_Class::Free_From_Global_Index()
{
   cout << "Releasing class " << Name << " from global index.\n";
   g_tree_remove(By_Name,&Name);
   g_tree_remove(By_ID,&Class);
   return(true);
}



//
// Expects:  Row_To_Process[0] == Class ID, Row_To_Process[1] == Name
//
bool Kludge_Class::Extract_From_MysqlRow(MYSQL_ROW Row_To_Process) 
{
 bool result = true;
 istrstream* tmp_ss = NULL; // for MY_SSCANF

   if (Row_To_Process != NULL)
   {
      // Set Kludge_Class.Name (of type `string' - hence the direct assignment)
      Name = Row_To_Process[1];
      // Set Kludge_Class.Class
      MY_SSCANF(Row_To_Process[0],Class);
      //Add_To_Index(); //moved to the calling ctors - probably a better way

   }
   else
   {
      cout << 
         "In Kludge_Class::Extract_From_MysqlRow():\n"
         " the row passed to this function is a NULL pointer!\n"
         " Extract_From_MysqlRow() failed.\n";
      result = false;
   }

   return(result);
}



int Kludge_Class::List_Properties(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   if (Properties == NULL) Load_Properties(CLI);
   
   if (Properties != NULL)
   {
      if (! inheritance_done) Inherit(CLI); // Load_Properties should do that, too...
                                            // A little bit of defensive programming here
      if (inheritance_done)
      {
         if (recursive)
         {
            g_tree_traverse(
                            Properties, 
                            Kludge_Class::List_Properties_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
            g_tree_traverse(
                            Inherited_Props,
                            Kludge_Class::List_Properties_Inh_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
         }
         else
         {
            g_tree_traverse(
                            Properties, 
                            Kludge_Class::List_Properties_Names_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
            g_tree_traverse(
                            Inherited_Props,
                            Kludge_Class::List_Properties_Inh_Names_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
         }
      }
      else
      {
         cout << "Kludge_Class::List_Properties(): failed to Inherit()!\n";
         result = 1;
      }
   }
   else
   {
      cout << "Kludge_Class::List_Properties(): failed to load properties!\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class::List_Children(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   if (Children == NULL) Load_Children(CLI, /*recursive?*/false);
   
   if (Children != NULL)
   {
      if (recursive)
      {
         g_tree_traverse(
                         Children, 
                         Kludge_Class::List_Children_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
      else
      {
         g_tree_traverse(
                         Children, 
                         Kludge_Class::List_Children_Names_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
   }
   else
   {
      cout << "Kludge_Class::List_Children(): failed to load children!\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class::List_Relationships(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   if (! relationships_loaded) Load_Relationships(CLI);
   
   if (relationships_loaded)
   {
      if (! inheritance_done) Inherit(CLI);  // Load_Relationships should do that, too...
                                             // A little bit of defensive programming here
      if (inheritance_done)
      {
         if (recursive)
         {
            g_tree_traverse(
                            Relationships, 
                            Kludge_Class::List_Relationships_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
            g_tree_traverse(
                            Inherited_Rels, 
                            Kludge_Class::List_Relationships_Inh_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
         }
         else
         {
            g_tree_traverse(
                            Relationships, 
                            Kludge_Class::List_Relationship_Names_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
            g_tree_traverse(
                            Inherited_Rels, 
                            Kludge_Class::List_Relationship_Inh_Names_Traverse,
                            G_IN_ORDER,
                            CLI
                           );
         }
      }
      else
      {
         cout << "Kludge_Class::List_Relationships(): failed to Inherit()!\n";
         result = 1;
      }
   }
   else
   {
      cout << "Kludge_Class::List_Relationships(): failed to load relationships!\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
   *CLI->output_stream << "CLASS \"" << Name << "\"\n{\n";
   List_Properties(CLI, recursive);
   List_Children(CLI, recursive);
   List_Relationships(CLI, recursive);
   *CLI->output_stream << "};\n";

   return(0);
}



int Kludge_Class::Describe_Self(Kludge_CLI* CLI)
{
   *CLI->output_stream << "CLASS \"" << Name << "\";\n";

   return(0);
}
   


Kludge_Class* Kludge_Class::Lookup_Child(string* name_to_lookup, Kludge_CLI* CLI)
{
 Kludge_Class *result = NULL;

   if (Children == NULL)
   {
      Load_Children(CLI, /*recursive?*/false);
   }

   if (Children != NULL)
   {
      result = (Kludge_Class *) g_tree_lookup(Children,name_to_lookup);
   }
   else
   {
      cout << "Kludge_Class::Lookup_Child : Oops! Could not load children!\n";
   }

   return(result);
}



Kludge_Class_Property* Kludge_Class::Lookup_Property(string* name_to_lookup, Kludge_CLI* CLI,
                         bool own_only = false, bool initiate_inheritance = true)
{
 Kludge_Class_Property *result = NULL;

   if (Properties == NULL)
   {
      Load_Properties(CLI,initiate_inheritance);
   }

   if (Properties != NULL)
   {
      if ((! inheritance_done) && initiate_inheritance) Inherit(CLI);
      if (Inherited_Props != NULL)
      {
         result = (Kludge_Class_Property *) g_tree_lookup(Properties,name_to_lookup);
         if ((result == NULL) && (! own_only))
         {
            result = (Kludge_Class_Property *) g_tree_lookup(
                                                 Inherited_Props,name_to_lookup);
         }
      }
   }
   else
   {
      cout << "Kludge_Class::Lookup_Property() : Oops! Couldn't load properties!\n";
   }

   return(result);
}



Kludge_Class_Rel_Stub* Kludge_Class::Lookup_Relationship_Raw(string* name_to_lookup, Kludge_CLI* CLI)
{
 Kludge_Class_Rel_Stub* result = NULL;

   if (Relationships != NULL)
   {
      result = (Kludge_Class_Rel_Stub*) g_tree_lookup(Relationships,name_to_lookup);
   }
   else
   {
      cout << "Kludge_Class::Lookup_Relationship_Raw(): Oops! Relationships not initiated!\n";
   }

   return(result);
}



Kludge_Class_Rel_Stub* Kludge_Class::Lookup_Relationship(string* name_to_lookup, Kludge_CLI* CLI,
                         bool own_only = false, bool initiate_inheritance = true)
{
 Kludge_Class_Rel_Stub* result = NULL;

   if (! relationships_loaded)
   {
      Load_Relationships(CLI,initiate_inheritance);
   }

   if (Relationships != NULL)
   {
      result = Lookup_Relationship_Raw(name_to_lookup, CLI);

      if ((! inheritance_done) && initiate_inheritance) Inherit(CLI);
      if (Inherited_Rels != NULL)
      {
         if ((result == NULL) && (! own_only))
         {
            result = (Kludge_Class_Rel_Stub*) g_tree_lookup(
                                                 Inherited_Rels,name_to_lookup);
         }
      }
      else
      {
         if (initiate_inheritance)
         {
            cout << "Kludge_Class::Lookup_Relationship(): Oops! Couldn't inherit relationships!\n";
         }
         // else allright, we were not requested to initiate inheritance,
         // don't bother anymore
      }
   }
   else
   {
      cout << "Kludge_Class::Lookup_Relationship(): Oops! Couldn't load relationships!\n";
   }

   return(result);
}



Kludge_Object* Kludge_Class::Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
   return(Resolve_Content_Name_Enh(Content_Name, CLI, own_only, /*initiate_inherit?*/true));
}



// own_only: don't try a lookup in the inherited indexes 
// (of properties and relationships) - only lookup own contents.
//
// "initiate_inheritance": if own_only is false, i.e. we should
// check inherited contents as well, do a lookup only if the respective
// index is initialized and don't care about inheritance_done
// (i.e., don't care whether the indexes are up to date or not).
// This option was required for Kludge_Class::Inherit() to function.
//
Kludge_Object* Kludge_Class::Resolve_Content_Name_Enh(string* Content_Name, Kludge_CLI* CLI,
                                           bool own_only = false, bool initiate_inheritance = true)
{
 Kludge_Object* result = NULL;

   result = Lookup_Child(Content_Name, CLI);        // a child Kludge_Class ?
   if (result == NULL)
   {                                                // a Kludge_Class_Property ?
      result = Lookup_Property(Content_Name, CLI, own_only, initiate_inheritance);
   }
   if (result == NULL)
   {                                                // a Kludge_Class_Relationship ?
      result = Lookup_Relationship(Content_Name, CLI, own_only, initiate_inheritance);
   }
   if (result == NULL)
   {
      if (
            (*Content_Name == "PARENT")
            || (*Content_Name == "..")
         )                                      // a parent Kludge_Class reference?
      {
         result = Superior();
      }
      else if (*Content_Name == ".")             // a "this" style reference?
      {
         result = this;
      }
   }

   return(result);
}



// *obj_to_insert is owned by its parent Kludge_CLI_Command that is deleted by
// its respective Kludge_CLI after use. We shan't delete it here.
int Kludge_Class::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
 int result = 0;
 string* tmp_pwd = NULL;
 Kludge_Object* tmp_obj_ptr = NULL;
 Kludge_Path* tmp_path = NULL;
 Kludge_Class* new_class = NULL;
 istrstream* tmp_stream = NULL;
 Kludge_Class_Property* new_cls_prop = NULL;
 Kludge_Class_Relationship* new_cls_rel = NULL;
 byte_buf* tmp_byte_buf = NULL; // for BLOB handling
 
   switch (obj_to_insert->type)
   {
      case PARM_CLASS :
      {
         // the proposed class name should not be an empty string in the first place
         if (obj_to_insert->contents[0]->content.size() != 0)
         {
            // check if such a name exists under `this' class
            tmp_obj_ptr = Resolve_Content_Name(&obj_to_insert->contents[0]->content,CLI,/*own only?*/true);
            // if not, let's also check if it's not an existing class name in general
            if (tmp_obj_ptr == NULL)
            {
               tmp_obj_ptr = Name2ptr(&obj_to_insert->contents[0]->content,CLI);
            }
            
            if (tmp_obj_ptr == NULL) // no such class name exists yet
            {
               new_class = new Kludge_Class(&obj_to_insert->contents[0]->content,this,CLI);
               if (new_class != NULL)
               {
                  if (! new_class->ctor_failed)
                  {
                     for (int i = obj_to_insert->contents.size()-1; i>0; i--)
                     // obj_to_insert->contents[0] == <class name> - that's why the terminal
                     // condition reads i>0, rather than i>=0. We simply don't want to iterate
                     // through the zero'th parameter
                     {
                        new_class->Insert(obj_to_insert->contents[i], CLI);
                     }
                  }
                  else
                  {
                     *CLI->output_stream <<
                     "Something failed while inserting a new Kludge_Class. Class malloc()'ed\n"
                     " but Kludge_Class::Kludge_Class() ctor failed to fully set it up.\n";

                     delete new_class; new_class = NULL; // no need to zap it in this case.
                     result = 1; // internal error
                  }
               }
               else
               {
                  cout << "Kludge_Class::Insert(): Failed to create (malloc?) a new Kludge_Class.\n";
                  result = 1; // internal error
               }
            }
            else
            {
               // this name already exists in this context (path) !
               tmp_pwd = Kludge_CLI_PWD(tmp_obj_ptr, CLI->output_stream);
               
               *CLI->output_stream <<
                  "I've been asked to insert a new Kludge Class called \""
                  << obj_to_insert->contents[0]->content << "\".\n"
                  " Such a name already exists under Class " << this->Name <<
                  ", it resolves to:\n "
                  << *tmp_pwd << endl <<
                  " Class names are kept unique on a system-wide scale, too.\n";
               
               delete tmp_pwd; tmp_pwd = NULL;
               tmp_obj_ptr = NULL;
               //result = 0; // not an internal error
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error: in Kludge_Class::Insert(<CLASS>):\n"
               " class name can't be an empty string.\n";
            //result = 0; // not an internal error
         }
      }
      break;
      case PARM_CLSPROP :
      {
         // the proposed property name should not be an empty string in the first place
         if (obj_to_insert->contents[0]->content.size() != 0)
         {
            // check if such a name exists under `this' class
            tmp_obj_ptr = Resolve_Content_Name(&obj_to_insert->contents[0]->content,CLI,/*own only*/true);
            // no need to check the global class names index - we're not inserting a class.

            if (tmp_obj_ptr == NULL) // no such name exists yet in this context
            {
               new_cls_prop = new Kludge_Class_Property(&obj_to_insert->contents[0]->content,
                                            this, &obj_to_insert->contents[1]->content, CLI);

               if (new_cls_prop != NULL)
               {
                  if (new_cls_prop->ctor_failed)
                  {
                     cout << "Kludge_Class::Insert(): Failed to create "
                             "a new Kludge_Class_Property (although malloc() OK).\n";
                     delete new_cls_prop; new_cls_prop = NULL;
                     result = 1;
                  }
               }
               else
               {
                  cout << "Kludge_Class::Insert(): Failed to create (malloc?) "
                          "a new Kludge_Class_Property.\n";
                  result = 1; // internal error
               }
            }
            else
            {
               // this name already exists in this context (path) !
               tmp_pwd = Kludge_CLI_PWD(tmp_obj_ptr, CLI->output_stream);
               
               *CLI->output_stream <<
                  "I've been asked to insert a new Kludge Class Property called \""
                  << obj_to_insert->contents[0]->content << "\".\n"
                  " Such a name already exists under Class " << this->Name <<
                  ", it resolves to:\n "
                  << *tmp_pwd << endl;
               
               delete tmp_pwd; tmp_pwd = NULL;
               tmp_obj_ptr = NULL;
               //result = 0; // not an internal error
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error: in Kludge_Class::Insert(<CLASS_PROPERTY>):\n"
               " class property name can't be an empty string.\n";
            //result = 0; // not an internal error
         }
      }
      break;
      case PARM_CLSREL :
      {
         // The other sections in this switch/case breakdown simply call 
         // the respective Kludge Objects's heavy-weight constructor.
         // With class Kludge_Class_Relationship, I decided to avoid the
         // heavy-weight style constructor to prevent unnecessary
         // copy-constructions. I simly read the data directly into an
         // embryonal Kludge_Class_Relationship (after its light-weight
         // constructor has entered reasonable defaults) and perform
         // semantic checks over that embryonal structure, before I proceed
         // to the heavy-weight initialization using the Init() method.

         new_cls_rel = new Kludge_Class_Relationship(); // enters some defaults that we rely on here

#define IS_VAR(VAR_NAME_STRING) obj_to_insert->contents[i]->contents[0]->content == VAR_NAME_STRING
         // read the variables into the uninitialized Kludge_Class_Relationship
         for (int i = obj_to_insert->contents.size()-1; (i>=0) && (new_cls_rel != NULL) ; i--)
         {
            // warn at variable names that we don't know (and skip them)
            
            if (IS_VAR("MASTER"))
            {
               // path/name not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_path = Kludge_CLI_String2path(
                                &obj_to_insert->contents[i]->contents[1]->content, //Value
                                CLI->output_stream
                             );
                  // does path/name string resolve fine to a path (vector of elements) ?
                  if (tmp_path != NULL)
                  {
                     tmp_obj_ptr = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
                     // does path (vector of elements) resolve fine into an object pointer ?
                     if (tmp_obj_ptr != NULL)
                     {
                        // is the resolved prospective master object really a Kludge_Class ? 
                        if (tmp_obj_ptr->Type == KLUDGE_CLASS)
                        {
                           new_cls_rel->Master_Stub.Owner = (Kludge_Class*)tmp_obj_ptr; // done with this one
                        }
                        else
                        {
                           *CLI->output_stream <<
                              "Error while inserting a new CLASS_RELATIONSHIP:\n"
                              " Path to MASTER resolves to a kludge object other than CLASS.\n";
                           delete new_cls_rel; new_cls_rel = NULL;
                           // result = 0; Not an internal error.
                        }

                        // do not delete tmp_obj_ptr - object neither created nor owned by us
                        tmp_obj_ptr = NULL;
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error while inserting a new CLASS_RELATIONSHIP:\n"
                           " Path2Ptr failed to follow the given path to MASTER - wrong path?\n";
                        delete new_cls_rel; new_cls_rel = NULL;
                        // result = 0; Not an internal error.
                     }

                     delete tmp_path; tmp_path = NULL;
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error while inserting a new CLASS_RELATIONSHIP:\n"
                        " String2path failed while parsing path to MASTER - probably a syntax error.\n";
                     delete new_cls_rel; new_cls_rel = NULL;
                     // result = 0; Not an internal error.
                  }
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new CLASS_RELATIONSHIP:\n"
                     " master name/path is an empty string.\n";
                  delete new_cls_rel; new_cls_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("SLAVE"))
            {
               // path/name not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_path = Kludge_CLI_String2path(
                                &obj_to_insert->contents[i]->contents[1]->content, //Value
                                CLI->output_stream
                             );
                  // does path/name string resolve fine to a path (vector of elements) ?
                  if (tmp_path != NULL)
                  {
                     tmp_obj_ptr = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
                     // does path (vector of elements) resolve fine into an object pointer ?
                     if (tmp_obj_ptr != NULL)
                     {
                        // is the resolved prospective slave object really a Kludge_Class ? 
                        if (tmp_obj_ptr->Type == KLUDGE_CLASS)
                        {
                           new_cls_rel->Slave_Stub.Owner = (Kludge_Class*)tmp_obj_ptr; // done with this one
                        }
                        else
                        {
                           *CLI->output_stream <<
                              "Error while inserting a new CLASS_RELATIONSHIP:\n"
                              " Path to SLAVE resolves to a kludge object other than CLASS.\n";
                           delete new_cls_rel; new_cls_rel = NULL;
                           // result = 0; Not an internal error.
                        }

                        // do not delete tmp_obj_ptr - object neither created nor owned by us
                        tmp_obj_ptr = NULL;
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error while inserting a new CLASS_RELATIONSHIP:\n"
                           " Path2Ptr failed to follow the given path to SLAVE - wrong path?\n";
                        delete new_cls_rel; new_cls_rel = NULL;
                        // result = 0; Not an internal error.
                     }

                     delete tmp_path; tmp_path = NULL;
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error while inserting a new CLASS_RELATIONSHIP:\n"
                        " String2path failed while parsing path to SLAVE - probably a syntax error.\n";
                     delete new_cls_rel; new_cls_rel = NULL;
                     // result = 0; Not an internal error.
                  }
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new CLASS_RELATIONSHIP:\n"
                     " slave name/path is an empty string.\n";
                  delete new_cls_rel; new_cls_rel = NULL;
                  // result = 0; Not an internal error.
               }

            }
            else if (IS_VAR("NAME_M"))
            {
               // name at master not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  new_cls_rel->Master_Stub.Name = obj_to_insert->contents[i]->contents[1]->content;
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new CLASS_RELATIONSHIP:\n"
                     " name at master is an empty string.\n";
                  delete new_cls_rel; new_cls_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("NAME_S"))
            {
               // name at slave not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  new_cls_rel->Slave_Stub.Name = obj_to_insert->contents[i]->contents[1]->content;
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new CLASS_RELATIONSHIP:\n"
                     " name at slave is an empty string.\n";
                  delete new_cls_rel; new_cls_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("TYPE"))
            {
               // really an ENUM-style value
               if (obj_to_insert->contents[i]->contents[1]->content == "OWN")
               {
                  new_cls_rel->RelType = 2;
               }
               else if (obj_to_insert->contents[i]->contents[1]->content == "M/S")
               {
                  new_cls_rel->RelType = 1;
               }
               else if (obj_to_insert->contents[i]->contents[1]->content == "P2P")
               {
                  new_cls_rel->RelType = 0;
               }
               else
               {
                  *CLI->output_stream <<
                     "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                     " TYPE is not one of permitted values: \"P2P\", \"M/S\", \"OWN\".\n"
                     " Assuming default value (M/S).";
               }
            }
            else if (IS_VAR("MINCOUNT_M"))
            {
               // min count at master not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_stream = new istrstream(obj_to_insert->contents[i]->contents[1]->content.c_str());
                  if (!(*tmp_stream >> new_cls_rel->Master_Stub.MinCount))
                  {
                     *CLI->output_stream <<
                        "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                        " failed to read an integer value from variable MINCOUNT_M."
                        " Assuming default value (0).";
                  }
                  // else value parsed successfully.
                  delete tmp_stream;
               }
               // else value is NULL - assuming default (0)
            }
            else if (IS_VAR("MAXCOUNT_M"))
            {
               // max count at master not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_stream = new istrstream(obj_to_insert->contents[i]->contents[1]->content.c_str());
                  if (!(*tmp_stream >> new_cls_rel->Master_Stub.MaxCount))
                  {
                     *CLI->output_stream <<
                        "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                        " failed to read an integer value from variable MAXCOUNT_M."
                        " Assuming default value (0).";
                  }
                  // else value parsed successfully.
                  delete tmp_stream;
               }
               // else value is NULL - assuming default (0)
            }
            else if (IS_VAR("DEFAULTCOUNT_M"))
            {
               // default count at master not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_stream = new istrstream(obj_to_insert->contents[i]->contents[1]->content.c_str());
                  if (!(*tmp_stream >> new_cls_rel->DefaultCount_M))
                  {
                     *CLI->output_stream <<
                        "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                        " failed to read an integer value from variable DEFAULTCOUNT_M."
                        " Assuming default value (0).";
                  }
                  // else value parsed successfully.
                  delete tmp_stream;
               }
               // else value is NULL - assuming default (0)
            }
            else if (IS_VAR("NAMINGGEN_M"))
            {
               new_cls_rel->Master_Stub.NamingGen = obj_to_insert->contents[i]->contents[1]->content;
            }
            else if (IS_VAR("NAMINGMATCH_M"))
            {
               new_cls_rel->Master_Stub.SetNamingMatch(
                       &obj_to_insert->contents[i]->contents[1]->content,
                       CLI
                     );
            }
            else if (IS_VAR("MINCOUNT_S"))
            {
               // min count at slave not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_stream = new istrstream(obj_to_insert->contents[i]->contents[1]->content.c_str());
                  if (!(*tmp_stream >> new_cls_rel->Slave_Stub.MinCount))
                  {
                     *CLI->output_stream <<
                        "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                        " failed to read an integer value from variable MINCOUNT_S."
                        " Assuming default value (0).";
                  }
                  // else value parsed successfully.
                  delete tmp_stream;
               }
               // else value is NULL - assuming default (0)
            }
            else if (IS_VAR("MAXCOUNT_S"))
            {
               // max count at slave not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_stream = new istrstream(obj_to_insert->contents[i]->contents[1]->content.c_str());
                  if (!(*tmp_stream >> new_cls_rel->Slave_Stub.MaxCount))
                  {
                     *CLI->output_stream <<
                        "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                        " failed to read an integer value from variable MAXCOUNT_S."
                        " Assuming default value (0).";
                  }
                  // else value parsed successfully.
                  delete tmp_stream;
               }
               // else value is NULL - assuming default (0)
            }
            else if (IS_VAR("NAMINGGEN_S"))
            {
               new_cls_rel->Slave_Stub.NamingGen = obj_to_insert->contents[i]->contents[1]->content;
            }
            else if (IS_VAR("NAMINGMATCH_S"))
            {
               new_cls_rel->Slave_Stub.SetNamingMatch(
                       &obj_to_insert->contents[i]->contents[1]->content,
                       CLI
                     );
            }
            else
            {
               *CLI->output_stream <<
                  "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                  " Unknown variable encountered, called \""
                  << obj_to_insert->contents[i]->contents[0]->content <<
                  "\".\n"
                  " Skipped.";
            }
         }
#undef IS_VAR(VAR_NAME_STRING)

         // perform a few basic semantic checks over the proto-K_C_Relationship loaded with data
         if (new_cls_rel != NULL)
         {
            if ((new_cls_rel->Master_Stub.Name.size() != 0) && (new_cls_rel->Slave_Stub.Name.size() != 0))
            {
               if ((new_cls_rel->Master_Stub.Owner == NULL) && (new_cls_rel->Slave_Stub.Owner == NULL))
               {
                  *CLI->output_stream <<
                     "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                     " both MASTER and SLAVE undefined - setting both to the INSERT destination.\n"
                     " I.e., the CLASS_RELATIONSHIP just inserted expresses relationship of\n"
                     " the single master/slave class to itself. Hope that's what you meant.\n";
               }
               if (new_cls_rel->Master_Stub.Owner == NULL) new_cls_rel->Master_Stub.Owner = this;
               if (new_cls_rel->Slave_Stub.Owner == NULL) new_cls_rel->Slave_Stub.Owner = this;

               if ((new_cls_rel->Master_Stub.Owner == this) || (new_cls_rel->Slave_Stub.Owner == this))
               {
                  if ( new_cls_rel->Master_Stub.Owner->Resolve_Content_Name(
                        &new_cls_rel->Master_Stub.Name,CLI,/*own only*/true) == NULL
                  )
                  {
                     if ( new_cls_rel->Slave_Stub.Owner->Resolve_Content_Name(
                           &new_cls_rel->Slave_Stub.Name,CLI,/*own only*/true) == NULL
                     )
                     {
                        // just a few final checks
                        if ((new_cls_rel->RelType == 2) && (new_cls_rel->Slave_Stub.MinCount > 1))
                        {
                           *CLI->output_stream <<
                              "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                              " TYPE = \"OWN\" and MINCOUNT_S > 1. MINCOUNT_S set to 1.\n";
                           new_cls_rel->Slave_Stub.MinCount = 1;
                        }
                        if ((new_cls_rel->Slave_Stub.MinCount > new_cls_rel->Slave_Stub.MaxCount))
                        {
                           *CLI->output_stream <<
                              "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                              " MINCOUNT_S > MAXCOUNT_S. MAXCOUNT_S set to MINCOUNT_S.\n";
                           new_cls_rel->Slave_Stub.MaxCount = new_cls_rel->Slave_Stub.MinCount;
                        }
                        if ((new_cls_rel->Master_Stub.MinCount > new_cls_rel->Master_Stub.MaxCount))
                        {
                           *CLI->output_stream <<
                              "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                              " MINCOUNT_M > MAXCOUNT_M. MAXCOUNT_M set to MINCOUNT_M.\n";
                           new_cls_rel->Master_Stub.MaxCount = new_cls_rel->Master_Stub.MinCount;
                        }
                        if ((new_cls_rel->DefaultCount_M > new_cls_rel->Master_Stub.MaxCount))
                        {
                           *CLI->output_stream <<
                              "Warning: while inserting a new CLASS_RELATIONSHIP:\n"
                              " DEFAULTCOUNT_M > MAXCOUNT_M. DEFAULTCOUNT_M set to MINCOUNT_M.\n";
                           new_cls_rel->DefaultCount_M = new_cls_rel->Master_Stub.MinCount;
                        }

                        // enough of semantic precautions - let's launch it
                        if (new_cls_rel->Init(CLI) != 0)
                        {
                           *CLI->output_stream <<
                              "Error: while inserting a new CLASS_RELATIONSHIP:\n"
                              " Parameter parsing OK, but Kludge_Class_Relationship::Init() failed.\n"
                              " INSERT canceled.\n";
                           delete new_cls_rel; new_cls_rel = NULL;
                           result = 1;
                        }
                        // else launch succeeded
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error: while inserting a new CLASS_RELATIONSHIP:\n"
                           " Name_S already exists under the prospective Slave.\n"
                           " INSERT canceled.\n";
                        delete new_cls_rel; new_cls_rel = NULL;
                     }
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error: while inserting a new CLASS_RELATIONSHIP:\n"
                        " Name_M already exists under the prospective Master.\n"
                        " INSERT canceled.\n";
                     delete new_cls_rel; new_cls_rel = NULL;
                  }
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: while inserting a new CLASS_RELATIONSHIP:\n"
                     " both MASTER and SLAVE are different from the INSERT destination.\n"
                     " This is not permitted for the moment - subject to further discussion.\n"
                     " INSERT canceled.\n"
                     " This issue is related to the current syntax and basic semantics of\n"
                     " the \"Kludge query language\".\n";
                  delete new_cls_rel; new_cls_rel = NULL;
               }
            }
            else
            {
               *CLI->output_stream <<
                  "Error: while inserting a new CLASS_RELATIONSHIP:\n"
                  " both NAME_M and NAME_S must be defined and not empty.\n"
                  " INSERT canceled.\n";
               delete new_cls_rel; new_cls_rel = NULL;
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error while reading the mandatory attributes of a CLASS_RELATIONSHIP\n"
               " being inserted from command line. Giving up - INSERT canceled.\n";
         }
      }
      break;
      case PARM_CLS_BLOB :  // the BLOB names are stored in the Class_Properties index.
      {
         // the proposed blob/property name should not be an empty string in the first place
         if (obj_to_insert->contents[0]->content.size() != 0)
         {
            // check if such a name exists under `this' class
            tmp_obj_ptr = Resolve_Content_Name(&obj_to_insert->contents[0]->content,CLI,/*own only*/true);
            // no need to check the global class names index - we're not inserting a class.

            if (tmp_obj_ptr == NULL) // no such name exists yet in this context
            {
               // the second parameter can be empty.
               if (obj_to_insert->contents.size() > 1)
               {
                  tmp_byte_buf = decode_base64(&obj_to_insert->contents[1]->content);
                  if (tmp_byte_buf == NULL)
                  {
                     result = 1;
                     *CLI->output_stream << 
                           "Error - while inserting a BLOB thingy:\n"
                           " error decoding base64 content.\n";
                  }
               }
               // else tmp_byte_buf == NULL
               
               if (result == 0)
               {
                  Kludge_Class_Property::Property_SubType tmp_subtype;

                  if (obj_to_insert->content == "CLS_BLOB")
                  {
                     tmp_subtype = Kludge_Class_Property::SUBTYPE_BLOB;
                  }
                  else if (obj_to_insert->content == "CLS_TXT_BLOB")
                  {
                     tmp_subtype = Kludge_Class_Property::SUBTYPE_TXT_BLOB;
                  }
                  else if (obj_to_insert->content == "METHOD")
                  {
                     tmp_subtype = Kludge_Class_Property::SUBTYPE_METHOD;
                  }
                  else
                  {
                     cout << "Internal error - in Kludge_Class::Insert(<BLOB thingy>):\n"
                             " unknown BLOB subtype: \"" << obj_to_insert->content << "\"\n";
                     result = 1;
                  }

                  if (result == 0)
                  {
                     new_cls_prop = new Kludge_Class_Property(
                                          &obj_to_insert->contents[0]->content, // name
                                          this,
                                          NULL,                                 // no straight content
                                          CLI,
                                          /* suppress? */ false,
                                          tmp_subtype,                          // blob flags
                                          tmp_byte_buf                          // binary BLOB content
                                                             );
                     
                     if (new_cls_prop != NULL)
                     {
                        if (new_cls_prop->ctor_failed)
                        {
                           cout << "Kludge_Class::Insert(): Failed to create "
                                   "a new Kludge_Class_Blob thingy (although malloc() OK).\n";
                           delete new_cls_prop; new_cls_prop = NULL;
                           result = 1;
                        }
                        // else OK - finished inserting the blob.
                     }
                     else
                     {
                        cout << "Kludge_Class::Insert(): Failed to create (malloc?) "
                                "a new Kludge_Class_Blob.\n";
                        result = 1; // internal error
                     }
                  }
                  // else error already reported

                  if (tmp_byte_buf != NULL)
                  {
                     delete tmp_byte_buf; tmp_byte_buf = NULL;
                  } // else OK - probably an empty BLOB (a valid thing to insert)
               }
               // else error already reported - perhaps base64 decoding failed
            }
            else
            {
               // this name already exists in this context (path) !
               tmp_pwd = Kludge_CLI_PWD(tmp_obj_ptr, CLI->output_stream);
               
               *CLI->output_stream <<
                  "I've been asked to insert a new Kludge Class BLOB thingy called \""
                  << obj_to_insert->contents[0]->content << "\".\n"
                  " Such a name already exists under Class " << this->Name <<
                  ", it resolves to:\n "
                  << *tmp_pwd << endl;
               
               delete tmp_pwd; tmp_pwd = NULL;
               tmp_obj_ptr = NULL;
               //result = 0; // not an internal error
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error: in Kludge_Class::Insert(<class BLOB style thingy>):\n"
               " BLOB name can't be an empty string.\n";
            //result = 0; // not an internal error
         }
      }
      break;
      default :
      {
         *CLI->output_stream <<
         "Can only insert one of the following into a CLASS:\n"
         " - a CLASS,\n"
         " - a CLASS_PROPERTY,\n"
         " - a CLASS_RELATIONSHIP,\n"
         " - or one of the BLOB brethren:\n"
         "    - a CLS_BLOB,\n"
         "    - a CLS_TXT_BLOB,\n"
         "    - a METHOD\n";
         //result = 0; // not an internal error
      }
      break;
   }

   return(result);
}



int Kludge_Class::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
      "Set a value to a class? Conceptually this could be justified for the potential\n"
      " occasions when someone might want to mask a class as a property for backwards\n"
      " compatibility with old methods' code - but for now this just doesn't fit into\n"
      " the picture (syntax and all) - so let's keep this a future possibility.\n"
      " Value NOT set.\n";
   return(0);
}



int Kludge_Class::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
 del_param del_parm_passed;
  del_parm_passed.CLI = CLI;
  del_parm_passed.objs_deleted.clear();
 int result = 0;
 unsigned int counter = 0;
 ostrstream qry_string;
 ostrstream check_inst, check_inst_prop; // , check_inst_rel;
 MYSQL_RES *mysql_query_result = NULL;
 bool inst_OK = false, inst_prop_OK = false; // , inst_rel_OK = false;

   if (Children == NULL) Load_Children(CLI);
   if (Children == NULL)
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class::DB_Delete() : failed to initialize the tree\n"
         " of children (used to check their existence/absence before deletion).\n";
      result = 1;
   }
   if (Properties == NULL) Load_Properties(CLI);
   if (Properties == NULL)
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class::DB_Delete() : failed to initialize the tree\n"
         " of properties (used to check their existence/absence before deletion).\n";
      result = 1;
   }
   if (! relationships_loaded) Load_Relationships(CLI);
   if ((! relationships_loaded) || (Relationships == NULL))
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class::DB_Delete() : failed to initialize the tree\n"
         " of relationships (used to check their existence/absence before deletion).\n";
      result = 1;
   }

   if (Name == "Root_Class")
   {
      *CLI->output_stream <<
         "In Kludge_Class::DB_Delete(): cannot delete the Root_Class.\n";
      result = 1;
   }

   if (result == 0)
   {
      if (   (g_tree_nnodes(Children) > 0)
          || (g_tree_nnodes(Properties) > 0)
          || (g_tree_nnodes(Relationships) > 0)
         )
      {
         if (! recursive)
         {
            *CLI->output_stream <<
            "Cannot delete class " << Name <<
            ". There are still contents inside.\n";
            result = 1;
         }
         else
         // Unfortunately we can't just delete'em'all in one SQL sweep - we have
         // to perform some checks on each of them. That's why we have to call
         // DB_Delete on every one of them. We only Zap those who have been
         // successfully DB_Delete()'d.
         {
            g_tree_traverse(          // DB_Delete() all the children
                  Children,
                  Kludge_Delete_Contents_Traverse,
                  G_POST_ORDER,
                  &del_parm_passed
                           );
            g_tree_traverse(          // DB_Delete() all the properties
                  Properties,
                  Kludge_Delete_Contents_Traverse,
                  G_POST_ORDER,
                  &del_parm_passed
                           );
            g_tree_traverse(          // DB_Delete() all the properties
                  Relationships,
                  Kludge_Delete_Contents_Traverse,
                  G_POST_ORDER,
                  &del_parm_passed
                           );
            // and Zap() them all
            for (counter = 0; counter < del_parm_passed.objs_deleted.size(); counter++)
            {
               del_parm_passed.objs_deleted[counter]->Zap();
            }
            del_parm_passed.objs_deleted.clear();

            // if the trees are empty, cut them down:
            if (
                (g_tree_nnodes(Children) == 0)
                && (g_tree_nnodes(Properties) == 0)
                && (g_tree_nnodes(Relationships) == 0)
               )
            {
               g_tree_destroy(Children);
               Children = NULL;
               
               g_tree_destroy(Properties);
               Properties = NULL;
               
               g_tree_destroy(Relationships);
               Relationships = NULL;
               relationships_loaded = false;

               UnInherit();
            }
            else result = 1;   // don't bother cutting down any of the trees
         }
      }
      else
      {  // no child classes, no properties, no relationships
         g_tree_destroy(Children);
         Children = NULL;
         
         g_tree_destroy(Properties);
         Properties = NULL;
         
         g_tree_destroy(Relationships);
         Relationships = NULL;
         relationships_loaded = false;

         UnInherit();
      }
   }

   // before attempting the deletion, check for dependent instance-side objects:
   check_inst      << "SELECT OID FROM Instances WHERE Class = \""
                   << Class << "\"" << ends;
   mysql_query_result = CLI->thr->DB_Query(check_inst.str());
   if (mysql_query_result != NULL)
   {
      if (mysql_fetch_row(mysql_query_result) == NULL) inst_OK = true;
      mysql_free_result(mysql_query_result);
   }
   // else inst_OK remains false
           
   check_inst_prop << "SELECT OID FROM Instance_Properties WHERE Class = \""
                   << Class << "\"" << ends;
   mysql_query_result = CLI->thr->DB_Query(check_inst_prop.str());
   if (mysql_query_result != NULL)
   {
      if (mysql_fetch_row(mysql_query_result) == NULL) inst_prop_OK = true;
      mysql_free_result(mysql_query_result);
   }
   // else inst_prop_OK remains false

   // KLUDGE currently doesn't implement relationship classes in the form of
   // a Kludge_Class
   /* 
   check_inst_rel  << "SELECT OID FROM Instance_Relationships WHERE ClsRel = \""
                   << RelCLSID << "\"" << ends;
   mysql_query_result = CLI->thr->DB_Query(check_inst.str());
   if (mysql_query_result != NULL)
   {
      if (mysql_fetch_row(mysql_query_result) == NULL) inst_rel_OK = true;
      mysql_free_result(mysql_query_result);
   }
   */
   // done with the necessary checks

   if (! (inst_OK && inst_prop_OK))
   {
      *CLI->output_stream <<
         "Cannot delete CLASS " << Name << "\n" <<
         " There are dependent instance-side objects.\n"
         " (Or, the backend DB query to check that has failed.)\n";
      result = 1;
   }

   // Delete this Kludge_Class from the DB.
   // Can't be deallocated here - this function is used from within
   // a traversal function. Deallocate ('delete') it in the calling function.
   if (result == 0)
   {
      qry_string << "DELETE FROM Classes WHERE Class = \""
                 << Class
                 << "\""
                 <<  ends;
      result = CLI->thr->DB_Cmd(qry_string.str());
   }

   return(result);
}



int Kludge_Class::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 0;
 //Kludge_Class* former_parent = NULL;
 ostrstream qry_string;
 
   // can only move under another Kludge Class
   if (destination->Type == KLUDGE_CLASS)
   {
      if (destination->Resolve_Content_Name(&Name,CLI,/*own only?*/true) == NULL)
      {
         if (Parent != NULL)
         {
            Add_To_Index((Kludge_Class*)destination);
            Free_From_Index();
            //former_parent = Parent;   // - for locking purposes only
            Parent = (Kludge_Class*)destination;
            
            qry_string << "UPDATE Classes SET Parent = \""
                       << Parent->Class
                       << "\" WHERE Class = \""
                       << Class
                       << "\""
                       <<  ends;
            result = CLI->thr->DB_Cmd(qry_string.str());
            if (result != 0)
            {
               *CLI->output_stream <<
                  "Error while storing the new parent into the backend database.\n";
            }
         }
         else
         {
            *CLI->output_stream <<
            "In Kludge_Class::Move(): can't move a class with NULL superior.\n"
            " Probably an internal error - an attempt to move an unanchored class.\n";
            // Kludge_CLI::Execute_Command()'s cycle-avoidance rules alone should
            // prevent any attempts to move the root class.
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Class::Move(): content with this name already exists\n"
            " at the desired destination. Move canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "In Kludge_Class::Move(): can only move under another kludge class.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   if (new_name->type == PARM_NAME)
   {
      if (Parent->Resolve_Content_Name(&new_name->content,CLI,/*own only?*/true) == NULL)
      {
         Free_From_Index();
         Free_From_Global_Index();
         Name = new_name->content;
         Add_To_Index();

         qry_string << "UPDATE Classes SET Name = \""
                    << Name 
                    << "\" WHERE Class = \""
                    << Class
                    << "\""
                    <<  ends;
         result = CLI->thr->DB_Cmd(qry_string.str());
         if (result != 0)
         {
            *CLI->output_stream <<
               "Error while storing the new name into the backend database.\n";
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Class::Rename(): content with this name already exists\n"
            " in the parent class. Rename operation cancelled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class::Rename():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_NAME.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Class::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   int result = DB_Delete(CLI, recursive);
   if (result == 0)
   {
      cout <<
         "Deleting class: " << Name << endl;
      result = Zap();
   }

   return(result);
}



int Kludge_Class::pwd(string* result)
{
   if (Parent == NULL)
   {
      Kludge_Root->pwd(result);
   }
   else
   {
      Parent->pwd(result);
   }
   result->append(Name);
   result->append("/");
   return(0);
}



Kludge_Object* Kludge_Class::Superior()
{
   if (Parent != NULL) return(Parent);
   else if (Name == "Root_Class") return(Kludge_Root);
   else return (NULL);
}



// Common guts of both versions of Kludge_Class::DB_Lookup.
// Returns a Kludge_MYSQL_RES with one Row (pointer into a MYSQL_RES structure)
// Row[0] = Class, Row[1] = Name, Row[2] = Parent
// See kludge_mysql.h for the declaration of class Kludge_MYSQL_RES.
// Delete the Kludge_MYSQL_RES* returned by either version of Kludge_Class::DB_Lookup
// after use.
//
//static
Kludge_MYSQL_RES* Kludge_Class::DB_Lookup_Internals(char* qry_string, Kludge_CLI* CLI)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 Kludge_MYSQL_RES* result = NULL;
 
   mysql_query_result = CLI->thr->DB_Query(qry_string);

   if (mysql_query_result != NULL)
   {
      result_row = mysql_fetch_row(mysql_query_result);
      if (result_row != NULL)
      {
         result = new Kludge_MYSQL_RES(result_row, mysql_query_result);
      }
      else
      {
         mysql_free_result(mysql_query_result);
         // else query result returned with no data rows - no such class in the DB.
         // That's not an internal error - don't report it here, just return NULL.
      }
   }
   else
   {
      cout << "Error in Kludge_Class::DB_Lookup(): query returned null MYSQL_RES* .\n";
   }

   return(result);
}

//static
Kludge_MYSQL_RES* Kludge_Class::DB_Lookup(string* Class_Name, Kludge_CLI* CLI)
{
 Kludge_MYSQL_RES* result = NULL;
 ostrstream qry_string;

   qry_string << "SELECT Class,Name,Parent FROM Classes WHERE Name = \"" 
              << *Class_Name
              << "\""
              <<  ends;
   result = DB_Lookup_Internals(qry_string.str(), CLI);
   
   return(result);
}

//static
Kludge_MYSQL_RES* Kludge_Class::DB_Lookup(OID_Type Class_ID, Kludge_CLI* CLI)
{
 Kludge_MYSQL_RES* result = NULL;
 ostrstream qry_string;

   qry_string << "SELECT Class,Name,Parent FROM Classes WHERE Class = \""
              << Class_ID
              << "\""
              <<  ends;
   result = DB_Lookup_Internals(qry_string.str(), CLI);

   return(result);
}



// expects row->Row[0] = int Class, row->Row[1] = char[] Name, row->Row[2] = int Parent
//static
Kludge_Class* Kludge_Class::Load_By_ID(Kludge_MYSQL_RES* row, Kludge_CLI* CLI)
{
 OID_Type tmp_ID = 0;
 vector<OID_Type> branch_to_load;
 Kludge_Class* result = NULL;
 bool dead_end = false;
 istrstream* tmp_ss = NULL; // for MY_SSCANF

   MY_SSCANF(row->Row[0],tmp_ID);                // read Class ID

   result = ID2ptr_Raw(&tmp_ID);
   if (result == NULL)
   {
      branch_to_load.push_back(tmp_ID);
      while ((result == NULL) && (! dead_end))
      {
         MY_SSCANF(row->Row[2],tmp_ID);          // read parent Class ID
         delete row; row = NULL;

         result = ID2ptr_Raw(&tmp_ID);
         if (result == NULL)
         {
            branch_to_load.push_back(tmp_ID);
            row = DB_Lookup(tmp_ID, CLI);
            if (row == NULL) dead_end = true;
         }
         // else the parent class is already loaded. Now that we have it by
         // its pointer, we can load its children and then look up our
         // desired child class.
      }

      if (row != NULL)  // clean up the mysql session before proceeding
      {
         delete row;
         row = NULL;
      }
      
      if (! dead_end)
      {
         for (int i = branch_to_load.size() - 1; (i >= 0) && (! dead_end); i--)
         {
            if (result->Children == NULL)           // children loaded?
            {
               result->Load_Children(CLI, false);   // just load them all
            }
            result = ID2ptr_Raw(&branch_to_load[i]);
            if (result == NULL) dead_end = true;
         }
      }
      branch_to_load.clear();
   }
   // else the Class is already loaded - which may or may not be an error,
   // depending on which function has called us.

   if (row != NULL)  // A final cleanup of the mysql session - takes effect if
   {                 // the class we were asked to load was already loaded.
      delete row;
      row = NULL;
   }

   return(result);
}



//static
Kludge_Class* Kludge_Class::Name2ptr_Raw(string* Class_Name)
{
 Kludge_Class *result = NULL;

   result = (Kludge_Class *) g_tree_lookup(By_Name,Class_Name);
   
   return(result);
}



//static
Kludge_Class* Kludge_Class::ID2ptr_Raw(OID_Type* ID)
{
 Kludge_Class *result = NULL;

   result = (Kludge_Class *) g_tree_lookup(By_ID,ID);
   
   return(result);
}



//static
Kludge_Class* Kludge_Class::Name2ptr(string* Class_Name, Kludge_CLI* CLI)
{
 Kludge_Class *result = NULL;
 Kludge_MYSQL_RES* tmp_mysql_res = NULL;

   result = Name2ptr_Raw(Class_Name);
   if (result == NULL)
   {
      tmp_mysql_res = DB_Lookup(Class_Name, CLI);
      if (tmp_mysql_res != NULL)
      {
         result = Load_By_ID(tmp_mysql_res, CLI);
      }
      // else no such class name in the DB - implies no such class name
      // in the system. Which is not an internal error => no report here.
   }
   
   return(result);
}



//static
Kludge_Class* Kludge_Class::ID2ptr(OID_Type* ID, Kludge_CLI* CLI)
{
 Kludge_Class *result = NULL;
 Kludge_MYSQL_RES* tmp_mysql_res = NULL;

   result = ID2ptr_Raw(ID);
   if (result == NULL)
   {
      tmp_mysql_res = DB_Lookup(*ID, CLI);
      if (tmp_mysql_res != NULL)
      {
         result = Load_By_ID(tmp_mysql_res, CLI);
      }
      // else no such class ID in the DB - implies no such class ID
      // in the system. Which is not an internal error => no report here.
   }
   
   return(result);
}



//static
int Kludge_Class::List_Children_Traverse(gpointer key, gpointer Child_Class, gpointer CLI)
{
   ((Kludge_Class *)Child_Class)->List_Contents((Kludge_CLI *)CLI, true);
   return(false);
}



//static
int Kludge_Class::List_Children_Names_Traverse(gpointer key, gpointer Child_Class, gpointer CLI)
{
   ((Kludge_Class *)Child_Class)->Describe_Self((Kludge_CLI *)CLI);

   return(false);
}



//static
int Kludge_Class::List_Properties_Traverse(gpointer key, gpointer Property, gpointer CLI)
{
   ((Kludge_Class_Property *)Property)->List_Contents((Kludge_CLI*)CLI);
   
   return(false);
}



//static
int Kludge_Class::List_Properties_Names_Traverse(gpointer key, gpointer Property, gpointer CLI)
{
   ((Kludge_Class_Property *)Property)->Describe_Self((Kludge_CLI*)CLI);
   
   return(false);
}



//static
int Kludge_Class::List_Properties_Inh_Traverse(gpointer key, gpointer Property, gpointer CLI)
{
   *((Kludge_CLI*)CLI)->output_stream << "INHERITED ";
   ((Kludge_Class_Property *)Property)->List_Contents((Kludge_CLI*)CLI);
   
   return(false);
}



//static
int Kludge_Class::List_Properties_Inh_Names_Traverse(gpointer key, gpointer Property, gpointer CLI)
{
   *((Kludge_CLI*)CLI)->output_stream << "INHERITED ";
   ((Kludge_Class_Property *)Property)->Describe_Self((Kludge_CLI*)CLI);
   
   return(false);
}



//static
int Kludge_Class::List_Relationships_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI)
{
   ((Kludge_Class_Rel_Stub*)Rel_Stub)->List_Contents((Kludge_CLI *)CLI);
   return(false);
}



//static
int Kludge_Class::List_Relationship_Names_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI)
{
   ((Kludge_Class_Rel_Stub*)Rel_Stub)->Describe_Self((Kludge_CLI *)CLI);

   return(false);
}



//static
int Kludge_Class::List_Relationships_Inh_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI)
{
   *((Kludge_CLI*)CLI)->output_stream << "INHERITED ";
   ((Kludge_Class_Rel_Stub*)Rel_Stub)->List_Contents((Kludge_CLI *)CLI);
   return(false);
}



//static
int Kludge_Class::List_Relationship_Inh_Names_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI)
{
   *((Kludge_CLI*)CLI)->output_stream << "INHERITED ";
   ((Kludge_Class_Rel_Stub*)Rel_Stub)->Describe_Self((Kludge_CLI *)CLI);

   return(false);
}



//static
int Kludge_Class::Free_Properties_Traverse(gpointer key, gpointer Property, gpointer data)
{
   delete (Kludge_Class_Property *)Property;
   return(false);
}



//static
int Kludge_Class::Free_Children_Traverse(gpointer key, gpointer Child_Class, gpointer data)
{
   delete (Kludge_Class *)Child_Class;
   return(false);
}



//static
int Kludge_Class::Free_Relationships_Traverse(gpointer key, gpointer Relationship, gpointer rels_to_free)
{
   ((vector<Kludge_Class_Rel_Stub*>*)rels_to_free)->push_back((Kludge_Class_Rel_Stub*)Relationship);
   return(false);
}



//static
int Kludge_Class::ReInherit_Children_Traverse(gpointer key, gpointer Child, gpointer CLI)
{
   ((Kludge_Class*)Child)->ReInherit((Kludge_CLI*)CLI);
   return(false);
}



//static
int Kludge_Class::UnInherit_Children_Traverse(gpointer key, gpointer Child, gpointer data)
{
   ((Kludge_Class*)Child)->UnInherit();
   return(false);
}



//static
int Kludge_Class::Build_Props_Array_Traverse(gpointer key, gpointer Content, gpointer array_to_build)
{
   ((vector<Kludge_Class_Property*>*)array_to_build)->push_back((Kludge_Class_Property*)Content);
   return(false);
}



//static
int Kludge_Class::Build_Rels_Array_Traverse(gpointer key, gpointer Content, gpointer array_to_build)
{
   ((vector<Kludge_Class_Rel_Stub*>*)array_to_build)->push_back((Kludge_Class_Rel_Stub*)Content);
   return(false);
}









/* ###    class Kludge_Class_Property    ### */


// This constructor is used for creating a new property from scratch.
// It cannot be used for loading a property from the DB.
// It actually stores the newly created property in the DB.
//
Kludge_Class_Property::Kludge_Class_Property(
  string* cls_prop_name,
  Kludge_Class* master_class,
  string* val, Kludge_CLI* CLI,
  bool sup = false,
  Property_SubType _type = SUBTYPE_PROPERTY,
  byte_buf* blob_value = NULL
) : 
  Class(master_class), 
  RDN(*cls_prop_name),
  //Value(""),
  Suppress(sup),
  ctor_failed(false),
  SubType(_type),
  Method_CV(NULL)
{
 int stor_success = 0;
 ostrstream* qry_string;
 ostrstream* tmp_ss;
 string blob_tbl_name;
 unsigned int tmp_flags = 0;
 char* tmp_escaped_data = NULL;

   Type = KLUDGE_CLASS_PROPERTY;

   if (SubType == SUBTYPE_PROPERTY)
   {  // not a blob
      if (val != NULL)
      {
         Value = *val;
      }
      else
      {
         cout << "Internal error - in Kludge_Class_Property:: ctor :\n"
                 " asked to create a regular Kludge_Class_Property, obtained a NULL PTR for value!\n";
         ctor_failed = true;
      }
   }
   else // SubType != SUBTYPE_PROPERTY  ==> some sort of a blob
   {
      switch (SubType)
      {
         case SUBTYPE_BLOB:
         {
            blob_tbl_name = "Blobs";
            tmp_flags = CLSPROP_IS_BLOB;
         }
         break;
         case SUBTYPE_TXT_BLOB:
         {
            blob_tbl_name = "Text_Blobs";
            tmp_flags = CLSPROP_IS_TXT_BLOB;
         }
         break;
         case SUBTYPE_METHOD:
         {
            blob_tbl_name = "Text_Blobs";
            tmp_flags = CLSPROP_IS_METHOD;
         }
         break;
         default:
         {
            cout << "Internal error - in Kludge_Class_Property:: ctor :\n"
                    " asked to create a BLOB thingy, obtained an invalid SubType!\n";
            ctor_failed = true;
         }
         break;
      }

      if (!ctor_failed)
      {
         if (blob_value != NULL)
         {
            tmp_escaped_data = new char[blob_value->len * 2 + 1];
            if (tmp_escaped_data == NULL)
            {
               cout <<
                  "Internal error - in Kludge_Class_Property:: ctor :\n"
                  " while storing BLOB contents, couldn't allocate memory for the escaped string!\n";
               ctor_failed = true;
            }
         }
         // else no problem - we're inserting an empty BLOB

         if (! ctor_failed)
         {
            CLI->thr->Open_DB(); // If this is not the first thread to look up all
                                 // the necessary data, the database session might 
                                 // still be closed in this thread.
            if (tmp_escaped_data != NULL)
            {
               mysql_real_escape_string(
                       CLI->thr->DB,            // requires an open DB session!
                       tmp_escaped_data,
                       (char*) blob_value->buf, // the cast suppresses a warning about signedness
                       blob_value->len
                                       );
            }
            
            qry_string = new ostrstream;
            *qry_string << "INSERT INTO " << blob_tbl_name << " VALUES (\""
                        << 0
                        << "\",\""
                        << ((tmp_escaped_data == NULL) ? "" : tmp_escaped_data)
                        << "\")"
                        <<  ends;   
            stor_success = CLI->thr->DB_Cmd(qry_string->str());
            if (stor_success == 0)
            {  // get the auto-generated blob ID and store it in the property's Value
               tmp_ss = new ostrstream;
               *tmp_ss << mysql_insert_id(CLI->thr->DB) << ends;
               Value = tmp_ss->str();
               delete tmp_ss; tmp_ss = NULL;
            }
            else
            {
               cout << "Internal error - in Kludge_Class_Property:: ctor :\n"
                       " DB op failed while storing BLOB content!\n";
               ctor_failed = true;
            }
            delete qry_string; qry_string = NULL;
            if (tmp_escaped_data != NULL)
            {
               delete tmp_escaped_data; tmp_escaped_data = NULL;
            }
         }
         // else error already reported
      }
      // else something wrong with the SubType - error already reported
   }
  
   if (! ctor_failed) // whatever the content really is, let's save the property body
   {
      if ( Class->Properties == NULL ) Class->Load_Properties(CLI);
     
      qry_string = new ostrstream;
      *qry_string << "INSERT INTO Class_Properties VALUES (\""
                  << Class->Class
                  << "\",\""
                  << RDN
                  << "\",\""
                  << Value
                  << "\",\""
                  << (Suppress ? "1" : "0")
                  << "\",\""
                  << tmp_flags   // Flags
                  << "\")"
                  <<  ends;   
      stor_success = CLI->thr->DB_Cmd(qry_string->str());
      delete qry_string; qry_string = NULL;

      if (stor_success == 0)
      {
         Add_To_Index();
      }
      else
      {
         ctor_failed = true;
      }
   }
   // else something went wrong - no need to bother storing the property.
}



// This constructor can only be used for loading an existing Kludge_Class_Property
// from a database. It is not intended for creating a new class property from scratch.
//
Kludge_Class_Property::Kludge_Class_Property(
  MYSQL_ROW DB_record,
  Kludge_Class* master_class,
  Kludge_CLI* CLI
) :
  Class(master_class),
  RDN(""),
  Value(""),
  Suppress(false),
  ctor_failed(false),
  SubType(SUBTYPE_PROPERTY),
  Method_CV(NULL)
{ 
   Type = KLUDGE_CLASS_PROPERTY; 

   if ( ! Extract_From_MysqlRow(DB_record) )
   {
      ctor_failed = true;
   }
   else Add_To_Index();
}



Kludge_Class_Property::~Kludge_Class_Property()
{
   cout << "Deallocating class property: " << RDN <<"\n";
   UnParse_Method();
}



//
// Expects:  Row_To_Process[0] == RDN, Row_To_Process[1] == Value, Row_To_Process[2] == Suppress
//
bool Kludge_Class_Property::Extract_From_MysqlRow(MYSQL_ROW Row_To_Process) 
{
 bool result = true;
 istrstream* tmp_ss = NULL;
 unsigned int tmp_flags = 0;

   if (Row_To_Process != NULL)
   {
      // Set Kludge_Class_Property.RDN (of type `string' - hence the direct assignment)
      RDN = Row_To_Process[0];
      // Set Kludge_Class_Property.Value (of type `string' - hence the direct assignment)
      Value = Row_To_Process[1];
      // Set Kludge_Class_Property.Suppress
      if ( (strcmp(Row_To_Process[2],"0")) || (strlen(Row_To_Process[2])==0) ) 
        Suppress = false;
      else Suppress = true;
      // Set Kludge_Class_Property.Flags
      tmp_ss = new istrstream(Row_To_Process[3]);
      *tmp_ss >> tmp_flags;
      delete tmp_ss; tmp_ss = NULL;
      if (tmp_flags == 0) SubType = SUBTYPE_PROPERTY;
      else if (tmp_flags & CLSPROP_IS_BLOB) SubType = SUBTYPE_BLOB;
      else if (tmp_flags & CLSPROP_IS_TXT_BLOB) SubType = SUBTYPE_TXT_BLOB;
      else if (tmp_flags & CLSPROP_IS_METHOD) SubType = SUBTYPE_METHOD;
   }
   else
   {
      cout << 
         "In Kludge_Class_Property::Extract_From_MysqlRow():\n"
         " the row passed to this function is a NULL pointer!\n"
         " Extract_From_MysqlRow() failed.\n";
      result = false;
   }

   return(result);
}



// Zap() does a nice memory cleanup.
// Only makes sense 
// a) for dynamic partial release from memory 
//     In that case, invalidate the list of properties.
// b) when a property is deleted
//     Requires deletion from the SQL-based object repository
//      - i.e., use DB_Delete() and then Zap() (or just Delete())
int Kludge_Class_Property::Zap()
{
 int result = 0;

   cout << "Zapping a Class_Property: " << Class->Name << "/" << RDN << "\n" ;
   if (Class->Properties != NULL)
   {
      Free_From_Index();
   }
   else
   {
      cout << 
         "Internal error - in Kludge_Class_Property::Zap():\n"
         " parent's Properties index not initialized!\n";
      result = 1;
   }  
   delete this;

   return(result);
}



int Kludge_Class_Property::Report_Subtype(Kludge_CLI* CLI)
{
 int result = 0;

   switch (SubType)
   {
      case SUBTYPE_PROPERTY :
      {
         *CLI->output_stream << "CLASS_PROPERTY ";
      }
      break;
      case SUBTYPE_BLOB :
      {
         *CLI->output_stream << "CLS_BLOB ";
      }
      break;
      case SUBTYPE_TXT_BLOB :
      {
         *CLI->output_stream << "CLS_TXT_BLOB ";
      }
      break;
      case SUBTYPE_METHOD :
      {
         *CLI->output_stream << "METHOD ";
      }
      break;
      default:
      {
         cout << "Internal error - in Kludge_Class_Property::Report_Subtype():\n"
                 " unknown class property subtype!\n";
         result = 1;
      }
      break;
   }

   return(result);
}


Kludge_MYSQL_RES* Kludge_Class_Property::Get_Blob_Content(Kludge_CLI* CLI)
{
 MYSQL_RES* mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 Kludge_MYSQL_RES* result = NULL;
 ostrstream qry_ss;
 string blob_tbl_name;
 bool failure = false;

   switch (SubType)
   {
      case SUBTYPE_BLOB:
      {
         blob_tbl_name = "Blobs";
      }
      break;
      case SUBTYPE_TXT_BLOB:
      {
         blob_tbl_name = "Text_Blobs";
      }
      break;
      case SUBTYPE_METHOD:
      {
         blob_tbl_name = "Text_Blobs";
      }
      break;
      default:
      {
         cout << "Internal error - in Kludge_Class_Property::Get_Blob_Content() :\n"
                 " invalid Kludge_Class_Property::SubType!\n";
         failure = true;
      }
      break;
   }

   if (!failure)
   {
      qry_ss << "SELECT Content FROM " << blob_tbl_name << " WHERE BlobID = \""
             << Value << "\"" << ends;
      if (( mysql_query_result = CLI->thr->DB_Query(qry_ss.str()) ))
      {
         if (( result_row = mysql_fetch_row(mysql_query_result) ))
         {
            result = new Kludge_MYSQL_RES(result_row, mysql_query_result);
         }
         else
         {
            cout << "Internal error - in Kludge_Class_Property::Get_Blob_Content():\n"
                    " while loading BLOB content, DB query returned empty result.\n";
            // result = NULL;
            mysql_free_result(mysql_query_result);
         }
      }
      else
      {
         cout << "Internal error - in Kludge_Class_Property::Get_Blob_Content():\n"
                 " DB query failed while loading BLOB. Query API function returned NULL result*.\n";
         // result = NULL;
      }
   }

   return(result);
}



int Kludge_Class_Property::List_Contents(Kludge_CLI* CLI, bool recursive = false)
// The recursive tag doesn't really make sense.
// It's here for the sake of compliance with the abstract interface
// of Kludge_Object.
{
 int result = 0;
 Kludge_MYSQL_RES* qry_result = NULL;

   if (SubType == SUBTYPE_PROPERTY) result = Describe_Self(CLI);
   else  // some sorta BLOB
   {
      result = Report_Subtype(CLI);
      
      if (result == 0)
      {
         *CLI->output_stream
               // << "CLASS_PROPERTY \""
               << "\""
               << RDN
               << "\" {\n";  // BLOB grammar lead-in done.
                               
         // Now dump BLOB data in Base64.
         if ((qry_result = Get_Blob_Content(CLI)))
         {
            byte_buf* tmp_buf = new byte_buf(
                                       qry_result->Row[0],
                                       mysql_fetch_lengths(qry_result->Result)[0],
                                       false, // deep copy? (false = shallow copy)
                                       false  // delete the buf* after use? (no, not here!)
                                            );
            string* tmp_base64_data = encode_base64(tmp_buf); // The result contains a trailing \n.
            delete tmp_buf;
            delete qry_result;

            *CLI->output_stream << *tmp_base64_data;

            delete tmp_base64_data;
         }
         else result = 1; // error already reported
         // Data dump done.
      }
      // else there was an error - don't bother. Error already reported - no report here.
                            
      *CLI->output_stream << "};\n"; // BLOB grammar complete.
   }

   return(result);
}



int Kludge_Class_Property::Describe_Self(Kludge_CLI* CLI)
{
   int result = Report_Subtype(CLI);

   *CLI->output_stream
         // << "CLASS_PROPERTY \""
         << "\""
         << RDN
         << "\" \"" 
         << Value
         << "\";\n";
   return(result);
}



Kludge_Object* Kludge_Class_Property::Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
 Kludge_Object* result = NULL;

   if (*Content_Name == "..")            // a higher-level "folder" in the path?
   {
      result = Class;
   }
   else if (*Content_Name == ".")        // a "this" style reference?
   {
      result = this;
   }
/* else
   {
      *error_output << "Error while resolving name \"" << Content_Name <<
                    << "\" within KLUDGE_CLASS_PROPERTY \"" << RDN << "\":\n"
                    << " a Kludge Class Property doesn't have further nested contents.\n"
   } */

   return(result);
}



int Kludge_Class_Property::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
   "Cannot insert anything in a Kludge Class Property.\n"
   " What would you want to insert into a \"leaf object\"?"
   " You can only insert things into classes and instances.\"";

   return(0);
}


int Kludge_Class_Property::Set_Blob(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
 int result = 0;
 byte_buf* blob_value = NULL;
 string blob_tbl_name;
 ostrstream* qry_string = NULL;

   if (new_value->type == PARM_BLOB_VALUE)
   {
      blob_value = decode_base64(&new_value->content);
   }
   else if (new_value->type == PARM_VALUE)
   {
      blob_value = new byte_buf(
                                (char*) new_value->content.c_str(),
                                new_value->content.size(),
                                false, // no copy construction - just a shallow copy
                                false  // don't dealloc the buffer after use
                               );
   }
   else
   {
      cout <<
         "Internal error - in Kludge_Class_Property::Set_Blob():\n"
         " obtained a value of type other than PARM_VALUE or PARM_BLOB_VALUE.\n";
      result = 1;
   }
   
   if (blob_value != NULL)
   {
      switch (SubType)
      {
         case SUBTYPE_BLOB:
         {
            blob_tbl_name = "Blobs";
         }
         break;
         case SUBTYPE_TXT_BLOB:
         {
            blob_tbl_name = "Text_Blobs";
         }
         break;
         case SUBTYPE_METHOD:
         {
            blob_tbl_name = "Text_Blobs";
         }
         break;
         case SUBTYPE_PROPERTY:
         {
            // this should never happen - filtered in the calling function
            cout << "Internal error - in Kludge_Class_Property::Set_Blob() :\n"
                    " asked to Set a BLOB value to a plain property !\n";
            result = 1;
         }
         break;
         default:
         {
            cout << "Internal error - in Kludge_Class_Property::Set_Blob() :\n"
                    " asked to Set a BLOB thingy, obtained an invalid SubType!\n";
            result = 1;
         }
         break;
      }

      if (result == 0)
      {
         char* tmp_escaped_data = new char[blob_value->len * 2 + 1];
         if (tmp_escaped_data != NULL)
         {
            CLI->thr->Open_DB(); // If this is not the first thread to look up all
                                 // the necessary data, the database session might 
                                 // still be closed in this thread.
            mysql_real_escape_string(
                            CLI->thr->DB,            // requires an open DB session!
                            tmp_escaped_data,
                            (char*) blob_value->buf, // the cast suppresses a warning about signedness
                            blob_value->len
                                    );
            // We could consider optimizing memory usage by deallocating
            // blob_value->buf right here - but it would be a terrible hack
            // that might later cause portability problems.
            // Subject to later optimization if necessary.

            qry_string = new ostrstream;
            *qry_string << "UPDATE " << blob_tbl_name << " SET Content = \""
                        << tmp_escaped_data
                        << "\" WHERE BlobID = "
                        << Value
                        <<  ends;   
            int stor_success = CLI->thr->DB_Cmd(qry_string->str());
            if (stor_success != 0)
            {
               cout << "Internal error - in Kludge_Class_Property::Set_Blob() :\n"
                       " DB op failed while storing BLOB content!\n";
               result = 1;
            }
            else // OK, done storing the data :)
            {
               if (SubType == SUBTYPE_METHOD) UnParse_Method();
            }
            
            delete qry_string; qry_string = NULL;
            delete tmp_escaped_data; tmp_escaped_data = NULL;
         }
         else
         {
            cout << "Internal error - in Kludge_Class_Property::Set_Blob() :\n"
                    " while storing BLOB contents, couldn't allocate memory for the escaped string!\n";
            result = 1;
         }
      }
      // else something wrong with the SubType - error already reported

      delete blob_value;
   }
   else
   {
      if (result == 0) // obtained a viable value type
      {
         *CLI->output_stream << 
                 "Error - in Kludge_Class_Property::Set_Blob():\n"
                 " failed to decode the Base64 BLOB value.\n";
      
         result = 1;
      }
      // else invalid value type - error already reported
   }

   return(result);
}



int Kludge_Class_Property::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   if (new_value->type == PARM_VALUE)
   {
      if (SubType == SUBTYPE_PROPERTY)
      {
         Value = new_value->content;

         qry_string << "UPDATE Class_Properties SET Value = \""
                    << Value
                    << "\" WHERE Class = \""
                    << Class->Class
                    << "\" AND RDN = \""
                    << RDN
                    << "\""
                    <<  ends;
         result = CLI->thr->DB_Cmd(qry_string.str());
         if (result != 0)
         {
            *CLI->output_stream <<
               "Error: in Kludge_Class_Property::Set():\n"
               " error while storing the new value into the backend database.\n";
         }
      }
      else // a blob of some kind - Set_Blob() will object against insane SubTypes.
      {
         result = Set_Blob(new_value, CLI);
      }
   }
   else if (new_value->type == PARM_BLOB_VALUE)
   {
      if (SubType != SUBTYPE_PROPERTY)
      {  // a blob of some kind - Set_Blob() will object against insane SubTypes.
         result = Set_Blob(new_value, CLI);
      }
      else
      {
         *CLI->output_stream <<
            "Error - in Kludge_Class_Property::Set():\n"
            " obtained a Kludge_CLI_CMD_PARM of type PARM_BLOB_VALUE.\n"
            " Cannot set a BLOB value to a plain CLASS_PROPERTY.\n"
            " Cannot upgrade a plain CLASS_PROPERTY to a BLOB - to achieve this effect,\n"
            " you have to explicitly delete the property and insert a BLOB thingy.\n"
            " Consult your documentation for proper syntax.\n";
         //result = 0; // not an internal error
      }
   }
   else if (new_value->type == PARM_VARIABLE)
   {
      *CLI->output_stream <<
         "Error - in Kludge_Class_Property::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type PARM_VARIABLE.\n"
         " The \"variable\" format of the SET command is used with relationships,\n"
         " to set relationship attributes. Can't use this format with a property.\n"
         " To a property, you can only set a value.\n"
         " Consult your documentation for proper syntax.\n";
      //result = 0; // not an internal error
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class_Property::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_VALUE or PARM_BLOB_VALUE.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Class_Property::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 0;
 Kludge_Class* former_class = NULL;
 ostrstream qry_string;
 
   // can only move under another Kludge Class
   if (destination->Type == KLUDGE_CLASS)
   {
      if (destination->Resolve_Content_Name(&RDN,CLI,/*own only*/true) == NULL)
      {                          //^^^ an inherited property shall be overriden
         if (Class != NULL)
         {
            Add_To_Index((Kludge_Class*)destination);
            Free_From_Index();
            former_class = Class;
            Class = (Kludge_Class*)destination;
            
            qry_string << "UPDATE Class_Properties SET Class = \""
                       << Class->Class
                       << "\" WHERE Class = \""
                       << former_class->Class
                       << "\" AND RDN = \""
                       << RDN
                       << "\""
                       <<  ends;
            result = CLI->thr->DB_Cmd(qry_string.str());
            if (result != 0)
            {
               *CLI->output_stream <<
                  "Error while storing the new master class into the backend database.\n";
            }
         }
         else
         {
            *CLI->output_stream <<
            "In Kludge_Class_Property::Move(): can't move a class property with\n"
            " NULL master class. Probably an internal error - an attempt to move\n"
            " an unanchored class property.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Class_Property::Move(): content with this name already exists\n"
            " at the desired destination. Move canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "In Kludge_Class_Property::Move(): can only move under another kludge class.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class_Property::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 0;
 string old_name;
 ostrstream qry_string;
 
   if (new_name->type == PARM_NAME)
   {
      if (Class->Resolve_Content_Name(&new_name->content,CLI) == NULL)
      {
         Free_From_Index();
         old_name = RDN;
         RDN = new_name->content;
         Add_To_Index();

         qry_string << "UPDATE Class_Properties SET RDN = \""
                    << RDN
                    << "\" WHERE Class = \""
                    << Class->Class
                    << "\" AND RDN = \""
                    << old_name
                    << "\""
                    <<  ends;
         result = CLI->thr->DB_Cmd(qry_string.str());
         if (result != 0)
         {
            *CLI->output_stream <<
               "Error while storing the new name into the backend database.\n";
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Class_Property::Rename(): content with this name already exists\n"
            " in the master class. Rename operation cancelled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class_Property::Rename():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_NAME.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Class_Property::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;
 ostrstream qry_ss, qry_ss2;
 string blob_tbl_name;

   // Delete this Kludge_ClassProperty from the DB.
   // Can't be deallocated here - this function is used from within
   // a traversal function. Deallocate ('delete') it in the calling function.

   if (SubType != SUBTYPE_PROPERTY)
   {
      switch (SubType)
      {
         case SUBTYPE_BLOB:
         {
            blob_tbl_name = "Blobs";
         }
         break;
         case SUBTYPE_TXT_BLOB:
         {
            blob_tbl_name = "Text_Blobs";
         }
         break;
         case SUBTYPE_METHOD:
         {
            blob_tbl_name = "Text_Blobs";
         }
         break;
         default:
         {
            cout << "Internal error - in Kludge_Class_Property::DB_Delete() :\n"
                    " invalid Kludge_Class_Property::SubType!\n";
            result = 1;
         }
         break;
      }
      qry_ss2 << "DELETE FROM " << blob_tbl_name << " WHERE BlobID = \""
              << Value
              << "\""
              <<  ends;
      result = CLI->thr->DB_Cmd(qry_ss2.str());
   }
   // else not a BLOB - no dependent BLOB entry to be deleted

   qry_ss << "DELETE FROM Class_Properties WHERE Class = \""
          << Class->Class
          << "\" AND RDN = \""
          << RDN
          << "\""
          <<  ends;
   result += CLI->thr->DB_Cmd(qry_ss.str());

   return(result);
}



int Kludge_Class_Property::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   int result = DB_Delete(CLI, recursive);
   if (result == 0)
   {
      cout <<
         "Deleting class property: " << RDN << endl;
      result = Zap();
   }
   return(result);
}



int Kludge_Class_Property::pwd(string* result)
{
   if (Class != NULL)
   {
      Class->pwd(result);
   }
   else
   {
      cout << "Kludge_Class_Property::pwd(): Class is null (property uninitialized).\n";
      result->append("Internal error - pointer towards root is null. //");
   }
   result->append(RDN);
   return(0);
}



int Kludge_Class_Property::Add_To_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Class;

   cout << "Adding a class property: " << RDN << "\n"
        << " into the index of class " << Where->Name << "\n";
   
   // load properties if necessary? Nah, I don't think so - that'd be defensive programming.
   if (Where->Properties != NULL)
   {
      g_tree_insert(Where->Properties, &RDN,this);
      Where->UnInherit();
   }
   else
   {
      cout << "Kludge_Class_Property::Add_To_Index(): property index not initalized (pointer null).\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class_Property::Free_From_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Class;

   cout << "Releasing a class property: " << RDN << "\n"
        << " from the index of class " << Where->Name << "\n";
        
   // load properties if necessary? Nah, I don't think so - that'd be defensive programming.
   if (Where->Properties != NULL)
   {
      g_tree_remove(Where->Properties, &RDN);
      Where->UnInherit();
   }
   else
   {
      cout << "Kludge_Class_Property::Free_From_Index(): property index not initalized (pointer null).\n";
      result = 1;
   }

 return(result);
}



Kludge_Object* Kludge_Class_Property::Superior()
{
   return(Class);
}



int Kludge_Class_Property::UnParse_Method()
{
   if (Method_CV != NULL)
   {
      SvREFCNT_dec((SV*)Method_CV); // Perl's garbage collector has to take care of deallocation
      Method_CV = NULL;
   }

   return(0);
}



int Kludge_Class_Property::Parse_Method(Kludge_CLI* CLI)
{
 int result = 0;
 Kludge_MYSQL_RES* qry_result = NULL;

   if (SubType == SUBTYPE_METHOD)
   {
      if (Method_CV == NULL)
      {
         if ((qry_result = Get_Blob_Content(CLI) ))
         {
            // Hopefully the strings in MYSQL_ROW are naturally null-terminated.
            // If not, on the next line we're in trouble.
            if (! (( Method_CV = Kludge_pl_eval_sub(qry_result->Row[0]) )) )
            {
               result = 1;
               *CLI->output_stream <<
                "Error in Parse_Method(): the Perl eval() failed for some reason.\n";
            }
            delete qry_result;
         }
         else result = 2;
      }
      // else this has already been evaluated - we're finished and successful.
      // The Method_CV can only be valid (refcnt >= 1) or NULL.
   }
   else
   {
      *CLI->output_stream <<
       "Error in Parse_Method(): asked to eval() a Kludge Class Property style\n"
       " thingy that's not a METHOD. Instead, the thingy in question is a\n  ";
      Report_Subtype(CLI);
      *CLI->output_stream << endl;
      result = 3;
   }
   
   return(result);
}



int Kludge_Class_Property::Run_Method(Kludge_CLI* CLI, Kludge_pl_params* args)
{
 int result = 0;

   if (! (( result = Parse_Method(CLI) )) )  // if parsed fine
   {
      if (Kludge_pl_call_CV(Method_CV, args) != 0)
      {
         *CLI->output_stream <<
         "Error in Run_Method(): method execution failed.\n";
         result = 4;
      }
      // else method executed just fine
   }
   // else result already contains the error code and error already reported
   
   return(result);
}










/* ###    class Kludge_Class_Rel_Stub    ### */

// The Kludge_Class_Relationship (a Kludge_Object look-alike) has two superiors,
// and sometimes we want its methods to return different results based on through
// which superior the method was called. The only way to implement this behavior
// is probably to have the Kludge_Class_Relationship contain two stubs, that are
// themselves true heirs of Kludge_Object (the Kludge_Class_Relationship is not)
// but otherwise only serve as wrappers to treat the ambiguities of class
// Kludge_Class_Relationship. The master and slave stubs have a single clearly
// superior Kludge_Class and will tell the member method of the respective
// Kludge_Class_Relationship which end (master/slave) is calling it - or add
// proper differentiated behavior.
// Most of the code is embedded in the methods of Kludge_Class_Relationship - 
// the stubs only host the necessary bivalent code.
// The two stubs are truly welded on to the relationship's "body" - they're 
// initialized and deleted along with the body.

Kludge_Class_Rel_Stub::Kludge_Class_Rel_Stub(Rel_Stub_Type stub_type, Kludge_Class_Relationship* rel_body) :
  Stub_Type(stub_type),
  Body(rel_body),
  Owner(NULL),
  Name(""),
  MinCount(0),
  MaxCount(1),
  NamingGen(""),
  NamingMatch(""),
  NamingMatchCompiled(NULL)
{
   Type = KLUDGE_CLASS_REL_STUB;
}



Kludge_Class_Rel_Stub::~Kludge_Class_Rel_Stub()
{
   if (NamingMatchCompiled != NULL) pcre_free( (void*)NamingMatchCompiled );
   // Come to think of it, this may not clean up everything it should.
   // I should really inspect this with some memory leak detection tool.
}



int Kludge_Class_Rel_Stub::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   *CLI->output_stream
      << "CLASS_RELATIONSHIP" 
      << "\n{\n";
      
   result = Body->List_Contents(CLI);

   *CLI->output_stream << "};\n";

   return(result);
}



int Kludge_Class_Rel_Stub::Describe_Self(Kludge_CLI* CLI)
{
   *CLI->output_stream 
      << "CLASS_RELATIONSHIP \""
      << Name
      << "\";\n";

   return(0);
}



Kludge_Object* Kludge_Class_Rel_Stub::Resolve_Content_Name(
             string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
 Kludge_Object* result = NULL;

   if (*Content_Name == "..")            // a higher-level "folder" in the path?
   {
      result = Owner;
   }
   else if (*Content_Name == ".")        // a "this" style reference?
   {
      result = this;
   }
   else if (*Content_Name == "MASTER")   // a pointer to the master Kludge Class?
   {
      result = Body->Master_Stub.Owner;
   }
   else if (*Content_Name == "SLAVE")    // a pointer to the slave Kludge Class?
   {
      result = Body->Slave_Stub.Owner;
   }
   else if (*Content_Name == "PEER")    // a pointer to the other party to the relationship?
   {
      result = ((Stub_Type == MASTER) ? Body->Slave_Stub.Owner : Body->Master_Stub.Owner);
   }

   return(result);
}



int Kludge_Class_Rel_Stub::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
   return(Body->DB_Delete(CLI));
}



int Kludge_Class_Rel_Stub::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   return(Body->Delete(CLI));
}



int Kludge_Class_Rel_Stub::Zap()
{
   return(Body->Zap());
}



int Kludge_Class_Rel_Stub::Init_Owner(Kludge_CLI* CLI)
{
 int result = 0;

   if (Owner != NULL)
   {
      if ( Owner->Relationships == NULL ) Owner->Init_Relationships();
   }
   else
   {
      cout << "Kludge_Class_Rel_Stub::Init(): asked to point to a null Owner!";
      result = 1;
   }

   return(result);
}



int Kludge_Class_Rel_Stub::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   return(Body->Insert(obj_to_insert,CLI));
}



int Kludge_Class_Rel_Stub::Add_To_Owner_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Owner;

   cout << "Adding class relationship stub " << Name << " to its Owner's index.\n";
   if (Where != NULL)
   {
      if (Where->Relationships != NULL)
      {
         g_tree_insert(Where->Relationships,&Name,this);
         Where->UnInherit();
      }
      else
      {
         cout << 
            "Internal error - in Kludge_Class_Rel_Stub::Add_To_Owner_Index(): \n"
            " the Owner's \"relationships\" index not initialized!\n";
         result = 1;
      }
   }
   else
   {
      cout <<
         "Internal error - In Kludge_Class_Rel_Stub::Add_To_Owner_Index():\n"
         " owner tap undefined! (pointer NULL)\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Class_Rel_Stub::Free_From_Owner_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Owner;

   cout << "Releasing class relationship stub" << Name
        << " from its Owner's index.\n";

   if (Where != NULL)
   {
      if (Where->Relationships != NULL)
      {
         g_tree_remove(Where->Relationships,&Name);
         Where->relationships_loaded = false; // not necessarily true when called from Delete(), but who cares
         Where->UnInherit();
      }
      else
      {
         cout << 
            "Internal error - In Kludge_Class_Rel_Stub::Free_From_Owner_Index(): \n"
            " the Owner's \"relationships\" index not initialized!\n";
         result = 1;
      }
   }
   else
   {
      cout << 
         "Internal error - In Kludge_Class_Rel_Stub::Free_From_Owner_Index(): \n"
         " owner tap undefined! (pointer NULL)\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class_Rel_Stub::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   // can only move under another Kludge Class
   if (destination->Type == Kludge_Object::KLUDGE_CLASS)
   {
      if (destination->Resolve_Content_Name(&Name,CLI,/*own only*/true) == NULL)
      {                        //^^^ inherited paraphernalia shall be overridden
         if (Owner != NULL)
         {
            Add_To_Owner_Index((Kludge_Class*)destination);
            Free_From_Owner_Index();
            Owner = (Kludge_Class*)destination;
            
            qry_string << "UPDATE Class_Relationships SET "
                       << ((Stub_Type == MASTER) ? "Master" : "Slave")
                       << " = \""
                       << Owner->Class
                       << "\" WHERE RelCLSID = \""
                       << Body->RelCLSID
                       << "\""
                       <<  ends;
            result = CLI->thr->DB_Cmd(qry_string.str());
            if (result != 0)
            {
               *CLI->output_stream <<
                  "Error while storing the new stub owner pointer into the backend database.\n";
            }
         }
         else
         {
            *CLI->output_stream <<
            "In Kludge_Class_Rel_Stub::Move(): can't move a class relationship stub\n"
            " with null Owner. Probably an internal error - an attempt to move\n"
            " an unanchored class relationship.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Class_Rel_Stub::Move(): content with this name already exists\n"
            " at the desired destination. Move canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "In Kludge_Class_Rel_Stub::Move(): can only move under another kludge class.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class_Rel_Stub::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   if (new_name->type == PARM_NAME)
   {
      if (Owner->Resolve_Content_Name(&new_name->content,CLI,/*own only?*/true) == NULL)
      {
         Free_From_Owner_Index();
         Name = new_name->content;
         Add_To_Owner_Index();

         qry_string << "UPDATE Class_Relationships SET "
                    << ((Stub_Type == MASTER) ? "Name_M" : "Name_S")
                    << " = \""
                    << Name
                    << "\" WHERE RelCLSID = \""
                    << Body->RelCLSID
                    << "\""
                    <<  ends;
         result = CLI->thr->DB_Cmd(qry_string.str());
         if (result != 0)
         {
            *CLI->output_stream <<
               "In Kludge_Class_Rel_Stub::Rename():\n"
               " error while storing the new name into the backend database.\n";
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Class_Rel_Stub::Rename(): content with this name already exists\n"
            " in the master class. Rename operation canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class_Rel_Stub::Rename():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_NAME.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Class_Rel_Stub::pwd(string* result)
{
   // Hmm. Perhaps this should better return the path to the master. I dunno.
   if (Owner != NULL)
   {
      Owner->pwd(result);
   }
   else
   {
      cout << "Kludge_Class_Rel_Stub::pwd(): Owner is null (Rel_Stub uninitialized).\n";
      result->append("Internal error - pointer towards root is null. //");
   }
   result->append(Name);
   
   return(0);
}



int Kludge_Class_Rel_Stub::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
   return(Body->Set(new_value, CLI));
}



Kludge_Object* Kludge_Class_Rel_Stub::Superior()
{
   return(Owner);
}



int Kludge_Class_Rel_Stub::SetNamingMatch(string* pattern, Kludge_CLI* CLI)
{
 int result = 0;
   
   NamingMatch = *pattern;
   result = Compile_Regex(CLI);
   if (result != 0)
   {
      NamingMatch = "";
      *CLI->output_stream << 
      "The suggested NAMINGMATCH pattern is not a valid regexp.\n"
      " The pattern failed to pass a compilation attempt.\n"
      " The respective NAMINGMATCH variable is reset to \"\" (an empty string).\n";
   }

   return(result);
}



int Kludge_Class_Rel_Stub::Compile_Regex(Kludge_CLI* CLI)
{
 int result = 0;
 const char *compile_error = NULL;
 int compile_err_offset;

   if (NamingMatch.size() != 0)
   {
      NamingMatchCompiled = pcre_compile(
                  NamingMatch.c_str(),  // the regex pattern
                  0,                    // default options
                  &compile_error,       // for error message
                  &compile_err_offset,  // for error offset
                  NULL                  // use default character tables
                 );

      if (NamingMatchCompiled == NULL)
      {
         *CLI->output_stream <<
            "NAMINGMATCH regexp compilation failed at character # "
            << compile_err_offset << ".\n"
            " Reason: "<< compile_error << "\n"
            << NamingMatch << "\n";
         for (int i=0; i < compile_err_offset - 1; i++)
         {
            *CLI->output_stream << " ";
         }
         *CLI->output_stream << "^\n";

         result = 1;
      }

      if (compile_error != NULL)
      {
         pcre_free((void*)compile_error);
      }
   }
   // else don't bother

   return(result);
}



int Kludge_Class_Rel_Stub::Match_Name(string* subject, Kludge_CLI* CLI)
{
#define MATCH_ARR_DIM 30    // should be a multiple of 3

 int result = 0;
 int array_of_matches[MATCH_ARR_DIM];
 int match_count;

   if (NamingMatch.size() != 0)
   {
      if (NamingMatchCompiled == NULL) Compile_Regex(CLI);

      if (NamingMatchCompiled != NULL)
      {
         match_count = pcre_exec(
                 NamingMatchCompiled,  // the compiled pattern
                 NULL,                 // no extra data - we didn't study the pattern
                 subject->c_str(),     // the subject string
                 subject->size(),      // the length of the subject
                 0,                    // start at offset 0 in the subject
                 0,                    // default options
                 array_of_matches,     // output vector for substring information
                 MATCH_ARR_DIM         // size of the output vector in sizeof(int)
         );

         if (match_count < 0)
         {
            switch (match_count)
            {
               case PCRE_ERROR_NOMATCH:
               {
                  *CLI->output_stream <<
                  "The suggested name does not match the applicable NAMINGMATCH restriction.\n";
               }
               break;
               default:
               {
                  *CLI->output_stream <<
                  "Internal error while applying the NAMINGMATCH regexp to the suggested name.\n"
                  " Regexp compiled fine and it's not that the name would not match -\n"
                  " - something else went wrong.\n";
               }
               break;
            }

            result = 1;
         }
         else if (match_count == 1)
         {
            // Exactly one match found. As good as it gets.
            // No need to comment on that - no message to the user.
         }
         else if (match_count == 0)
         {
            cout << "Warning - while applying the NAMINGMATCH regexp to the suggested name:\n"
                    " several matches found. This would suggest a poorly designed matching\n"
                    " rule in the relevant CLASS_RELATIONSHIP. Ask the class model maintainer.\n";
            // Not really an error.
         }
      }
      else
      {
         // Compile_Regex() failed - and most certainly reported an error.
         result = 1;
      }
   }
   // else don't bother - anything matches.

   return(result);
}








/* ###    class Kludge_Class_Relationship    ### */

// This constructor is used for creating a new class relationship from scratch.
// It cannot be used for loading a class relationship from the DB.
//
// At this constructor, I had to strike a reasonable tradeoff.
// The transformation of Kludge_CLI_CMD_PARM* CLASS_RELATIONSHIP into
// a Kludge_Class_Relationship is non-trivial and, if I wanted to use
// a heavy-weight constructor as with Kludge_Class or K_C_Property, I'd
// end up passing some fourteen variables - or a prototype-style object.
// So I decided to use a lightweight constructor, do the transform over
// the uninitialized object and use a stand-alone initializer method
// to set the finished Kludge_Class_Relationship airborne.
//
// To store the newly created class relationship in the DB and hook it up
// in the trees and indexes, use Kludge_Class_Relationship::Init().
//
Kludge_Class_Relationship::Kludge_Class_Relationship() :
  Master_Stub(Kludge_Class_Rel_Stub::MASTER, this),
  Slave_Stub(Kludge_Class_Rel_Stub::SLAVE, this),
  RelCLSID(0),  // will be established upon storage to the database
  RelType(1), // 0 = peer2peer, 1 = master/slave, 2 = ownership
  DefaultCount_M(0),
  Instances(NULL),
  Suppress(false),
  ctor_failed(false)
{
}

int Kludge_Class_Relationship::Init(Kludge_CLI* CLI)
{
 int stor_success = 0, result = 0;
 ostrstream qry_string;

   //Type = KLUDGE_CLASS_RELATIONSHIP; // not a Kludge_Object heir anymore
   result = Master_Stub.Init_Owner(CLI) + Slave_Stub.Init_Owner(CLI); // :)
   
   if (result == 0)
   {
      qry_string << "INSERT INTO Class_Relationships VALUES (NULL,\""
                 << Master_Stub.Owner->Class << "\",\""
                 << Slave_Stub.Owner->Class << "\",\""
                 << RelType << "\",\""
                 << Master_Stub.Name << "\",\""
                 << Slave_Stub.Name << "\",\""
                 << Master_Stub.MinCount << "\",\""
                 << Master_Stub.MaxCount << "\",\""
                 << DefaultCount_M << "\",\""
                 << Master_Stub.NamingGen << "\",\""
                 << Master_Stub.NamingMatch << "\",\""
                 << Slave_Stub.MinCount << "\",\""
                 << Slave_Stub.MaxCount << "\",\""
                 << Slave_Stub.NamingGen << "\",\""
                 << Slave_Stub.NamingMatch << "\",\""
                 << Suppress << "\")"
                 <<  ends ; 

      stor_success = CLI->thr->DB_Cmd(qry_string.str());
      if (stor_success == 0)
      {
         RelCLSID = mysql_insert_id(CLI->thr->DB);
         Add_To_Index();
      }
      else
      {
         cout << "Storage of a newly created class relationship "
              << Master_Stub.Name << "/" << Slave_Stub.Name << " has failed!\n";
         result = 1;
      }
   }

   return(result);
}



// This constructor can only be used for loading an existing 
// Kludge_Class_Relationship from a database. It is not intended for creating
// a new class relationship from scratch.
//
Kludge_Class_Relationship::Kludge_Class_Relationship(
  MYSQL_ROW DB_record,
  Kludge_CLI* CLI
) : 
  Master_Stub(Kludge_Class_Rel_Stub::MASTER, this),
  Slave_Stub(Kludge_Class_Rel_Stub::SLAVE, this),
  RelCLSID(0),
  RelType(-1), // 0 = peer2peer, 1 = master/slave, 2 = ownership
  DefaultCount_M(0),
  Instances(NULL),
  Suppress(false),
  ctor_failed(false)
{
   //Type = KLUDGE_CLASS_RELATIONSHIP;

   if ( ! Extract_From_MysqlRow(DB_record,CLI) )
   {
      cout << "Extraction of a class relationship out of a DB record has failed!\n";
      ctor_failed = true;
   }
   else
   {
      // Master and Slave are both loaded (by Extract_From_MysqlRow()
      Master_Stub.Init_Owner(CLI);
      Slave_Stub.Init_Owner(CLI);
      Add_To_Index();
   }
}



Kludge_Class_Relationship::~Kludge_Class_Relationship()
{
   cout << "Deallocating class relationship (name not available)\n";
   Free_Instances();
   Free_From_Global_Index();
}



// Zap() does a nice memory cleanup.
// Only makes sense 
// a) for dynamic partial release from memory 
//     In that case, invalidate the list of relationships.
// b) when a class relationship is deleted
//     Requires deletion from the SQL-based object repository
//      - i.e., use DB_Delete() and then Zap() (or just Delete())
int Kludge_Class_Relationship::Zap()
{
 int result = 0;
 bool inst_OK = true;

   cout << "Zapping a class relationship: " << Master_Stub.Name << "<==>" << Slave_Stub.Name << "\n" ;

   if (Instances != NULL)
   {
      if (g_tree_nnodes(Instances) > 0)
      {
         cout <<
            "Cannot zap CLASS_RELATIONSHIP "
            << Master_Stub.Owner->Name << "/" << Master_Stub.Name << "\n" <<
            " There are dependent objects loaded in memory.\n";

         inst_OK = false;
         result = 1;
      }
   }

   if (inst_OK)
   {
      result = Free_From_Index();
      
      delete this;
   }
   
   return(result);
}



//
// Extract_From_MysqlRow()
//
// Expects:
// Row_To_Process[0] == RelCLSID
// Row_To_Process[1] == Master
// Row_To_Process[2] == Slave
// Row_To_Process[3] == Type
// Row_To_Process[4] == Name_M
// Row_To_Process[5] == Name_S
// Row_To_Process[6] == MinCount_M
// Row_To_Process[7] == MaxCount_M
// Row_To_Process[8] == DefaultCount_M
// Row_To_Process[9] == NamingGen_M
// Row_To_Process[10] == NamingMatch_M
// Row_To_Process[11] == MinCount_S
// Row_To_Process[12] == MaxCount_S
// Row_To_Process[13] == NamingGen_S
// Row_To_Process[14] == NamingMatch_S
// Row_To_Process[15] == Suppress
//
bool Kludge_Class_Relationship::Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI) 
{
 bool result = true;
 OID_Type master_ID = 0, slave_ID = 0;
 istrstream* tmp_ss = NULL; // for MY_SSCANF

   if (Row_To_Process != NULL)
   {
      MY_SSCANF(Row_To_Process[0],RelCLSID);
      MY_SSCANF(Row_To_Process[1],master_ID);
      MY_SSCANF(Row_To_Process[2],slave_ID);
      Master_Stub.Owner = Kludge_Class::ID2ptr(&master_ID, CLI);
      Slave_Stub.Owner = Kludge_Class::ID2ptr(&slave_ID, CLI);
      if (Master_Stub.Owner == NULL)
      {
         cout << 
            "In Kludge_Class_Relationship::Extract_From_MysqlRow():\n"
            " Couldn't resolve the Master class ID (dangling reference).\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }
      if (Slave_Stub.Owner == NULL)
      {
         cout << 
            "In Kludge_Class_Relationship::Extract_From_MysqlRow():\n"
            " Couldn't resolve the Slave class ID (dangling reference).\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }
      MY_SSCANF(Row_To_Process[3],RelType);
      Master_Stub.Name = Row_To_Process[4];
      Slave_Stub.Name = Row_To_Process[5];
      MY_SSCANF(Row_To_Process[6],Master_Stub.MinCount);
      MY_SSCANF(Row_To_Process[7],Master_Stub.MaxCount);
      MY_SSCANF(Row_To_Process[8],DefaultCount_M);
      Master_Stub.NamingGen = Row_To_Process[9];
      Master_Stub.NamingMatch = Row_To_Process[10];
      MY_SSCANF(Row_To_Process[11],Slave_Stub.MinCount);
      MY_SSCANF(Row_To_Process[12],Slave_Stub.MaxCount);
      Slave_Stub.NamingGen = Row_To_Process[13];
      Slave_Stub.NamingMatch = Row_To_Process[14];
      // Set Kludge_Class_Relationship.Suppress
      if ( (strcmp(Row_To_Process[15],"0")) || (strlen(Row_To_Process[15])==0) ) 
         Suppress = false;
      else Suppress = true;
   }
   else
   {
      cout << 
         "In Kludge_Class_Relationship::Extract_From_MysqlRow():\n"
         " the row passed to this function is a NULL pointer!\n"
         " Extract_From_MysqlRow() failed.\n";
      result = false;
   }

   return(result);
}



int Kludge_Class_Relationship::Init_Instances()
{
   if (Instances == NULL)
   {
      Instances = g_tree_new(Kludge_Compare_IDs);
   }
   // else already initialized - nothing to be done.

   return(0);
}



bool Kludge_Class_Relationship::Free_Instances()
{
   if (Instances != NULL)
   {
      if (g_tree_nnodes(Instances) > 0)
      {
         cout << "In Kludge_Class_Relationship::Free_Instances():\n"
                 " there are still instances left on the heap!\n";
         // I don't plan to free instances recursively from their classes.
         // On shutdown, instances should be torn down first, then classes.
         // I guess this is more civilized.
         // Let's make this a serious warning.
      }
      
      g_tree_destroy(Instances);
      Instances = NULL;
   }
   // else instances not initialized - probably there was no reason

   return(true);
}



int Kludge_Class_Relationship::Add_To_Index()
{
 int result = 0;
 
   result = Master_Stub.Add_To_Owner_Index();
   result += Slave_Stub.Add_To_Owner_Index();
   result += Add_To_Global_Index();
   
   return(result);
}



int Kludge_Class_Relationship::Free_From_Index()
{
 int result = 0;
 
   result = Master_Stub.Free_From_Owner_Index();
   result += Slave_Stub.Free_From_Owner_Index();
   //Free_From_Global_Index(); // this is done by the dtor

   return(result);
}



int Kludge_Class_Relationship::Add_To_Global_Index()
{
   cout << "Adding class relationship " << Master_Stub.Name << "/" << Slave_Stub.Name << " to global index.\n";
   g_tree_insert(Index,&RelCLSID,this);
   return(0);
}



int Kludge_Class_Relationship::Free_From_Global_Index()
{
   cout << "Releasing class relationship " << Master_Stub.Name << "/" << Slave_Stub.Name << " from global index.\n";
   g_tree_remove(Index,&RelCLSID);
   return(0);
}



int Kludge_Class_Relationship::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
 string *tmp_pwd_m = NULL, *tmp_pwd_s = NULL;
 string rel_type;

   tmp_pwd_m = Kludge_CLI_PWD(Master_Stub.Owner,CLI->output_stream);
   tmp_pwd_s = Kludge_CLI_PWD(Slave_Stub.Owner,CLI->output_stream);
   switch (RelType)
   {
      case 0 : rel_type = "P2P";
      break;
      case 1 : rel_type = "M/S";
      break;
      case 2 : rel_type = "OWN";
      break;
      default :
         cout << "In Kludge_Class_Relationship::List_Contents(): unknown relationship type.\n";
      break;
   }

   *CLI->output_stream <<
      "MASTER=\""         << *tmp_pwd_m                 << "\";\n"
      "SLAVE=\""          << *tmp_pwd_s                 << "\";\n"
      "TYPE=\""           << rel_type                   << "\";\n"
      "NAME_M=\""         << Master_Stub.Name           << "\";\n"
      "NAME_S=\""         << Slave_Stub.Name            << "\";\n"
      "MINCOUNT_M=\""     << Master_Stub.MinCount       << "\";\n"
      "MAXCOUNT_M=\""     << Master_Stub.MaxCount       << "\";\n"
      "DEFAULTCOUNT_M=\"" << DefaultCount_M             << "\";\n"
      "NAMINGGEN_M=\""    << Master_Stub.NamingGen      << "\";\n"
      "NAMINGMATCH_M=\""  << Master_Stub.NamingMatch    << "\";\n"
      "MINCOUNT_S=\""     << Slave_Stub.MinCount        << "\";\n"
      "MAXCOUNT_S=\""     << Slave_Stub.MaxCount        << "\";\n"
      "NAMINGGEN_S=\""    << Slave_Stub.NamingGen       << "\";\n"
      "NAMINGMATCH_S=\""  << Slave_Stub.NamingMatch     << "\";\n";
      //"SUPPRESS="

   if (tmp_pwd_m != NULL) delete tmp_pwd_m;
   if (tmp_pwd_s != NULL) delete tmp_pwd_s;

   return(0);
}



int Kludge_Class_Relationship::DB_Delete(Kludge_CLI* CLI)
{
 int result = 0;
 MYSQL_RES *mysql_query_result = NULL;
 ostrstream qry_string, check_inst, check_inst_prop, check_inst_rel;
 bool inst_OK = false, inst_prop_OK = false, inst_rel_OK = false;

   // Delete this Kludge_Class_Relationship from the DB.
   // Can't be deallocated here - this function is used from within
   // a traversal function. Deallocate ('delete') it in the calling function.

   // before attempting the deletion, check for dependent instance-side objects:
   check_inst      << "SELECT OID FROM Instances WHERE ClsRel = \""
                   << RelCLSID << "\"" << ends;
   mysql_query_result = CLI->thr->DB_Query(check_inst.str());
   if (mysql_query_result != NULL)
   {
      if (mysql_fetch_row(mysql_query_result) == NULL) inst_OK = true;
      mysql_free_result(mysql_query_result);
   }

   check_inst_prop << "SELECT OID FROM Instance_Properties WHERE ClsRel = \""
                   << RelCLSID << "\"" << ends;
   mysql_query_result = CLI->thr->DB_Query(check_inst_prop.str());
   if (mysql_query_result != NULL)
   {
      if (mysql_fetch_row(mysql_query_result) == NULL) inst_prop_OK = true;
      mysql_free_result(mysql_query_result);
   }
   
   check_inst_rel  << "SELECT RelOID FROM Instance_Relationships WHERE ClsRel = \""
                   << RelCLSID << "\"" << ends;
   mysql_query_result = CLI->thr->DB_Query(check_inst_rel.str());
   if (mysql_query_result != NULL)
   {
      if (mysql_fetch_row(mysql_query_result) == NULL) inst_rel_OK = true;
      mysql_free_result(mysql_query_result);
   }
   // done with the necessary checks

   if (inst_OK && inst_prop_OK && inst_rel_OK)
   {
      qry_string << "DELETE FROM Class_Relationships WHERE RelCLSID = \""
                 << RelCLSID << "\""
                 <<  ends;
      result = CLI->thr->DB_Cmd(qry_string.str());
   }
   else
   {
      *CLI->output_stream <<
         "Cannot delete the following CLASS_RELATIONSHIP:\n "
         << Master_Stub.Owner->Name << "/" << Master_Stub.Name << "\n" <<
         " There are dependent instance-side objects.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Class_Relationship::Delete(Kludge_CLI* CLI)
{
   int result = DB_Delete(CLI);
   if (result == 0)
   {
      cout <<
         "Deleting class relationship: " << Master_Stub.Name << 
         "<==>" << Slave_Stub.Name << endl;
      result = Zap();
   }

   return(result);
}



int Kludge_Class_Relationship::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
   "Cannot insert anything in a Kludge Class Relationship.\n"
   " For the moment, class relationships are relatively dumb leaf objects.\n"
   " You can only insert things into classes and instances.\"";

   return(0);
}



// If someone asks for the PWD to a Kludge_Class_Relationship,
// let's take the master route (even with P2P relationships).
int Kludge_Class_Relationship::pwd(string* result)
{
   return(Master_Stub.pwd(result));
}



int Kludge_Class_Relationship::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
 int result = 0;
 bool user_error = false;
 ostrstream qry_string;
 istrstream* tmp_stream;
 string name;
 ostrstream value;

   if (new_value->type == PARM_VARIABLE)
   {
      if ((new_value->contents[0]->type == PARM_NAME) && (new_value->contents[1]->type == PARM_VALUE))
      {
         // parse attribute name and value
         if (new_value->contents[0]->content == "MINCOUNT_M")
         {
            // value should not be an empty string
            if (new_value->contents[1]->content.size() != 0)
            {
               tmp_stream = new istrstream(new_value->contents[1]->content.c_str());
               if (!(*tmp_stream >> Master_Stub.MinCount))
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Class_Relationship::Set():\n"
                     " MINCOUNT_M must be a decimal number. Value NOT set.\n";
                  user_error = true;
               }
               // else value parsed successfully.

               delete tmp_stream;
            }
            else
            {
               // value is NULL - assuming default (0)
               Master_Stub.MinCount = 0;
            }

            if (! user_error)
            {
               name = "MinCount_M";
               value << Master_Stub.MinCount << ends;
            }
         }
         else if (new_value->contents[0]->content == "MAXCOUNT_M")
         {
            // value should not be an empty string
            if (new_value->contents[1]->content.size() != 0)
            {
               tmp_stream = new istrstream(new_value->contents[1]->content.c_str());
               if (!(*tmp_stream >> Master_Stub.MaxCount))
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Class_Relationship::Set():\n"
                     " MAXCOUNT_M must be a decimal number. Value NOT set.\n";
                  user_error = true;
               }
               // else value parsed successfully.

               delete tmp_stream;
            }
            else
            {
               // value is NULL - assuming default (0)
               Master_Stub.MaxCount = 0;
            }

            if (! user_error)
            {
               name = "MaxCount_M";
               value << Master_Stub.MaxCount << ends;
            }
         }
         else if (new_value->contents[0]->content == "DEFAULTCOUNT_M")
         {
            // value should not be an empty string
            if (new_value->contents[1]->content.size() != 0)
            {
               tmp_stream = new istrstream(new_value->contents[1]->content.c_str());
               if (!(*tmp_stream >> DefaultCount_M))
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Class_Relationship::Set():\n"
                     " DEFAULTCOUNT_M must be a decimal number. Value NOT set.\n";
                  user_error = true;
               }
               // else value parsed successfully.

               delete tmp_stream;
            }
            else
            {
               // value is NULL - assuming default (0)
               DefaultCount_M = 0;
            }

            if (! user_error)
            {
               name = "DefaultCount_M";
               value << DefaultCount_M << ends;
            }
         }
         else if (new_value->contents[0]->content == "NAMINGGEN_M")
         {
            Master_Stub.NamingGen = new_value->contents[1]->content;
            name = "NamingGen_M";
            value << Master_Stub.NamingGen << ends;
         }
         else if (new_value->contents[0]->content == "NAMINGMATCH_M")
         {
            if (Master_Stub.SetNamingMatch(&new_value->contents[1]->content, CLI) == 0)
            {
               name = "NamingMatch_M";
               value << Master_Stub.NamingMatch << ends;
            }
            else user_error = true;
         }
         else if (new_value->contents[0]->content == "MINCOUNT_S")
         {
            // value should not be an empty string
            if (new_value->contents[1]->content.size() != 0)
            {
               tmp_stream = new istrstream(new_value->contents[1]->content.c_str());
               if (!(*tmp_stream >> Slave_Stub.MinCount))
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Class_Relationship::Set():\n"
                     " MINCOUNT_S must be a decimal number. Value NOT set.\n";
                  user_error = true;
               }
               // else value parsed successfully.

               delete tmp_stream;
            }
            else
            {
               // value is NULL - assuming default (0)
               Slave_Stub.MinCount = 0;
            }

            if (! user_error)
            {
               name = "MinCount_S";
               value << Slave_Stub.MinCount << ends;
            }
         }
         else if (new_value->contents[0]->content == "MAXCOUNT_S")
         {
            // value should not be an empty string
            if (new_value->contents[1]->content.size() != 0)
            {
               tmp_stream = new istrstream(new_value->contents[1]->content.c_str());
               if (!(*tmp_stream >> Slave_Stub.MaxCount))
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Class_Relationship::Set():\n"
                     " MAXCOUNT_S must be a decimal number. Value NOT set.\n";
                  user_error = true;
               }
               // else value parsed successfully.

               delete tmp_stream;
            }
            else
            {
               // value is NULL - assuming default (0)
               Slave_Stub.MaxCount = 0;
            }

            if (! user_error)
            {
               name = "MaxCount_S";
               value << Slave_Stub.MaxCount << ends;
            }
         }
         else if (new_value->contents[0]->content == "NAMINGGEN_S")
         {
            Slave_Stub.NamingGen = new_value->contents[1]->content;
            name = "NamingGen_S";
            value << Slave_Stub.NamingGen << ends;
         }
         else if (new_value->contents[0]->content == "NAMINGMATCH_S")
         {
            if (Slave_Stub.SetNamingMatch(&new_value->contents[1]->content, CLI) == 0)
            {
               name = "NamingMatch_S";
               value << Slave_Stub.NamingMatch << ends;
            }
            else user_error = true;
         }
         else if (new_value->contents[0]->content == "TYPE")
         {
            if (new_value->contents[1]->content == "P2P")
            {
               RelType = 0;
            }
            else if (new_value->contents[1]->content == "M/S")
            {
               RelType = 1;
            }
            else if (new_value->contents[1]->content == "OWN")
            {
               RelType = 2;
            }
            else
            {
               *CLI->output_stream <<
                  "Error - in Kludge_Class_Relationship::Set():\n"
                  " unknown TYPE. Known TYPEs are \"P2P\", \"M/S\" and \"OWN\".\n";
               user_error = true;
            }

            if (! user_error)
            {
               name = "Type";
               value << RelType << ends;
            }
         }
         else if (new_value->contents[0]->content == "NAME_M")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Class_Relationship::Set():\n"
               " can't modify NAME_M - use the RENAME command.\n";
            user_error = true;
         }
         else if (new_value->contents[0]->content == "NAME_S")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Class_Relationship::Set():\n"
               " can't modify NAME_S - use the RENAME command.\n";
            user_error = true;
         }
         else if (new_value->contents[0]->content == "MASTER")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Class_Relationship::Set():\n"
               " can't modify MASTER - use the MOVE command.\n";
            user_error = true;
         }
         else if (new_value->contents[0]->content == "SLAVE")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Class_Relationship::Set():\n"
               " can't modify SLAVE - use the MOVE command.\n";
            user_error = true;
         }
         else
         {
            *CLI->output_stream <<
               "Error - in Kludge_Class_Relationship::Set():\n"
               " unknown attribute name: \"" << new_value->contents[0]->content <<
               "\".\n";
            user_error = true;
         }
         // done parsing attribute name and value

         if (user_error = false)
         {
            qry_string << "UPDATE Class_Relationships SET "
                       << name 
                       << " = \""
                       << value.str() 
                       << "\" WHERE RelCLSID = \""
                       << RelCLSID
                       << "\""
                       <<  ends;

            result = CLI->thr->DB_Cmd(qry_string.str());
            if (result != 0)
            {
               *CLI->output_stream <<
                  "Error while storing the new attriute value into the backend database.\n";
               result = 1;
            }
         }
         //else nothing to do, error already reported.
      }
      else
      {
         *CLI->output_stream <<
            "Internal error - in Kludge_Class_Relationship::Set():\n"
            " obtained a PARM_VARIABLE with improper contents.\n";
         result = 1;
      }
   }
   else if (new_value->type == PARM_VALUE)
   {
      *CLI->output_stream <<
         "Error - in Kludge_Class_Relationship::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type PARM_VALUE.\n"
         " The \"value\" format of the SET command is used with properties,\n"
         " to set property value. Can't use this format with a relationship.\n"
         " You can only set an attribute=value with a relationship.\n"
         " Consult your documentation for proper syntax.\n";
      //result = 0; // not an internal error
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Class_Relationship::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_VARIABLE or PARM_VALUE.\n";
      result = 1;
   }
   
   return(result);
}



//static
Kludge_Class_Relationship* Kludge_Class_Relationship::ID2ptr_Raw(OID_Type* ID)
{
 Kludge_Class_Relationship* result = NULL;

   result = (Kludge_Class_Relationship *) g_tree_lookup(Index,ID);
   
   return(result);
}



//static
Kludge_Class_Relationship* Kludge_Class_Relationship::ID2ptr(OID_Type* ID, Kludge_CLI* CLI)
{
 Kludge_Class_Relationship* result = NULL;
 Kludge_MYSQL_RES* tmp_mysql_res = NULL;

   result = ID2ptr_Raw(ID);
   if (result == NULL)
   {
      tmp_mysql_res = DB_Lookup(*ID, CLI);
      if (tmp_mysql_res != NULL)
      {
         result = new Kludge_Class_Relationship(tmp_mysql_res->Row,CLI);
         
         if (result != NULL)
         {
            if (! result->ctor_failed)
            {
               // do nothing - just return result
            }
            else
            {
               *CLI->output_stream <<
                  "Internal error - in Kludge_C_R::ID2ptr():\n"
                  " failed to load a Kludge_Class_Relationship \n"
                  " - ctor probably failed to interpret the passed DB record.\n";
               delete result; result = NULL;
            }
         }
         else
         {
            *CLI->output_stream <<
               "Internal error - in Kludge_C_R::ID2ptr():\n"
               " failed to construct a Kludge_Class_Relationship,\n"
               " probably due to a failed malloc().\n";
         }
         
         delete tmp_mysql_res; tmp_mysql_res = NULL;
      }
      else
      {
         cout << "Internal error - in Kludge_C_R::ID2ptr: couldn't find the requested ID in the database!\n";
      }
   }
   
   return(result);
}



//static
Kludge_MYSQL_RES* Kludge_Class_Relationship::DB_Lookup(OID_Type rel_clsid, Kludge_CLI* CLI)
{
 Kludge_MYSQL_RES* result = NULL;
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;

   qry_string << "SELECT * FROM Class_Relationships WHERE RelCLSID = \""
              << rel_clsid
              << "\""
              <<  ends;

   mysql_query_result = CLI->thr->DB_Query(qry_string.str());

   if (mysql_query_result != NULL)
   {
      result_row = mysql_fetch_row(mysql_query_result);
      if (result_row != NULL)
      {
         result = new Kludge_MYSQL_RES(result_row, mysql_query_result);
      }
      else
      {
         mysql_free_result(mysql_query_result);
         // Query result returned with no data rows - no such K_C_R RelCLSID in the DB.
         // That's not an internal error to this function - don't report it here, 
         // just return NULL. It'll be reported elsewhere, in proper context.
      }
   }
   else
   {
      cout << "Error in Kludge_C_R::DB_Lookup(): query returned null MYSQL_RES* .\n";
   }

   return(result);
}








/* ###    class Kludge_Instance    ### */


// This constructor is used for creating a new instance from scratch.
// It cannot be used for loading an instance from the DB.
// It actually stores the newly created instance in the DB.
//
Kludge_Instance::Kludge_Instance(
  string* instance_name,
  Kludge_Instance* owner_instance,
  Kludge_Class* inst_class,
  Kludge_Class_Relationship* cls_rel,
  Kludge_CLI* CLI
) :
  OID(0),
  Class(inst_class),
  ClsRel(cls_rel),
  RDN(*instance_name),
  Owner(owner_instance),
  Instances_Owned(NULL),
  Properties(NULL),
  Relationships(NULL),
  relationships_loaded(false),
  ctor_failed(false)
{
 int stor_success = 0;
 ostrstream qry_string;

   // first let's try a few referential sanity checks
   if (*ClsRel->Slave_Stub.Owner >= *Class)
   {
      if (*ClsRel->Master_Stub.Owner >= *Owner->Class)
      {
         if (ClsRel->RelType == 2)
         {
            // okay :) The Owner/Class/ClsRel references seem sane.
            if (ClsRel->Master_Stub.Match_Name(&RDN, CLI) != 0)
            {
               *CLI->output_stream <<
                  "Suggested instance name not accepted. Instance not created.\n";
               ctor_failed = true;
            }
            // else fine - sanity checks passed, including the regexp name check.
         }
         else
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance::Kludge_Instance():\n"
               " the indicated class relationship is of inappropriate relationship type.\n"
               " Its type should be \"OWN\" (ownership - the \"primary key\" relationship).\n"
               " Instance NOT inserted.\n";
            ctor_failed = true;
         }
      }
      else
      {
         *CLI->output_stream <<
            "Error - in Kludge_Instance::Kludge_Instance():\n"
            " the indicated owner instance is not a valid master for the indicated\n"
            " class relationship.\n"
            " Instance NOT inserted.\n";
         ctor_failed = true;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Error - in Kludge_Instance::Kludge_Instance():\n"
         " the indicated class is not a valid slave for the indicated\n"
         " class relationship.\n"
         " Instance NOT inserted.\n";
      ctor_failed = true;
   }

   if (! ctor_failed) // if the referential sanity is fine,
   {
      Type = KLUDGE_INSTANCE;
      if ( Owner->Instances_Owned == NULL ) Owner->Load_Instances_Owned(CLI);
      
      qry_string << "INSERT INTO Instances VALUES (NULL,\""
                 << Class->Class
                 << "\",\""
                 << ClsRel->RelCLSID
                 << "\",\""
                 << RDN
                 << "\",\""
                 << Owner->OID
                 << "\")"
                 <<  ends;
      stor_success = CLI->thr->DB_Cmd(qry_string.str());
      if (stor_success == 0)
      {
         OID = mysql_insert_id(CLI->thr->DB);
         Add_To_Index();
      }
      else
      {
         ctor_failed = true;
      }
   }
   // else no need to bother.
}


// This constructor can only be used for loading an existing Kludge_Instance
// from a database. It is not intended for creating a new instance from scratch.
//
Kludge_Instance::Kludge_Instance(
  const char *instance_name,
  Kludge_Instance* owner_instance,
  Kludge_CLI* CLI
) : 
  OID(0),
  Class(NULL),
  ClsRel(NULL),
  //RDN(*instance_name),
  Owner(owner_instance),
  Instances_Owned(NULL),
  Properties(NULL),
  Relationships(NULL),
  relationships_loaded(false),
  ctor_failed(false)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;

   Type = KLUDGE_INSTANCE;

   qry_string << "SELECT * FROM Instances WHERE RDN = \""
              << instance_name
              << "\" AND Owner = \""
              << ((Owner == NULL) ? 0 : Owner->OID)
              << "\""
              <<  ends;
      
   mysql_query_result = CLI->thr->DB_Query(qry_string.str());

   if (( result_row = mysql_fetch_row(mysql_query_result) ))
   {
      cout << "Got an instance body to read, found by Name.\n";
      if (! Extract_From_MysqlRow(result_row, CLI) )
      {
         ctor_failed = true;
      }
      else Add_To_Index();
   }
   else
   {
      ctor_failed = true;
   }
   mysql_free_result(mysql_query_result);
}



// This constructor can only be used for loading an existing Kludge_Instance
// from a database. It is not intended for creating a new instance from scratch.
//
Kludge_Instance::Kludge_Instance(
  MYSQL_ROW DB_record,
  Kludge_Instance* owner_instance,
  Kludge_CLI* CLI
) : 
  OID(0),
  Class(NULL),
  ClsRel(NULL),
  Owner(owner_instance),
  Instances_Owned(NULL),
  Properties(NULL),
  Relationships(NULL),
  relationships_loaded(false),
  ctor_failed(false)
{
   Type = KLUDGE_INSTANCE;

   if ( ! Extract_From_MysqlRow(DB_record,CLI) )
   {
      ctor_failed = true;
   }
   else Add_To_Index();
}



//
// Extract_From_MysqlRow()
//
// Expects:
// Row_To_Process[0] == OID
// Row_To_Process[1] == Class
// Row_To_Process[2] == ClsRel
// Row_To_Process[3] == RDN
// Row_To_Process[4] == Owner
//
bool Kludge_Instance::Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI) 
{
 bool result = true;
 OID_Type tmp_cls = 0, tmp_cls_rel = 0;
 istrstream* tmp_str = NULL;

   if (Row_To_Process != NULL)
   {
      tmp_str = new istrstream(Row_To_Process[0]);
      *tmp_str >> OID;
      delete tmp_str;

      tmp_str = new istrstream(Row_To_Process[1]);
      *tmp_str >> tmp_cls;
      delete tmp_str;
      
      tmp_str = new istrstream(Row_To_Process[2]);
      *tmp_str >> tmp_cls_rel;
      delete tmp_str;

      RDN = Row_To_Process[3];

      Class = Kludge_Class::ID2ptr(&tmp_cls,CLI);
      if (Class == NULL)
      {
         cout <<
            "In Kludge_Instance::Extract_From_MysqlRow():\n"
            " Class ID could not be resolved to a Kludge_Class pointer!\n"
            " Inconsistency in the backend SQL database?\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }
         
      if (Owner != NULL)
      {
         ClsRel = Kludge_Class_Relationship::ID2ptr(&tmp_cls_rel,CLI);
         if (ClsRel == NULL)
         {
            cout <<
               "In Kludge_Instance::Extract_From_MysqlRow():\n"
               " ClsRel ID could not be resolved to a K_C_R pointer!\n"
               " Inconsistency in the backend SQL database?\n"
               " Extract_From_MysqlRow() failed.\n";
            result = false;
         }
      }
      // else this is probably the Root_Instance - let's leave ClsRel to be NULL
   }
   else
   {
      cout <<
         "In Kludge_Instance::Extract_From_MysqlRow():\n"
         " the row passed to this function is a NULL pointer!\n"
         " Extract_From_MysqlRow() failed.\n";
      result = false;
   }

   return(result);
}



Kludge_Instance::~Kludge_Instance()
{
   cout << "Deallocating instance: " << RDN << "\n" ;
   Free_Properties();
   Free_Instances_Owned();
   Free_Relationships();
   // we have to be able to delete while traversing Instances_Owned
   // ==> no Free_From_Owner_Index() here. Zap() does that for us.
   Free_From_Class_Index();
   Free_From_ClsRel_Index();
   Free_From_Global_Index();
}



// Zap() does a nice memory cleanup.
// Only makes sense 
// a) for dynamic partial release from memory 
//     In that case, invalidate the list of owned instances.
// b) when an instance is deleted
//     Requires deletion from the SQL-based object repository
//      - i.e., use DB_Delete() and then Zap() (or just Delete())
int Kludge_Instance::Zap()
{
 int result = 0;

   cout << "Zapping an instance: " << RDN << "\n" ;
   
   if (Owner != NULL)
   {
      if (Owner->Instances_Owned != NULL)
      {
         Free_From_Owner_Index();
      }
      else
      {
         cout <<
            "Internal error - in Kludge_Instance::Zap():\n"
            " owners's \"instances_owned\" index not initialized!\n";
         result = 1;
      }

      delete this;
   }
   else
   {
      if (RDN == "Root_Instance")
      {
         cout << 
            "In Kludge_Instance::Zap(): cannot delete the Root Instance.\n";
         result = 1;
      }
      else
      {
         cout << 
            "In Kludge_Instance::Zap(): zapping instance " << RDN << ".\n"
            " This instance doesn't seem to have an owner!\n";
         delete this;
         result = 1;
      }
   }
   
   return(result);
}



int Kludge_Instance::Load_Instances_Owned(Kludge_CLI* CLI, bool recursive = false)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;
 Kludge_Instance *New_Inst;
 int result = 0;

   if ( (OID != 0) && (Instances_Owned == NULL) )
   {
      Instances_Owned = g_tree_new(Kludge_Compare_Names);

      qry_string << "SELECT * FROM Instances WHERE Owner = \""
                 << OID
                 << "\""
                 <<  ends;

      mysql_query_result = CLI->thr->DB_Query(qry_string.str());

      while (( result_row = mysql_fetch_row(mysql_query_result) ))
      {
         New_Inst = new Kludge_Instance(result_row, this, CLI);
         if (New_Inst != NULL)
         {
            if (! New_Inst->ctor_failed)
            {
               if (recursive)
               {
                  New_Inst->Load_Instances_Owned(CLI, /*recursive?*/true);
               }
            }
            else
            {
               cout << "Creation of an owned instance from a MYSQL_ROW failed,"
                       " although raw malloc() of the instance succeeded.\n" ;
               delete New_Inst; New_Inst = NULL;
               result = 1;
            }
         }
         else
         {
            cout << "Creation of an owned instance from a MYSQL_ROW failed,"
                    " probably on malloc() of the new instance.\n" ;
            result = 1;
         }
      }
      mysql_free_result(mysql_query_result);
   }
   else
   {
      cout << "In Kludge_Instance::Load_Instances_Owned():\n"
      << "the OID is equal to zero,\n"
      << "or the Instances_Owned are already loaded.\n";
      result = 1;
   }

   return(result);
}



bool Kludge_Instance::Free_Instances_Owned()
{
   if (Instances_Owned != NULL)
   {
      g_tree_traverse(          // Delete all the children
            Instances_Owned,
            Kludge_Instance::Free_Instances_Owned_Traverse,
            G_POST_ORDER,
            NULL);

      g_tree_destroy(Instances_Owned);
      Instances_Owned = NULL;
   }
/*
   else
   {
      cout << "Releasing owned instances - instances not loaded!\n";
   }
*/

   return(true);
}



int Kludge_Instance::Load_Properties(Kludge_CLI* CLI)
{
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;
 Kludge_Instance_Property *New_Inst_Prop = NULL;
 int result = 0;

   if (Properties == NULL)
   {
      Properties = g_tree_new(Kludge_Compare_Names);

      qry_string << "SELECT * FROM Instance_Properties WHERE Instance = \""
                 << OID
                 << "\""
                 <<  ends;

      mysql_query_result = CLI->thr->DB_Query(qry_string.str());

      while (( result_row = mysql_fetch_row(mysql_query_result) ))
      {
         New_Inst_Prop = new Kludge_Instance_Property(result_row, this, CLI);
         if (New_Inst_Prop != NULL)
         { 
            if (New_Inst_Prop->ctor_failed)
            {
               cout << "Creation of an instance property from a MYSQL_ROW failed,"
                       " for reasons other than raw malloc().\n" ;
               delete New_Inst_Prop; New_Inst_Prop = NULL;
               result = 1;
            }
         }
         else
         {
            cout <<
               "Error: in Kludge_Instance::Load_Properties():\n"
               " creation of an instance property from a MYSQL_ROW failed.\n"
               " Probably a failed malloc().\n";
            result = 1;
         }
      }
      mysql_free_result(mysql_query_result);
   }
   else
   {
      cout << "In Kludge_Instance::Load_Properties:\n"
           << " properties already loaded!\n";
   }

   return(result);
}



bool Kludge_Instance::Free_Properties()
{
   if (Properties != NULL)
   {
      g_tree_traverse(          // Delete all the properties
               Properties,
               Kludge_Instance::Free_Properties_Traverse,
               G_POST_ORDER,
               NULL);

      g_tree_destroy(Properties);
      Properties = NULL;
   }
   // else properties not loaded

   return(true);
}



int Kludge_Instance::Init_Relationships()
{
   if (Relationships == NULL)
   {
      Relationships = g_tree_new(Kludge_Compare_Names);
   }
   // else already initialized - nothing to be done.

   return(0);
}



int Kludge_Instance::Load_Relationships(Kludge_CLI* CLI)
{
 int result = 0;
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;
 Kludge_Instance_Relationship* new_inst_rel = NULL;
 istrstream *tmp_master = NULL, *tmp_slave = NULL;
 OID_Type tmp_master_ID = 0, tmp_slave_ID = 0;
 string tmp_name_M, tmp_name_S;

   if (! relationships_loaded)
   {
      // If the relationships are partially loaded, the easiest way to prevent
      // cloned relationship entries in memory is to free any previously loaded
      // relationships and reload from scratch.
      // This way however, the relationships index at the other peer of any
      // previously loaded relationship would get invalidated upon loading
      // full index at our end - thus, the approach described above is a no-go.
      // We must not delete previously loaded rels from an incomplete index.

      if (Relationships == NULL) Init_Relationships();

      if (Relationships != NULL)
      {
         qry_string << 
            "SELECT * FROM Instance_Relationships WHERE Master = \""
            << OID
            << "\" OR Slave = \""
            << OID
            << "\""
            <<  ends;
            
         mysql_query_result = CLI->thr->DB_Query(qry_string.str());

         while (( result_row = mysql_fetch_row(mysql_query_result) ))
         {
            // check that this relationship is not loaded yet
            // first fetch some data to check from the Mysql Row
            tmp_master = new istrstream(result_row[1]);
            tmp_slave = new istrstream(result_row[2]);
            *tmp_master >> tmp_master_ID;
            *tmp_slave >> tmp_slave_ID;
            tmp_name_M = result_row[4];
            tmp_name_S = result_row[5];

            if // check based on relative name and Master/Slave OID's
            (!(
               (( tmp_master_ID == OID ) && (Lookup_Relationship_Raw(&tmp_name_M,CLI) != NULL))
               || (( tmp_slave_ID == OID ) && (Lookup_Relationship_Raw(&tmp_name_S,CLI) != NULL))
            ))
            {
               // relationship not loaded yet
               new_inst_rel = new Kludge_Instance_Relationship(result_row, CLI);
               if (new_inst_rel != NULL)
               { 
                  if (new_inst_rel->ctor_failed)
                  {
                     cout << "Creation of an instance relationship from a MYSQL_ROW failed,"
                             " for reasons other than raw malloc() failure.\n" ;
                     delete new_inst_rel; new_inst_rel = NULL;
                     result = 1;
                  }
                  // else OK, relationship loaded
               }
               else
               {
                  cout <<
                     "Error: in Kludge_Instance::Load_Relationships():\n"
                     " creation of an instance relationship from a MYSQL_ROW failed.\n" ;
                  result = 1;
               }
            }
            // else relationship already loaded - skip it.

            // clean up the local test variables for another cycle
            tmp_master_ID = 0;
            tmp_slave_ID = 0;
            tmp_name_M = "";
            tmp_name_S = "";
            delete tmp_master; tmp_master = NULL;
            delete tmp_slave; tmp_slave = NULL;
         }

         mysql_free_result(mysql_query_result);
         
         if (result == 0)
         {
            relationships_loaded = true;
         }
      }
      else
      {
         cout << "In Kludge_Instance::Load_Relationships():\n"
              << " couldn't Init_Relationships().\n";
         result = 1;
      }
   }
   else
   {
      cout << "In Kludge_Instance::Load_Relationships():\n"
           << " relationships already loaded!\n";
   }

   return(result);
}



bool Kludge_Instance::Free_Relationships()
{
 vector<Kludge_Inst_Rel_Stub*> rels_to_free;
 GTree* tmp_duplicates_index = NULL;
 bool skip_zap_for_now = false;

   if (Relationships != NULL)
   {
      // First, stuff all the relationship pointers into a throwaway vector
      // (we can't zap them while traversing  btree).
      // No matter if they're completely loaded or not - we're just clearing up space on the heap.
      g_tree_traverse(
               Relationships,
               Kludge_Instance::Free_Relationships_Traverse,
               G_POST_ORDER,
               &rels_to_free);

      // We need to keep track of self-referral relationships.
      // Such relationships would have two index entries, both of which
      // will be zapped in a single run of Zap(). An attempt to Zap() 
      // the second entry in our derived throwaway vector would cause a SIGSEGV.
      tmp_duplicates_index = g_tree_new(Kludge_Compare_Names);
      
      for (int i=rels_to_free.size()-1; i>=0; i--) // now zap all the relationships one by one
      {
         skip_zap_for_now = false;
            
         // detect self-referral relationships
         if (rels_to_free[i]->Body->Master_Stub.Owner
               == rels_to_free[i]->Body->Slave_Stub.Owner)
         {
            // check that it's not already in the duplicates index
            if (g_tree_lookup(tmp_duplicates_index, &rels_to_free[i]->Name) == NULL)
            {
               // OK, we didn't know about this one yet.
               // For now, let's just write it down into the index of duplicates
               // - we'll zap it later, when we encounter the other end.
               skip_zap_for_now = true;

               // find the other end (stub) and store it in the duplicates index
               if (rels_to_free[i]->Stub_Type == Kludge_Inst_Rel_Stub::MASTER)
               {
                  g_tree_insert(
                     tmp_duplicates_index,
                     &rels_to_free[i]->Body->Slave_Stub.Name,
                     &rels_to_free[i]->Body->Slave_Stub
                  );
               }
               else if (rels_to_free[i]->Stub_Type == Kludge_Inst_Rel_Stub::SLAVE)
               {
                  g_tree_insert(
                     tmp_duplicates_index,
                     &rels_to_free[i]->Body->Master_Stub.Name,
                     &rels_to_free[i]->Body->Master_Stub
                  );
               }
               else
               {
                  cout <<
                     "Internal error - in Kludge_Instance::Free_Relationships():\n"
                     " self-referencing relationship encountered, but couldn't detect the other end.\n";
               }
            }
            else
            {
               // This self-referral relationship has already been noticed.
               // Therefore we can zap it by now. But before we do that, we
               // remove it from the index of duplicates (BTree) to prevent
               // its corruption.
               g_tree_remove(tmp_duplicates_index, &rels_to_free[i]->Name);
            }
         }

         if (skip_zap_for_now = false)
         {
            // Detection of duplicates done - now zap the current relationship.

            // Before actually clearing the struct from memory, invalidate the 
            // list of relationships at the peer Kludge Instance
            if (rels_to_free[i]->Stub_Type == Kludge_Inst_Rel_Stub::MASTER)
            {
               rels_to_free[i]->Body->Slave_Stub.Owner->relationships_loaded = false;
            }
            else if (rels_to_free[i]->Stub_Type == Kludge_Inst_Rel_Stub::SLAVE)
            {
               rels_to_free[i]->Body->Master_Stub.Owner->relationships_loaded = false;
            }
            else
            {
               cout <<
                  "Internal error - in Kludge_Instance::Free_Relationships():\n"
                  " couldn't detect the other end.\n";
            }
            rels_to_free[i]->Zap();
         }   
      }
      g_tree_destroy(tmp_duplicates_index);

      // As we need the relationships tree to exist while Zap()ping,
      // we can only destroy it after we've iterated through the throwaway vector 
      g_tree_destroy(Relationships);
      Relationships = NULL;
      relationships_loaded = false;
   }
   // else relationships not loaded

   return(true);
}



int Kludge_Instance::Add_To_Owner_Index(Kludge_Instance* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Owner;

   cout << "Adding instance " << RDN << " to its owner's index.\n";
   if (Where != NULL)
   {
      if (Where->Instances_Owned != NULL)
      {
         g_tree_insert(Where->Instances_Owned,&RDN,this);
      }
      else
      {
         cout << "Kludge_Instance::Add_To_Owner_Index(): siblings not loaded.\n";
         result = 1;
      }
   }
   // else this instance is probably the Root_Instance (or something's broken)
   
   return(result);
}



int Kludge_Instance::Add_To_Class_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Class;

   cout << "Adding instance " << RDN << " to its Class'es index.\n";
   if (Where != NULL)
   {
      if (Where->Instances == NULL) Where->Init_Instances();
      
      if (Where->Instances != NULL)
      {
         g_tree_insert(Where->Instances,&OID,this);
      }
      else
      {
         cout << "Kludge_Instance::Add_To_Class_Index():\n"
                 " could not initialize the Class'es index of instances.\n";
         result = 1;
      }
   }
   // else something's likely broken (the Root_Instance is an instance of the Root_Instance_Class)
   
   return(result);
}



int Kludge_Instance::Add_To_ClsRel_Index(Kludge_Class_Relationship* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = ClsRel;

   cout << "Adding instance " << RDN << " to its class relationship's index.\n";
   if (Where != NULL)
   {
      if (Where->Instances == NULL) Where->Init_Instances();
      
      if (Where->Instances != NULL)
      {
         g_tree_insert(Where->Instances,&OID,this);
      }
      else
      {
         cout << "Kludge_Instance::Add_To_ClsRel_Index():\n"
                 " could not initialize the Class Relationship's index of instances.\n";
         result = 1;
      }
   }
   // else this instance is probably the Root_Instance (or something's broken)

   return(result);
}



int Kludge_Instance::Add_To_Index()
// local, global, Class, ClsRel
{
 int result = 0;

   result = Add_To_Owner_Index();
   if (result == 0) result = Add_To_Class_Index();
   if (result == 0) result = Add_To_ClsRel_Index();
   if (result == 0) g_tree_insert(Index,&OID,this); // global index
   
   return(result);
}



int Kludge_Instance::Free_From_Owner_Index(Kludge_Instance* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Owner;

   cout << "Releasing instance " << RDN << " from Owner's index.\n";
   if (Where != NULL)
   {
      if (Where->Instances_Owned != NULL)
      {
         g_tree_remove(Where->Instances_Owned,&RDN);
      }
      else
      {
         cout << "Kludge_Instance::Free_From_Owner_Index():\n"
                 " Owner's index not initialized.\n";
         result = 1;
      }
   }
   // else this instance is probably the Root_Instance (or something's broken)
   
   return(result);
}



int Kludge_Instance::Free_From_Class_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Class;

   cout << "Releasing instance " << RDN << " from its Class'es index.\n";
   if (Where != NULL)
   {
      if (Where->Instances != NULL)
      {
         g_tree_remove(Where->Instances,&OID);
      }
      else
      {
         cout << "Kludge_Instance::Free_From_Class_Index():\n"
                 " the Class'es \"Instances\" index not initialized.\n";
         result = 1;
      }
   }
   // else something's broken (the Root_Instance is an instance of Root_Instance_Class)
   
   return(result);
}



int Kludge_Instance::Free_From_ClsRel_Index(Kludge_Class_Relationship* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = ClsRel;

   cout << "Releasing instance " << RDN << " from its Class Relationship's index.\n";
   if (Where != NULL)
   {
      if (Where->Instances != NULL)
      {
         g_tree_remove(Where->Instances,&OID);
      }
      else
      {
         cout << "Kludge_Instance::Free_From_ClsRel_Index():\n"
                 " the Class Relationship's \"Instances\" index not initialized.\n";
         result = 1;
      }
   }
   // else this instance is probably the Root_Instance (or something's broken)
   
   return(result);
}



bool Kludge_Instance::Free_From_Global_Index()
{
   g_tree_remove(Index,&OID);
   return(true);
}



int Kludge_Instance::List_Properties(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   if (Properties == NULL) Load_Properties(CLI);
   
   if (Properties != NULL)
   {
      if (recursive)
      {
         g_tree_traverse(
                         Properties, 
                         Kludge_Instance::List_Properties_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
      else
      {
         g_tree_traverse(
                         Properties, 
                         Kludge_Instance::List_Properties_Names_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
   }
   else
   {
      cout << "Kludge_Instance::List_Properties(): failed to load properties!\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance::List_Instances_Owned(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   if (Instances_Owned == NULL) Load_Instances_Owned(CLI, /*recursive?*/false);
   
   if (Instances_Owned != NULL)
   {
      if (recursive)
      {
         g_tree_traverse(
                         Instances_Owned, 
                         Kludge_Instance::List_Instances_Owned_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
      else
      {
         g_tree_traverse(
                         Instances_Owned, 
                         Kludge_Instance::List_Instances_Owned_Names_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
   }
   else
   {
      cout << "Kludge_Instance::List_Instances_Owned():\n"
           << " failed to load owned instances!\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance::List_Relationships(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   if (! relationships_loaded) Load_Relationships(CLI);
   
   if (relationships_loaded)
   {
      if (recursive)
      {
         g_tree_traverse(
                         Relationships,
                         Kludge_Instance::List_Relationships_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
      else
      {
         g_tree_traverse(
                         Relationships,
                         Kludge_Instance::List_Relationship_Names_Traverse,
                         G_IN_ORDER,
                         CLI
                        );
      }
   }
   else
   {
      cout << "Kludge_Instance::List_Relationships(): failed to load relationships!\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;
 string* tmp_pwd = NULL;
 string tmp_cls_name;

   if (ClsRel != NULL)
   {
      tmp_pwd = Kludge_CLI_PWD(&ClsRel->Master_Stub,CLI->output_stream);
   }
   else tmp_pwd = new string; // probably the root instance

   if (Class != NULL) tmp_cls_name = Class->Name;
   // else tmp_cls_name remains empty - this is probably the root instance

   if (tmp_pwd != NULL)
   {
      *CLI->output_stream
            << "INSTANCE \""
            << RDN << "\" \""
            << tmp_cls_name << "\" \""
            << *tmp_pwd << "\"\n{\n";
      List_Properties(CLI,recursive);
      List_Instances_Owned(CLI, recursive);
      List_Relationships(CLI, recursive);

      *CLI->output_stream << "};\n";

      delete tmp_pwd;
   }
   else
   {
      cout << "Error: In Kludge_Instance::List_Contents():\n"
              " something's horribly wrong.\n";
   }

   return(result);
}



int Kludge_Instance::Describe_Self(Kludge_CLI* CLI)
{
 int result = 0;
 string* tmp_pwd = NULL;
 string tmp_cls_name;

   if (ClsRel != NULL)
   {
      tmp_pwd = Kludge_CLI_PWD(&ClsRel->Master_Stub,CLI->output_stream);
   }
   else tmp_pwd = new string; // probably the root instance

   if (Class != NULL) tmp_cls_name = Class->Name;
   // else tmp_cls_name remains empty - this is probably the root instance

   if (tmp_pwd != NULL)
   {
      *CLI->output_stream
            << "INSTANCE \""
            << RDN << "\" \""
            << tmp_cls_name << "\" \""
            << *tmp_pwd << "\";\n";
      delete tmp_pwd;
   }
   else
   {
      cout << "Error: In Kludge_Instance::Describe_Self():\n"
              " something's horribly wrong.\n";
   }

   return(result);
}



Kludge_Instance* Kludge_Instance::Lookup_Instance_Owned(string* name_to_lookup, Kludge_CLI* CLI)
{
 Kludge_Instance *result = NULL;

   if (Instances_Owned == NULL)
   {
      Load_Instances_Owned(CLI, /*recursive?*/false);
   }

   if (Instances_Owned != NULL)
   {
      result = (Kludge_Instance *) g_tree_lookup(Instances_Owned,name_to_lookup);
   }
   else
   {
      cout << "Error in Kludge_Instance::Lookup_Instance_Owned():\n"
              " Could not load the instances owned!\n";
   }

   return(result);
}



Kludge_Instance_Property* Kludge_Instance::Lookup_Property(string* name_to_lookup, Kludge_CLI* CLI)
{
 Kludge_Instance_Property *result = NULL;

   if (Properties == NULL)
   {
      Load_Properties(CLI);
   }

   if (Properties != NULL)
   {
      result = (Kludge_Instance_Property *) g_tree_lookup(Properties,name_to_lookup);
   }
   else
   {
      cout << "Error in Kludge_Instance::Lookup_Property():\n"
              " Couldn't load properties!\n";
   }

   return(result);
}



Kludge_Inst_Rel_Stub* Kludge_Instance::Lookup_Relationship_Raw(string* name_to_lookup, Kludge_CLI* CLI)
{
 Kludge_Inst_Rel_Stub* result = NULL;

   if (Relationships != NULL)
   {
      result = (Kludge_Inst_Rel_Stub*) g_tree_lookup(Relationships,name_to_lookup);
   }
   else
   {
      cout << "Kludge_Instance::Lookup_Relationship_Raw(): Oops! Relationships not initiated!\n";
   }

   return(result);
}



Kludge_Inst_Rel_Stub* Kludge_Instance::Lookup_Relationship(string* name_to_lookup, Kludge_CLI* CLI)
{
 Kludge_Inst_Rel_Stub* result = NULL;

   if (! relationships_loaded)
   {
      Load_Relationships(CLI);
   }

   if (Relationships != NULL)   // perhaps this should better read "if (relationships_loaded)"
   {
      result = Lookup_Relationship_Raw(name_to_lookup,CLI);
   }
   else
   {
      cout << "Error in Kludge_Instance::Lookup_Relationship():\n"
              " Couldn't load relationships!\n";
   }

   return(result);
}



Kludge_Object* Kludge_Instance::Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
 Kludge_Object* result = NULL;

   result = Lookup_Instance_Owned(Content_Name, CLI);        // an owned instance  ?
   if (result == NULL)
   {                                                // a Kludge_Instance_Property ?
      result = Lookup_Property(Content_Name, CLI);
   }
   if (result == NULL)
   {                                                // a Kludge_Instance_Relationship ?
      result = Lookup_Relationship(Content_Name, CLI);
      //+ There should be a fallthrough by default!
   }
   if (result == NULL)
   {
      if (
            (*Content_Name == "OWNER")
            || (*Content_Name == "..")
         )                                      // an owner Kludge_Instance reference?
      {
         result = Superior();
      }
      else if (*Content_Name == ".")             // a "this" style reference?
      {
         result = this;
      }
      else if (*Content_Name == "CLASS")
      {
         result = Class;
      }
      else if (*Content_Name == "CLASS_RELATIONSHIP")
      {
         result = &ClsRel->Master_Stub;
      }
   }

   return(result);
}



// *obj_to_insert is owned by its parent Kludge_CLI_Command that is deleted by
// its respective Kludge_CLI after use. We shan't delete it here.
int Kludge_Instance::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
 int result = 0;
 bool user_level_error = false;
 string* tmp_pwd = NULL;
 Kludge_Object* tmp_obj_ptr = NULL;
 Kludge_Path* tmp_path = NULL;
 Kludge_Object *tmp_class = NULL, *tmp_cls_rel_stub = NULL;
 Kludge_Instance* new_instance = NULL;
 Kludge_Instance_Property* new_inst_prop = NULL;
 Kludge_Instance_Relationship* new_inst_rel = NULL;
 byte_buf* tmp_byte_buf = NULL; // for BLOB handling
 
   switch (obj_to_insert->type)
   {
      case PARM_INSTANCE :
      {
         // Let's start with a few checks of user input.
         
         // internal sanity check - first parameter = PARM_NAME <name> 
         if (obj_to_insert->contents[0]->type != PARM_NAME)
         {
            cout <<
               "Internal error: in Kludge_Instance::Insert(<INSTANCE>):\n"
               " the first parameter should be of type PARM_NAME.\n";
            result = 1;
         }
         
         // the proposed instance name should not be an empty string in the first place
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[0]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " instance name can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }
         
         // check if the proposed instance name exists under `this' instance 
         if ((! user_level_error) && (result == 0))
         {
            tmp_obj_ptr = Resolve_Content_Name(&obj_to_insert->contents[0]->content,CLI,/*own only?*/true);

            if (tmp_obj_ptr != NULL)
            {
               // this name already exists in this context (path) !
               tmp_pwd = Kludge_CLI_PWD(tmp_obj_ptr, CLI->output_stream);
               
               *CLI->output_stream <<
                  "I've been asked to insert a new Kludge Instance called \""
                  << obj_to_insert->contents[0]->content << "\".\n"
                  " Such a name already exists under Instance " << this->RDN <<
                  ", it resolves to:\n "
                  << *tmp_pwd << endl;
               
               delete tmp_pwd; tmp_pwd = NULL;
               tmp_obj_ptr = NULL;
               //result = 0; // not an internal error
               user_level_error = true;
            }
            //else OK, we can use this new name
         }
         
         // internal sanity check - second parameter = PARM_OBJ_CLASS <class> 
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[1]->type != PARM_OBJ_CLASS)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " the second parameter should be of type PARM_OBJ_CLASS.\n";
               result = 1;
            }
         }

         // the proposed class pathname should not be an empty string
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[1]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " class pathname can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // check if the proposed class pathname exists
         if ((! user_level_error) && (result == 0))
         {
            tmp_path = Kludge_CLI_String2path(
                          &obj_to_insert->contents[1]->content,
                          CLI->output_stream
                       );
            
            if (tmp_path != NULL)
            {
               tmp_class = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
               
               if (tmp_class != NULL)
               {
                  if (tmp_class->Type != KLUDGE_CLASS)
                  {
                     *CLI->output_stream <<
                        "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                        " the specified class pathname resolves to an object that\n"
                        " is not a Kludge CLASS.\n";
                     //result = 0; // not an internal error
                     user_level_error = true;
                  }
                  // else the second parameter is indeed a valid Kludge Class. Check passed OK.
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                     " couldn't find a Kludge Object with the specified class pathname.\n";
                  //result = 0; // not an internal error
                  user_level_error = true;
               }

               delete tmp_path; tmp_path = NULL;
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " couldn't decompose (string) class pathname into Kludge_Path elements.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // internal sanity check - third parameter = PARM_OBJ_CLSREL <class relationship> 
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[2]->type != PARM_OBJ_CLSREL)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " the third parameter should be of type PARM_OBJ_CLSREL.\n";
               result = 1;
            }
         }

         // the proposed class relationship pathname should not be an empty string
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[2]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " class relationship pathname can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // check if the proposed class relationship pathname exists
         if ((! user_level_error) && (result == 0))
         {
            tmp_path = Kludge_CLI_String2path(
                          &obj_to_insert->contents[2]->content,
                          CLI->output_stream
                       );
            
            if (tmp_path != NULL)
            {
               tmp_cls_rel_stub = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
               
               if (tmp_cls_rel_stub != NULL)
               {
                  if (tmp_cls_rel_stub->Type != KLUDGE_CLASS_REL_STUB)
                  {
                     *CLI->output_stream <<
                        "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                        " the specified class relationship pathname resolves to an object that\n"
                        " is not a Kludge CLASS_RELATIONSHIP.\n";
                     //result = 0; // not an internal error
                     user_level_error = true;
                  }
                  // else the third parameter is indeed a valid Kludge Class Relationship. Check passed OK.
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                     " couldn't find a Kludge Object with the specified class_rel pathname.\n";
                  //result = 0; // not an internal error
                  user_level_error = true;
               }

               delete tmp_path;
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE>):\n"
                  " couldn't decompose (string) class_rel pathname into Kludge_Path elements.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // Done with the checks. Now we can attempt to insert the new instance.
         if ((! user_level_error) && (result == 0)) // passed all the checks just fine
         {
            new_instance = new Kludge_Instance(
                                 &obj_to_insert->contents[0]->content,
                                 this,
                                 (Kludge_Class*)tmp_class,
                                 ((Kludge_Class_Rel_Stub*)tmp_cls_rel_stub)->Body,
                                 CLI
                               );
            if (new_instance != NULL)
            {
               if (! new_instance->ctor_failed) // success!
               {
                  Kludge_pl_params* ctor_params = NULL;
                  unsigned int params_offset = 3;
                  if (obj_to_insert->contents.size() > params_offset)
                  {
                     if (obj_to_insert->contents[params_offset]->type == PARM_ARGLIST)
                     {
                        ctor_params = Kludge_CLI::Prepare_Method_Params(
                                              obj_to_insert->contents[params_offset]);
                        params_offset++;
                     }
                  }
                  
                  // let's insert the nested contents.
                  for (unsigned int i = params_offset; i < obj_to_insert->contents.size(); i++)
                  // obj_to_insert->contents[0] == <instance name> 
                  // obj_to_insert->contents[1] == <class name|pathname>
                  // obj_to_insert->contents[2] == <cls_rel name|pathname>
                  // - that's why the terminal condition reads i>2, rather than i>=0. 
                  // We simply don't want to iterate through the mandatory parameters
                  {
                     new_instance->Insert(obj_to_insert->contents[i], CLI);
                  }

                  // try to run the Perl-space CTOR
                  new_instance->Try_Method("CTOR", ctor_params, CLI, /*ignore inherited CTORs*/true);
               }
               else
               {
                  *CLI->output_stream <<
                  "Something failed while inserting a new Kludge_Instance. Instance malloc()'ed\n"
                  " but Kludge_Instance::Kludge_Instance() ctor failed to fully set it up.\n";

                  delete new_instance; new_instance = NULL; // no need to zap it in this case.
                  // result = 0; // not necessarily an internal error
                  user_level_error = true;
               }
            }
            else
            {
               cout << "Kludge_instance::Insert(): Failed to create (malloc?) a new Kludge_Instance.\n";
               result = 1; // internal error
            }
         }
         //else there was an error - no insertion is to take place. Error already reported.
      }
      break;
      case PARM_INSTPROP :
      {
         // Let's start with a few checks of user input.
         
         // internal sanity check - first parameter = PARM_NAME <name> 
         if (obj_to_insert->contents[0]->type != PARM_NAME)
         {
            cout <<
               "Internal error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
               " the first parameter should be of type PARM_NAME.\n";
            result = 1;
         }
         
         // the proposed instance property name should not be an empty string in the first place
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[0]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " instance property name can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }
         
         // check if the proposed instance property name exists under `this' instance 
         if ((! user_level_error) && (result == 0))
         {
            tmp_obj_ptr = Resolve_Content_Name(&obj_to_insert->contents[0]->content,CLI,/*own only?*/true);

            if (tmp_obj_ptr != NULL)
            {
               // this name already exists in this context (path) !
               tmp_pwd = Kludge_CLI_PWD(tmp_obj_ptr, CLI->output_stream);
               
               *CLI->output_stream <<
                  "I've been asked to insert a new Kludge Instance Property called \""
                  << obj_to_insert->contents[0]->content << "\".\n"
                  " Such a name already exists under Instance " << this->RDN <<
                  ", it resolves to:\n "
                  << *tmp_pwd << endl;
               
               delete tmp_pwd; tmp_pwd = NULL;
               tmp_obj_ptr = NULL;
               //result = 0; // not an internal error
               user_level_error = true;
            }
            //else OK, we can use this new name
         }
         
         // internal sanity check - second parameter = PARM_OBJ_CLASS <class> 
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[1]->type != PARM_OBJ_CLASS)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " the second parameter should be of type PARM_OBJ_CLASS.\n";
               result = 1;
            }
         }

         // the proposed class pathname should not be an empty string
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[1]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " class pathname can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // check if the proposed class pathname exists
         if ((! user_level_error) && (result == 0))
         {
            tmp_path = Kludge_CLI_String2path(
                          &obj_to_insert->contents[1]->content,
                          CLI->output_stream
                       );
            
            if (tmp_path != NULL)
            {
               // check that such an object exists and that it's a KLUDGE_CLASS
               tmp_class = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
               
               if (tmp_class != NULL)
               {
                  if (tmp_class->Type != KLUDGE_CLASS)
                  {
                     *CLI->output_stream <<
                        "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                        " the specified class pathname resolves to an object that\n"
                        " is not a Kludge CLASS.\n";
                     //result = 0; // not an internal error
                     user_level_error = true;
                  }
                  // else the second parameter is indeed a valid Kludge Class. Check passed OK.
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                     " couldn't find a Kludge Object with the specified class pathname.\n";
                  //result = 0; // not an internal error
                  user_level_error = true;
               }

               delete tmp_path; tmp_path = NULL;
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " couldn't decompose (string) class pathname into Kludge_Path elements.\n";
               //result = 0; // not necessarily an internal error
               user_level_error = true;
            }
         }

         // internal sanity check - third parameter = PARM_OBJ_CLSREL <class relationship> 
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[2]->type != PARM_OBJ_CLSREL)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " the third parameter should be of type PARM_OBJ_CLSREL.\n";
               result = 1;
            }
         }

         // the proposed class relationship pathname should not be an empty string
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[2]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " class relationship pathname can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // check if the proposed class relationship pathname exists
         if ((! user_level_error) && (result == 0))
         {
            tmp_path = Kludge_CLI_String2path(
                          &obj_to_insert->contents[2]->content,
                          CLI->output_stream
                       );
            
            if (tmp_path != NULL)
            {
               // check that such an object exists and that it's a KLUDGE_CLASS_RELATIONSHIP
               tmp_cls_rel_stub = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
               
               if (tmp_cls_rel_stub != NULL)
               {
                  if (tmp_cls_rel_stub->Type != KLUDGE_CLASS_REL_STUB)
                  {
                     *CLI->output_stream <<
                        "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                        " the specified class relationship pathname resolves to an object that\n"
                        " is not a Kludge CLASS_RELATIONSHIP.\n";
                     //result = 0; // not an internal error
                     user_level_error = true;
                  }
                  // else the third parameter is indeed a valid Kludge Class Relationship. Check passed OK.
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                     " couldn't find a Kludge Object with the specified class_rel pathname.\n";
                  //result = 0; // not an internal error
                  user_level_error = true;
               }

               delete tmp_path;
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " couldn't decompose (string) class_rel pathname into Kludge_Path elements.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // internal sanity check - fourth parameter = PARM_VALUE <value>
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[3]->type != PARM_VALUE)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INSTANCE_PROPERTY>):\n"
                  " the fourth parameter should be of type PARM_VALUE.\n";
               result = 1;
            }
         }

         // Done with the checks. Now we can attempt to insert the new instance property.
         if ((! user_level_error) && (result == 0)) // passed all the checks just fine
         {
            new_inst_prop = new Kludge_Instance_Property(
                                   &obj_to_insert->contents[0]->content, //name
                                   this,
                                   (Kludge_Class*)tmp_class,
                                   ((Kludge_Class_Rel_Stub*)tmp_cls_rel_stub)->Body,
                                   &obj_to_insert->contents[3]->content, //value
                                   CLI
                                );

            if (new_inst_prop != NULL)
            {
               if (new_inst_prop->ctor_failed)
               {
                  cout << "Internal error: Kludge_Instance::Insert():\n"
                          " failed to create a new Kludge_Instance_Property (although malloc() OK).\n";
                  delete new_inst_prop; new_inst_prop = NULL;
                  result = 1;
               }
               // else OK, the new property is airborne
            }
            else
            {
               cout << "Internal error: in Kludge_Instance::Insert():\n"
                       " failed to create (malloc?) a new Kludge_Instance_Property.\n";
               result = 1; // internal error
            }
         }
         //else there was an error - no insertion is to take place. Error already reported.
      }
      break;
      case PARM_INSTREL :
      {
         // The other sections in this switch/case breakdown simply call 
         // the respective Kludge Objects's heavy-weight constructor.
         // With class Kludge_Instance_Relationship, I decided to avoid the
         // heavy-weight style constructor to prevent unnecessary
         // copy-constructions. I simply read the data directly into an
         // embryonal Kludge_Instance_Relationship (after its light-weight
         // constructor has entered reasonable defaults) and perform
         // semantic checks over that embryonal structure, before I proceed
         // to the heavy-weight initialization using the Init() method.

         new_inst_rel = new Kludge_Instance_Relationship(); // enters some defaults that we rely on here

#define IS_VAR(VAR_NAME_STRING) obj_to_insert->contents[i]->contents[0]->content == VAR_NAME_STRING
         // read the variables into the uninitialized Kludge_Instance_Relationship
         for (int i = obj_to_insert->contents.size()-1; (i>=0) && (new_inst_rel != NULL) ; i--)
         {
            // warn at variable names that we don't know (and skip them)
            
            if (IS_VAR("MASTER"))
            {
               // path/name not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_path = Kludge_CLI_String2path(
                                &obj_to_insert->contents[i]->contents[1]->content, //Value
                                CLI->output_stream
                             );
                  // does path/name string resolve fine to a path (vector of elements) ?
                  if (tmp_path != NULL)
                  {
                     tmp_obj_ptr = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
                     // does path (vector of elements) resolve fine into an object pointer ?
                     if (tmp_obj_ptr != NULL)
                     {
                        // is the resolved prospective master object really a Kludge_Instance ? 
                        if (tmp_obj_ptr->Type == KLUDGE_INSTANCE)
                        {
                           new_inst_rel->Master_Stub.Owner = (Kludge_Instance*)tmp_obj_ptr; // done with this one
                        }
                        else
                        {
                           *CLI->output_stream <<
                              "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                              " Path to MASTER resolves to a kludge object other than INSTANCE.\n";
                           delete new_inst_rel; new_inst_rel = NULL;
                           // result = 0; Not an internal error.
                        }

                        // do not delete tmp_obj_ptr - object neither created nor owned by us
                        tmp_obj_ptr = NULL;
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                           " Path2Ptr failed to follow the given path to MASTER - wrong path?\n";
                        delete new_inst_rel; new_inst_rel = NULL;
                        // result = 0; Not an internal error.
                     }

                     delete tmp_path; tmp_path = NULL;
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                        " String2path failed while parsing path to MASTER - probably a syntax error.\n";
                     delete new_inst_rel; new_inst_rel = NULL;
                     // result = 0; Not an internal error.
                  }
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                     " master name/path is an empty string.\n";
                  delete new_inst_rel; new_inst_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("SLAVE"))
            {
               // path/name not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_path = Kludge_CLI_String2path(
                                &obj_to_insert->contents[i]->contents[1]->content, //Value
                                CLI->output_stream
                             );
                  // does path/name string resolve fine to a path (vector of elements) ?
                  if (tmp_path != NULL)
                  {
                     tmp_obj_ptr = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
                     // does path (vector of elements) resolve fine into an object pointer ?
                     if (tmp_obj_ptr != NULL)
                     {
                        // is the resolved prospective slave object really a Kludge_Instance ? 
                        if (tmp_obj_ptr->Type == KLUDGE_INSTANCE)
                        {
                           new_inst_rel->Slave_Stub.Owner = (Kludge_Instance*)tmp_obj_ptr; // done with this one
                        }
                        else
                        {
                           *CLI->output_stream <<
                              "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                              " Path to SLAVE resolves to a kludge object other than INSTANCE.\n";
                           delete new_inst_rel; new_inst_rel = NULL;
                           // result = 0; Not an internal error.
                        }

                        // do not delete tmp_obj_ptr - object neither created nor owned by us
                        tmp_obj_ptr = NULL;
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                           " Path2Ptr failed to follow the given path to SLAVE - wrong path?\n";
                        delete new_inst_rel; new_inst_rel = NULL;
                        // result = 0; Not an internal error.
                     }

                     delete tmp_path; tmp_path = NULL;
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                        " String2path failed while parsing path to SLAVE - probably a syntax error.\n";
                     delete new_inst_rel; new_inst_rel = NULL;
                     // result = 0; Not an internal error.
                  }
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                     " slave name/path is an empty string.\n";
                  delete new_inst_rel; new_inst_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("CLS_REL"))
            {
               // path/name not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  tmp_path = Kludge_CLI_String2path(
                                &obj_to_insert->contents[i]->contents[1]->content, //Value
                                CLI->output_stream
                             );
                  // does path/name string resolve fine to a path (vector of elements) ?
                  if (tmp_path != NULL)
                  {
                     tmp_obj_ptr = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
                     // does path (vector of elements) resolve fine into an object pointer ?
                     if (tmp_obj_ptr != NULL)
                     {
                        // is the resolved prospective CLS_REL object really a Kludge_Class_Relationship ? 
                        if (tmp_obj_ptr->Type == KLUDGE_CLASS_REL_STUB)
                        {
                           new_inst_rel->ClsRel = ((Kludge_Class_Rel_Stub*)tmp_obj_ptr)->Body; // done with this one
                        }
                        else
                        {
                           *CLI->output_stream <<
                              "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                              " Path to CLS_REL resolves to a kludge object other than CLASS_RELATIONSHIP.\n";
                           delete new_inst_rel; new_inst_rel = NULL;
                           // result = 0; Not an internal error.
                        }

                        // do not delete tmp_obj_ptr - object neither created nor owned by us
                        tmp_obj_ptr = NULL;
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                           " Path2Ptr failed to follow the given path to CLS_REL - wrong path?\n";
                        delete new_inst_rel; new_inst_rel = NULL;
                        // result = 0; Not an internal error.
                     }

                     delete tmp_path; tmp_path = NULL;
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                        " String2path failed while parsing path to CLS_REL - probably a syntax error.\n";
                     delete new_inst_rel; new_inst_rel = NULL;
                     // result = 0; Not an internal error.
                  }
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                     " CLS_REL name/path is an empty string.\n";
                  delete new_inst_rel; new_inst_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("NAME_M"))
            {
               // name at master not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  new_inst_rel->Master_Stub.Name = obj_to_insert->contents[i]->contents[1]->content;
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                     " name at master is an empty string.\n";
                  delete new_inst_rel; new_inst_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else if (IS_VAR("NAME_S"))
            {
               // name at slave not a null string?
               if (obj_to_insert->contents[i]->contents[1]->content.size() != 0)
               {
                  new_inst_rel->Slave_Stub.Name = obj_to_insert->contents[i]->contents[1]->content;
               }
               else
               {
                  *CLI->output_stream <<
                     "Error while inserting a new INSTANCE_RELATIONSHIP:\n"
                     " name at slave is an empty string.\n";
                  delete new_inst_rel; new_inst_rel = NULL;
                  // result = 0; Not an internal error.
               }
            }
            else
            {
               *CLI->output_stream <<
                  "Warning: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " Unknown variable encountered, called \""
                  << obj_to_insert->contents[i]->contents[0]->content <<
                  "\".\n"
                  " Skipped.";
            }
         }
#undef IS_VAR(VAR_NAME_STRING)

         // perform a few basic semantic checks over the proto-K_I_Relationship loaded with data
         if (new_inst_rel != NULL)
         {
            if (
                 (new_inst_rel->Master_Stub.Owner == NULL)
                 && (new_inst_rel->Slave_Stub.Owner == NULL)
               )
            {
               *CLI->output_stream <<
                  "Warning: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " both MASTER and SLAVE undefined - setting both to the INSERT destination.\n"
                  " I.e., the INSTANCE_RELATIONSHIP just inserted expresses relationship of\n"
                  " the single master/slave instance to itself. Hope that's what you meant.\n";
            }

            if (new_inst_rel->Master_Stub.Owner == NULL)
                new_inst_rel->Master_Stub.Owner = this;
            if (new_inst_rel->Slave_Stub.Owner == NULL)
                new_inst_rel->Slave_Stub.Owner = this;

            if (
                (new_inst_rel->Master_Stub.Owner != this)
                && (new_inst_rel->Slave_Stub.Owner != this)
               )
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " both MASTER and SLAVE are different from the INSERT destination.\n"
                  " This is not permitted for the moment - subject to further discussion.\n"
                  " INSERT canceled.\n"
                  " This issue is related to the current syntax and basic semantics of\n"
                  " the \"Kludge query language\".\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if (new_inst_rel->Master_Stub.Name.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " Name_M is null.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if (new_inst_rel->Slave_Stub.Name.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " Name_S is null.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if ( new_inst_rel->Master_Stub.Owner->Resolve_Content_Name(
                  &new_inst_rel->Master_Stub.Name,CLI,/*own only*/true) != NULL
            )
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " Name_M already exists under the prospective Master.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if ( new_inst_rel->Slave_Stub.Owner->Resolve_Content_Name(
                  &new_inst_rel->Slave_Stub.Name,CLI,/*own only*/true) != NULL
            )
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " Name_S already exists under the prospective Slave.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if (new_inst_rel->ClsRel == NULL)
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " CLS_REL is null. CLS_REL must be specified. INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if (new_inst_rel->ClsRel->RelType == 2)
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " The indicated class relationship is of inappropriate type.\n"
                  " \"OWN\"-type class relationships can not be instantiated into\n"
                  " instance relationships - rather, they give birth to the primary\n"
                  " ownership of instances and relationships. Thus, you can only refer\n"
                  " to them when inserting new instances and instance relationships.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if ( ! (*new_inst_rel->ClsRel->Master_Stub.Owner
                      >= *new_inst_rel->Master_Stub.Owner->Class))
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " the prospective master instance's class is neither identical to\n"
                  " nor a heir of the specified CLS_REL's master class.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if ( ! (*new_inst_rel->ClsRel->Slave_Stub.Owner
                      >= *new_inst_rel->Slave_Stub.Owner->Class))
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " the prospective slave instance's class is neither identical to\n"
                  " nor a heir of the specified CLS_REL's master class.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if ( new_inst_rel->ClsRel->Master_Stub.Match_Name(
                        &new_inst_rel->Master_Stub.Name, CLI) != 0 )
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " The suggested name at master not accepted.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else if ( new_inst_rel->ClsRel->Slave_Stub.Match_Name(
                        &new_inst_rel->Slave_Stub.Name, CLI) != 0 )
            {
               *CLI->output_stream <<
                  "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                  " The suggested name at slave not accepted.\n"
                  " INSERT canceled.\n";
               delete new_inst_rel; new_inst_rel = NULL;
            }
            else
            {
               // enough of semantic precautions - let's launch it
               if (new_inst_rel->Init(CLI) != 0)
               {
                  *CLI->output_stream <<
                     "Error: while inserting a new INSTANCE_RELATIONSHIP:\n"
                     " Parameter parsing OK, but Kludge_Instance_Relationship::Init() failed.\n"
                     " INSERT canceled.\n";
                  delete new_inst_rel; new_inst_rel = NULL;
                  result = 1;
               }
               // else it's airborne
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error while reading the mandatory attributes of an INSTANCE_RELATIONSHIP\n"
               " being inserted from command line. Giving up - INSERT canceled.\n";
         }
      }
      break;
      case PARM_INST_BLOB :
      {
         // Let's start with a few checks of user input.
         
         // internal sanity check - first parameter = PARM_NAME <name> 
         if (obj_to_insert->contents[0]->type != PARM_NAME)
         {
            cout <<
               "Internal error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
               " the first parameter should be of type PARM_NAME.\n";
            result = 1;
         }
         
         // the proposed instance BLOB name should not be an empty string in the first place
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[0]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " BLOB name can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }
         
         // check if the proposed instance BLOB name exists under `this' instance 
         if ((! user_level_error) && (result == 0))
         {
            tmp_obj_ptr = Resolve_Content_Name(&obj_to_insert->contents[0]->content,
                                               CLI,/*own only?*/true);

            if (tmp_obj_ptr != NULL)
            {
               // this name already exists in this context (path) !
               tmp_pwd = Kludge_CLI_PWD(tmp_obj_ptr, CLI->output_stream);
               
               *CLI->output_stream <<
                  "I've been asked to insert a new BLOB called \""
                  << obj_to_insert->contents[0]->content << "\".\n"
                  " Such a name already exists under Instance " << this->RDN <<
                  ", it resolves to:\n "
                  << *tmp_pwd << endl;
               
               delete tmp_pwd; tmp_pwd = NULL;
               tmp_obj_ptr = NULL;
               //result = 0; // not an internal error
               user_level_error = true;
            }
            //else OK, we can use this new name
         }
         
         // internal sanity check - second parameter = PARM_OBJ_CLASS <class> 
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[1]->type != PARM_OBJ_CLASS)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " the second parameter should be of type PARM_OBJ_CLASS.\n";
               result = 1;
            }
         }

         // the proposed class pathname should not be an empty string
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[1]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " class pathname can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // check if the proposed class pathname exists
         if ((! user_level_error) && (result == 0))
         {
            tmp_path = Kludge_CLI_String2path(
                          &obj_to_insert->contents[1]->content,
                          CLI->output_stream
                       );
            
            if (tmp_path != NULL)
            {
               // check that such an object exists and that it's a KLUDGE_CLASS
               tmp_class = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
               
               if (tmp_class != NULL)
               {
                  if (tmp_class->Type != KLUDGE_CLASS)
                  {
                     *CLI->output_stream <<
                        "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                        " the specified class pathname resolves to an object that\n"
                        " is not a Kludge CLASS.\n";
                     //result = 0; // not an internal error
                     user_level_error = true;
                  }
                  // else the second parameter is indeed a valid Kludge Class. Check passed OK.
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                     " couldn't find a Kludge Object with the specified class pathname.\n";
                  //result = 0; // not an internal error
                  user_level_error = true;
               }

               delete tmp_path; tmp_path = NULL;
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " couldn't decompose (string) class pathname into Kludge_Path elements.\n";
               //result = 0; // not necessarily an internal error
               user_level_error = true;
            }
         }

         // internal sanity check - third parameter = PARM_OBJ_CLSREL <class relationship> 
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[2]->type != PARM_OBJ_CLSREL)
            {
               cout <<
                  "Internal error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " the third parameter should be of type PARM_OBJ_CLSREL.\n";
               result = 1;
            }
         }

         // the proposed class relationship pathname should not be an empty string
         if ((! user_level_error) && (result == 0))
         {
            if (obj_to_insert->contents[2]->content.size() == 0)
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " class relationship pathname can't be an empty string.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // check if the proposed class relationship pathname exists
         if ((! user_level_error) && (result == 0))
         {
            tmp_path = Kludge_CLI_String2path(
                          &obj_to_insert->contents[2]->content,
                          CLI->output_stream
                       );
            
            if (tmp_path != NULL)
            {
               // check that such an object exists and that it's a KLUDGE_CLASS_RELATIONSHIP
               tmp_cls_rel_stub = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
               
               if (tmp_cls_rel_stub != NULL)
               {
                  if (tmp_cls_rel_stub->Type != KLUDGE_CLASS_REL_STUB)
                  {
                     *CLI->output_stream <<
                        "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                        " the specified class relationship pathname resolves to an object that\n"
                        " is not a Kludge CLASS_RELATIONSHIP.\n";
                     //result = 0; // not an internal error
                     user_level_error = true;
                  }
                  // else the third parameter is indeed a valid Kludge Class Relationship. Check passed OK.
               }
               else
               {
                  *CLI->output_stream <<
                     "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                     " couldn't find a Kludge Object with the specified class_rel pathname.\n";
                  //result = 0; // not an internal error
                  user_level_error = true;
               }

               delete tmp_path;
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance::Insert(<INST_BLOB>):\n"
                  " couldn't decompose (string) class_rel pathname into Kludge_Path elements.\n";
               //result = 0; // not an internal error
               user_level_error = true;
            }
         }

         // Done with the checks. Now we can attempt to insert the new instance BLOB
         if ((! user_level_error) && (result == 0))
         {
            // the fourth parameter can be empty.
            if (obj_to_insert->contents.size() > 3)
            {
               tmp_byte_buf = decode_base64(&obj_to_insert->contents[3]->content);
               if (tmp_byte_buf == NULL)
               {
                  result = 1;
                  *CLI->output_stream <<
                        "Error - while inserting a BLOB thingy:\n"
                        " error decoding base64 content.\n";
               }
            }
            // else tmp_byte_buf == NULL

            if (result == 0)
            {
               Kludge_Instance_Property::Property_SubType tmp_subtype;

               if (obj_to_insert->content == "INST_BLOB")
               {
                  tmp_subtype = Kludge_Instance_Property::SUBTYPE_BLOB;
               }
               else if (obj_to_insert->content == "INST_TXT_BLOB")
               {
                  tmp_subtype = Kludge_Instance_Property::SUBTYPE_TXT_BLOB;
               }
               else
               {
                  cout << "Internal error - in Kludge_Instance::Insert(<BLOB thingy>):\n"
                          " unknown BLOB subtype: \"" << obj_to_insert->content << "\"\n";
                  result = 1;
               }

               if (result == 0)
               {
                  new_inst_prop = new Kludge_Instance_Property(
                                       &obj_to_insert->contents[0]->content, //name
                                       this,
                                       (Kludge_Class*)tmp_class,
                                       ((Kludge_Class_Rel_Stub*)tmp_cls_rel_stub)->Body,
                                       NULL,                                 // no straight content
                                       CLI,
                                       tmp_subtype,                          // blob flags
                                       tmp_byte_buf                          // binary BLOB content
                                      );

                  if (new_inst_prop != NULL)
                  {
                     if (new_inst_prop->ctor_failed)
                     {
                        cout << "Kludge_Instance::Insert(): Failed to create "
                                "a new INST_BLOB thingy (although malloc() OK).\n";
                        delete new_inst_prop; new_inst_prop = NULL;
                        result = 1;
                     }
                     // else OK - finished inserting the blob.
                  }
                  else
                  {
                     cout << "Kludge_Instance::Insert(): Failed to create (malloc?) "
                             "a new INST_BLOB thing.\n";
                     result = 1; // internal error
                  }
               }
               // else error already reported

               if (tmp_byte_buf != NULL)
               {
                  delete tmp_byte_buf; tmp_byte_buf = NULL;
               } // else OK - probably an empty BLOB (a valid thing to insert)
            }
            // else error already reported - perhaps base64 decoding failed
         }
         // else error already reported - something wrong with the previous checks
      }
      break;
      default :
      {
         *CLI->output_stream <<
         "Can only insert one of the following into an INSTANCE:\n"
         " - an INSTANCE,\n"
         " - an INSTANCE_PROPERTY,\n"
         " - an INSTANCE_RELATIONSHIP,\n"
         " - or one of the BLOB brethren:\n"
         "    - an INST_BLOB,\n"
         "    - an INST_TXT_BLOB\n";
         //result = 0; // not an internal error
      }
      break;
   }

   return(result);
}



int Kludge_Instance::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
      "Set a value to an instance? Conceptually this could be justified for the potential\n"
      " occasions when someone might want to mask an instance as a property for backwards\n"
      " compatibility with old methods' code - but for now this just doesn't fit into\n"
      " the picture (syntax and all) - so let's keep this a future possibility.\n"
      " Value NOT set.\n";
   return(0);
}



int Kludge_Instance::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 0;
 //Kludge_Class* former_parent = NULL;
 ostrstream qry_string;

   if (destination != NULL)
   {
      // can only move under another Kludge Instance
      if (destination->Type == KLUDGE_INSTANCE)
      {
         if (destination->Resolve_Content_Name(&RDN,CLI,/*own only?*/true) == NULL)
         {
            if ( *((Kludge_Instance*)destination)->Class <= *ClsRel->Master_Stub.Owner )
            {
               if (Owner != NULL)
               {
                  Add_To_Owner_Index((Kludge_Instance*)destination);
                  Free_From_Owner_Index();
                  //former_owner = Owner;   // - for locking purposes only
                  Owner = (Kludge_Instance*)destination;
                  
                  qry_string << "UPDATE Instances SET Owner = \""
                             << Owner->OID
                             << "\" WHERE OID = \""
                             << OID
                             << "\""
                             <<  ends;
                  result = CLI->thr->DB_Cmd(qry_string.str());
                  if (result != 0)
                  {
                     *CLI->output_stream <<
                        "In Kludge_Instance::Move():\n"
                        " Error while storing the new owner into the backend database.\n";
                  }
               }
               else
               {
                  *CLI->output_stream <<
                  "In Kludge_Instance::Move(): can't move an instance with NULL owner.\n"
                  " Probably an internal error - an attempt to move an unanchored instance.\n";
                  // Kludge_CLI::Execute_Command()'s cycle-avoidance rules alone should
                  // prevent any attempts to move the root instance.
                  result = 1;
               }
            }
            else
            {
               *CLI->output_stream <<
                  "In Kludge_Instance::Move(): the destination is not\n"
                  " a valid Owner as per the current ClsRel. Move canceled.\n";
               result = 1;
            }
         }
         else
         {
            *CLI->output_stream <<
               "In Kludge_Instance::Move(): content with this name already exists\n"
               " at the desired destination. Move canceled.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Instance::Move(): can only move under another kludge instance.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "In Kludge_Instance::Move(): internal error: obtained a NULL destination pointer.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   if (new_name->type == PARM_NAME)
   {
      if (new_name->content.size() != 0)
      {
         if (Owner->Resolve_Content_Name(&new_name->content,CLI,/*own only?*/true) == NULL)
         {
            if (ClsRel->Master_Stub.Match_Name(&new_name->content,CLI) == 0)
            {
               Free_From_Owner_Index();
               RDN = new_name->content;
               Add_To_Owner_Index();

               qry_string << "UPDATE Instances SET RDN = \""
                          << RDN 
                          << "\" WHERE OID = \""
                          << OID
                          << "\""
                          <<  ends;

               result = CLI->thr->DB_Cmd(qry_string.str());
               if (result != 0)
               {
                  *CLI->output_stream <<
                     "In Kludge_Instance::Rename():\n"
                     " Error while storing the new name into the backend database.\n";
               }
            }
            else
            {
               *CLI->output_stream <<
                  "In Kludge_Instance::Rename(): the requested name doesn't match NAMINGMATCH_M\n"
                  " specified in the CLS_REL of this instance. Rename operation canceled.\n";
               result = 1;
            }
         }
         else
         {
            *CLI->output_stream <<
               "In Kludge_Instance::Rename(): content with this name already exists\n"
               " in the owner instance. Rename operation canceled.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Instance::Rename(): the new name cannot be an empty string.\n"
            " Rename operation canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance::Rename():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_NAME.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
 del_param del_parm_passed;
  del_parm_passed.CLI = CLI;
  del_parm_passed.objs_deleted.clear();
 int result = 0;
 unsigned int counter = 0;
 ostrstream qry_string;

   // try to run the Perl-space destructor
   Try_Method("DTOR", NULL, CLI, /*ignore inherited DTORs*/true);

   if (Instances_Owned == NULL) Load_Instances_Owned(CLI);
   if (Instances_Owned == NULL)
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance::DB_Delete() : failed to initialize the index\n"
         " of instances owned (used to check their existence/absence before deletion).\n";
      result = 1;
   }
   if (Properties == NULL) Load_Properties(CLI);
   if (Properties == NULL)
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance::DB_Delete() : failed to initialize the index\n"
         " of properties (used to check their existence/absence before deletion).\n";
      result = 1;
   }
   if (! relationships_loaded) Load_Relationships(CLI);
   if ((! relationships_loaded) || (Relationships == NULL))
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance::DB_Delete() : failed to initialize the index\n"
         " of relationships (used to check their existence/absence before deletion).\n";
      result = 1;
   }

   if (RDN == "Root_Instance")
   {
      *CLI->output_stream <<
         "In Kludge_Instance::DB_Delete(): cannot delete the Root_Instance.\n";
      result = 1;
   }

   if (result == 0)
   {
      if (   (g_tree_nnodes(Instances_Owned) > 0)
          || (g_tree_nnodes(Properties) > 0)
          || (g_tree_nnodes(Relationships) > 0)
         )
      {
         if (! recursive)
         {
            *CLI->output_stream <<
            "Cannot delete instance " << RDN <<
            ". There are still contents inside.\n";
            result = 1;
         }
         else
         // Unfortunately we can't just delete'em'all in one SQL sweep - we have
         // to perform some checks on each of them. That's why we have to call
         // DB_Delete on every one of them. We only Zap those that have been
         // successfully DB_Delete()'d.
         {
            g_tree_traverse(          // DB_Delete() all the owned instances
                  Instances_Owned,
                  Kludge_Delete_Contents_Traverse,
                  G_POST_ORDER,
                  &del_parm_passed
                           );
            g_tree_traverse(          // DB_Delete() all the properties
                  Properties,
                  Kludge_Delete_Contents_Traverse,
                  G_POST_ORDER,
                  &del_parm_passed
                           );
            g_tree_traverse(          // DB_Delete() all the properties
                  Relationships,
                  Kludge_Delete_Contents_Traverse,
                  G_POST_ORDER,
                  &del_parm_passed
                           );
            // and Zap() them all
            for (counter = 0; counter < del_parm_passed.objs_deleted.size(); counter++)
            {
               del_parm_passed.objs_deleted[counter]->Zap();
            }
            del_parm_passed.objs_deleted.clear();

            // if the trees are empty, cut them down:
            if (g_tree_nnodes(Instances_Owned) == 0) 
            {
               g_tree_destroy(Instances_Owned);
               Instances_Owned = NULL;
            }
            else result = 1;

            if (g_tree_nnodes(Properties) == 0)
            {
               g_tree_destroy(Properties);
               Properties = NULL;
            }
            else result = 1;

            if (g_tree_nnodes(Relationships) == 0)
            {
               g_tree_destroy(Relationships);
               Relationships = NULL;
               relationships_loaded = false;
            }
            else result = 1;
         }
      }
      else
      {  // no instances owned, no properties, no relationships
         g_tree_destroy(Instances_Owned);
         Instances_Owned = NULL;
         g_tree_destroy(Properties);
         Properties = NULL;
         g_tree_destroy(Relationships);
         Relationships = NULL;
         relationships_loaded = false;
      }
   }
   //else the error has already been reported.

   // Delete this Kludge_Instance from the DB.
   // Can't be deallocated here - this function is used from within
   // a traversal function. Deallocate ('delete') it in the calling function.
   if (result == 0)
   {
      qry_string << "DELETE FROM Instances WHERE OID = \""
                 << OID
                 << "\""
                 <<  ends;
      result = CLI->thr->DB_Cmd(qry_string.str());
   }

   return(result);
}



int Kludge_Instance::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   int result = DB_Delete(CLI, recursive);
   if (result == 0)
   {
      cout <<
         "Deleting instance: " << RDN << endl;
      result = Zap();
   }

   return(result);
}



int Kludge_Instance::pwd(string* result)
{
   if (Owner == NULL)
   {
      Kludge_Root->pwd(result);
   }
   else
   {
      Owner->pwd(result);
   }
   result->append(RDN);
   result->append("/");
   return(0);
}



Kludge_Object* Kludge_Instance::Superior()
{
   if (Owner != NULL) return(Owner);
   else if (RDN == "Root_Instance") return(Kludge_Root);
   else return (NULL);
}



int Kludge_Instance::Try_Method(char* method_name, Kludge_pl_params* args, 
                                Kludge_CLI* CLI, bool own_only = false)
{
 int result = 0;
 string tmp_name = method_name;
 Kludge_Object* tmp_obj_ptr = NULL;
   
   if (( tmp_obj_ptr = Class->Resolve_Content_Name(&tmp_name, CLI, own_only) ))
   {
      if (tmp_obj_ptr->Type == Kludge_Object::KLUDGE_CLASS_PROPERTY)
      {
         if ( ((Kludge_Class_Property*)tmp_obj_ptr)->SubType == Kludge_Class_Property::SUBTYPE_METHOD)
         {
            if (args == NULL) args = new Kludge_pl_params;     // If no args were supplied,
            args->push_back(new Kludge_pl_param("this",this)); // at least provide "this".
            result = ((Kludge_Class_Property*)tmp_obj_ptr)->Run_Method(CLI, args);
         }
         // else no such method available (not an error) - the name exists but is not a METHOD
      }
      // else no such method available (not an error) - the name exists but is not a METHOD
   }
   // else no such name under the respective Kludge_Class
   // ==> no such method available (not an error)

   return(result);
}



//static
Kludge_MYSQL_RES* Kludge_Instance::DB_Lookup(OID_Type OID, Kludge_CLI* CLI)
{
 Kludge_MYSQL_RES* result = NULL;
 MYSQL_RES *mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 ostrstream qry_string;

   qry_string << "SELECT OID,Owner FROM Instances WHERE OID = \""
              << OID
              << "\""
              <<  ends;
   mysql_query_result = CLI->thr->DB_Query(qry_string.str());

   if (mysql_query_result != NULL)
   {
      result_row = mysql_fetch_row(mysql_query_result);
      if (result_row != NULL)
      {
         result = new Kludge_MYSQL_RES(result_row, mysql_query_result);
      }
      else
      {
         mysql_free_result(mysql_query_result);
         // query result returned with no data rows - no such instance in the DB.
         // That's not necessarily an internal error - don't report it here, 
         // just return NULL.
      }
   }
   else
   {
      cout << "Internal error in Kludge_Instance::DB_Lookup():\n"
              " query returned null MYSQL_RES* .\n";
   }

   return(result);
}



// expects row->Row[0] = int OID, row->Row[1] = int Owner
//static
Kludge_Instance* Kludge_Instance::Load_By_ID(Kludge_MYSQL_RES* row, Kludge_CLI* CLI)
{
 OID_Type tmp_ID = 0;
 vector<OID_Type> branch_to_load;
 Kludge_Instance* result = NULL;
 bool dead_end = false;
 istrstream* tmp_ss = NULL; // for MY_SSCANF

   MY_SSCANF(row->Row[0],tmp_ID);                // read instance ID

   result = ID2ptr_Raw(&tmp_ID);
   if (result == NULL)                    // if this instance is not loaded yet
   {
      // identify its superior instance ID's bottom-up
      // until the first one that's loaded.
      branch_to_load.push_back(tmp_ID);
      while ((result == NULL) && (! dead_end))
      {
         MY_SSCANF(row->Row[1],tmp_ID);          // read owner instance ID
         delete row; row = NULL;

         result = ID2ptr_Raw(&tmp_ID);
         if (result == NULL)
         {
            branch_to_load.push_back(tmp_ID);
            row = DB_Lookup(tmp_ID, CLI);
            if (row == NULL)
            {
               cout <<
                "Internal error in Kludge_Instance::Load_By_ID():\n"
                " referenced owner instance ID could not be found in the DB.\n";
               dead_end = true;
            }
         }
         // else the owner instance is already loaded. Now that we have it by
         // its pointer, we can load its owned instances and then look up our
         // desired owned instance.
      }

      if (row != NULL)  // clean up the mysql session before proceeding
      {
         delete row;
         row = NULL;
      }

      // Now descend the chain of identified superior instance ID's
      // and load them, one by one, top-to-bottom
      if (! dead_end) // if ready to go
      {
         for (int i = branch_to_load.size() - 1; (i >= 0) && (! dead_end); i--)
         {
            if (result->Instances_Owned == NULL)    // instances owned loaded?
            {
               result->Load_Instances_Owned(CLI, false); // just load them all
            }
            result = ID2ptr_Raw(&branch_to_load[i]);
            if (result == NULL)
            {
               cout <<
                "Internal error in Kludge_Instance::Load_By_ID():\n"
                " an instance owned has mysteriously disappeared before it could be loaded.\n";
               dead_end = true;
            }
         }
      }
      // else there was a failure. No need to bother.

      branch_to_load.clear();
   }
   // else the instance is already loaded - which may or may not be an error,
   // depending on which function has called us.
   
   if (row != NULL)  // A final cleanup of the mysql session - takes effect if
   {                 // the instance we were asked to load was already loaded.
      delete row;
      row = NULL;
   }

   return(result);
}



//static
Kludge_Instance* Kludge_Instance::ID2ptr_Raw(OID_Type* ID)
{
 Kludge_Instance *result = NULL;

   result = (Kludge_Instance *) g_tree_lookup(Index,ID);
   
   return(result);
}



//static
Kludge_Instance* Kludge_Instance::ID2ptr(OID_Type* ID, Kludge_CLI* CLI)
{
 Kludge_Instance *result = NULL;
 Kludge_MYSQL_RES* tmp_mysql_res = NULL;

   result = ID2ptr_Raw(ID);
   if (result == NULL)
   {
      tmp_mysql_res = DB_Lookup(*ID, CLI);
      if (tmp_mysql_res != NULL)
      {
         result = Load_By_ID(tmp_mysql_res, CLI);
      }
      // else no such instance in the DB - implies no such instance
      // in the system. Not necessarily an internal error => no report here.
   }
   // else instance already loaded
   
   return(result);
}



//static
int Kludge_Instance::List_Properties_Traverse(gpointer key, gpointer Property, gpointer CLI)
{
   ((Kludge_Instance_Property *)Property)->List_Contents((Kludge_CLI*)CLI);
   
   return(false);
}



//static
int Kludge_Instance::List_Properties_Names_Traverse(gpointer key, gpointer Property, gpointer CLI)
{
   ((Kludge_Instance_Property *)Property)->Describe_Self((Kludge_CLI*)CLI);
   
   return(false);
}



//static
int Kludge_Instance::List_Instances_Owned_Traverse(gpointer key, gpointer Instance_Owned, gpointer CLI)
{
   ((Kludge_Instance *)Instance_Owned)->List_Contents((Kludge_CLI *)CLI, true);
   return(false);
}



//static
int Kludge_Instance::List_Instances_Owned_Names_Traverse(gpointer key, gpointer Instance_Owned, gpointer CLI)
{
   ((Kludge_Instance *)Instance_Owned)->Describe_Self((Kludge_CLI *)CLI);

   return(false);
}



//static
int Kludge_Instance::List_Relationships_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI)
{
   ((Kludge_Inst_Rel_Stub*)Rel_Stub)->List_Contents((Kludge_CLI *)CLI);
   return(false);
}



//static
int Kludge_Instance::List_Relationship_Names_Traverse(gpointer key, gpointer Rel_Stub, gpointer CLI)
{
   ((Kludge_Inst_Rel_Stub*)Rel_Stub)->Describe_Self((Kludge_CLI *)CLI);

   return(false);
}



//static
int Kludge_Instance::Free_Properties_Traverse(gpointer key, gpointer Property, gpointer data)
{
   delete (Kludge_Instance_Property *)Property;
   return(false);
}



//static
int Kludge_Instance::Free_Instances_Owned_Traverse(gpointer key, gpointer Owned_Instance, gpointer data)
{
   delete (Kludge_Instance *)Owned_Instance;
   return(false);
}



//static
int Kludge_Instance::Free_Relationships_Traverse(gpointer key, gpointer Relationship, gpointer rels_to_free)
{
   ((vector<Kludge_Inst_Rel_Stub*>*)rels_to_free)->push_back((Kludge_Inst_Rel_Stub*)Relationship);
   return(false);
}










/* ###    class Kludge_Instance_Property    ### */


// This constructor is used for creating a new property from scratch.
// It cannot be used for loading a property from the DB.
// It actually stores the newly created property in the DB.
//
Kludge_Instance_Property::Kludge_Instance_Property(
  string* inst_prop_name,
  Kludge_Instance* owner_instance,
  Kludge_Class* inst_prop_class,
  Kludge_Class_Relationship* cls_rel,
  string* val,
  Kludge_CLI* CLI,
  Property_SubType _type = SUBTYPE_PROPERTY,
  byte_buf* blob_value = NULL
) :
  OID(0),
  Instance(owner_instance),
  Class(inst_prop_class),
  ClsRel(cls_rel),
  RDN(*inst_prop_name),
  //Value(*val),
  ctor_failed(false),
  SubType(_type)
{
 int stor_success = 0;
 ostrstream* qry_string;
 ostrstream* tmp_ss;
 string blob_tbl_name;
 unsigned int tmp_flags = 0;
 char* tmp_escaped_data = NULL;
 
   // first let's try a few referential sanity checks
   if (*ClsRel->Slave_Stub.Owner >= *Class)
   {
      if (*ClsRel->Master_Stub.Owner >= *Instance->Class)
      {
         if (ClsRel->RelType == 2)
         {
            // okay :) The Owner/Class/ClsRel references seem sane.
            if (ClsRel->Master_Stub.Match_Name(&RDN, CLI) != 0)
            {
               *CLI->output_stream <<
                  "Suggested instance property name not accepted. Property not created.\n";
               ctor_failed = true;
            }
            // else fine - sanity checks passed, including the regexp name check.
         }
         else
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance_Property::Kludge_Instance_Property():\n"
               " the indicated class relationship is of inappropriate relationship type.\n"
               " Its type should be \"OWN\" (ownership - the \"primary key\" relationship).\n"
               " Instance NOT inserted.\n";
            ctor_failed = true;
         }
      }
      else
      {
         *CLI->output_stream <<
            "Error - in Kludge_Instance_Property::Kludge_Instance_Property():\n"
            " the indicated owner instance is not a valid master for the indicated\n"
            " class relationship.\n"
            " Property NOT inserted.\n";
         ctor_failed = true;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Error - in Kludge_Instance_Property::Kludge_Instance_Property():\n"
         " the indicated class is not a valid slave for the indicated\n"
         " class relationship.\n"
         " Property NOT inserted.\n";
      ctor_failed = true;
   }

   if (! ctor_failed) // if the referential sanity is fine, let's try to store the BLOB content
   {
      Type = KLUDGE_INSTANCE_PROPERTY;

      if (SubType == SUBTYPE_PROPERTY)
      {  // not a blob
         if (val != NULL)
         {
            Value = *val;
         }
         else
         {
            cout << "Internal error - in Kludge_Instance_Property:: ctor :\n"
                    " asked to create a regular Kludge_Instance_Property,\n"
                    " obtained a NULL PTR for value!\n";
            ctor_failed = true;
         }
      }
      else // SubType != SUBTYPE_PROPERTY  ==> some sort of a blob
      {
         switch (SubType)
         {
            case SUBTYPE_BLOB:
            {
               blob_tbl_name = "Blobs";
               tmp_flags = INSTPROP_IS_BLOB;
            }
            break;
            case SUBTYPE_TXT_BLOB:
            {
               blob_tbl_name = "Text_Blobs";
               tmp_flags = INSTPROP_IS_TXT_BLOB;
            }
            break;
            default:
            {
               cout << "Internal error - in Kludge_Instance_Property:: ctor :\n"
                       " asked to create a BLOB thingy, obtained an invalid SubType!\n";
               ctor_failed = true;
            }
            break;
         }

         if (!ctor_failed)
         {
            if (blob_value != NULL)
            {
               tmp_escaped_data = new char[blob_value->len * 2 + 1];
               if (tmp_escaped_data == NULL)
               {
                  cout <<
                     "Internal error - in Kludge_Instance_Property:: ctor :\n"
                     " while storing BLOB contents, couldn't allocate memory for the escaped string!\n";
                  ctor_failed = true;
               }
            }
            // else no problem - we're inserting an empty BLOB

            if (! ctor_failed)
            {
               CLI->thr->Open_DB(); // If this is not the first thread to look up all
                                    // the necessary data, the database session might 
                                    // still be closed in this thread.
               if (tmp_escaped_data != NULL)
               {
                  mysql_real_escape_string(
                          CLI->thr->DB,            // requires an open DB session!
                          tmp_escaped_data,
                          (char*) blob_value->buf, // the cast suppresses a warning about signedness
                          blob_value->len
                                          );
               }
               
               qry_string = new ostrstream;
               *qry_string << "INSERT INTO " << blob_tbl_name << " VALUES (\""
                           << 0
                           << "\",\""
                           << ((tmp_escaped_data == NULL) ? "" : tmp_escaped_data)
                           << "\")"
                           <<  ends;   
               stor_success = CLI->thr->DB_Cmd(qry_string->str());
               if (stor_success == 0)
               {  // get the auto-generated blob ID and store it in the property's Value
                  tmp_ss = new ostrstream;
                  *tmp_ss << mysql_insert_id(CLI->thr->DB) << ends;
                  Value = tmp_ss->str();
                  delete tmp_ss; tmp_ss = NULL;
               }
               else
               {
                  cout << "Internal error - in Kludge_Instance_Property:: ctor :\n"
                          " DB op failed while storing BLOB content!\n";
                  ctor_failed = true;
               }
               delete qry_string; qry_string = NULL;
               if (tmp_escaped_data != NULL)
               {
                  delete tmp_escaped_data; tmp_escaped_data = NULL;
               }
            }
            // else error already reported
         }
         // else something wrong with the SubType - error already reported
      }
   }
   // else something went wrong. Error already reported.

   if (! ctor_failed) // referential sanity is fine, BLOB content stored fine
   {
      if ( Instance->Properties == NULL ) Instance->Load_Properties(CLI);

      qry_string = new ostrstream;      
      *qry_string << "INSERT INTO Instance_Properties VALUES (NULL,\""
                  << Instance->OID
                  << "\",\""
                  << RDN
                  << "\",\""
                  << Class->Class
                  << "\",\""
                  << ClsRel->RelCLSID
                  << "\",\""
                  << Value
                  << "\",\""
                  << tmp_flags
                  << "\")"
                  <<  ends;   
      stor_success = CLI->thr->DB_Cmd(qry_string->str());
      delete qry_string; qry_string = NULL;

      if (stor_success == 0)
      {
         OID = mysql_insert_id(CLI->thr->DB);
         Add_To_Index();
      }
      else
      {
         ctor_failed = true;
      }
   }
   // else no need to bother.
}



// This constructor can only be used for loading an existing Kludge_Instance_Property
// from a database. It is not intended for creating a new instance property from scratch.
//
Kludge_Instance_Property::Kludge_Instance_Property(
  MYSQL_ROW DB_record,
  Kludge_Instance* owner_instance,
  Kludge_CLI* CLI
) :
  OID(0),
  Instance(owner_instance),
  Class(NULL),
  ClsRel(NULL),
  RDN(""),
  Value(""),
  ctor_failed(false),
  SubType(SUBTYPE_PROPERTY)
{ 
   Type = KLUDGE_INSTANCE_PROPERTY; 

   if ( ! Extract_From_MysqlRow(DB_record,CLI) )
   {
      ctor_failed = true;
   }
   else Add_To_Index();
}



Kludge_Instance_Property::~Kludge_Instance_Property()
{
   cout << "Deallocating instance property: " << RDN <<"\n";
   Free_From_Class_Index();
   Free_From_ClsRel_Index();
   // we have to be able to delete while traversing Properties
   // ==> no Free_From_Owner_Index() here. Zap() does that for us.
}



//
// Expects:  
// Row_To_Process[0] == OID
// Row_To_Process[1] == Instance
// Row_To_Process[2] == RDN
// Row_To_Process[3] == Class
// Row_To_Process[4] == ClsRel
// Row_To_Process[5] == Value
//
bool Kludge_Instance_Property::Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI) 
{
 bool result = true;
 OID_Type tmp_cls = 0, tmp_cls_rel = 0;
 istrstream* tmp_ss = NULL; // for MY_SSCANF
 unsigned int tmp_flags = 0;

   if (Row_To_Process != NULL)
   {
      // Set Kludge_Instance_Property.RDN (of type `string' - hence the direct assignment)
      RDN = Row_To_Process[2];
      // Set Kludge_Instance_Property.Value (of type `string' - hence the direct assignment)
      Value = Row_To_Process[5];

      MY_SSCANF(Row_To_Process[0],OID);
      MY_SSCANF(Row_To_Process[3],tmp_cls);
      MY_SSCANF(Row_To_Process[4],tmp_cls_rel);
      MY_SSCANF(Row_To_Process[6],tmp_flags);

      Class = Kludge_Class::ID2ptr(&tmp_cls,CLI);
      if (Class == NULL)
      {
         cout <<
            "In Kludge_Instance_Property::Extract_From_MysqlRow():\n"
            " Class ID could not be resolved to a Kludge_Class pointer!\n"
            " Inconsistency in the backend SQL database?\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }
      
      ClsRel = Kludge_Class_Relationship::ID2ptr(&tmp_cls_rel,CLI);
      if (ClsRel == NULL)
      {
         cout <<
            "In Kludge_Instance_Property::Extract_From_MysqlRow():\n"
            " ClsRel ID could not be resolved to a K_C_R pointer!\n"
            " Inconsistency in the backend SQL database?\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }

      if (tmp_flags == 0) SubType = SUBTYPE_PROPERTY;
      else if (tmp_flags & INSTPROP_IS_BLOB) SubType = SUBTYPE_BLOB;
      else if (tmp_flags & INSTPROP_IS_TXT_BLOB) SubType = SUBTYPE_TXT_BLOB;
   }
   else
   {
      cout << 
         "In Kludge_Instance_Property::Extract_From_MysqlRow():\n"
         " the row passed to this function is a NULL pointer!\n"
         " Extract_From_MysqlRow() failed.\n";
      result = false;
   }

   return(result);
}



int Kludge_Instance_Property::Add_To_Owner_Index(Kludge_Instance* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Instance;

   cout << "Adding instance property: " << RDN << "\n"
        << " into the index of instance " << Where->RDN << "\n";
   
   // load properties if necessary? Nah, I don't think so - that'd be defensive programming.
   if (Where != NULL)
   {
      if (Where->Properties != NULL)
      {
         g_tree_insert(Where->Properties, &RDN,this);
      }
      else
      {
         cout << "Kludge_Instance_Property::Add_To_Index(): property index not initalized (pointer null).\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error: in Kludge_Instance_Property::Add_To_Owner_Index():\n"
               " pointer to the target instance is NULL.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance_Property::Add_To_Class_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Class;

   cout << "Adding instance property " << RDN << " to its Class'es index.\n";
   if (Where != NULL)
   {
      if (Where->Instances == NULL) Where->Init_Instances();
      
      if (Where->Instances != NULL)
      {
         g_tree_insert(Where->Instances,&OID,this);
      }
      else
      {
         cout << "Internal error: in Kludge_Instance_Property::Add_To_Class_Index():\n"
                 " could not initialize the Class'es index of instances.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error: in Kludge_Instance_Property::Add_To_Class_Index():\n"
               " pointer to the target class is NULL.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Property::Add_To_ClsRel_Index(Kludge_Class_Relationship* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = ClsRel;

   cout << "Adding instance property " << RDN << " to its class relationship's index.\n";
   if (Where != NULL)
   {
      if (Where->Instances == NULL) Where->Init_Instances();
      
      if (Where->Instances != NULL)
      {
         g_tree_insert(Where->Instances,&OID,this);
      }
      else
      {
         cout << "Kludge_Instance_Property::Add_To_ClsRel_Index():\n"
                 " could not initialize the Class Relationship's index of instances.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error: in Kludge_Instance_Property::Add_To_ClsRel_Index():\n"
               " pointer to the target class relationship is NULL.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance_Property::Free_From_Owner_Index(Kludge_Instance* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Instance;

   cout << "Releasing instance property " << RDN << " from Owner's index.\n";
   if (Where != NULL)
   {
      if (Where->Properties != NULL)
      {
         g_tree_remove(Where->Properties,&RDN);
      }
      else
      {
         cout << "Internal error: in Kludge_Instance_Property::Free_From_Owner_Index():\n"
                 " Owner's index not initialized.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error: in Kludge_Instance_Property::Free_From_Owner_Index():\n"
               " pointer to the target instance is NULL.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Property::Free_From_Class_Index(Kludge_Class* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Class;

   cout << "Releasing instance property " << RDN << " from its Class'es index.\n";
   if (Where != NULL)
   {
      if (Where->Instances != NULL)
      {
         g_tree_remove(Where->Instances,&OID);
      }
      else
      {
         cout << "Internal error: in Kludge_Instance_Property::Free_From_Class_Index():\n"
                 " the Class'es \"Instances\" index not initialized.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error: in Kludge_Instance_Property::Free_From_Class_Index():\n"
               " pointer to the target class is NULL.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Property::Free_From_ClsRel_Index(Kludge_Class_Relationship* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = ClsRel;

   cout << "Releasing instance property " << RDN << " from its Class Relationship's index.\n";
   if (Where != NULL)
   {
      if (Where->Instances != NULL)
      {
         g_tree_remove(Where->Instances,&OID);
      }
      else
      {
         cout << "Internal error: in Kludge_Instance_Property::Free_From_ClsRel_Index():\n"
                 " the Class Relationship's \"Instances\" index not initialized.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error: in Kludge_Instance_Property::Free_From_ClsRel_Index():\n"
               " pointer to the target class relationship is NULL.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Property::Add_To_Index()
// local, Class, ClsRel
{
 int result = 0;

   result = Add_To_Owner_Index();
   if (result == 0) result = Add_To_Class_Index();
   if (result == 0) result = Add_To_ClsRel_Index();
   
   return(result);
}



// Zap() does a nice memory cleanup.
// Only makes sense 
// a) for dynamic partial release from memory 
//     In that case, invalidate the list of properties.
// b) when a property is deleted
//     Requires deletion from the SQL-based object repository
//      - i.e., use DB_Delete() and then Zap() (or just Delete())
int Kludge_Instance_Property::Zap()
{
 int result = 0;

   cout << "Zapping an Instance Property: " << Instance->RDN << "/" << RDN << "\n" ;
   if (Instance->Properties != NULL)
   {
      Free_From_Owner_Index();  // the other indexes are taken care of by the dtor
   }
   else
   {
      cout << 
         "Internal error - in Kludge_Instance_Property::Zap():\n"
         " parent's Properties index not initialized!\n";
      result = 1;
   }  
   delete this;

   return(result);
}



int Kludge_Instance_Property::Report_Subtype(Kludge_CLI* CLI)
{
 int result = 0;

   switch (SubType)
   {
      case SUBTYPE_PROPERTY :
      {
         *CLI->output_stream << "INSTANCE_PROPERTY ";
      }
      break;
      case SUBTYPE_BLOB :
      {
         *CLI->output_stream << "INST_BLOB ";
      }
      break;
      case SUBTYPE_TXT_BLOB :
      {
         *CLI->output_stream << "INST_TXT_BLOB ";
      }
      break;
      default:
      {
         cout << "Internal error - in Kludge_Instance_Property::Report_Subtype():\n"
                 " unknown instance property subtype!\n";
         result = 1;
      }
      break;
   }

   return(result);
}



Kludge_MYSQL_RES* Kludge_Instance_Property::Get_Blob_Content(Kludge_CLI* CLI)
{
 MYSQL_RES* mysql_query_result = NULL;
 MYSQL_ROW result_row = NULL;
 Kludge_MYSQL_RES* result = NULL;
 ostrstream qry_ss;
 string blob_tbl_name;
 bool failure = false;

   switch (SubType)
   {
      case SUBTYPE_BLOB:
      {
         blob_tbl_name = "Blobs";
      }
      break;
      case SUBTYPE_TXT_BLOB:
      {
         blob_tbl_name = "Text_Blobs";
      }
      break;
      default:
      {
         cout << "Internal error - in Kludge_Instance_Property::Get_Blob_Content() :\n"
                 " invalid Kludge_Instance_Property::SubType!\n";
         failure = true;
      }
      break;
   }

   if (!failure)
   {
      qry_ss << "SELECT Content FROM " << blob_tbl_name << " WHERE BlobID = \""
             << Value << "\"" << ends;
      if (( mysql_query_result = CLI->thr->DB_Query(qry_ss.str()) ))
      {
         if (( result_row = mysql_fetch_row(mysql_query_result) ))
         {
            result = new Kludge_MYSQL_RES(result_row, mysql_query_result);
         }
         else
         {
            cout << "Internal error - in Kludge_Instance_Property::Get_Blob_Content():\n"
                    " while loading BLOB content, DB query returned empty result.\n";
            // result = NULL;
            mysql_free_result(mysql_query_result);
         }
      }
      else
      {
         cout << "Internal error - in Kludge_Instance_Property::Get_Blob_Content():\n"
                 " DB query failed while loading BLOB. Query API function returned NULL result*.\n";
         // result = NULL;
      }
   }

   return(result);
}



int Kludge_Instance_Property::List_Contents(Kludge_CLI* CLI, bool recursive = false)
// The recursive tag doesn't really make sense.
// It's here for the sake of compliance with the abstract interface
// of Kludge_Object.
{
 string* tmp_pwd = NULL;
 int result = 0;
 Kludge_MYSQL_RES* qry_result = NULL;

   if (SubType == SUBTYPE_PROPERTY) result = Describe_Self(CLI);
   else  // some sorta BLOB
   {
      if (ClsRel != NULL)
      {
         tmp_pwd = Kludge_CLI_PWD(&ClsRel->Master_Stub,CLI->output_stream);
         if (tmp_pwd == NULL)
         {
            cout << "Error: In Kludge_Instance_Property::List_Contents():\n"
                    " something's horribly wrong - ClsRel not null but Kludge_CLI_PWD() failed.\n";
            result = 1;
         }
      }
      else
      {
         cout << "Error: In Kludge_Instance_Property::List_Contents():\n"
                 " ClsRel pointer is NULL!\n";
         result = 1;
      }

      if (Class == NULL)
      {
         cout << "Error: In Kludge_Instance_Property::List_Contents():\n"
                 " Class pointer is NULL!\n";
         result = 1;
      }

      if (result == 0)
      {
         result = Report_Subtype(CLI);

         if (result == 0)
         {
            *CLI->output_stream
                   // << "INSTANCE_PROPERTY \""
                   << "\""
                   << RDN
                   << "\" \""
                   << Class->Name
                   << "\" \""
                   << *tmp_pwd
                   << "\" {\n";  // BLOB grammar lead-in done.
            delete tmp_pwd;
                               
            // Now dump BLOB data in Base64.
            if ((qry_result = Get_Blob_Content(CLI)))
            {
               byte_buf* tmp_buf = new byte_buf(
                                          qry_result->Row[0],
                                          mysql_fetch_lengths(qry_result->Result)[0],
                                          false, // deep copy? (false = shallow copy)
                                          false  // delete the buf* after use? (no, not here!)
                                               );
               string* tmp_base64_data = encode_base64(tmp_buf); // The result contains a trailing \n.
               delete tmp_buf;
               delete qry_result;

               *CLI->output_stream << *tmp_base64_data;

               delete tmp_base64_data;
            }
            else result = 1; // error already reported
            // Data dump done.
         }
         // else there was an error - don't bother. Error already reported - no report here.
                               
         *CLI->output_stream << "};\n"; // BLOB grammar complete.
      }
      // else error already found and reported
   }

   return(result);
}



int Kludge_Instance_Property::Describe_Self(Kludge_CLI* CLI)
{
 string* tmp_pwd = NULL;
 string tmp_cls_name;
 int result = 0;

   if (ClsRel != NULL)
   {
      tmp_pwd = Kludge_CLI_PWD(&ClsRel->Master_Stub,CLI->output_stream);
      if (tmp_pwd == NULL)
      {
         cout << "Internal error: In Kludge_Instance_Property::Describe_Self():\n"
                 " something's horribly wrong - ClsRel not null but Kludge_CLI_PWD() failed.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Error: In Kludge_Instance_Property::Describe_Self():\n"
              " ClsRel pointer is NULL!\n";
      result = 1;
   }

   if (Class == NULL)
   {
      cout << "Error: In Kludge_Instance_Property::Describe_Self():\n"
              " Class pointer is NULL!\n";
      result = 1;
   }

   if (result == 0)
   {
      result = Report_Subtype(CLI);

      *CLI->output_stream 
                // << "INSTANCE_PROPERTY \""
                << "\""
                << RDN
                << "\" \""
                << Class->Name
                << "\" \""
                << *tmp_pwd
                << "\" \""
                << Value
                << "\";\n";
      delete tmp_pwd;
   }
   // else error already reported

   return(result);
}



Kludge_Object* Kludge_Instance_Property::Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
 Kludge_Object* result = NULL;

   if (                                  // a higher-level "folder" in the path?
        (*Content_Name == "..")
        || (*Content_Name == "OWNER")
      )
   {
      result = Instance;
   }
   else if (*Content_Name == ".")        // a "this" style reference?
   {
      result = this;
   }
   else if (*Content_Name == "CLASS")
   {
      result = Class;
   }
   else if (*Content_Name == "CLASS_RELATIONSHIP")
   {
      result = &ClsRel->Master_Stub;
   }
/* else
   {
      *error_output << "Error while resolving name \"" << Content_Name <<
                    << "\" within KLUDGE_INSTANCE_PROPERTY \"" << RDN << "\":\n"
                    << " a Kludge Instance Property doesn't have further nested contents.\n"
   } */

   return(result);
}



int Kludge_Instance_Property::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
   "Cannot insert anything in a Kludge Instance Property.\n"
   " What would you want to insert into a \"leaf object\"?"
   " You can only insert things into classes and instances.\"";

   return(0);
}



int Kludge_Instance_Property::Set_Blob(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
 int result = 0;
 byte_buf* blob_value = NULL;
 string blob_tbl_name;
 ostrstream* qry_string = NULL;

   if (new_value->type == PARM_BLOB_VALUE)
   {
      blob_value = decode_base64(&new_value->content);
   }
   else if (new_value->type == PARM_VALUE)
   {
      blob_value = new byte_buf(
                                (char*) new_value->content.c_str(),
                                new_value->content.size(),
                                false, // no copy construction - just a shallow copy
                                false  // don't dealloc the buffer after use
                               );
   }
   else
   {
      cout <<
         "Internal error - in Kludge_Instance_Property::Set_Blob():\n"
         " obtained a value of type other than PARM_VALUE or PARM_BLOB_VALUE.\n";
      result = 1;
   }
   
   if (blob_value != NULL)
   {
      switch (SubType)
      {
         case SUBTYPE_BLOB:
         {
            blob_tbl_name = "Blobs";
         }
         break;
         case SUBTYPE_TXT_BLOB:
         {
            blob_tbl_name = "Text_Blobs";
         }
         break;
         case SUBTYPE_PROPERTY:
         {
            // this should never happen - filtered in the calling function
            cout << "Internal error - in Kludge_Instance_Property::Set_Blob() :\n"
                    " asked to Set a BLOB value to a plain property !\n";
            result = 1;
         }
         break;
         default:
         {
            cout << "Internal error - in Kludge_Instance_Property::Set_Blob() :\n"
                    " asked to Set a BLOB thingy, obtained an invalid SubType!\n";
            result = 1;
         }
         break;
      }

      if (result == 0)
      {
         char* tmp_escaped_data = new char[blob_value->len * 2 + 1];
         if (tmp_escaped_data != NULL)
         {
            CLI->thr->Open_DB(); // If this is not the first thread to look up all
                                 // the necessary data, the database session might 
                                 // still be closed in this thread.
            mysql_real_escape_string(
                            CLI->thr->DB,            // requires an open DB session!
                            tmp_escaped_data,
                            (char*) blob_value->buf, // the cast suppresses a warning about signedness
                            blob_value->len
                                    );
            // We could consider optimizing memory usage by deallocating
            // blob_value->buf right here - but it would be a terrible hack
            // that might later cause portability problems.
            // Subject to later optimization if necessary.

            qry_string = new ostrstream;
            *qry_string << "UPDATE " << blob_tbl_name << " SET Content = \""
                        << tmp_escaped_data
                        << "\" WHERE BlobID = "
                        << Value
                        <<  ends;   
            int stor_success = CLI->thr->DB_Cmd(qry_string->str());
            if (stor_success != 0)
            {
               cout << "Internal error - in Kludge_Instance_Property::Set_Blob() :\n"
                       " DB op failed while storing BLOB content!\n";
               result = 1;
            }
            //else OK, done storing the data :)
            
            delete qry_string; qry_string = NULL;
            delete tmp_escaped_data; tmp_escaped_data = NULL;
         }
         else
         {
            cout << "Internal error - in Kludge_Instance_Property::Set_Blob() :\n"
                    " while storing BLOB contents, couldn't allocate memory for the escaped string!\n";
            result = 1;
         }
      }
      // else something wrong with the SubType - error already reported

      delete blob_value;
   }
   else
   {
      if (result == 0) // obtained a viable value type
      {
         *CLI->output_stream << 
                 "Error - in Kludge_Instance_Property::Set_Blob():\n"
                 " failed to decode the Base64 BLOB value.\n";
      
         result = 1;
      }
      // else invalid value type - error already reported
   }

   return(result);
}



int Kludge_Instance_Property::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   if (new_value->type == PARM_VALUE)
   {
      if (SubType == SUBTYPE_PROPERTY)
      {
         Value = new_value->content;

         qry_string << "UPDATE Instance_Properties SET Value = \""
                    << Value
                    << "\" WHERE Instance = \""
                    << Instance->OID
                    << "\" AND RDN = \""
                    << RDN
                    << "\""
                    <<  ends;
         result = CLI->thr->DB_Cmd(qry_string.str());
         if (result != 0)
         {
            *CLI->output_stream <<
               "Error: in Kludge_Instance_Property::Set():\n"
               " error while storing the new value into the backend database.\n";
         }
      }
      else // a blob of some kind - Set_Blob() will object against insane SubTypes.
      {
         result = Set_Blob(new_value, CLI);
      }
   }
   else if (new_value->type == PARM_BLOB_VALUE)
   {
      if (SubType != SUBTYPE_PROPERTY)
      {  // a blob of some kind - Set_Blob() will object against insane SubTypes.
         result = Set_Blob(new_value, CLI);
      }
      else
      {
         *CLI->output_stream <<
            "Error - in Kludge_Instance_Property::Set():\n"
            " obtained a Kludge_CLI_CMD_PARM of type PARM_BLOB_VALUE.\n"
            " Cannot set a BLOB value to a plain INSTANCE_PROPERTY.\n"
            " Cannot upgrade a plain INSTANCE_PROPERTY to a BLOB - to achieve this effect,\n"
            " you have to explicitly delete the property and insert a BLOB thingy.\n"
            " Consult your documentation for proper syntax.\n";
         //result = 0; // not an internal error
      }
   }
   else if (new_value->type == PARM_VARIABLE)
   {
      *CLI->output_stream <<
         "Error - in Kludge_Instance_Property::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type PARM_VARIABLE.\n"
         " The \"variable\" format of the SET command is used with relationships,\n"
         " to set relationship attributes. Can't use this format with a property.\n"
         " To a property, you can only set a value.\n"
         " Consult your documentation for proper syntax.\n";
      //result = 0; // not an internal error
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance_Property::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_VALUE or PARM_BLOB_VALUE.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Property::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 0;
 Kludge_Instance* former_owner = NULL;
 ostrstream qry_string;

   if (destination != NULL)
   {
      // can only move under another Kludge Instance
      if (destination->Type == KLUDGE_INSTANCE)
      {
         if (destination->Resolve_Content_Name(&RDN,CLI,/*own only*/true) == NULL)
         {
            if ( *((Kludge_Instance*)destination)->Class <= *ClsRel->Master_Stub.Owner )
            {
               if (Instance != NULL)
               {
                  Add_To_Owner_Index((Kludge_Instance*)destination);
                  Free_From_Owner_Index();
                  former_owner = Instance;
                  Instance = (Kludge_Instance*)destination;
                  
                  qry_string << "UPDATE Instance_Properties SET Instance = \""
                             << Instance->OID
                             << "\" WHERE Instance = \""
                             << former_owner->OID
                             << "\" AND RDN = \""
                             << RDN
                             << "\""
                             <<  ends;
                  result = CLI->thr->DB_Cmd(qry_string.str());
                  if (result != 0)
                  {
                     *CLI->output_stream <<
                        "Error while storing the new owner instance into the backend database.\n";
                  }
               }
               else
               {
                  *CLI->output_stream <<
                  "Internal error: in Kludge_Instance_Property::Move():\n"
                  " can't move an instance property with a NULL owner instance.\n"
                  " Probably an internal error - an attempt to move an unanchored property.\n";
                  result = 1;
               }
            }
            else
            {
               *CLI->output_stream <<
                  "Error: in Kludge_Instance_Property::Move():\n"
                  " the destination is not a valid Owner as per the current ClsRel.\n"
                  " Move canceled.\n";
               result = 1;
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error: in Kludge_Instance_Property::Move():\n"
               " content with this name already exists at the desired destination.\n"
               " Move canceled.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "Error: in Kludge_Instance_Property::Move():\n"
            " can only move under another kludge instance.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error: in Kludge_Instance_Property::Move():\n"
         " obtained a null destination pointer.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance_Property::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 0;
 string old_name;
 ostrstream qry_string;
 
   if (new_name->type == PARM_NAME)
   {
      if (new_name->content.size() != 0)
      {
         if (Class->Resolve_Content_Name(&new_name->content,CLI) == NULL)
         {
            if (ClsRel->Master_Stub.Match_Name(&new_name->content,CLI) == 0)
            {
               Free_From_Owner_Index();
               old_name = RDN;
               RDN = new_name->content;
               Add_To_Owner_Index();

               qry_string << "UPDATE Instance_Properties SET RDN = \""
                          << RDN
                          << "\" WHERE Instance = \""
                          << Instance->OID
                          << "\" AND RDN = \""
                          << old_name
                          << "\""
                          <<  ends;

               result = CLI->thr->DB_Cmd(qry_string.str());
               if (result != 0)
               {
                  *CLI->output_stream <<
                     "In Kludge_Instance_Property::Rename():\n"
                     " error while storing the new name into the backend database.\n";
                  result = 1;
               }
            }
            else
            {
               *CLI->output_stream <<
                  "In Kludge_Instance_Property::Rename(): the requested name doesn't match\n"
                  " the NAMINGMATCH_M specified in the CLS_REL of this instance.\n"
                  " Rename operation canceled.\n";
               result = 1;
            }
         }
         else
         {
            *CLI->output_stream <<
               "In Kludge_Instance_Property::Rename(): content with this name already exists\n"
               " in the owner instance. Rename operation canceled.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Instance_Property::Rename(): the new name cannot be an empty string.\n"
            " Rename operation canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance_Property::Rename():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_NAME.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Property::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;
 ostrstream qry_ss, qry_ss2;
 string blob_tbl_name;

   // Delete this Kludge_Instance_Property from the DB.
   // Can't be deallocated here - this function is used from within
   // a traversal function. Deallocate ('delete') it in the calling function.

   if (SubType != SUBTYPE_PROPERTY)
   {
      switch (SubType)
      {
         case SUBTYPE_BLOB:
         {
            blob_tbl_name = "Blobs";
         }
         break;
         case SUBTYPE_TXT_BLOB:
         {
            blob_tbl_name = "Text_Blobs";
         }
         break;
         default:
         {
            cout << "Internal error - in Kludge_Instance_Property::DB_Delete() :\n"
                    " invalid Kludge_Instance_Property::SubType!\n";
            result = 1;
         }
         break;
      }
      qry_ss2 << "DELETE FROM " << blob_tbl_name << " WHERE BlobID = \""
              << Value
              << "\""
              <<  ends;
      result = CLI->thr->DB_Cmd(qry_ss2.str());
   }
   // else not a BLOB - no dependent BLOB entry to be deleted

   qry_ss << "DELETE FROM Instance_Properties WHERE Instance = \""
          << Instance->OID
          << "\" AND RDN = \""
          << RDN
          << "\""
          <<  ends;
   result += CLI->thr->DB_Cmd(qry_ss.str());

   return(result);
}



int Kludge_Instance_Property::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   int result = DB_Delete(CLI, recursive);
   if (result == 0)
   {
      cout <<
         "Deleting instance property: " << RDN << endl;
      result = Zap();
   }
   return(result);
}



int Kludge_Instance_Property::pwd(string* result)
{
   if (Instance != NULL)
   {
      Instance->pwd(result);
   }
   else
   {
      cout << "Kludge_Instance_Property::pwd(): Instance is null (property uninitialized).\n";
      result->append("Internal error - pointer towards root is null. //");
   }
   result->append(RDN);
   return(0);
}



Kludge_Object* Kludge_Instance_Property::Superior()
{
   return(Instance);
}














/* ###    class Kludge_Inst_Rel_Stub    ### */

// The Kludge_Instance_Relationship (a Kludge_Object look-alike) has two superiors,
// and sometimes we want its methods to return different results based on through
// which superior the method was called. The only way to implement this behavior
// is probably to have the Kludge_Instance_Relationship contain two stubs, that are
// themselves true heirs of Kludge_Object (the Kludge_Instance_Relationship is not)
// but otherwise only serve as wrappers to treat the ambiguities of class
// Kludge_Instance_Relationship. The master and slave stubs have a single clearly
// superior Kludge_Instance and will tell the member method of the respective
// Kludge_Instance_Relationship which end (master/slave) is calling it - or add
// proper differentiated behavior.
// The methods' code is distributed between the stubs and the body based on relevance
// - which happens to be a rather vague criterion.
// The two stubs are truly welded on to the relationship's "body" - they're 
// initialized and deleted along with the body.

Kludge_Inst_Rel_Stub::Kludge_Inst_Rel_Stub(Rel_Stub_Type stub_type, Kludge_Instance_Relationship* rel_body) 
: Stub_Type(stub_type), Body(rel_body), Owner(NULL), Name("")
{
   Type = KLUDGE_INST_REL_STUB;
}



int Kludge_Inst_Rel_Stub::Init_Owner(Kludge_CLI* CLI)
{
 int result = 0;

   if (Owner != NULL)
   {
      if ( Owner->Relationships == NULL ) Owner->Init_Relationships();
   }
   else
   {
      cout << "Kludge_Inst_Rel_Stub::Init(): asked to point to a null Owner!";
      result = 1;
   }

   return(result);
}



int Kludge_Inst_Rel_Stub::Add_To_Owner_Index(Kludge_Instance* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Owner;

   cout << "Adding instance relationship stub " << Name << " to its Owner's index.\n";
   if (Where != NULL)
   {
      if (Where->Relationships != NULL)
      {
         g_tree_insert(Where->Relationships,&Name,this);
      }
      else
      {
         cout << 
            "Internal error - in Kludge_Inst_Rel_Stub::Add_To_Owner_Index(): \n"
            " the Owner's \"relationships\" index not initialized!\n";
         result = 1;
      }
   }
   else
   {
      cout <<
         "Internal error - In Kludge_Inst_Rel_Stub::Add_To_Owner_Index():\n"
         " owner tap undefined! (pointer NULL)\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Inst_Rel_Stub::Free_From_Owner_Index(Kludge_Instance* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = Owner;

   cout << "Releasing instance relationship stub" << Name
        << " from its Owner's index.\n";

   if (Where != NULL)
   {
      if (Where->Relationships != NULL)
      {
         g_tree_remove(Where->Relationships,&Name);
         Where->relationships_loaded = false; // not necessarily true when called from Delete(), but who cares
      }
      else
      {
         cout << 
            "Internal error - In Kludge_Inst_Rel_Stub::Free_From_Owner_Index(): \n"
            " the Owner's \"relationships\" index not initialized!\n";
         result = 1;
      }
   }
   else
   {
      cout << 
         "Internal error - In Kludge_Inst_Rel_Stub::Free_From_Owner_Index(): \n"
         " owner tap undefined! (pointer NULL)\n";
      result = 1;
   }

   return(result);
}



int Kludge_Inst_Rel_Stub::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
 int result = 0;

   *CLI->output_stream <<
         "INSTANCE_RELATIONSHIP" 
         "\n{\n";
      
   result = Body->List_Contents(CLI);

   *CLI->output_stream << "};\n";

   return(result);
}



int Kludge_Inst_Rel_Stub::Describe_Self(Kludge_CLI* CLI)
{
   *CLI->output_stream 
      << "INSTANCE_RELATIONSHIP \""
      << Name
      << "\";\n";

   return(0);
}



Kludge_Object* Kludge_Inst_Rel_Stub::Resolve_Content_Name(
             string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
 Kludge_Object* result = NULL;

   if (*Content_Name == "..")            // a higher-level "folder" in the path?
   {
      result = Owner;
   }
   else if (*Content_Name == ".")        // a "this" style reference?
   {
      result = this;
   }
   else if (*Content_Name == "MASTER")   // a pointer to the master Kludge Instance?
   {
      result = Body->Master_Stub.Owner;
   }
   else if (*Content_Name == "SLAVE")    // a pointer to the slave Kludge Instance?
   {
      result = Body->Slave_Stub.Owner;
   }
   else if (*Content_Name == "PEER")     // a pointer to the other party to the relationship?
   {
      result = ((Stub_Type == MASTER) ? Body->Slave_Stub.Owner : Body->Master_Stub.Owner);
   }
   else if (*Content_Name == "CLS_REL")  // a pointer to the class relationship (relationship class)?
   {
      result = &Body->ClsRel->Master_Stub;
   }

   return(result);
}



int Kludge_Inst_Rel_Stub::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
   return(Body->DB_Delete(CLI));
}



int Kludge_Inst_Rel_Stub::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   return(Body->Delete(CLI));
}



int Kludge_Inst_Rel_Stub::Zap()
{
   return(Body->Zap());
}



int Kludge_Inst_Rel_Stub::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   return(Body->Insert(obj_to_insert,CLI));
}



int Kludge_Inst_Rel_Stub::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   // can only move under another Kludge Instance
   if (destination->Type == Kludge_Object::KLUDGE_INSTANCE)
   {
      if (destination->Resolve_Content_Name(&Name,CLI) == NULL)
      {
         if (Owner != NULL)
         {
            Add_To_Owner_Index((Kludge_Instance*)destination);
            Free_From_Owner_Index();
            Owner = (Kludge_Instance*)destination;
            
            qry_string << "UPDATE Instance_Relationships SET "
                       << ((Stub_Type == MASTER) ? "Master" : "Slave")
                       << " = \""
                       << Owner->OID
                       << "\" WHERE RelOID = \""
                       << Body->RelOID
                       << "\""
                       <<  ends;
            result = CLI->thr->DB_Cmd(qry_string.str());
            if (result != 0)
            {
               *CLI->output_stream <<
                  "Error while storing the new destination reference into the backend database.\n";
            }
         }
         else
         {
            *CLI->output_stream <<
            "Error - in Kludge_Inst_Rel_Stub::Move():\n"
            " can't move an instance relationship stub with null Owner. Probably an internal\n"
            " error - an attempt to move an unanchored instance relationship.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "Error - in Kludge_Inst_Rel_Stub::Move():\n"
            " content with this name already exists at the desired destination. Move canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Error - in Kludge_Inst_Rel_Stub::Move():\n"
         " can only move under another kludge instance.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Inst_Rel_Stub::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;
 
   if (new_name->type == PARM_NAME)
   {
      if (new_name->content.size() != 0)
      {
         if (Owner->Resolve_Content_Name(&new_name->content,CLI) == NULL)
         {
            if (
                  (Stub_Type == MASTER) ?
                  (Body->ClsRel->Master_Stub.Match_Name(&new_name->content,CLI) == 0)
                  : (Body->ClsRel->Slave_Stub.Match_Name(&new_name->content,CLI) == 0)
               )
            {
               Free_From_Owner_Index();
               Name = new_name->content;
               Add_To_Owner_Index();

               qry_string << "UPDATE Instance_Relationships SET "
                          << ((Stub_Type == MASTER) ? "Name_M" : "Name_S")
                          << " = \"" 
                          << Name
                          << "\" WHERE RelCLSID = \""
                          << Body->RelOID
                          << "\""
                          <<  ends;

               result = CLI->thr->DB_Cmd(qry_string.str());
               if (result != 0)
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Inst_Rel_Stub::Rename():\n"
                     " error while storing the new name into the backend database.\n";
               }
            }
            else
            {
               *CLI->output_stream <<
                  "In Kludge_Inst_Rel_Stub::Rename(): the requested name doesn't match the respective\n"
                  " NAMINGMATCH parameter as specified in the CLS_REL of this instance relationship.\n"
                  " Rename operation canceled.\n";
               result = 1;
            }
         }
         else
         {
            *CLI->output_stream <<
               "Error - in Kludge_Inst_Rel_Stub::Rename():\n"
               " content with this name already exists in the owner instance.\n"
               " Rename operation canceled.\n";
            result = 1;
         }
      }
      else
      {
         *CLI->output_stream <<
            "In Kludge_Inst_Rel_Stub::Rename(): the new name cannot be an empty string.\n"
            " Rename operation canceled.\n";
         result = 1;
      }
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Inst_Rel_Stub::Rename():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_NAME.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Inst_Rel_Stub::pwd(string* result)
{
   if (Owner != NULL)
   {
      Owner->pwd(result);
   }
   else
   {
      cout << "Error - in Kludge_Inst_Rel_Stub::pwd():\n"
              " Owner is null (Rel_Stub uninitialized).\n";
      result->append("Internal error - pointer towards root is null. //");
   }
   result->append(Name);
   
   return(0);
}



int Kludge_Inst_Rel_Stub::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
   return(Body->Set(new_value, CLI));
}



Kludge_Object* Kludge_Inst_Rel_Stub::Superior()
{
   return(Owner);
}











/* ###    class Kludge_Instance_Relationship    ### */


// This constructor is used for creating a new instance relationship from scratch.
// It cannot be used for loading an instance relationship from the DB.
//
// At this constructor, I had to strike a reasonable tradeoff.
// The transformation of Kludge_CLI_CMD_PARM* INSTANCE_RELATIONSHIP into
// a Kludge_Instance_Relationship is non-trivial and, if I wanted to use
// a heavy-weight constructor as with Kludge_Class or K_C_Property, I'd
// end up passing quite a few variables - or a prototype-style object.
// So I decided to use a lightweight constructor, do the transform over
// the uninitialized object and use a stand-alone initializer method
// to set the finished Kludge_Instance_Relationship airborne.
//
// To store the newly created instance relationship in the DB and hook it up
// in the trees and indexes, use Kludge_Instance_Relationship::Init().
//
Kludge_Instance_Relationship::Kludge_Instance_Relationship() :
  Master_Stub(Kludge_Inst_Rel_Stub::MASTER, this),
  Slave_Stub(Kludge_Inst_Rel_Stub::SLAVE, this),
  RelOID(0),  // will be established upon storage to the database
  ClsRel(NULL),
  ctor_failed(false)
{
}

int Kludge_Instance_Relationship::Init(Kludge_CLI* CLI)
{
 int stor_success = 0, result = 0;
 ostrstream qry_string;

   //Type = KLUDGE_INSTANCE_RELATIONSHIP; // not a Kludge_Object heir anymore

   result = Master_Stub.Init_Owner(CLI) + Slave_Stub.Init_Owner(CLI); // :)
   if (ClsRel == NULL) result = 1;
  
   if (result == 0)
   {
      qry_string << "INSERT INTO Instance_Relationships VALUES (NULL,\""
                 << Master_Stub.Owner->OID << "\",\""
                 << Slave_Stub.Owner->OID << "\",\""
                 << ClsRel->RelCLSID << "\",\""
                 << Master_Stub.Name << "\",\""
                 << Slave_Stub.Name << "\")"
                 <<  ends ; 

      stor_success = CLI->thr->DB_Cmd(qry_string.str());
      if (stor_success == 0)
      {
         RelOID = mysql_insert_id(CLI->thr->DB);
         Add_To_Index();
      }
      else
      {
         cout << "Storage of a newly created instance relationship "
              << Master_Stub.Name << "/" << Slave_Stub.Name << " has failed!\n";
         result = 1;
      }
   }

   return(result);
}



// This constructor can only be used for loading an existing 
// Kludge_Instance_Relationship from a database. It is not intended for creating
// a new instance relationship from scratch.
//
Kludge_Instance_Relationship::Kludge_Instance_Relationship(
  MYSQL_ROW DB_record,
  Kludge_CLI* CLI
) : 
  Master_Stub(Kludge_Inst_Rel_Stub::MASTER, this),
  Slave_Stub(Kludge_Inst_Rel_Stub::SLAVE, this),
  RelOID(0),
  // RelType(-1), // 0 = peer2peer, 1 = master/slave
  ClsRel(NULL),
  ctor_failed(false)
{
   //Type = KLUDGE_INSTANCE_RELATIONSHIP;

   if ( ! Extract_From_MysqlRow(DB_record,CLI) )
   {
      cout << "Extraction of an instance relationship out of a DB record has failed!\n";
      ctor_failed = true;
   }
   else
   {
      // Master and Slave are both loaded (by Extract_From_MysqlRow()
      Master_Stub.Init_Owner(CLI);
      Slave_Stub.Init_Owner(CLI);
      Add_To_Index();
   }
}



Kludge_Instance_Relationship::~Kludge_Instance_Relationship()
{
   cout << "Deallocating instance relationship (name not available).\n";
}



// Zap() does a nice memory cleanup.
// Only makes sense 
// a) for dynamic partial release from memory 
//     In that case, invalidate the list of relationships.
// b) when an instance relationship is deleted
//     Requires deletion from the SQL-based object repository
//      - i.e., use DB_Delete() and then Zap() (or just Delete())
int Kludge_Instance_Relationship::Zap()
{
 int result = 0;

   cout << "Zapping an instance relationship: "
        << Master_Stub.Name << "/" << Slave_Stub.Name << "\n" ;

   result = Free_From_Index();
   
   delete this;
   
   return(result);
}



//
// Extract_From_MysqlRow()
//
// Expects:
// Row_To_Process[0] == RelOID
// Row_To_Process[1] == Master
// Row_To_Process[2] == Slave
// Row_To_Process[3] == ClsRel
// Row_To_Process[4] == Name_M
// Row_To_Process[5] == Name_S
//
bool Kludge_Instance_Relationship::Extract_From_MysqlRow(MYSQL_ROW Row_To_Process, Kludge_CLI* CLI) 
{
 bool result = true;
 istrstream* tmp_str = NULL;
 OID_Type master_ID = 0, slave_ID = 0, cls_rel_ID = 0;

   if (Row_To_Process != NULL)
   {
      tmp_str = new istrstream(Row_To_Process[0]);
      *tmp_str >> RelOID;
      delete tmp_str;

      tmp_str = new istrstream(Row_To_Process[1]);
      *tmp_str >> master_ID;
      delete tmp_str;
      
      tmp_str = new istrstream(Row_To_Process[2]);
      *tmp_str >> slave_ID;
      delete tmp_str;

      tmp_str = new istrstream(Row_To_Process[3]);
      *tmp_str >> cls_rel_ID;
      delete tmp_str;

      Master_Stub.Name = Row_To_Process[4];
      Slave_Stub.Name = Row_To_Process[5];

      Master_Stub.Owner = Kludge_Instance::ID2ptr(&master_ID, CLI);
      if (Master_Stub.Owner == NULL)
      {
         cout << 
            "In Kludge_Instance_Relationship::Extract_From_MysqlRow():\n"
            " Couldn't resolve the Master class ID (dangling reference).\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }

      Slave_Stub.Owner = Kludge_Instance::ID2ptr(&slave_ID, CLI);
      if (Slave_Stub.Owner == NULL)
      {
         cout << 
            "In Kludge_Instance_Relationship::Extract_From_MysqlRow():\n"
            " Couldn't resolve the Slave class ID (dangling reference).\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }

      ClsRel = Kludge_Class_Relationship::ID2ptr(&cls_rel_ID,CLI);
      if (ClsRel == NULL)
      {
         cout <<
            "In Kludge_Instance_Relationship::Extract_From_MysqlRow():\n"
            " ClsRel ID could not be resolved to a K_C_R pointer!\n"
            " Inconsistency in the backend SQL database?\n"
            " Extract_From_MysqlRow() failed.\n";
         result = false;
      }
   }
   else
   {
      cout << 
         "In Kludge_Instance_Relationship::Extract_From_MysqlRow():\n"
         " the row passed to this function is a NULL pointer!\n"
         " Extract_From_MysqlRow() failed.\n";
      result = false;
   }

   return(result);
}



int Kludge_Instance_Relationship::Add_To_ClsRel_Index(Kludge_Class_Relationship* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = ClsRel;

   cout << "Adding instance relationship "
        << Master_Stub.Name 
        << "/"
        << Slave_Stub.Name 
        << " to its class relationship's index.\n";

   if (Where != NULL)
   {
      if (Where->Instances == NULL) Where->Init_Instances();
      
      if (Where->Instances != NULL)
      {
         g_tree_insert(Where->Instances,&RelOID,this);
      }
      else
      {
         cout << "Internal error - in Kludge_Instance_Relationship::Add_To_ClsRel_Index():\n"
                 " could not initialize the Class Relationship's index of instances.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error - in Kludge_Instance_Relationship::Add_To_ClsRel_Index():\n"
               " pointer to the target class relationship is NULL.\n";
      result = 1;
   }

   return(result);
}



int Kludge_Instance_Relationship::Add_To_Index()
{
 int result = 0;
 
   result = Master_Stub.Add_To_Owner_Index();
   if (result == 0) result = Slave_Stub.Add_To_Owner_Index();
   if (result == 0) result = Add_To_ClsRel_Index();
   
   return(result);
}



int Kludge_Instance_Relationship::Free_From_ClsRel_Index(Kludge_Class_Relationship* Where = NULL)
{
 int result = 0;
 if (Where == NULL) Where = ClsRel;

   cout << "Releasing instance relationship "
        << Master_Stub.Name 
        << "/"
        << Slave_Stub.Name 
        << " from its Class Relationship's index.\n";

   if (Where != NULL)
   {
      if (Where->Instances != NULL)
      {
         g_tree_remove(Where->Instances,&RelOID);
      }
      else
      {
         cout << "Internal error - in Kludge_Instance_Relationship::Free_From_ClsRel_Index():\n"
                 " the Class Relationship's \"Instances\" index not initialized.\n";
         result = 1;
      }
   }
   else
   {
      cout << "Internal error - in Kludge_Instance_Relationship::Free_From_ClsRel_Index():\n"
               " pointer to the target class relationship is NULL.\n";
      result = 1;
   }
   
   return(result);
}



int Kludge_Instance_Relationship::Free_From_Index()
{
 int result = 0;
 
   result = Master_Stub.Free_From_Owner_Index()
          + Slave_Stub.Free_From_Owner_Index()
          + Free_From_ClsRel_Index();

   return(result);
}



int Kludge_Instance_Relationship::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
 string *tmp_pwd_m = NULL, *tmp_pwd_s = NULL, *tmp_pwd_c = NULL;

   tmp_pwd_m = Kludge_CLI_PWD(Master_Stub.Owner,CLI->output_stream);
   tmp_pwd_s = Kludge_CLI_PWD(Slave_Stub.Owner,CLI->output_stream);
   tmp_pwd_c = Kludge_CLI_PWD(&ClsRel->Master_Stub,CLI->output_stream);

   *CLI->output_stream <<
      "MASTER=\""         << *tmp_pwd_m        << "\";\n"
      "SLAVE=\""          << *tmp_pwd_s        << "\";\n"
      "NAME_M=\""         << Master_Stub.Name << "\";\n"
      "NAME_S=\""         << Slave_Stub.Name  << "\";\n"
      "CLS_REL=\""        << *tmp_pwd_c        << "\";\n";

   if (tmp_pwd_m != NULL) delete tmp_pwd_m;
   if (tmp_pwd_s != NULL) delete tmp_pwd_s;
   if (tmp_pwd_c != NULL) delete tmp_pwd_c;

   return(0);
}



int Kludge_Instance_Relationship::DB_Delete(Kludge_CLI* CLI)
{
 int result = 0;
 ostrstream qry_string;

   // Delete this Kludge_Instance_Relationship from the DB.
   // Can't be deallocated here - this function is used from within
   // a traversal function. Deallocate ('delete') it in the calling function.

   qry_string << "DELETE FROM Instance_Relationships WHERE RelOID = \""
              << RelOID << "\""
              <<  ends;
   result = CLI->thr->DB_Cmd(qry_string.str());

   return(result);
}



int Kludge_Instance_Relationship::Delete(Kludge_CLI* CLI)
{
   int result = DB_Delete(CLI);
   if (result == 0)
   {
      cout <<
         "Deleting instance relationship: " << Master_Stub.Name << 
         "<==>" << Slave_Stub.Name << endl;
      result = Zap();
   }
   return(result);
}



int Kludge_Instance_Relationship::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
   "Cannot insert anything in a Kludge Instance Relationship.\n"
   " For the moment, instance relationships are relatively dumb leaf objects.\n"
   " You can only insert things into classes and instances.\"";

   return(0);
}



// If someone asks for the PWD to a Kludge_Instance_Relationship,
// let's take the master route (even with P2P relationships).
int Kludge_Instance_Relationship::pwd(string* result)
{
   if (Master_Stub.Owner != NULL)
   {
      Master_Stub.Owner->pwd(result);
   }
   else
   {
      cout << 
         "Internal error - in Kludge_Instance_Relationship::pwd():\n"
         " Master is null (Relationship uninitialized).\n";
      result->append("Internal error - pointer towards root is null. //");
   }
   result->append(Master_Stub.Name);
   
   return(0);
}



int Kludge_Instance_Relationship::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
 int result = 0;
 bool user_error = false;
 Kludge_Path* tmp_path = NULL;
 Kludge_Object* tmp_obj_ptr = NULL;
 ostrstream qry_string;
 string name;
 ostrstream value;

   if (new_value->type == PARM_VARIABLE)
   {
      if ((new_value->contents[0]->type == PARM_NAME) && (new_value->contents[1]->type == PARM_VALUE))
      {
         // parse attribute name and value
         if (new_value->contents[0]->content == "CLS_REL")
         {
            // value should not be an empty string
            if (new_value->contents[1]->content.size() != 0)
            {
               tmp_path = Kludge_CLI_String2path(
                             &new_value->contents[1]->content, //Value
                             CLI->output_stream
                          );
               // does path/name string resolve fine to a path (vector of elements) ?
               if (tmp_path != NULL)
               {
                  tmp_obj_ptr = Kludge_CLI_Path2Ptr(tmp_path, CLI, CLI->pwd);
                  // does path (vector of elements) resolve fine into an object pointer ?
                  if (tmp_obj_ptr != NULL)
                  {
                     // is the resolved prospective ClsRel object really a Kludge_Class_Relationship ? 
                     if (tmp_obj_ptr->Type == Kludge_Object::KLUDGE_CLASS_REL_STUB)
                     {
                        if
                        (
                           (
                              *((Kludge_Class_Rel_Stub*)tmp_obj_ptr)->Body->Master_Stub.Owner 
                              >= *Master_Stub.Owner->Class
                           )
                           &&
                           (
                              *((Kludge_Class_Rel_Stub*)tmp_obj_ptr)->Body->Slave_Stub.Owner
                              >= *Slave_Stub.Owner->Class
                           )
                        )
                        {
                           ClsRel = ((Kludge_Class_Rel_Stub*)tmp_obj_ptr)->Body;
                           name = "ClsRel";
                           value << ClsRel->RelCLSID;
                        }
                        else
                        {
                           *CLI->output_stream <<
                              "Error - in Kludge_Instance_Relationship::Set():\n"
                              " the desired CLS_REL doesn't match the master and slave instances' classes.\n";
                           user_error = true;
                           // result = 0; Not an internal error.
                        }
                     }
                     else
                     {
                        *CLI->output_stream <<
                           "Error - in Kludge_Instance_Relationship::Set():\n"
                           " Path to CLS_REL resolves to a kludge object other than CLASS_RELATIONSHIP.\n";
                        user_error = true;
                        // result = 0; Not an internal error.
                     }

                     // do not delete tmp_obj_ptr - object neither created nor owned by us
                     tmp_obj_ptr = NULL;
                  }
                  else
                  {
                     *CLI->output_stream <<
                        "Error - in Kludge_Instance_Relationship::Set():\n"
                        " Path2Ptr failed to follow the given path to CLS_REL - wrong path?\n";
                     user_error = true;
                     // result = 0; Not an internal error.
                  }

                  delete tmp_path; tmp_path = NULL;
               }
               else
               {
                  *CLI->output_stream <<
                     "Error - in Kludge_Instance_Relationship::Set():\n"
                     " String2path failed while parsing path to CLS_REL - probably a syntax error.\n";
                  user_error = true;
                  // result = 0; Not an internal error.
               }
            }
            else
            {
               *CLI->output_stream <<
                  "Error - in Kludge_Instance_Relationship::Set():\n"
                  " CLS_REL cannot be empty. Empty value NOT set.\n";
               user_error = true;
            }
         }
         else if (new_value->contents[0]->content == "NAME_M")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance_Relationship::Set():\n"
               " can't modify NAME_M - use the RENAME command.\n";
            user_error = true;
         }
         else if (new_value->contents[0]->content == "NAME_S")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance_Relationship::Set():\n"
               " can't modify NAME_S - use the RENAME command.\n";
            user_error = true;
         }
         else if (new_value->contents[0]->content == "MASTER")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance_Relationship::Set():\n"
               " can't modify MASTER - use the MOVE command.\n";
            user_error = true;
         }
         else if (new_value->contents[0]->content == "SLAVE")
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance_Relationship::Set():\n"
               " can't modify SLAVE - use the MOVE command.\n";
            user_error = true;
         }
         else
         {
            *CLI->output_stream <<
               "Error - in Kludge_Instance_Relationship::Set():\n"
               " unknown attribute name: \"" << new_value->contents[0]->content <<
               "\".\n";
            user_error = true;
         }
         // done parsing attribute name and value

         if (user_error = false)
         {
            qry_string << "UPDATE Instance_Relationships SET "
                       << name 
                       << " = \""
                       << value.str() 
                       << "\" WHERE RelOID = \""
                       << RelOID
                       << "\""
                       <<  ends;

            result = CLI->thr->DB_Cmd(qry_string.str());
            if (result != 0)
            {
               *CLI->output_stream <<
                  "Error while storing the new attriute value into the backend database.\n";
               result = 1;
            }
         }
         //else nothing to do, error already reported.
      }
      else
      {
         *CLI->output_stream <<
            "Internal error - in Kludge_Instance_Relationship::Set():\n"
            " obtained a PARM_VARIABLE with improper contents.\n";
         result = 1;
      }
   }
   else if (new_value->type == PARM_VALUE)
   {
      *CLI->output_stream <<
         "Error - in Kludge_Instance_Relationship::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type PARM_VALUE.\n"
         " The \"value\" format of the SET command is used with properties,\n"
         " to set property value. Can't use this format with a relationship.\n"
         " You can only set an attribute=value to a relationship.\n"
         " Consult your documentation for proper syntax.\n";
      //result = 0; // not an internal error
   }
   else
   {
      *CLI->output_stream <<
         "Internal error - in Kludge_Instance_Relationship::Set():\n"
         " obtained a Kludge_CLI_CMD_PARM of type other than PARM_VARIABLE or PARM_VALUE.\n";
      result = 1;
   }
   
   return(result);
}











/* ###    class Kludge_Root_Object    ### */


Kludge_Root_Object::Kludge_Root_Object()
 : Name("Kludge_Root")
{
   Type = KLUDGE_ROOT;
   Classes = new Kludge_Class("Root_Class", NULL, Pseudo_Master_Thread->CLI);
   Instances = new Kludge_Instance("Root_Instance", NULL, Pseudo_Master_Thread->CLI);
}



Kludge_Root_Object::~Kludge_Root_Object()
{
   delete Instances;
   delete Classes;
}



int Kludge_Root_Object::Zap()
{
   cout <<
   "In Kludge_Root_Object::Zap(): cannot delete the root pseudo-object :)\n";
   return(1);
}



int Kludge_Root_Object::List_Contents(Kludge_CLI* CLI, bool recursive = false)
{
   *CLI->output_stream << "ROOT\n{\n";
   
   if (recursive)
   {
      Classes->List_Contents(CLI, recursive);
      Instances->List_Contents(CLI, recursive);
   }
   else
   {
      Classes->Describe_Self(CLI);
      Instances->Describe_Self(CLI);
   }
   
   *CLI->output_stream << "};\n";

   return(0);
}



Kludge_Object* Kludge_Root_Object::Resolve_Content_Name(string* Content_Name, Kludge_CLI* CLI, bool own_only = false)
{
 Kludge_Object* result = NULL;
 
   if ((*Content_Name == "Root_Class") || (*Content_Name == "Classes"))
   {
      result = Classes;
   }
   else if ((*Content_Name == "Root_Instance") || (*Content_Name == "Instances"))
   {
      result = Instances;
   }
   else if (*Content_Name == ".")
   {
      result = this;
   }
   
   return(result);
}



int Kludge_Root_Object::Insert(Kludge_CLI_CMD_PARM* obj_to_insert, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
   "Cannot insert anything into the Kludge Root.\n"
   " It already contains everything it should.\n"
   " You can insert your own items under the Root Class and the Root Instance.\n";

   return(0);
}



int Kludge_Root_Object::Set(Kludge_CLI_CMD_PARM* new_value, Kludge_CLI* CLI)
{
   *CLI->output_stream <<
      "It doesn't make sense to set a value to the root pseudo-object.\n";
   return(0);
}



int Kludge_Root_Object::Move(Kludge_Object* destination, Kludge_CLI* CLI)
{
 int result = 1;

   *CLI->output_stream <<
   "In Kludge_Root_Object::Move(): cannot move the root pseudo-object.\n";

   return(result);
}



int Kludge_Root_Object::Rename(Kludge_CLI_CMD_PARM* new_name, Kludge_CLI* CLI)
{
 int result = 1;

   *CLI->output_stream <<
   "In Kludge_Root_Object::Rename(): cannot rename the root pseudo-object.\n";

   return(result);
}



int Kludge_Root_Object::DB_Delete(Kludge_CLI* CLI, bool recursive = false)
{
   *CLI->output_stream <<
   "In Kludge_Root_Object::DB_Delete(): cannot delete the root pseudo-object :)\n";
   return(1);
}



int Kludge_Root_Object::Delete(Kludge_CLI* CLI, bool recursive = false)
{
   *CLI->output_stream <<
   "In Kludge_Root_Object::Delete(): cannot delete the root pseudo-object :)\n";
   return(1);
}



int Kludge_Root_Object::pwd(string* result)
{
   result->append("/");
   return(0);
}



Kludge_Object* Kludge_Root_Object::Superior()
{
   return(NULL);
}









/* ### global functions ### */

// GTree traversal and comparison functions


// Thanks to polymorphism and the virtual mechanism, this traversal helper 
// function can be used for classes, class properties and class relationships :)
// The deletion is recursive.
int Kludge_Delete_Contents_Traverse(gpointer key, gpointer Content, gpointer del_parm_passed)
{
 string* tmp_pwd = NULL;
 Kludge_CLI* CLI = ((Kludge_Object::del_param*)del_parm_passed)->CLI;

   // if the content has Delete()d itself cleanly, add it to the vector<> for
   // deallocation ( done by the calling object's Delete() )
   if ( ((Kludge_Object*)Content)->DB_Delete(CLI,true) == 0)
   {
      ((Kludge_Object::del_param*)del_parm_passed)->objs_deleted.push_back((Kludge_Object*)Content);
   }
   // else report that the content cannot be deleted. 
   else
   {
      tmp_pwd = Kludge_CLI_PWD((Kludge_Object*)Content,CLI->output_stream);
      *CLI->output_stream
         << "Content \"" << *tmp_pwd << "\" could not be deleted.\n";
      delete tmp_pwd;
   }
   
   return(false);
}



// for string comparisons in GTrees
int Kludge_Compare_Names(gconstpointer a, gconstpointer b)
{
   if (*(string*) a < *(string*) b) return(-1);
   else
   {
      if (*(string*) a > *(string*) b) return(1);
      else return(0);
   }
}



// for integer comparisons in GTrees
int Kludge_Compare_IDs(gconstpointer a, gconstpointer b)
{
   if ( *(unsigned int *)a < *(unsigned int *)b ) return(-1);
   else if ( *(unsigned int *)a == *(unsigned int *)b ) return(0);
   else return(1);
}



// a cleanup for global index trees that are declared static
// and are therefore defined in the mandatory static members definition section
int Clean_Up_Global_Indexes()
{
   g_tree_destroy(Kludge_Class::By_Name);
   g_tree_destroy(Kludge_Class::By_ID);
   g_tree_destroy(Kludge_Class_Relationship::Index);
   g_tree_destroy(Kludge_Instance::Index);

   return(0);
}

#undef MY_SSCANF
